#ifndef TTCL_CL_SOFT_CLUSTERING_HXX
#define TTCL_CL_SOFT_CLUSTERING_HXX

// TTCL: The Template Clustering Library

/** @file
    Clustering - Soft clustering results
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/exception.hxx>
#include <ttcl/cl/hard_clustering.hxx>
#include <ttcl/co/boost_matrix.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Clustering Namespace
  namespace cl {

    /// Hard clustering results
    template < typename Container >
    class hard_clustering;

    /// Soft clustering results
    /** @param [Container] Underlying container
     */
    template < typename Container = co::boost_matrix<double> >
    class soft_clustering : public Container {
    private:
      /// Base type
      typedef Container base_type;

    public:
      /// Element type
      typedef typename base_type::element element;

      /// Size type
      typedef typename base_type::size_type size_type;

      /// Default constructor
      soft_clustering() :
	base_type() {
      }

      /// Constructor
      /** @param _size Size
	  @param _clusters Number of clusters
      */
      soft_clustering(size_type _size,
		      size_type _clusters) :
	base_type(_size, _clusters) {
      }

      /// Copy constructor
      /** @param _other Source soft_clustering
       */
      soft_clustering(const soft_clustering& _other) :
	base_type(static_cast<const base_type&>(_other)) {
      }

      /// Conversion constructor
      /** @param _other Source hard_clustering
       */
      template <typename Cont>
      soft_clustering(const hard_clustering<Cont>& _other) :
	base_type(_other.size(), _other.clusters()) {
	// Other value type
	typedef typename hard_clustering<Cont>::value_type other_value_type;

	// Set each
	for (size_type row = 0; row < this->size(); ++row)
	  for (size_type col = 0; col < clusters(); ++col)
	    (*this)[row][col] =
	      (_other[row] == other_value_type(col) ? element(1) : element(0));
      }

      /// Assignment operator
      /** @param _other Source soft_clustering
       */
      soft_clustering& operator=(const soft_clustering& _other) {
	// Assign as a container
	this->base_type::operator=(_other);

	// Return this
	return *this;
      }

      /// Assignment operator
      /** @param _other Source soft_clustering
       */
      template <typename Cont>
      soft_clustering& operator=(const hard_clustering<Cont>& _other) {
	// Other value type
	typedef typename hard_clustering<Cont>::value_type other_value_type;

	// Resize
	this->resize(_other.size(), _other.clusters());

	// Set each
	for (size_type row = 0; row < this->size(); ++row)
	  for (size_type col = 0; col < clusters(); ++col)
	    (*this)[row][col] =
	      (_other[row] == other_value_type(col) ? element(1) : element(0));

	// Return this
	return *this;
      }

      /// Number of clusters
      size_type clusters() const {
	// Numer of columns
	return this->columns();
      }

      /// Most probable cluster
      ssize_t mpc(size_type _idx) const {
	// Probability
	ssize_t high_cl = -1;
	element high_pr = element();
	for (size_type cl = 0; cl < clusters(); ++cl) {
	  if ((*this)[_idx][cl] > high_pr) {
	    high_cl = cl;
	    high_pr = (*this)[_idx][cl];
	  }
	}

	// Return it
	return high_cl;
      }
    };
  }
}

#endif
