#ifndef TTCL_CL_SOFT_DATA_MODEL_CLUSTERER_HXX
#define TTCL_CL_SOFT_DATA_MODEL_CLUSTERER_HXX

// TTCL: The Template Clustering Library

/** @file
    Clustering - Soft model data clusterer
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/cl/choices.hxx>
#include <ttcl/cl/soft_clustering.hxx>
#include <ttcl/cl/soft_data_clusterer.hxx>
#include <ttcl/cl/soft_data_model.hxx>
#include <ttcl/tu/choices.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Clustering Namespace
  namespace cl {

    /// Soft data model clusterer
    /** @param Data The data type
     */
    template <typename Data,
	      typename Choices = tu::default_choices>
    class soft_data_model_clusterer :
      public soft_data_clusterer<Data, Choices> {
    public:
      /// Base type
      typedef soft_data_clusterer<Data, Choices> base_type;

      /// Data type
      typedef typename base_type::data_type data_type;

      /// Hard clustering type
      typedef typename base_type::hard_clustering_type hard_clustering_type;

      /// Soft clustering type
      typedef typename base_type::soft_clustering_type soft_clustering_type;

      /// Model type
      typedef soft_data_model<data_type, Choices> model_type;

      /// Create a model
      virtual model_type*
      create_model(const data_type& _data,
		   const uint _n_clusters) const = 0;

      /// Model
      /** @param _soft_clustering Target soft clustering
	  @param _data Data
	  @param _n_clusters Number of clusters
      */
      virtual model_type*
      soft_model_on(soft_clustering_type& _soft_clustering,
		    const data_type& _data,
		    const uint _n_clusters) const = 0;

      /// Model
      /** @param _hard_clustering Target hard clustering
	  @param _data Data
	  @param _clusters Number of clusters
      */
      virtual model_type*
      hard_model_on(hard_clustering_type& _hard_clustering,
		    const data_type& _data,
		    const uint _clusters) const {
	// Create soft clustering
	soft_clustering_type soft_clustering(_data.size(), _clusters);

	// Soft model on it
	model_type* model = soft_model_on(soft_clustering, _data, _clusters);

	// Assign clustering
	_hard_clustering = soft_clustering;

	// Return the model
	return model;
      }

      /// Model
      /** @param _data Data
	  @param _n_clusters Number of clusters
      */
      model_type*
      model(const data_type& _data,
	    const uint _n_clusters) const {
	// Cluster and return
	soft_clustering_type soft_clustering(_data.size(), _n_clusters);
	return soft_model_on(soft_clustering, _data, _n_clusters);
      }

      /// Perform the soft clustering
      /** @param _soft_clustering Target soft clustering
	  @param _data Data
	  @param _n_clusters Number of clusters
      */
      virtual void
      soft_cluster_on(soft_clustering_type& _soft_clustering,
		      const data_type& _data,
		      const uint _n_clusters) const {
	// Cluster using a model, and ignore it
	model_type* model = soft_model_on(_soft_clustering, _data, _n_clusters);
	delete model;
      }
    };
  }
}

#endif
