#ifndef TTCL_CL_WR_HARD_DATA_MODEL_CLUSTERER_WRAP_HXX
#define TTCL_CL_WR_HARD_DATA_MODEL_CLUSTERER_WRAP_HXX

// TTCL: The Template Clustering Library

/** @file
    Clustering - Hard data model clusterer wrapper
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/cl/wr/data_wrap.hxx>
#include <ttcl/cl/wr/hard_data_model_wrap.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/wr/typed_shared_referee.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Clustering Namespace
  namespace cl {

    /// Wrappers Namespace
    namespace wr {

      /// Hard data model clusterer interface
      template <typename Choices>
      class _hard_data_model_clusterer_interface :
	public _hard_data_clusterer_interface<Choices> {
      public:
	/// Base type
	typedef _hard_data_clusterer_interface<Choices> base_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Model type
	typedef hard_data_model_wrap<Choices> model_wrap_type;

	/// Constructor
	_hard_data_model_clusterer_interface(bool _owned = true,
					     uint _references = 0) :
	  base_type(_owned, _references) {
	}

	/// Create a model
	virtual model_wrap_type
	create_model(const data_wrap& _data,
		     const uint _n_clusters) const = 0;

	/// Model
	virtual model_wrap_type
	hard_model_on(hard_clustering_type& _hard_clustering,
		      const data_wrap& _data,
		      const uint _n_clusters) const = 0;

	/// Model
	virtual model_wrap_type
	model(const data_wrap& _data,
	      const uint _n_clusters) const = 0;
      };

      /// Hard data clusterer referee
      template <typename T,
		typename Interface>
      class _hard_data_model_clusterer_referee :
	public _hard_data_clusterer_referee<T, Interface> {
      private:
	// Data type
	typedef typename T::data_type data_type;

	// Model type
	typedef typename T::model_type model_type;

	/// Hard clustering type
	typedef typename Interface::hard_clustering_type hard_clustering_type;

	/// Model wrap type
	typedef typename Interface::model_wrap_type model_wrap_type;

      public:
	/// Constructor
	_hard_data_model_clusterer_referee(T* _ptr = 0, bool _owned = true) :
	  _hard_data_clusterer_referee<T, Interface>(_ptr, _owned) {
	}

	/// Create a model
	virtual model_wrap_type
	create_model(const data_wrap& _data,
		     const uint _n_clusters) const {
	  // Convert to data type
	  const data_type* data = _data.get<data_type>();

	  // Call and return
	  return model_wrap_type
	    (this->ptr_->create_model(*data, _n_clusters));
	}

	/// Model
	virtual model_wrap_type
	hard_model_on(hard_clustering_type& _hard_clustering,
		      const data_wrap& _data,
		      const uint _n_clusters) const {
	  // Convert data type
	  const data_type* data = _data .get<data_type> ();

	  // Call and return
	  return model_wrap_type
	    (this->ptr_->hard_model_on(_hard_clustering,
				       *data, _n_clusters));
	}

	/// Model
	virtual model_wrap_type
	model(const data_wrap& _data,
	      const uint _n_clusters) const {
	  // Convert to data type
	  const data_type* data = _data.get<data_type>();

	  // Call and return
	  return model_wrap_type
	    (this->ptr_->model(*data, _n_clusters));
	}
      };

      /// Base hard data model clusterer wrapper
      template <typename Interface>
      class _hard_data_model_clusterer_wrap :
	public _hard_data_clusterer_wrap<Interface> {
      public:
	/// Base type
	typedef _hard_data_clusterer_wrap<Interface> base_type;

	/// Hard clustering type
	typedef typename Interface::hard_clustering_type hard_clustering_type;

	/// Model wrap type
	typedef typename Interface::model_wrap_type model_wrap_type;

      protected:
	/// Empty constructor
	_hard_data_model_clusterer_wrap() :
	  base_type() {
	}

	/// Constructor
	_hard_data_model_clusterer_wrap(Interface* _ptr) :
	  base_type(_ptr) {
	}

      public:
	/// Create a model
	virtual model_wrap_type
	create_model(const data_wrap& _data,
		     const uint _n_clusters) const {
	  // Call
	  return this->ptr_->create_model(_data, _n_clusters);
	}

	/// Model
	virtual model_wrap_type
	hard_model_on(hard_clustering_type& _hard_clustering,
		      const data_wrap& _data,
		      const uint _n_clusters) const {
	  // Call
	  return this->ptr_->hard_model_on(_hard_clustering,
					   _data, _n_clusters);
	}

	/// Model
	virtual model_wrap_type
	model(const data_wrap& _data,
	      const uint _n_clusters) const {
	  // Call
	  return this->ptr_->model(_data, _n_clusters);
	}
      };

      /// Hard data clusterer wrapper
      template <typename Choices = tu::default_choices>
      class hard_data_model_clusterer_wrap :
	public _hard_data_model_clusterer_wrap
          <_hard_data_model_clusterer_interface<Choices> > {
      private:
	/// Interface
	typedef _hard_data_model_clusterer_interface<Choices> interface;

	/// Base type
	typedef _hard_data_model_clusterer_wrap<interface> base_type;

      public:
	/// Empty Constructor
	hard_data_model_clusterer_wrap() :
	  base_type() {
	}

	/// Constructor
	template <typename T>
	hard_data_model_clusterer_wrap(T* _value, bool _owned = true) :
	  base_type(new _hard_data_model_clusterer_referee
		      <T, interface>(_value, _owned)) {
	}

	/// Constructor
	template <typename T>
	hard_data_model_clusterer_wrap(T& _value, bool _owned = false) :
	  base_type(new _hard_data_model_clusterer_referee
		      <T, interface>(&_value, _owned)) {
	}
      };
    }
  }
};

#endif
