#ifndef TTCL_CL_WR_HARD_MATRIX_CLUSTERER_WRAP_HXX
#define TTCL_CL_WR_HARD_MATRIX_CLUSTERER_WRAP_HXX

// TTCL: The Template Clustering Library

/** @file
    Clustering - Hard matrix clusterer wrapper
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/cl/choices.hxx>
#include <ttcl/cl/hard_clustering.hxx>
#include <ttcl/cl/wr/matrix_wrap.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/wr/interface_wrap.hxx>
#include <ttcl/wr/shared_referee.hxx>
#include <ttcl/wr/typed_shared_referee.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Clustering Namespace
  namespace cl {

    /// Wrappers Namespace
    namespace wr {

      /// Hard matrix clusterer interface
      template <typename Choices>
      class _hard_matrix_clusterer_interface :
	public ttcl::wr::shared_referee {
      public:
	/// Hard clustering container
	typedef ttcl_tu_get_choice(Choices, hard_clustering_container)
	  hard_clustering_container_type;

	/// Hard clustering type
	typedef hard_clustering<hard_clustering_container_type>
	hard_clustering_type;

	/// Constructor
	_hard_matrix_clusterer_interface(bool _owned = true,
					 uint _references = 0) :
	  ttcl::wr::shared_referee(_owned, _references) {
	}

	/// Perform the hard clustering
	virtual void
	hard_cluster_on(hard_clustering_type& _hard_clustering,
			const matrix_wrap& _matrix,
			const uint _n_clusters) const = 0;

	/// Perform the clustering
	virtual hard_clustering_type
	hard_cluster(const matrix_wrap& _matrix,
		     const uint _n_clusters) const = 0;
      };

      /// Hard matrix clusterer referee
      template <typename T,
		typename Interface>
      class _hard_matrix_clusterer_referee :
	public ttcl::wr::typed_shared_referee<T, Interface> {
      private:
	// Matrix type
	typedef typename T::matrix_type matrix_type;

	/// Hard clustering type
	typedef typename Interface::hard_clustering_type hard_clustering_type;

      public:
	/// Constructor
	_hard_matrix_clusterer_referee(T* _ptr = 0, bool _owned = true) :
	  ttcl::wr::typed_shared_referee<T, Interface>(_ptr, _owned) {
	}

	/// Perform the hard clustering
	virtual void
	hard_cluster_on(hard_clustering_type& _hard_clustering,
			const matrix_wrap& _matrix,
			const uint _n_clusters) const {
	  // Convert to matrix type
	  const matrix_type* matrix = _matrix.get<matrix_type>();

	  // Call
	  this->ptr_->hard_cluster_on(_hard_clustering, *matrix, _n_clusters);
	}

	/// Perform the clustering
	virtual hard_clustering_type
	hard_cluster(const matrix_wrap& _matrix,
		     const uint _n_clusters) const {
	  // Convert to matrix type
	  const matrix_type* matrix = _matrix.get<matrix_type>();

	  // Call
	  return this->ptr_->hard_cluster(*matrix, _n_clusters);
	}
      };

      /// Base hard matrix clusterer wrapper
      template <typename Interface>
      class _hard_matrix_clusterer_wrap :
	public ttcl::wr::interface_wrap<Interface> {
      public:
	/// Base type
	typedef ttcl::wr::interface_wrap<Interface> base_type;

	/// Hard clustering type
	typedef typename Interface::hard_clustering_type hard_clustering_type;

      protected:
	/// Empty constructor
	_hard_matrix_clusterer_wrap() :
	  base_type() {
	}

	/// Constructor
	_hard_matrix_clusterer_wrap(Interface* _ptr) :
	  base_type(_ptr) {
	}

      public:
	/// Perform the hard clustering
	void
	hard_cluster_on(hard_clustering_type& _hard_clustering,
			const matrix_wrap& _matrix,
			const uint _n_clusters) const {
	  // Call
	  this->ptr_->hard_cluster_on(_hard_clustering, _matrix, _n_clusters);
	}

	/// Perform the clustering
	hard_clustering_type
	hard_cluster(const matrix_wrap& _matrix,
		     const uint _n_clusters) const {
	  // Call
	  return this->ptr_->hard_cluster(_matrix, _n_clusters);
	}
      };

      /// Hard matrix clusterer wrapper
      template <typename Choices = tu::default_choices>
      class hard_matrix_clusterer_wrap :
	public _hard_matrix_clusterer_wrap
          <_hard_matrix_clusterer_interface<Choices> > {
      private:
	/// Interface
	typedef _hard_matrix_clusterer_interface<Choices> interface;

	/// Base type
	typedef _hard_matrix_clusterer_wrap<interface> base_type;

      public:
	/// Empty Constructor
	hard_matrix_clusterer_wrap() :
	  base_type() {
	}

	/// Constructor
	template <typename T>
	hard_matrix_clusterer_wrap(T* _value, bool _owned = true) :
	  base_type(new _hard_matrix_clusterer_referee<T, interface>(_value,
								     _owned)) {
	}

	/// Constructor
	template <typename T>
	hard_matrix_clusterer_wrap(T& _value, bool _owned = false) :
	  base_type(new _hard_matrix_clusterer_referee<T, interface>(&_value,
								     _owned)) {
	}
      };
    }
  }
};

#endif
