#ifndef TTCL_CO_ARRAY_HXX
#define TTCL_CO_ARRAY_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Fixed-size Array
    @author Edgar Gonzlez i Pellicer
*/

#include <algorithm>
#include <iterator>
#include <memory>

#include <boost/compressed_pair.hpp>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Fixed-size array
    /** (Almost) a model of STL Random Access Container
	@param T Value type
    */
    template <typename T, typename Alloc = std::allocator<T> >
    class array {
    public:
      /// Value type
      typedef T value_type;

      /// Pointer type
      typedef T* pointer;

      /// Const pointer type
      typedef const T* const_pointer;

      /// Reference type
      typedef T& reference;

      /// Const reference
      typedef const T& const_reference;

      /// Iterator type
      typedef T* iterator;

      /// Const iterator type
      typedef const T* const_iterator;

      /// Reverse iterator type
      typedef std::reverse_iterator<iterator> reverse_iterator;

      /// Const reverse iterator type
      typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

      /// Distance type
      typedef ptrdiff_t difference_type;

      /// Size type
      typedef size_t size_type;
    
      /// Allocator type
      typedef Alloc allocator_type;

    protected:
      /// Size
      size_type size_;

      /// Data
      /** With Empty Base Class optimization
       */
      boost::compressed_pair<allocator_type, value_type*> data_;

    public:
      /// Default constructor
      array() :
	size_(0), data_(allocator_type(), 0) {
      }

      /// Constructor
      /** @param _size Size
	  @param _value Value
       */
      array(size_type _size, const value_type& _value = value_type()) :
	size_(_size), data_(allocator_type(), 0) {
	// Reserve data
	data_.second() = data_.first().allocate(size_);

	// Fill
	value_type* p = data_.second();
	for (size_type i = 0; i < size_; ++i, ++p)
	  data_.first().construct(p, _value);
      }

      /// Constructor
      /** @param _size Size
	  @param _data Data
       */
      array(size_type _size, const value_type* _data) :
	size_(_size) {
	// Reserve data
	data_.second() = data_.first().allocate(size_);

	// Copy
	const value_type* s = _data;
	value_type* t = data_.second();
	for (size_type i = 0; i < size_; ++i, ++s, ++t)
	  data_.first().construct(t, *s);
      }

      /// Copy constructor
      /** @param _other Source array
       */
      array(const array& _other) :
	size_(_other.size_), data_(_other.data_.first(), 0) {
	// Reserve data
	data_.second() = data_.first().allocate(size_);

	// Copy
	const value_type* s = _other.data_.second();
	value_type* t = data_.second();
	for (size_type i = 0; i < size_; ++i, ++s, ++t)
	  data_.first().construct(t, *s);
      }

      /// Destructor
      ~array() {
	// Destroy each object
	value_type* p = data_.second();
	for (size_type i = 0; i < size_; ++i, ++p)
	  data_.first().destroy(p);

	// Free data
	data_.first().deallocate(data_.second(), size_);
      }

      /// Assignment operator
      array& operator=(const array& _other) {
	// Check for self-assignment
	if (this != &_other) {
	  // Reserve a new block of data
	  value_type* new_data = data_.first().allocate(_other.size_);

	  // Copy
	  const value_type* s = _other.data_.second();
	  value_type* t = new_data;
	  for (size_type i = 0; i < _other.size_; ++i, ++s, ++t)
	    data_.first().construct(t, *s);

	  // Destroy
	  value_type* p = data_.second();
	  for (size_type i = 0; i < size_; ++i, ++p)
	    data_.first().destroy(p);

	  // Replace
	  data_.first().deallocate(data_.second(), size_);
	  data_.second() = new_data;

	  // Copy rest
	  size_ = _other.size_;
	}

	// Return this
	return *this;
      }

      /// Beginning of range
      /** Non-const version
       */
      iterator begin() {
	// Pointer to the data
	return data_.second();
      }

      /// Beginning of range
      /** Const version
       */
      const_iterator begin() const {
	// Pointer to the data
	return data_.second();
      }

      /// End of range
      /** Non-const version
       */
      iterator end() {
	// Pointer past the data
	return data_.second() + size_;
      }

      /// End of range
      /** Const version
       */
      const_iterator end() const {
	// Pointer past the data
	return data_.second() + size_;
      }

      /// Begin of reverse range
      /** Non-const version
       */
      reverse_iterator rbegin() {
	// Reverse of the end
	return reverse_iterator(end());
      }

      /// Begin of reverse range
      /** Const version
       */
      const_reverse_iterator rbegin() const {
	// Reverse of the end
	return reverse_iterator(end());
      }

      /// End of reverse range
      /** Non-const version
       */
      reverse_iterator rend() {
	// Reverse of the begin
	return reverse_iterator(begin());
      }

      /// End of reverse range
      /** Const version
       */
      const_reverse_iterator rend() const {
	// Reverse of the begin
	return reverse_iterator(begin());
      }
      
      /// Size
      size_type size() const {
	// Size
	return size_;
      }

      /// Maximum size
      size_type max_size() const {
	// Size
	return size_;
      }

      /// Resize
      void resize(size_type _new_size,
		  const value_type& _value = value_type()) {
	// Reserve new data
	value_type* new_data = data_.first().allocate(_new_size);

	// Copy
	const value_type* s = data_.second();
	value_type* t = new_data;

	// Sizes?
	if (_new_size < size_) {
	  // Copy
	  for (size_type i = 0; i < _new_size; ++i, ++s, ++t)
	    data_.first().construct(t, *s);
	}
	else {
	  // Copy
	  for (size_type i = 0; i < size_; ++i, ++s, ++t)
	    data_.first().construct(t, *s);

	  // Fill
	  for (size_type i = size_; i < _new_size; ++i, ++t)
	    data_.first().construct(t, _value);
	}
	
	// Destroy
	value_type* p = data_.second();
	for (size_type i = 0; i < size_; ++i, ++p)
	  data_.first().destroy(p);
	
	// Replace
	data_.first().deallocate(data_.second(), size_);
	data_.second() = new_data;
	
	// Copy rest
	size_ = _new_size;
      }

      /// Empty container
      bool empty() const {
	// Size is zero?
	return size_ == 0;
      }

      /// Clear
      void clear() {
	// Destroy
	value_type* p = data_.second();
	for (size_type i = 0; i < size_; ++i, ++p)
	  data_.first().destroy(p);

	// Free data and empty
	data_.first().deallocate(data_.second(), size_);
	data_.second() = 0;
	size_ = 0;
      }

      /// Element access
      /** Non-const version
       */
      reference operator[](size_type _n) {
	// Access the data
	return data_.second()[_n];
      }

      /// Element access
      /** const version
       */
      const_reference operator[](size_type _n) const {
	// Access the data
	return data_.second()[_n];
      }
    
      /// Front
      /** Non-const version
       */
      reference front() {
	// Access the data
	return data_.second()[0];
      }

      /// Front
      /** const version
       */
      const_reference front() const {
	// Access the data
	return data_.second()[0];
      }

      /// Back
      /** Non-const version
       */
      reference back() {
	// Access the data
	return data_.second()[size_ - 1];
      }

      /// Back
      /** const version
       */
      const_reference back() const {
	// Access the data
	return data_.second()[size_ - 1];
      }

      /// Swap
      /** @param _other Other array
       */
      void swap(array& _other) {
	// Swap sizes and data
	std::swap(size_, _other.size_);
	std::swap(data_, _other.data_);
      }
    };

    /// Equality
    /** @param _a First array
	@param _b Second array
    */
    template <typename T, typename Alloc>
    inline bool operator==(const array<T, Alloc>& _a,
			   const array<T, Alloc>& _b) {
      // Equal in size and content
      return (_a.size() == _b.size() and
	      std::equal(_a.begin(), _a.end(), _b.begin()));
    }

    /// Inequality
    /** Based on operator==
	@param _a First array
	@param _b Second array
    */
    template <typename T, typename Alloc>
    inline bool operator!=(const array<T, Alloc>& _a,
			   const array<T, Alloc>& _b) {
      // Not equal
      return not (_a == _b);
    }

    /// Less
    /** @param _a First array
	@param _b Second array
    */
    template <typename T, typename Alloc>
    inline bool operator<(const array<T, Alloc>& _a,
			  const array<T, Alloc>& _b) {
      // Lexicographical comparison
      return std::lexicographical_compare(_a.begin(), _a.end(),
					  _b.begin(), _b.end());
    }

    /// Greater
    /** Based on operator<
	@param _a First array
	@param _b Second array
    */
    template <typename T, typename Alloc>
    inline bool operator>(const array<T, Alloc>& _a,
			  const array<T, Alloc>& _b) {
      // Inverted less
      return _b < _a;
    }

    /// Less or equal
    /** Based on operator<
	@param _a First array
	@param _b Second array
    */
    template <typename T, typename Alloc>
    inline bool operator<=(const array<T, Alloc>& _a,
			   const array<T, Alloc>& _b) {
      // Not inverted less
      return not (_b < _a);
    }

    /// Greater or equal
    /** Based on operator<
	@param _a First array
	@param _b Second array
    */
    template <typename T, typename Alloc>
    inline bool operator>=(const array<T, Alloc>& _a,
			   const array<T, Alloc>& _b) {
      // Not less
      return not (_a < _b);
    }
  }
};

/// Standard namespace
namespace std {
  /// Swap
  /** Template specialization
      @param _a First array
      @param _b Second array
  */
  template<typename T, typename Alloc>
  void swap(ttcl::co::array<T, Alloc>& _a,
	    ttcl::co::array<T, Alloc>& _b) {
    // Swap
    _a.swap(_b);
  }
}

#endif
