#ifndef TTCL_CO_BIT_SET_HXX
#define TTCL_CO_BIT_SET_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Bit Set
    @author Edgar Gonzalez i Pellicer
*/

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Bit Set
    class bit_set {
    public:
      /// Value type
      typedef bool value_type;

      /// Size type
      typedef size_t size_type;
      
    private:
      /// On bit counts
      static unsigned char bits_on_[256];

      /// First bit
      static unsigned char first_bit_[256];

      /// Size
      size_type data_size_;

      /// Data
      unsigned char* data_;

    public:
      /// Proxy class
      class bit_proxy {
      private:
	/// Position in data
	unsigned char* ptr_;

	/// Bit number
	unsigned char bit_;

	/// Constructor
	bit_proxy(unsigned char* _ptr, unsigned char _bit) :
	  ptr_(_ptr), bit_(_bit) {
	}

      public:
	/// Access
	operator bool() const {
	  return *ptr_ & (1 << bit_);
	}

	/// Assignment
	bit_proxy& operator=(const bool _value) {
	  if (_value)
	    *ptr_ |=   1 << bit_;
	  else
	    *ptr_ &= ~(1 << bit_);
	  return *this;
	}

	// Friendship
	friend class bit_set;
      };

    private:
      /// Constructor
      /** @param _data_size Data size
       */
      bit_set(size_type _data_size, unsigned char* _data) :
	data_size_(_data_size), data_(_data) {
      }

    public:
      /// Constructor
      /** @param _size Size
       */
      bit_set(size_type _size) :
	data_size_((_size + 7) / 8) {
	// Reserve
	data_ = new unsigned char[data_size_];
	std::fill(data_, data_ + data_size_, 0);
      }

      /// Copy constructor
      /** @param _other Source set
       */
      bit_set(const bit_set& _other) :
	data_size_(_other.data_size_) {
	// Reserve and copy
	data_ = new unsigned char[data_size_];
	std::copy(_other.data_, _other.data_ + _other.data_size_,
		  data_);
      }

      /// Destructor
      virtual ~bit_set() {
	// Free
	delete[] data_;
      }

      /// Assignment operator
      /** @param _other Source set
       */
      bit_set& operator=(const bit_set& _other) {
	// Check for self-assignment
	if (this != &_other) {
	  // Reserve a new block of data
	  unsigned char* new_data = new unsigned char[_other.data_size_];
	  std::copy(_other.data_, _other.data_ + _other.data_size_,
		    data_);

	  // Replace
	  delete[] data_;
	  data_ = new_data;

	  // Copy rest
	  data_size_ = _other.data_size_;
	}

	// Return this
	return *this;
      }

      /// Empty
      bool empty() const {
	for (size_type i = 0; i < data_size_; ++i)
	  if (data_[i])
	    return false;
	return true;
      }

      /// Size
      size_type size() const {
	size_type total = size_type();
	for (size_type i = 0; i < data_size_; ++i)
	  total += bits_on_[data_[i]];
	return total;
      }

      /// Maximum Size
      size_type max_size() const {
	return 8 * data_size_;
      }

      /// Clear
      void clear() {
	std::fill(data_, data_ + data_size_, 0);
      }

      /// Front
      /** @return max_size() if empty
       */
      size_type front() const {
	size_type first = 0;
	for (size_type s = 0; s < data_size_; ++s, first += 8)
	  if (data_[s])
	    return first + first_bit_[data_[s]];
	return first;
      }

      /// Get the value
      bool operator[](size_type _elem) const {
	return data_[_elem / 8] & (1 << (7 - _elem % 8));
      }

      /// Get the value
      /** Through a proxy
       */
      bit_proxy operator[](size_type _elem) {
	return bit_proxy(data_ + _elem / 8, 7 - _elem % 8);
      }

      /// Equality operator
      bool operator==(const bit_set& _other) const {
	// Little, big
	if (data_size_ < _other.data_size_) {
	  // Look for differences
	  size_type s = 0;
	  for (; s < data_size_; ++s)
	    if (data_[s] != _other.data_[s])
	      return false;
	  for (; s < _other.data_size_; ++s)
	    if (_other.data_[s])
	      return false;
	}
	else {
	  // Look for differences
	  size_type s = 0;
	  for (; s < _other.data_size_; ++s)
	    if (data_[s] != _other.data_[s])
	      return false;
	  for (; s < data_size_; ++s)
	    if (data_[s])
	      return false;
	}

	// Equal
	return true;
      }

      /// Inequality operator
      bool operator!=(const bit_set& _other) const {
	return not (*this == _other);
      }

      /// Less than operator
      bool operator<(const bit_set& _other) const {
	// Little, big
	if (data_size_ < _other.data_size_) {
	  // Look for differences
	  size_type s = 0;
	  for (; s < data_size_; ++s)
	    if (data_[s] > _other.data_[s])
	      return true;
	    else if (data_[s] < _other.data_[s])
	      return false;
	  for (; s < _other.data_size_; ++s)
	    if (_other.data_[s])
	      return true;
	}
	else {
	  // Look for differences
	  size_type s = 0;
	  for (; s < _other.data_size_; ++s)
	    if (data_[s] > _other.data_[s])
	      return true;
	    else if (data_[s] < _other.data_[s])
	      return false;
	  for (; s < data_size_; ++s)
	    if (data_[s])
	      return false;
	}

	// Equal
	return false;
      }
      
      /// Less than or equal operator
      bool operator<=(const bit_set& _other) const {
	return not (_other < *this);
      }
      
      /// Greater than operator
      bool operator>(const bit_set& _other) const {
	return (_other < *this);
      }

      /// Greater than or equal operator
      bool operator>=(const bit_set& _other) const {
	return not (*this < _other);
      }

      // Friend functions
      friend bit_set set_union       (const bit_set& _a, const bit_set& _b);
      friend bit_set set_intersection(const bit_set& _a, const bit_set& _b);
      friend bit_set set_difference  (const bit_set& _a, const bit_set& _b);
      friend bit_set set_symmetric_difference(const bit_set& _a,
					      const bit_set& _b);
    };

    /// Set union
    bit_set set_union(const bit_set& _a, const bit_set& _b);

    /// Set intersection
    bit_set set_intersection(const bit_set& _a, const bit_set& _b);

    /// Set difference
    bit_set set_difference(const bit_set& _a, const bit_set& _b);

    /// Set symmetric difference
    bit_set set_symmetric_difference(const bit_set& _a, const bit_set& _b);
  }
}

// Static vars
#define ttcl_co_bit_set_static						\
  /* On bit counts */							\
  unsigned char ttcl::co::bit_set::bits_on_[256] =			\
    { 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,			\
      1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,			\
      1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,			\
      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,			\
      1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,			\
      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,			\
      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,			\
      3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,			\
      1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,			\
      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,			\
      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,			\
      3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,			\
      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,			\
      3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,			\
      3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,			\
      4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8 };			\
									\
  /* First bit */							\
  unsigned char ttcl::co::bit_set::first_bit_[256] =			\
    { 0, 7, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4,			\
      3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,			\
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,			\
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,			\
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,			\
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,			\
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,			\
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,			\
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			\
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			\
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			\
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			\
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			\
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			\
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			\
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };			\
									\
  /* Set union */							\
  ttcl::co::bit_set							\
  ttcl::co::set_union(const ttcl::co::bit_set& _a,			\
		      const ttcl::co::bit_set& _b) {			\
    /* Size and data */							\
    bit_set::size_type data_size;					\
    unsigned char* new_data;						\
									\
    /* Little, big */							\
    if (_a.data_size_ < _b.data_size_) {				\
      /* Make the data */						\
      data_size = _b.data_size_;					\
      new_data  = new unsigned char[data_size];				\
									\
      /* Fill it */							\
      bit_set::size_type i = 0;						\
      for (; i < _a.data_size_; ++i)					\
	new_data[i] = _a.data_[i] | _b.data_[i];			\
      for (; i < _b.data_size_; ++i)					\
	new_data[i] = _b.data_[i];					\
    }									\
    else {								\
      /* Make the data */						\
      data_size = _a.data_size_;					\
      new_data  = new unsigned char[data_size];				\
									\
      /* Fill it */							\
      bit_set::size_type i = 0;						\
      for (; i < _b.data_size_; ++i)					\
	new_data[i] = _a.data_[i] | _b.data_[i];			\
      for (; i < _a.data_size_; ++i)					\
	new_data[i] = _a.data_[i];					\
    }									\
									\
    /* Return */							\
    return bit_set(data_size, new_data);				\
  }									\
									\
  /* Set intersection */						\
  ttcl::co::bit_set							\
  ttcl::co::set_intersection(const ttcl::co::bit_set& _a,		\
			     const ttcl::co::bit_set& _b) {		\
    /* Size and data */							\
    bit_set::size_type data_size;					\
									\
    /* Little, big */							\
    if (_a.data_size_ < _b.data_size_)					\
      data_size = _a.data_size_;					\
    else								\
      data_size = _b.data_size_;					\
									\
    /* Make the data */							\
    unsigned char* new_data = new unsigned char[data_size];		\
									\
    /* Fill it */							\
    for (bit_set::size_type i = 0; i < data_size; ++i)			\
      new_data[i] = _a.data_[i] & _b.data_[i];				\
									\
    /* Return */							\
    return bit_set(data_size, new_data);				\
  }									\
									\
  /* Set difference */							\
  ttcl::co::bit_set							\
  ttcl::co::set_difference(const ttcl::co::bit_set& _a,			\
			   const ttcl::co::bit_set& _b) {		\
    /* Make the data */							\
    unsigned char* new_data = new unsigned char[_a.data_size_];		\
									\
    /* Little, big */							\
    if (_a.data_size_ < _b.data_size_) {				\
      /* Fill it */							\
      for (bit_set::size_type i = 0; i < _a.data_size_; ++i)		\
	new_data[i] = _a.data_[i] & ~_b.data_[i];			\
    }									\
    else {								\
      /* Fill it */							\
      bit_set::size_type i = 0;						\
      for (; i < _b.data_size_; ++i)					\
	new_data[i] = _a.data_[i] & ~_b.data_[i];			\
      for (; i < _a.data_size_; ++i)					\
	new_data[i] = _a.data_[i];					\
    }									\
									\
    /* Return */							\
    return bit_set(_a.data_size_, new_data);				\
  }									\
									\
  /* Set symmetric difference */					\
  ttcl::co::bit_set							\
  ttcl::co::set_symmetric_difference(const ttcl::co::bit_set& _a,	\
				     const ttcl::co::bit_set& _b) {	\
    /* Size and data; */						\
    bit_set::size_type data_size;					\
    unsigned char* new_data;						\
									\
    /* Little, big */							\
    if (_a.data_size_ < _b.data_size_) {				\
      /* Make the data */						\
      data_size = _b.data_size_;					\
      new_data  = new unsigned char[data_size];				\
									\
      /* Fill it */							\
      bit_set::size_type i = 0;						\
      for (; i < _a.data_size_; ++i)					\
	new_data[i] = _a.data_[i] ^ _b.data_[i];			\
      for (; i < _b.data_size_; ++i)					\
	new_data[i] = _b.data_[i];					\
    }									\
    else {								\
      /* Make the data */						\
      data_size = _a.data_size_;					\
      new_data  = new unsigned char[data_size];				\
									\
      /* Fill it */							\
      bit_set::size_type i = 0;						\
      for (; i < _b.data_size_; ++i)					\
	new_data[i] = _a.data_[i] ^ _b.data_[i];			\
      for (; i < _a.data_size_; ++i)					\
	new_data[i] = _a.data_[i];					\
    }									\
									\
    /* Return */							\
    return bit_set(data_size, new_data);				\
  }

#endif
