#ifndef TTCL_CO_BOOST_CUBE_HXX
#define TTCL_CO_BOOST_CUBE_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - boost::multi_array adapter for a cube
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>
#include <memory>

#include <boost/multi_array.hpp>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// boost::multi_array adapter for a cube
    template <typename T, typename Alloc = std::allocator<T> >
    class boost_cube :
      public boost::multi_array<T, 3, Alloc> {
    public:
      /// Base type
      typedef boost::multi_array<T, 3, Alloc> base_type;

      /// Element type
      typedef typename base_type::element element;

      /// Size type
      typedef typename base_type::size_type size_type;

      /// Element reference
      typedef element& element_reference;

      /// Const element reference
      typedef const element& const_element_reference;

      /// Default Constructor
      boost_cube() :
	base_type() {
      }

      /// Constructor
      /** @param _rows   Rows
	  @param _cols   Cols
	  @param _floors Floors
      */
      boost_cube(const size_type _rows,
		 const size_type _cols,
		 const size_type _floors) :
	base_type(boost::extents[_rows][_cols][_floors]) {
      }

      /// Resize
      /** @param _rows  Rows
	  @param _cols  Cols
      */
      void resize(const size_type _rows,
		  const size_type _cols,
		  const size_type _floors) {
	// Resize
	base_type::resize(boost::extents[_rows][_cols][_floors]);
      }

      /// Rows
      size_type rows() const {
	// Synonym for size
	return this->size();
      }

      /// Columns
      size_type columns() const {
	// Second dimension of shape
	return this->shape()[1];
      }

      /// Floors
      size_type floors() const {
	// Third dimension of shape
	return this->shape()[2];
      }

      /// Element access
      /** Non-const version
	  @param _row Row
	  @param _col Column
	  @param _flo Floor
      */  
      element_reference operator()(size_type _row, size_type _col,
				   size_type _flo) {
	return (*this)[_row][_col][_flo];
      }

      /// Element access
      /** const version
	  @param _row Row
	  @param _col Column
	  @param _flo Floor
      */
      const_element_reference operator()(size_type _row, size_type _col,
					 size_type _flo) const {
	return (*this)[_row][_col][_flo];
      }

      /// Clear
      void clear(const element& _value = element()) {
	std::fill(this->data(),
		  this->data() + rows() * columns() * floors(),
		  _value);
      }
    };
  }
}

#endif
