#ifndef TTCL_CO_BOOST_MATRIX_HXX
#define TTCL_CO_BOOST_MATRIX_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - boost::multi_array adapter for a matrix
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>
#include <memory>

#include <boost/multi_array.hpp>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// boost::multi_array adapter for a matrix
    template <typename T, typename Alloc = std::allocator<T> >
    class boost_matrix :
      public boost::multi_array<T, 2, Alloc> {
    public:
      /// Base type
      typedef boost::multi_array<T, 2, Alloc> base_type;

      /// Element type
      typedef typename base_type::element element;

      /// Size type
      typedef typename base_type::size_type size_type;

      /// Element reference
      typedef element& element_reference;

      /// Const element reference
      typedef const element& const_element_reference;

      /// Default Constructor
      boost_matrix() :
	base_type() {
      }

      /// Constructor
      /** @param _rows Rows
	  @param _cols Cols
      */
      boost_matrix(const size_type _rows,
		   const size_type _cols) :
	base_type(boost::extents[_rows][_cols]) {
      }

      /// Constructor
      /** @param _rows Rows
	  @param _cols Cols
	  @param _value Value
      */
      boost_matrix(const size_type _rows,
		   const size_type _cols,
		   const element& _value) :
	base_type(boost::extents[_rows][_cols]) {
	clear(_value);
      }

      /// Resize
      /** @param _rows  Rows
	  @param _cols  Cols
      */
      void resize(const size_type _rows,
		  const size_type _cols) {
	// Resize
	base_type::resize(boost::extents[_rows][_cols]);
      }

      /// Rows
      size_type rows() const {
	// Synonym for size
	return this->size();
      }

      /// Columns
      size_type columns() const {
	// Last dimension of shape
	return this->shape()[1];
      }

      /// Element access
      /** Non-const version
	  @param _row Row
	  @param _col Column
      */  
      element_reference operator()(size_type _row, size_type _col) {
	return (*this)[_row][_col];
      }

      /// Element access
      /** const version
	  @param _row Row
	  @param _col Column
      */
      const_element_reference operator()(size_type _row, size_type _col) const {
	return (*this)[_row][_col];
      }

      /// Clear
      void clear(const element& _value = element()) {
	std::fill(this->data(),
		  this->data() + rows() * columns(),
		  _value);
      }
    };
  }
}

#endif
