#ifndef TTCL_CO_CENTROID_SET_HXX
#define TTCL_CO_CENTROID_SET_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Centroid set
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>
#include <memory>

#include <boost/mpl/not.hpp>
#include <boost/multi_array.hpp>
#include <boost/utility/enable_if.hpp>

#include <ttcl/types.hxx>
#include <ttcl/co/traits.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Centroid set
    template <typename T, typename Alloc = std::allocator<T> >
    class centroid_set :
      public boost::multi_array<T, 2, Alloc> {
    public:
      /// Base type
      typedef boost::multi_array<T, 2, Alloc> base_type;

      /// Value type
      typedef typename base_type::element value_type;

      /// Centroid type
      typedef typename base_type::reference centroid_type;

      /// Centroid type
      typedef typename base_type::const_reference const_centroid_type;

    public:
      /// Empty Constructor
      centroid_set() :
	base_type() {
      }

      /// Constructor
      centroid_set(const uint _clusters, const uint _features) :
	base_type(boost::extents[_clusters][_features]) {
      }

      /// Copy constructor
      centroid_set(const centroid_set& _other) :
	base_type(_other) {
      }

      /// Resize
      void resize(uint _clusters, uint _features) {
	base_type::resize(boost::extents[_clusters][_features]);
      }

      /// Reset all centroids
      void reset() {
	std::fill(this->data(), this->data() + this->num_elements(),
		  value_type());
      }

      /// Reset centroid
      void reset(uint _cluster) {
	// Get the centroid
	centroid_type c = (*this)[_cluster];
	std::fill(c.begin(), c.end(), value_type());
      }

      /// Update centroid
      /** Specialization for non-sparse samples
       */
      template <typename Sample>
      typename boost::enable_if
        < boost::mpl::not_<co::is_sparse_sample<Sample> > >::type
      update(const uint _cluster,
	     const Sample& _sample,
	     const value_type _expectation) {
	// Just add to each feature
	for (uint f = 0; f < _sample.size(); ++f)
	  (*this)[_cluster][f] += _expectation * value_type(_sample[f]);
      }

      /// Update centroid
      /** Specialization for non-sparse samples and expectation = 1.0
       */
      template <typename Sample>
      typename boost::enable_if
        < boost::mpl::not_<co::is_sparse_sample<Sample> > >::type
      update(const uint _cluster,
	     const Sample& _sample) {
	// Just add to each feature
	for (uint f = 0; f < _sample.size(); ++f)
	  (*this)[_cluster][f] += value_type(_sample[f]);
      }

      /// Update centroid
      /** Specialization for sparse samples
       */
      template <typename Sample>
      typename boost::enable_if< co::is_sparse_sample<Sample> >::type
      update(const uint _cluster,
	     const Sample& _sample,
	     const value_type _expectation) {
	// Add to present features
	for (typename Sample::const_iterator it = _sample.begin();
	     it != _sample.end(); ++it)
	  (*this)[_cluster][it->first] += _expectation * value_type(it->second);
      }

      /// Update centroid
      /** Specialization for sparse samples and expectation = 1.0
       */
      template <typename Sample>
      typename boost::enable_if< co::is_sparse_sample<Sample> >::type
      update(const uint _cluster,
	     const Sample& _sample) {
	// Add to present features
	for (typename Sample::const_iterator it = _sample.begin();
	     it != _sample.end(); ++it)
	  (*this)[_cluster][it->first] += value_type(it->second);
      }

      /// Normalize centroid
      void normalize(const uint _cluster,
		     const value_type _counts) {
	// Get the centroid
	centroid_type c = (*this)[_cluster];
	for (typename centroid_type::iterator it = c.begin();
	     it != c.end(); ++it)
	  *it /= _counts;
      }
    };
  }
}

#endif
