#ifndef TTCL_DISK_VECTOR_HXX
#define TTCL_DISK_VECTOR_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Vector on disk
    @author Edgar Gonzlez i Pellicer
*/

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>

#include <algorithm>
#include <iterator>

#include <ttcl/exception.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Base class of vectors on disk
    class _disk_vector {
    public:
      /// Block size
      static const size_t block_size = 65536;
      
      /// Default cache size
      static const size_t default_cache_size = 65536;
    };

    /// Vector on disk
    /** A model of STL Random Access Container
	@param T Value type. Must be a simple type.
    */
    template <typename T>
    class disk_vector :
      public _disk_vector {
    public:
      /// Value type
      typedef T value_type;

      /// Pointer type
      typedef T* pointer;

      /// Const pointer type
      typedef const T* const_pointer;

      /// Reference type
      typedef T& reference;

      /// Const reference
      typedef const T& const_reference;

      /// Iterator type
      typedef T* iterator;

      /// Const iterator type
      typedef const T* const_iterator;

      /// Reverse iterator type
      typedef std::reverse_iterator<iterator> reverse_iterator;

      /// Const reverse iterator type
      typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

      /// Distance type
      typedef ptrdiff_t difference_type;

      /// Size type
      typedef size_t size_type;

    private:
      /// Elements per block
      static const size_t elements_block =
	_disk_vector::block_size / sizeof(T);

      /// Effective block size
      static const size_t effective_size =
	elements_block * sizeof(T);

      /// Temporary file name
      char tmp_name_[20];
      
      /// File descriptor
      int fd_;

      /// Size
      size_type size_;

      /// Cache size
      size_type cache_size_;
      
      /// Cache data
      mutable value_type** cache_data_;

      /// Cache offsets
      mutable size_type* cache_offsets_;

      /// Last block
      size_type last_block_;

    public:
      /// Constructor
      disk_vector(const size_type _size = 0,
		  const value_type _value = value_type(),
		  const size_type _cache_size =
		  _disk_vector::default_cache_size) 
	throw (exception) :
	fd_(0), size_(_size), cache_size_(_cache_size),
	cache_data_(0), cache_offsets_(0), last_block_(0) {
	// Open a temporary file
	std::strcpy(tmp_name_, "/tmp/vector.XXXXXX");
	if ((fd_ = mkstemp(tmp_name_)) == -1) {
	  // Exception
	  ttcl_fire("Cannot obtain temporary file");
	}

	// Reserve memory for the cache
	cache_data_    = new value_type*[cache_size_];
	cache_offsets_ = new size_type  [cache_size_];

	// Set them to zero
	std::fill(cache_data_, cache_data_ + cache_size_, (value_type*)(0));

	// Fill the blocks
	size_type n_blocks = size_ / elements_block;
	for (size_type i = 0; i < n_blocks; ++i) {
	  value_type* start = gimme_block(i);
	  std::fill(start, start + elements_block, _value);
	}

	// Any left?
	size_type n_remain = size_ % elements_block;
	if (n_remain > 0) {
	  // Fill the last block
	  value_type* start = gimme_block(n_blocks);
	  std::fill(start, start + n_remain, _value);
	  last_block_ = n_blocks;
	}
	else {
	  // Just remember
	  last_block_ = n_blocks - 1;
	}
      }

      /// Copy constructor
      disk_vector(const disk_vector& _other) {
      }

      /// Destructor
      ~disk_vector() {
	// Release the used cache_data_ entries
	for (size_type i = 0; i < cache_size_; ++i)
	  if (cache_data_[i])
	    ::operator delete(cache_data_[i]);

	// Release the arrays
	delete[] cache_data_;
	delete[] cache_offsets_;

	// Delete the file
	close(fd_);
	unlink(tmp_name_);
      }

      /// Assignment operator
      disk_vector& operator=(const disk_vector& _other) {
	// Return this
	return *this;
      }

      /// Beginning of range
      /** Non-const version
       */
      iterator begin() {
	return 0;
      }

      /// Beginning of range
      /** Const version
       */
      const_iterator begin() const {
	return 0;
      }

      /// End of range
      /** Non-const version
       */
      iterator end() {
	return 0;
      }

      /// End of range
      /** Const version
       */
      const_iterator end() const {
	return 0;
      }

      /// Begin of reverse range
      /** Non-const version
       */
      reverse_iterator rbegin() {
	// Reverse of the end
	return reverse_iterator(end());
      }

      /// Begin of reverse range
      /** Const version
       */
      const_reverse_iterator rbegin() const {
	// Reverse of the end
	return reverse_iterator(end());
      }

      /// End of reverse range
      /** Non-const version
       */
      reverse_iterator rend() {
	// Reverse of the begin
	return reverse_iterator(begin());
      }

      /// End of reverse range
      /** Const version
       */
      const_reverse_iterator rend() const {
	// Reverse of the begin
	return reverse_iterator(begin());
      }
      
      /// Size
      size_type size() const {
	// Size
	return size_;
      }

      /// Maximum size
      size_type max_size() const {
	// ?
	return 0;
      }

      /// Resize
      void resize(size_type _new_size,
		  value_type _value = value_type()) {
      }

      /// Empty container
      bool empty() const {
	// Size is zero?
	return size_ == 0;
      }

      /// Element access
      /** Non-const version
       */
      reference operator[](size_type _n) {
	// Find the block and element
	size_type idx = _n / elements_block;
	size_type ele = _n % elements_block;
	return *gimme_element(idx, ele);
      }

      /// Element access
      /** const version
       */
      const_reference operator[](size_type _n) const {
	// Find the block and element
	size_type idx = _n / elements_block;
	size_type ele = _n % elements_block;
	return *gimme_element(idx, ele);
      }

      /// Front
      /** Non-const version
       */
      reference front() {
	return *gimme_element(0, 0);
      }

      /// Front
      /** Const version
       */
      const_reference front() const {
	return *gimme_element(0, 0);
      }

      /// Back
      /** Non-const version
       */
      reference back() {
	return (*this)[size_ - 1];
      }

      /// Back
      /** Const version
       */
      const_reference back() const {
	return (*this)[size_ - 1];
      }

      /// Push back
      void push_back(const value_type& _value) {
      }

      /// Pop back
      void pop_back() {
      }

      /// Swap
      void swap(const disk_vector& _other) {
      }

      /// Clear
      void clear() {
      }

    private:
      /// Find an element
      /** It brings the block to memory if needed
	  Non-const version
       */
      value_type*
      gimme_element(const size_type _block, const size_type _element) {
	// Find the block and add the element
	return gimme_block(_block) + _element;
      }

      /// Find an element
      /** It brings the block to memory if needed
	  Non-const version
       */
      const value_type*
      gimme_element(const size_type _block, const size_type _element) const {
	// Find the block and add the element
	return gimme_block(_block) + _element;
      }

      /// Find a block
      /** It brings the block to memory if needed
       */
      value_type*
      gimme_block(const size_type _block) const
	throw (exception) {
	// Find the position in the cache
	size_type idx = _block % cache_size_;
	
	// Is it used?
	if (not cache_data_[idx]) {
	  // Create it
	  cache_offsets_[idx] = _block;
	  cache_data_   [idx] = 
	    static_cast<value_type*>(::operator new(effective_size));

	  // Load it
	  load_block(idx);

	}
	// Is it the right one?
	else if (cache_offsets_[idx] != _block) {
	  // Save old block if needed
	  if (cache_offsets_[idx] <= last_block_)
	    save_block(idx);
	  
	  // Load new block
	  cache_offsets_[idx] = _block;
	  load_block(idx);
	}

	// Return
	return cache_data_[idx];
      }

      /// Save a block
      void save_block(const size_type _idx) const
	throw (exception) {
	// Position
	if (lseek(fd_, cache_offsets_[_idx] * effective_size,
		  SEEK_SET) == off_t(-1))
	  ttcl_fire("Cannot position file offset for saving");

	// Write
	ssize_t to_write = effective_size;
	char*   data     = reinterpret_cast<char*>(cache_data_[_idx]);
	while (to_write) {
	  ssize_t written = write(fd_, data, to_write);
	  if (written == -1)
	    ttcl_fire("Cannot save data");

	  to_write -= written;
	  data     += written;
	}
      }

      /// Load a block
      void load_block(const size_type _idx) const
	throw (exception) {
	// Position
	if (lseek(fd_, cache_offsets_[_idx] * effective_size,
		  SEEK_SET) == off_t(-1))
	  ttcl_fire("Cannot position file offset for loading");

	// Read
	ssize_t to_read = effective_size;
	char*   data    = reinterpret_cast<char*>(cache_data_[_idx]);
	while (to_read) {
	  ssize_t readen = read(fd_, data, to_read);
	  if (readen == -1) {
	    // Error
	    ttcl_fire("Cannot load data");
	  }
	  else if (readen == 0) {
	    // End-of-file
	    to_read  = 0;
	  }
	  else {
	    // OK
	    to_read -= readen;
	    data    += readen;
	  }
	}
      }
    };
  }
}

/// Standard namespace
namespace std {
  /// Swap
  /** Template specialization
      @param _a First vector
      @param _b Second vector
  */
  template<typename T>
  void swap(ttcl::co::disk_vector<T>& _a, ttcl::co::disk_vector<T>& _b) {
    // Swap
    _a.swap(_b);
  }
}

#endif
