#ifndef TTCL_CO_EXTENDED_PRIORITY_QUEUE_HXX
#define TTCL_CO_EXTENDED_PRIORITY_QUEUE_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Extendeded priority queue
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>
#include <functional>
#include <vector>

#include <ttcl/exception.hxx>
#include <ttcl/types.hxx>
#include <ttcl/co/traits.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co { 

    /// Extendeded priority queue
    /** @param T Value type
	@param Sequence Sequence
	@param Compare Comparison function
     */
    template <typename T,
	      typename Sequence = std::vector<T>,
	      typename Compare = std::less<T>,
	      typename ISequence = std::vector<
		typename signed_type<typename Sequence::size_type>::type> >
    class extended_priority_queue {
    public:
      // Value type
      typedef typename Sequence::value_type value_type;

      // Size type
      typedef typename Sequence::size_type size_type;

    protected:
      /// Sequence type
      typedef Sequence sequence_type;

      /// Comparison function type
      typedef Compare compare_type;

      /// Index sequence type
      typedef ISequence index_sequence_type;

      /// Index type
      typedef typename index_sequence_type::value_type index_type;

      /// Comparison function
      compare_type comp_;

      /// Data
      sequence_type data_;

      /// Forward index
      index_sequence_type forward_;

      /// Backwards index
      index_sequence_type backwards_;

      /// Proxy
      class proxy {
      private:
	/// Reference
	extended_priority_queue& queue_;

	/// Element
	index_type element_;

	/// Constructor
	proxy(extended_priority_queue& _queue, index_type _element) :
	  queue_(_queue), element_(_element) {
	}

      public:
	/// Copy constructor
	proxy(const proxy& _other) :
	  queue_(_other.queue_), element_(_other.element_) {
	}

	/// Access
	operator const value_type&() const {
	  // Get
	  return queue_.value(element_);
	}

	/// Update
	proxy& operator=(const value_type& _value) {
	  // Set and return this
	  queue_.set_value(element_, _value);
	  return *this;
	}

	// Friend
	friend class extended_priority_queue;
      };

    public:
      /// Constructor
      /** @param _comp  Comparison function
      */
      explicit extended_priority_queue(const Compare& _comp = Compare()) :
	comp_(_comp), data_(), forward_(), backwards_() {
      }

      /// Copy constructor
      /** @param _other Source queue
       */
      extended_priority_queue(const extended_priority_queue& _other) :
	comp_(_other.comp_), data_(_other.data_),
	forward_(_other.forward_), backwards_(_other.backwards_) {
      }

      /// Destructor
      ~extended_priority_queue() {
      }

      /// Assignment operator
      /** @param _other Source queue
       */
      extended_priority_queue& operator=(const extended_priority_queue&
					   _other) {
	// Check for self-assignment
	if (this != &_other) {
	  // Copy field by field
	  comp_      = _other.comp_;
	  data_      = _other.data_;
	  forward_   = _other.forward_;
	  backwards_ = _other.backwards_;
	}

	// Return this
	return *this;
      }

      /// Is it empty?
      bool empty() const {
	// The container is empty
	return data_.empty();
      }

      /// Size
      size_type size() const {
	// The container size
	return data_.size();
      }

      /// Get the top value
      const value_type& top() const {
	// Return the top
        return data_[0];
      }

      /// Get the top element
      size_type top_element() const {
	// Return the top element
	return backwards_[0];
      }

      /// Pop the top
      void pop() {
	// Just one left?
	if (size() == 1) {
	  // Clear all
	  clear();
	}
	else {
	  // Move the last element to the root
	  swap_two(0, data_.size() - 1);

	  // Delete last
	  data_.pop_back();
	  forward_[backwards_.back()] = -1;
	  backwards_.pop_back();
	  
	  // Sink the root
	  do_sink(0);
	}
      }

      /// Insert an element
      /** @param  _value Value
	  @return Element index
       */
      size_type push(const value_type& _value) {
	// Add it to the end
	data_     .push_back(_value);
	forward_  .push_back(data_.size() - 1);
	backwards_.push_back(forward_.size() - 1);

	// Float
	do_float(data_.size() - 1);

	// Return the index
	return forward_.size() - 1;
      }

      /// Clear the queue
      void clear() {
	// Clear all
	data_     .clear();
	forward_  .clear();
	backwards_.clear();
      }

      /// Exists
      bool exists(const size_type _element) const {
	// OK
	return _element < forward_.size() and forward_[_element] != -1;
      }

      /// Value
      const value_type& value(const size_type _element) const {
	// Get the value
	return data_[forward_[_element]];
      }
	
      /// Set a value
      void set_value(const size_type _element, const value_type& _value) {
	// Set the value
	index_type index = forward_[_element];
	data_[index] = _value;

	// Float or sink?
	if (index > 0 and comp_(data_[(index - 1) / 2], data_[index]))
	  // Float
	  do_float(index);
	else
	  // Sink
	  do_sink(index);
      }

      /// Erase an element
      void erase(const size_type _element) {
	// Swap with the last
	int index = forward_[_element]; 
	swap_two(index, data_.size() - 1);

	// Delete last
	data_.pop_back();
	forward_[backwards_.back()] = -1;
	backwards_.pop_back();
	  
	// Float or sink?
	if (index > 0 and comp_(data_[(index - 1) / 2], data_[index]))
	  // Float
	  do_float(index);
	else
	  // Sink
	  do_sink(index);
      }

      /// Const Access
      const value_type& operator[](const size_type _element) const {
	// Get the value
	return value(_element);
      }

      /// Access
      proxy operator[](const size_type _element) {
	// Return a proxy
	return proxy(*this, _element);
      }

    protected:
      /// Swap two elements
      void swap_two(const size_type _pos1, const size_type _pos2) {
	// Swap data and backwards
	std::swap(data_     [_pos1], data_     [_pos2]);
	std::swap(backwards_[_pos1], backwards_[_pos2]);

	// Update the forward
	forward_[backwards_[_pos1]] = _pos1;
	forward_[backwards_[_pos2]] = _pos2;
      }

      // Do float
      void do_float(size_type _pos) {
	// Float
	size_type parent;
	while (_pos > 0) {
	  parent = (_pos - 1) >> 1;

	  if (comp_(data_[parent], data_[_pos])) {
	    // Float
	    swap_two(parent, _pos);
	    _pos = parent;
	  }
	  else {
	    // No more floatin'
	    break;
	  }
	}
      }

      // Do sink
      void do_sink(size_type _pos) {
	// Sink it
	size_type lpos = (_pos << 1) + 1;
	size_type rpos = lpos + 1;
	while (lpos < data_.size()) {
	  if (rpos < data_.size()) {
	    // Two children
	    if (comp_(data_[rpos], data_[lpos])) {
	      if (comp_(data_[_pos], data_[lpos])) {
		// Swap
		swap_two(_pos, lpos);
		_pos = lpos;
	      }
	      else {
		// No more sinkin'
		break;
	      }
	    }
	    else {
	      if (comp_(data_[_pos], data_[rpos])) {
		// Swap
		swap_two(_pos, rpos);
		_pos = rpos;
	      }
	      else {
		// No more sinkin'
		break;
	      }
	    }
	  }
	  else {
	    // A single child
	    if (comp_(data_[_pos], data_[lpos])) {
	      // Swap
	      swap_two(_pos, lpos);
	    }
	      
	    // No more sinkin'
	    break;
	  }
	  
	  // Update lpos and rpos
	  lpos = (_pos << 1) + 1;
	  rpos = lpos + 1;
	}
      }
    };
  }
}

#endif
