#ifndef TTCL_CO_FEATURE_CONTAINER_HXX
#define TTCL_CO_FEATURE_CONTAINER_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Feature mixin for containers
    @author Edgar Gonzlez i Pellicer
*/

#include <boost/utility/enable_if.hpp>

#include <ttcl/types.hxx>
#include <ttcl/co/traits.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Feature mixin base class for containers
    template <typename Base, typename Child>
    class _feature_container :
      public Base {
    protected:
      /// Cached value
      mutable uint features_;

      /// Is cached value OK?
      mutable bool cached_ok_;

    public:
      /// Base type
      typedef Base base_type;

      /// Size type
      typedef typename base_type::size_type size_type;

      /// Constructor
      _feature_container() :
	features_(0), cached_ok_(false) {
      }

      /// Constructor
      explicit _feature_container(size_type _n) :
	base_type(_n), features_(0), cached_ok_(false) {
      }

      /// Features
      uint features() const {
	// Find cached value
	static_cast<const Child*>(this)->find_cached_value();

	// Return cached value
	return features_;
      }

      /// Manually set features
      void set_features(uint _features) {
	// Set
	features_  = _features;
	cached_ok_ = true;
      }

      /// Invalidate cached value
      void invalidate_cache() const {
	// Invalidate
	cached_ok_ = false;
      }
    };

    /// Feature mixin for containers
    /** Uses the Curiously Recurring Template Pattern
     */
    template <typename Base, typename Enable = void>
    class feature_container :
      public _feature_container<Base, feature_container<Base, Enable> > {
    private:
      /// Real base type
      typedef _feature_container<Base, feature_container> _base_type;

    public:
      /// Base type
      typedef typename _base_type::base_type base_type;

      /// Size type
      typedef typename base_type::size_type size_type;

      /// Const iterator type
      typedef typename base_type::const_iterator const_iterator;
      
      /// Constructor
      feature_container() :
	_base_type() {
      }

      /// Constructor
      feature_container(size_type _n) :
	_base_type(_n) {
      }

    protected:
      // Find cached value
      void find_cached_value() const {
	// Valid?
	if (not this->cached_ok_) {
	  // Follow the samples
	  this->features_ = 0;
	  for (const_iterator dit = this->begin(); dit != this->end(); ++dit)
	    // Larger than the rest?
	    if (dit->size() > this->features_)
	      this->features_ = dit->size();

	  // It is OK now
	  this->cached_ok_ = true;
	}
      }

      // Friend class
      friend class  _feature_container<Base, feature_container>;
    };

    /// Feature mixin for containers
    /** Specialization for sparse containers
     */
    template <typename Base>
    class feature_container
      <Base, typename boost::enable_if<is_sparse_container<Base> >::type> :
      public _feature_container<Base, feature_container
	<Base, typename boost::enable_if<is_sparse_container<Base> >::type> > {
    private:
      /// Real base type
      typedef _feature_container<Base, feature_container> _base_type;

    public:
      /// Base type
      typedef typename _base_type::base_type base_type;

      /// Size type
      typedef typename base_type::size_type size_type;

      /// Const iterator type
      typedef typename base_type::const_iterator const_iterator;

      /// Feature const iterator type
      typedef typename base_type::value_type::const_iterator
        feature_const_iterator;

      /// Constructor
      feature_container() :
	_base_type() {
      }

      /// Constructor
      explicit feature_container(size_type _n) :
	_base_type(_n) {
      }

    protected:
      // Find cached value
      void find_cached_value() const {
	// Valid?
	if (not this->cached_ok_) {
	  // Follow the samples
	  this->features_ = 0;
	  for (const_iterator dit = this->begin(); dit != this->end(); ++dit)
	    // Follow the features
	    for (feature_const_iterator fit = dit->begin();
		 fit != dit->end(); ++fit)
	      // Larger than the rest?
	      if (fit->first >= this->features_)
		this->features_ = fit->first + 1;

	  // It is OK now
	  this->cached_ok_ = true;
	}
      }

      // Friend class
      friend class _feature_container<Base, feature_container>;
    };
  }
}

#endif
