#ifndef TTCL_CO_FIXED_FEATURE_CONTAINER_HXX
#define TTCL_CO_FIXED_FEATURE_CONTAINER_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Fixed feature mixin for containers
    @author Edgar Gonzlez i Pellicer
*/

#include <ttcl/types.hxx>
#include <ttcl/co/boost_matrix.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Fixed feature mixin for containers
    template < typename Base = boost_matrix<uint> >
    class fixed_feature_container : public Base {
    private:
      /// Number of features
      /** Cached value
       */
      mutable uint features_;

      /// Feature cardinalities
      /** Cached value
       */
      mutable uint* feature_cardinalities_;

    public:
      /// Base type
      typedef Base base_type;

      /// Const iterator type
      typedef typename base_type::const_iterator const_iterator;

      /// Constructor
      fixed_feature_container() :
	base_type(), features_(0), feature_cardinalities_(0) {
      }

      /// Sizes constructor
      fixed_feature_container(const uint _rows, const uint _cols) :
	base_type(_rows, _cols), features_(0), feature_cardinalities_(0) {
	// Will be initialized when first consulted
      }

      /// Copy constructor
      fixed_feature_container(const fixed_feature_container& _other) :
	base_type(_other), features_(0), feature_cardinalities_(0) {
	// Will be initialized when first consulted
      }

      /// Destructor
      ~fixed_feature_container() {
	delete[] feature_cardinalities_;
      }

      /// Assignment operator
      fixed_feature_container&
      operator=(const fixed_feature_container& _other) {
	// Assign as base
	base_type::operator=(_other);

	// Will be initialized when first consulted
	features_              = 0;
	feature_cardinalities_ = 0;
      }

      /// Number of features
      uint features() const {
	// Find cached values
	find_cached_values();

	// Return cached value
	return features_;
      }

      /// Feature cardinalities
      const uint* feature_cardinalities() const {
	// Find cached values
	find_cached_values();

	// Return cached value
	return feature_cardinalities_;
      }

      /// Set the feature cardinalities
      void set_feature_cardinalities(uint _features, uint* _cardinalities) {
	// Delete
	delete feature_cardinalities_;

	// Set
	features_              = _features;
	feature_cardinalities_ = _cardinalities;
      }

    protected:
      /// Find cached values
      void find_cached_values() const {
	// Valid?
	if (not feature_cardinalities_) {
	  // Empty?
	  if (this->empty()) {
	    // Create empty
	    features_              = 0;
	    feature_cardinalities_ = new uint[1];
	  }
	  else {
	    // Create with the size
	    features_              = (*this)[0].size();
	    feature_cardinalities_ = new uint[features_];

	    // Reset
	    std::fill(feature_cardinalities_,
		      feature_cardinalities_ + features_, 0);

	    // Follow data and features
	    for (const_iterator it = this->begin(); it != this->end(); ++it)
	      for (uint f = 0; f < features_; ++f)
		if ((*it)[f] >= feature_cardinalities_[f])
		  feature_cardinalities_[f] = (*it)[f] + 1;
	  }
	}
      }

    public:
      /// Invalidate cached value
      void invalidate_cache() const {
	// Invalidate
	delete feature_cardinalities_;
	feature_cardinalities_ = 0;
      }
    };
  }
}


#endif
