#ifndef TTCL_CO_FOLLOWABLE_MF_SET_HXX
#define TTCL_CO_FOLLOWABLE_MF_SET_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Followable Merge-Find Set
    @author Edgar Gonzalez i Pellicer
*/

#include <utility>

#include <ttcl/co/mf_set.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Followable Merge-Find Set
    class followable_mf_set :
      public mf_set {
    public:
      /// Base type
      typedef mf_set base_type;
      
      /// Value type
      typedef base_type::value_type value_type;

      /// Size type
      typedef base_type::size_type size_type;

      /// Follow type
      typedef uint follow_type;

      /// Const follow iterator
      /** A model of forward iterator
       */
      class const_follow_iterator {
      private:
	/// Next pointer
	const follow_type* next_;

	/// Current element
	follow_type current_;

      public:
	/// Default constructor
	const_follow_iterator() :
	  next_(0), current_(-1) {
	}

	/// Constructor from values
	const_follow_iterator(const follow_type* _next, follow_type _current) :
	  next_(_next), current_(_current) {
	}

	/// Dereference
	follow_type operator*() const {
	  return current_;
	}

	/// Preincrement
	const_follow_iterator& operator++() {
	  // Advance and return this
	  current_ = next_[current_];
	  return *this;
	}

	/// Postincrement
	const_follow_iterator operator++(int) {
	  // Copy
	  const_follow_iterator res(*this);

	  // Advance and return the res
	  current_ = next_[current_];
	  return res;
	}

	/// Equality
	bool operator==(const const_follow_iterator& _other) {
	  // Are next and current equal?
	  return next_ == _other.next_ and current_ == _other.current_;
	}

	/// Inequality
	bool operator!=(const const_follow_iterator& _other) {
	  // Not equal
	  return not (*this == _other);
	}
      };

    private:
      /// Next element
      follow_type* next_;

      /// Fist element
      follow_type* first_;

      /// Last element
      follow_type* last_;

    public:
      /// Constructor
      /** @param _size Size
       */
      followable_mf_set(size_type _size) :
	base_type(_size) {
	// Reserve memory
	next_  = new follow_type[size_];
	first_ = new follow_type[size_];
	last_  = new follow_type[size_];

	// Make everyone its father
	for (size_type i = 0; i < size_; ++i) {
	  next_ [i] = -1;
	  first_[i] = i;
	  last_ [i] = i;
	}
      }

      /// Copy constructor
      /** @param _other Source set
       */
      followable_mf_set(const followable_mf_set& _other) :
	base_type(_other) {
	// Reserve and copy next_
	next_ = new follow_type[size_];
	std::copy(_other.next_, _other.next_ + _other.size_,
		  next_);

	// Reserve and copy first_
	first_ = new follow_type[size_];
	std::copy(_other.first_, _other.first_ + _other.size_,
		  first_);

	// Reserve and copy last_
	last_ = new follow_type[size_];
	std::copy(_other.last_, _other.last_ + _other.size_,
		  last_);
      }

      /// Destructor
      ~followable_mf_set() {
	// Free
	delete[] next_;
	delete[] first_;
	delete[] last_;
      }
      
      /// Assignment operator
      /** @param _other Source set
       */
      followable_mf_set& operator=(const followable_mf_set& _other) {
	// Check for self-assignment
	if (this != &_other) {
	  // Reserve new block of data for next_
	  follow_type* new_next = new follow_type[_other.size_];
	  std::copy(_other.next_, _other.next_ + _other.size_,
		    new_next);

	  // Reserve new block of data for first_
	  follow_type* new_first = new follow_type[size_];
	  std::copy(_other.first_, _other.first_ + _other.size_,
		    new_first);

	  // Reserve new block of data for last_
	  follow_type* new_last = new follow_type[size_];
	  std::copy(_other.last_, _other.last_ + _other.size_,
		    new_last);

	  // Perform base_type assignment
	  base_type::operator=(_other);

	  // Replace
	  delete[] next_;
	  delete[] first_;
	  delete[] last_;
	  next_  = new_next;
	  first_ = new_first;
	  last_  = new_last;
	}

	// Return this
	return *this;
      }

      /// Reset
      virtual void reset() {
	// Call parent version
	mf_set::reset();

	// Make everyone its own father
	for (size_type i = 0; i < size_; ++i) {
	  next_ [i] = -1;
	  first_[i] = i;
	  last_ [i] = i;
	}
      }

      /// Merge
      virtual value_type merge(value_type _elem1, value_type _elem2) {
	// Find the parents
	value_type p1 = find_top(_elem1);
	value_type p2 = find_top(_elem2);

	// Is the parent the same?
	if (p1 == p2)
	  return p1;

	// Merge them to the lowest one
	parents_[_elem2] = p1;
	parents_[p2]     = p1;
    
	// Merge the followers
	next_ [last_[p1]] = first_[p2];
	last_ [p1]        = last_[p2];
	first_[p2]        = -1;
	last_ [p2]        = -1;

	// One set less
	--sets_;

	// Return the new common parent
	return p1;
      }

      /// Beginning of a follower
      const_follow_iterator begin(size_type _set) const {
	return const_follow_iterator(next_, first_[_set]);
      }

      /// End of a follower
      const_follow_iterator end(size_type /* _set */) const {
	return const_follow_iterator(next_, -1);
      }
    };
  }
}

#endif
