#ifndef TTCL_CO_FUNCTION_ADAPTER_HXX
#define TTCL_CO_FUNCTION_ADAPTER_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Function adapter
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Function adapter
    /** A model of STL Read-Only Sequence
     */
    template <typename Container, typename Function>
    class function_adapter {
    public:
      /// Container type
      typedef Container container_type;

      /// Function type
      typedef Function function_type;

      /// Value type
      typedef typename Function::result_type value_type;

      /// Reference type
      typedef value_type& reference;

      /// Const reference
      typedef const value_type& const_reference;

      /// Pointer
      typedef value_type* pointer;

      /// Difference type
      typedef ptrdiff_t difference_type;

      /// Size type
      typedef typename container_type::size_type size_type;

      /// Inner const iterator
      typedef typename container_type::const_iterator inner_const_iterator;
      
      /// Const iterator
      class const_iterator {
	/// Inner iterator
	typename inner_const_iterator inner_;

	/// Function
	const function_type* function_;

      public:
	/// Constructor
	const_iterator(inner_const_iterator _inner,
		       const function_type* _function) :
	  inner_(_inner), function_(_function) {
	}

	/// Indirection
	value_type operator*() {
	  return (*function_)(*inner_);
	}

	/// Addition
	const_iterator& operator+=(const difference_type _n) {
	  inner_ += _n;
	  return *this;
	}

	/// Addition
	const_iterator operator+(const difference_type _n) const {
	  return const_iterator(inner_ + _n, function_);
	}

	/// Preincrement
	const_iterator& operator++() {
	  ++inner_;
	  return *this;
	}

	/// Postincrement
	const_iterator operator++(int) {
	  const_iterator ret = *this;
	  ++inner_;
	  return ret;
	}

	/// Substraction
	const_iterator& operator-=(const difference_type _n) {
	  inner_ += _n;
	  return *this;
	}

	/// Substraction
	const_iterator operator-(const difference_type _n) const {
	  return const_iterator(inner_ - _n, function_);
	}

	/// Predecrement
	const_iterator& operator--() {
	  --inner_;
	  return *this;
	}

	/// Postdecrement
	const_iterator operator--(int) {
	  const_iterator ret = *this;
	  --inner_;
	  return ret;
	}

	/// Equality comparison
	bool operator==(const const_iterator& _other) const {
	  return inner_ == _other.inner_;
	}

	/// Inequality comparison
	bool operator!=(const const_iterator& _other) const {
	  return inner_ != _other.inner_;
	}

	/// Less than comparison
	bool operator<(const const_iterator& _other) const {
	  return inner_ < _other.inner_;
	}

	/// Less than or equal comparison
	bool operator<=(const const_iterator& _other) const {
	  return inner_ <= _other.inner_;
	}

	/// Greater than comparison
	bool operator>(const const_iterator& _other) const {
	  return inner_ > _other.inner_;
	}

	/// Greater than or equal comparison
	bool operator>=(const const_iterator& _other) const {
	  return inner_ >= _other.inner_;
	}

	/// Element
	value_type operator[](const difference_type _n) const {
	  return (*function_)(inner_[_n]);
	}

	/// Difference
	difference_type operator-(const const_iterator& _other) const {
	  return inner_ - _other.inner_;
	}
      };

    private:
      /// Pointer to the container
      const container_type* container_;

      /// Function object
      function_type function_;

    public:
      /// Constructor
      function_adapter(const container_type& _container,
		       const function_type& _function = function_type()) :
	container_(&_container), function_(_function) {
      }

      /// Begin
      const_iterator begin() const {
	return const_iterator(container_->begin(), &function_);
      }

      /// End
      const_iterator end() const {
	return const_iterator(container_->end(), &function_);
      }

      /// Size
      size_type size() const {
	return container_->size();
      }

      /// Max size
      size_type max_size() const {
	return container_->max_size();
      }

      /// Empty?
      bool empty() const {
	return container_->empty();
      }

      /// Swap
      void swap(function_adapter& _other) {
	std::swap(container_, _other.container_);
	std::swap(function_,  _other.function_);
      }

      /// Front
      value_type front() const {
	return function_(container_->front());
      }

      /// Back
      value_type back() const {
	return function_(container_->back());
      }
    };

    /// Adapt with function
    /** Helper function
     */
    template <typename Container, typename Function>
    function_adapter<Container, Function>
    function_adapt(const Container& _container, const Function& _function) {
      return function_adapter<Container, Function>(_container, _function);
    }
  }
}

#endif
