#ifndef TTCL_CO_INDEXED_ADAPTER_HXX
#define TTCL_CO_INDEXED_ADAPTER_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Indexed adapter
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Indexed adapter
    /** A model of STL Sequence
     */
    template <typename Container, typename Indexer>
    class indexed_adapter {
    public:
      /// Container type
      typedef Container container_type;

      /// Indexer type
      typedef Indexer indexer_type;

      /// Value type
      typedef typename container_type::value_type value_type;

      /// Reference type
      typedef typename container_type::reference reference;

      /// Const reference
      typedef typename container_type::const_reference const_reference;

      /// Pointer
      typedef typename container_type::pointer pointer;

      /// Difference type
      typedef typename container_type::difference_type difference_type;

      /// Size type
      typedef typename container_type::size_type size_type;

      /// Inner iterator
      typedef typename indexer_type::iterator inner_iterator;

      /// Inner const iterator
      typedef typename indexer_type::const_iterator inner_const_iterator;

      /// Iterator
      class iterator {
	/// Constainer
	container_type* container_;

	/// Inner iterator
	inner_iterator inner_;

      public:
	/// Constructor
	iterator(container_type& _container,
		 inner_iterator _inner) :
	  container_(&_container), inner_(_inner) {
	}

	/// Indirection
	value_type& operator*() {
	  return (*container_)[*inner_];
	}

	/// Indirection
	value_type* operator->() {
	  return &((*container_)[*inner_]);
	}

	/// Addition
	iterator& operator+=(const difference_type _n) {
	  inner_ += _n;
	  return *this;
	}

	/// Addition
	iterator operator+(const difference_type _n) const {
	  return iterator(*container_, inner_ + _n);
	}

	/// Preincrement
	iterator& operator++() {
	  ++inner_;
	  return *this;
	}

	/// Postincrement
	iterator operator++(int) {
	  iterator ret = *this;
	  ++inner_;
	  return ret;
	}

	/// Substraction
	iterator& operator-=(const difference_type _n) {
	  inner_ += _n;
	  return *this;
	}

	/// Substraction
	iterator operator-(const difference_type _n) const {
	  return iterator(*container_, inner_ - _n);
	}

	/// Predecrement
	iterator& operator--() {
	  --inner_;
	  return *this;
	}

	/// Postdecrement
	iterator operator--(int) {
	  iterator ret = *this;
	  --inner_;
	  return ret;
	}

	/// Equality comparison
	bool operator==(const iterator& _other) const {
	  return inner_ == _other.inner_;
	}

	/// Inequality comparison
	bool operator!=(const iterator& _other) const {
	  return inner_ != _other.inner_;
	}

	/// Less than comparison
	bool operator<(const iterator& _other) const {
	  return inner_ < _other.inner_;
	}

	/// Less than or equal comparison
	bool operator<=(const iterator& _other) const {
	  return inner_ <= _other.inner_;
	}

	/// Greater than comparison
	bool operator>(const iterator& _other) const {
	  return inner_ > _other.inner_;
	}

	/// Greater than or equal comparison
	bool operator>=(const iterator& _other) const {
	  return inner_ >= _other.inner_;
	}

	/// Element
	value_type operator[](const difference_type _n) const {
	  return (*container_)[inner_[_n]];
	}

	/// Difference
	difference_type operator-(const iterator& _other) const {
	  return inner_ - _other.inner_;
	}
      };

      /// Const Iterator
      class const_iterator {
	/// Constainer
	const container_type* container_;

	/// Inner iterator
	inner_const_iterator inner_;

      public:
	/// Constructor
	const_iterator(container_type& _container,
		       inner_const_iterator _inner) :
	  container_(&_container), inner_(_inner) {
	}

	/// Indirection
	const value_type& operator*() {
	  return (*container_)[*inner_];
	}

	/// Indirection
	const value_type* operator->() {
	  return &((*container_)[*inner_]);
	}

	/// Addition
	const_iterator& operator+=(const difference_type _n) {
	  inner_ += _n;
	  return *this;
	}

	/// Addition
	const_iterator operator+(const difference_type _n) const {
	  return const_iterator(*container_, inner_ + _n);
	}

	/// Preincrement
	const_iterator& operator++() {
	  ++inner_;
	  return *this;
	}

	/// Postincrement
	const_iterator operator++(int) {
	  const_iterator ret = *this;
	  ++inner_;
	  return ret;
	}

	/// Substraction
	const_iterator& operator-=(const difference_type _n) {
	  inner_ += _n;
	  return *this;
	}

	/// Substraction
	const_iterator operator-(const difference_type _n) const {
	  return const_iterator(*container_, inner_ - _n);
	}

	/// Predecrement
	const_iterator& operator--() {
	  --inner_;
	  return *this;
	}

	/// Postdecrement
	const_iterator operator--(int) {
	  const_iterator ret = *this;
	  --inner_;
	  return ret;
	}

	/// Equality comparison
	bool operator==(const const_iterator& _other) const {
	  return inner_ == _other.inner_;
	}

	/// Inequality comparison
	bool operator!=(const const_iterator& _other) const {
	  return inner_ != _other.inner_;
	}

	/// Less than comparison
	bool operator<(const const_iterator& _other) const {
	  return inner_ < _other.inner_;
	}

	/// Less than or equal comparison
	bool operator<=(const const_iterator& _other) const {
	  return inner_ <= _other.inner_;
	}

	/// Greater than comparison
	bool operator>(const const_iterator& _other) const {
	  return inner_ > _other.inner_;
	}

	/// Greater than or equal comparison
	bool operator>=(const const_iterator& _other) const {
	  return inner_ >= _other.inner_;
	}

	/// Element
	value_type operator[](const difference_type _n) const {
	  return (*container_)[inner_[_n]];
	}

	/// Difference
	difference_type operator-(const const_iterator& _other) const {
	  return inner_ - _other.inner_;
	}
      };

    private:
      /// Pointer to the container
      container_type* container_;

      /// Pointer to the indexer
      const indexer_type* indexer_;

    public:
      /// Constructor
      indexed_adapter(container_type& _container,
		      const indexer_type& _indexer) :
	container_(&_container), indexer_(&_indexer) {
      }
      
      /// Begin
      iterator begin() {
	return iterator(*container_, indexer_->begin());
      }

      /// Begin
      const_iterator begin() const {
	return const_iterator(*container_, indexer_->begin());
      }

      /// End
      iterator end() {
	return iterator(*container_, indexer_->end());
      }

      /// End
      const_iterator end() const {
	return const_iterator(*container_, indexer_->end());
      }

      /// Size
      size_type size() const {
	return indexer_->size();
      }

      /// Max size
      size_type max_size() const {
	return indexer_->max_size();
      }

      /// Empty?
      bool empty() const {
	return indexer_->empty();
      }

      /// Swap
      void swap(indexed_adapter& _other) {
	std::swap(container_, _other.container_);
	std::swap(indexer_,   _other.indexer_);
      }

      /// Front
      reference front() {
	return (*container_)[indexer_->front()];
      }

      /// Front
      const_reference front() const {
	return (*container_)[indexer_->front()];
      }

      /// Back
      reference back() {
	return (*container_)[indexer_->back()];
      }

      /// Back
      const_reference back() const {
	return (*container_)[indexer_->back()];
      }
    };
      
    /// Adapt with indexer
    /** Helper function
     */
    template <typename Container, typename Indexer>
    indexed_adapter<Container, Indexer>
    indexed_adapt(Container& _container, const Indexer& _indexer) {
      return indexed_adapter<Container, Indexer>(_container, _indexer);
    }
  }
}

#endif
