#ifndef TTCL_CO_LIMIT_PRIORITY_QUEUE_HXX
#define TTCL_CO_LIMIT_PRIORITY_QUEUE_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Limited priority queue
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>
#include <functional>
#include <vector>

#include <ttcl/exception.hxx>
#include <ttcl/types.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co { 

    /// Base limited priority queue
    class _limit_priority_queue {
    public:
      // Size type
      typedef size_t size_type;

      /// Default maximum size
      /** As a rule of thumb, a value of 2^20 is OK
       */
      static const uint default_max_size = 1048576;
    };

    /// Limited priority queue
    /** (Almost) a replacement for std::priority_queue
	but with limit on its size and values
	@param T Value type
	@param Sequence Sequence
	@param Compare Comparison function
     */
    template <typename T,
	      typename Sequence = std::vector<T>,
	      typename Compare = std::less<T> >
    class limit_priority_queue : public _limit_priority_queue {
    public:
      // Value type
      typedef typename Sequence::value_type value_type;

      // Size type
      typedef typename Sequence::size_type size_type;

    protected:
      /// Max size of the queue
      size_type max_size_;

      /// Minimum value
      value_type minimum_;
  
      /// Growing
      bool growing_;

      /// Comparison function
      Compare comp_;

      /// Data
      Sequence data_;

    public:
      /// Constructor
      /** @param _max_size Maximum size of the queue
	  @param _comp  Comparison function
      */
      explicit limit_priority_queue(size_type _max_size = default_max_size,
				    const Compare& _comp = Compare()) :
	max_size_(_max_size), minimum_(value_type()),
	growing_(true), comp_(_comp), data_() {
      }

      /// Copy constructor
      /** @param _other Source queue
       */
      limit_priority_queue(const limit_priority_queue& _other) :
	max_size_(_other.max_size_), minimum_(_other.minimum_),
	growing_(_other.growing_), comp_(_other.comp_),
	data_(_other.data_) {
      }

      /// Destructor
      ~limit_priority_queue() {
      }

      /// Assignment operator
      /** @param _other Source queue
       */
      limit_priority_queue& operator=(const limit_priority_queue& _other) {
	// Check for self-assignment
	if (this != &_other) {
	  // Copy field by field
	  max_size_ = _other.max_size_;
	  minimum_  = _other.minimum_;
	  growing_  = _other.growing_;
	  comp_     = _other.comp_;
	  data_     = _other.data_;
	}

	// Return this
	return *this;
      }

      /// Is it empty?
      bool empty() const {
	// Nothing used or exhausted after growing
	// (The top is not larger than the minimum)
	return (data_.empty()) or
	  (not growing_ and not comp_(minimum_, data_[0]));
      }

      /// Size
      /** If the queue is growing, it is the size of the sequence
	  @bug If the queue is not growing, the size is far too difficult to estimate, so we just return the size of the sequence
      */
      size_type size() const {
	// Empty?
	if (empty())
	  return 0;

	// Currently used
	return data_.size();
      }

      /// Get the top
#ifdef TTCL_MORE_CHECKS
      const value_type& top() const
	throw (exception) {
	// Check not empty
	if (empty())
	  ttcl_fire("Exhausted queue");
#else
      const value_type& top() const {
#endif
	// Return the top
        return data_[0];
      }

      /// Pop the top
#ifdef TTCL_MORE_CHECKS
      void pop()
	throw (exception) {
	// Check not empty
	if (empty())
	  ttcl_fire("Exhausted queue");
#else
      void pop() {
#endif
	// Move the last element to the root
	std::swap(data_[0], data_.back());
	data_.pop_back();

	// Need to float?
	if (data_.size() > 0) {
	  // Sink it
	  size_type pos  = 0;
	  size_type lpos = (pos << 1) + 1;
	  size_type rpos = lpos + 1;
	  while (lpos < data_.size()) {
	    if (rpos < data_.size()) {
	      // Two children
	      if (comp_(data_[rpos], data_[lpos])) {
		if (comp_(data_[pos], data_[lpos])) {
		  // Swap
		  std::swap(data_[pos], data_[lpos]);
		  pos = lpos;
		  
		} else {
		  // No more sinkin'
		  break;
		}
		
	      } else {
		if (comp_(data_[pos], data_[rpos])) {
		  // Swap
		  std::swap(data_[pos], data_[rpos]);
		  pos = rpos;
		  
		} else {
		  // No more sinkin'
		  break;
		}
	      }
	      
	    } else {
	      // A single child
	      if (comp_(data_[pos], data_[lpos])) {
		// Swap
		std::swap(data_[pos], data_[lpos]);
	      }
	      
	      // No more sinkin'
	      break;
	    }
	    
	    // Update lpos and rpos
	    lpos = (pos << 1) + 1;
	    rpos = lpos + 1;
	  }
	}
      }

      /// Insert an element
      /** @param _value Value
       */
      void push(value_type _value)
	throw (exception) {
	// Many cases...
	if (not growing_) {
	  if (empty()) {
	    // Cannot insert anything
	    //  -> Must reset and reload
	    ttcl_fire("Exhausted queue");
	  }
    
	  if (not comp_(minimum_, _value)) {
	    // Discarded
	    return;
	  }
	}

	// In all other cases...
	// Insert at the end
	data_.push_back(_value);
  
	// Float
	size_type pos    = data_.size() - 1;
	size_type parent;
	while (pos > 0) {
	  parent = (pos - 1) >> 1;

	  if (comp_(data_[parent], data_[pos])) {
	    // Float
	    std::swap(data_[parent], data_[pos]);
	    pos = parent;
      
	  } else {
	    // No more floatin'
	    break;
	  }
	}

	// If used exceeds max size
	if (data_.size() > max_size_) {
	  // Remove the last
	  // We keep the highest of the minima
	  if (growing_ or comp_(minimum_, data_[max_size_])) {
	    minimum_ = data_[max_size_];
	  }
    
	  // Truncate
	  data_.pop_back();
	
	  // We cannot grow further
	  growing_ = false;
	}
      }
      
      /// Clear the queue
      void clear() {
	growing_ = true;
	data_.clear();
      }
    };
  }
}

#endif
