#ifndef TTCL_CO_MF_SET_HXX
#define TTCL_CO_MF_SET_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Merge-Find Set
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>
#include <map>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Merge-Find Set
    /** With Quick-Find
     */
    class mf_set {
    public:
      /// Value type
      typedef size_t value_type;

      /// Pointer type
      typedef size_t* pointer;

      /// Const pointer type
      typedef const size_t* const_pointer;

      /// Reference type
      typedef size_t& reference;

      /// Const reference
      typedef const size_t& const_reference;

      /// Iterator type
      typedef size_t* iterator;

      /// Const iterator type
      typedef const size_t* const_iterator;

      /// Reverse iterator type
      typedef std::reverse_iterator<iterator> reverse_iterator;

      /// Const reverse iterator type
      typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

      /// Distance type
      typedef ptrdiff_t difference_type;

      /// Size type
      typedef size_t size_type;
      
    protected:
      /// Size
      size_type size_;

      /// Sets
      size_type sets_;

      /// Parent vector
      mutable value_type* parents_;

    public:
      /// Constructor
      /** @param _size Size
       */
      mf_set(size_type _size) :
	size_(_size), sets_(_size) {
	// Reserve
	parents_ = new value_type[size_];

	// Make everyone its father
	for (size_type i = 0; i < size_; ++i)
	  parents_[i] = i;
      }

      /// Copy constructor
      /** @param _other Source set
       */
      mf_set(const mf_set& _other) :
	size_(_other.size_), sets_(_other.sets_) {
	// Reserve and copy
	parents_ = new value_type[size_];
	std::copy(_other.parents_, _other.parents_ + _other.size_,
		  parents_);
      }

      /// Destructor
      virtual ~mf_set() {
	// Free
	delete[] parents_;
      }

      /// Assignment operator
      /** @param _other Source set
       */
      mf_set& operator=(const mf_set& _other) {
	// Check for self-assignment
	if (this != &_other) {
	  // Reserve a new block of data
	  value_type* new_parents = new value_type[_other.size_];
	  std::copy(_other.parents_, _other.parents_ + _other.size_,
		    new_parents);

	  // Replace
	  delete[] parents_;
	  parents_ = new_parents;

	  // Copy rest
	  size_ = _other.size_;
	  sets_ = _other.sets_;
	}

	// Return this
	return *this;
      }

      /// Size
      size_type size() const {
	return size_;
      }

      /// Number of sets
      size_type sets() const {
	return sets_;
      }

      /// Reset
      virtual void reset() {
	// Make everyone its own father
	sets_ = size_;
	for (size_type i = 0; i < size_; ++i)
	  parents_[i] = i;
      }

      /// Get the set
      value_type operator[](size_type _element) const {
	return find_top(_element);
      }

      /// Merge
      virtual value_type merge(value_type _elem1, value_type _elem2) {
	// Find the parents
	value_type p1 = find_top(_elem1);
	value_type p2 = find_top(_elem2);

	// Is the parent the same?
	if (p1 == p2)
	  return p1;

	// Merge them to the lowest one
	parents_[_elem2] = p1;
	parents_[p2]     = p1;
    
	// One set less
	--sets_;

	// Return the new common parent
	return p1;
      }

      /*
      /// Reindex
      value_type reindex_sets() {
	// Mapping
	value_type next_id = 0;
	std::map<size_type, value_type> mapping;

	// Follow the elements
	for (size_type i = 0; i < size_; ++i) {
	  value_type top = find_top(i); 
	  std::map<value_type, value_type>::iterator it = mapping.find(top);
	  if (it == mapping.end())
	    mapping.insert(std::make_pair(top, next_id++));
	}

	// Convert
	for (size_type i = 0; i < size_; ++i)
	  parents_[i] = mapping[parents_[i]];

	// Return the number of ids
	return next_id;
      }
      */

      /// Begin Iterator
      iterator begin() {
	// Pointer to the data
	return parents_;
      }

      /// Const Begin Iterator
      const_iterator begin() const {
	// Pointer to the data
	return parents_;
      }

      /// End Iterator
      iterator end() {
	// Pointer past the data
	return parents_ + size_;
      }

      /// Const End Iterator
      const_iterator end() const {
	// Pointer past the data
	return parents_ + size_;
      }

    protected:
      /// Get the top parent
      value_type find_top(value_type _element) const {
	// While we are not our own parent
	if (parents_[_element] == _element)
	  return _element;

	// Recursively call and compress
	return (parents_[_element] = find_top(parents_[_element]));
      }
    };
  }
}

#endif
