#ifndef TTCL_MMAP_VECTOR_HXX
#define TTCL_MMAP_VECTOR_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Vector through mmap
    @author Edgar Gonzlez i Pellicer
*/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <cstring>
#include <iterator>

#include <ttcl/sys_exception.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Vector through mmap
    /** A model of STL Random Access Container
	@param T Value type. Must be a simple type.
    */
    template <typename T>
    class mmap_vector {
    public:
      /// Value type
      typedef T value_type;

      /// Pointer type
      typedef T* pointer;

      /// Const pointer type
      typedef const T* const_pointer;

      /// Reference type
      typedef T& reference;

      /// Const reference
      typedef const T& const_reference;

      /// Iterator type
      typedef T* iterator;

      /// Const iterator type
      typedef const T* const_iterator;

      /// Reverse iterator type
      typedef std::reverse_iterator<iterator> reverse_iterator;

      /// Const reverse iterator type
      typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

      /// Distance type
      typedef ptrdiff_t difference_type;

      /// Size type
      typedef size_t size_type;

    private:
      /// Page size
      static const size_type page_size = 4096;
      
      /// Is file temporary?
      bool is_temp_;

      /// File descriptor
      int fd_;

      /// Temporary file name
      char temp_name_[20];

      /// Size
      size_type size_;

      /// Capacity
      size_type capacity_;

      /// File size
      size_type file_size_;

      /// Data pointer
      value_type* data_;

    public:
      /// Constructor
      mmap_vector(size_type _size = size_type(),
		  const value_type _value = value_type(),
		  int _fd = -1) 
	throw (exception) :
	is_temp_(_fd == -1), fd_(_fd), size_(_size),
	capacity_(0), file_size_(0), data_(0) {
	// Do we need a temporary file?
	if (is_temp_) {
	  // Allocate it
	  strcpy(temp_name_, "/tmp/vector.XXXXXX");
	  if ((fd_ = mkstemp(temp_name_)) == -1)
	    ttcl_sys_fire("Cannot create temporary file");
	}
	else {
	  // Check mode
	  int flags = fcntl(fd_, F_GETFL);
	  if (flags == -1)
	    ttcl_sys_fire("Cannot access file flags");
	  if ((flags & O_ACCMODE) != O_RDWR)
	    ttcl_fire("File is not open in read/write mode");

	  // Get the file size
	  off_t signed_size = lseek(fd_, 0, SEEK_END);
	  if (signed_size == off_t(-1))
	    ttcl_sys_fire("Cannot find file size");
	  file_size_ = signed_size;
	}
	
	// Adjust size
	size_type n_pages  = sizeof(value_type) * _size / page_size + 1;
	size_type req_size = n_pages * page_size;
	if (req_size > file_size_) {
	  if (lseek(fd_, req_size - 1, SEEK_SET) == -1)
	    ttcl_sys_fire("Cannot resize file");
	  char null = '\0';
	  if (write(fd_, &null, 1) != 1)
	    ttcl_sys_fire("Cannot resize file");
	  file_size_ = req_size;
	}
	capacity_ = file_size_ / sizeof(value_type);

	// Mmap it
	void* raw = mmap(0, file_size_,
			 PROT_READ | PROT_WRITE, MAP_SHARED, fd_, 0);
	if (raw == (void*)-1)
	  ttcl_sys_fire("Cannot map file");
	data_ = static_cast<value_type*>(raw);

	// Fill it
	value_type* fill = data_;
	for (size_type i = 0; i < size_; ++i)
	  *fill++ = _value;
      }

      /// Copy constructor
      mmap_vector(const mmap_vector& _other) :
	is_temp_(_other.is_temp_), fd_(0),
	size_(_other.size_), capacity_(_other.capacity_),
	file_size_(_other.file_size_), data_(0) {
	// Do we need a temporary file?
	if (is_temp_) {
	  // Allocate it
	  strcpy(temp_name_, "/tmp/vector.XXXXXX");
	  if ((fd_ = mkstemp(temp_name_)) == -1)
	    ttcl_sys_fire("Cannot create temporary file");

	  // Resize it
	  if (lseek(fd_, file_size_ - 1, SEEK_SET) == -1)
	    ttcl_sys_fire("Cannot resize file");
	  char null = '\0';
	  if (write(fd_, &null, 1) != 1)
	    ttcl_sys_fire("Cannot resize file");
	}
	else {
	  // Copy file descriptor
	  fd_ = _other.fd_;
	}

	// Mmap it
	void* raw = mmap(0, file_size_,
			 PROT_READ | PROT_WRITE, MAP_SHARED, fd_, 0);
	if (raw == (void*)-1)
	  ttcl_sys_fire("Cannot map file");
	data_ = static_cast<value_type*>(raw);

	// Copy contents if it is a temporary
	if (is_temp_) {
	  value_type* src = _other.data_;
	  value_type* dst = data_;
	  for (size_type i = 0; i < size_; ++i)
	    *dst++ = *src++;
	}
      }

      /// Destructor
      ~mmap_vector()
      throw (exception) {
	// Unmap
	if (munmap(data_, file_size_) == -1)
	  ttcl_sys_fire("Cannot unmap file");

	// Need to close and unlink?
	if (is_temp_) {
	  if (close(fd_) == -1)
	    ttcl_sys_fire("Cannot close descriptor");
	  if (unlink(temp_name_) == -1)
	    ttcl_sys_fire("Cannot unlink temporary file");
	}
      }

      /// Assignment operation
      mmap_vector& operator=(const mmap_vector& _other) {
	// Check for auto-assignment
	if (this != &_other) {
	  // Unmap
	  if (munmap(data_, file_size_) == -1)
	    ttcl_sys_fire("Cannot unmap file");
	  
	  // Need to close and unlink?
	  if (is_temp_) {
	    if (close(fd_) == -1)
	      ttcl_sys_fire("Cannot close descriptor");
	    if (unlink(temp_name_) == -1)
	      ttcl_sys_fire("Cannot unlink temporary file");
	  }
	  
	  // Field copy
	  is_temp_   = _other.is_temp_;
	  size_      = _other.size_;
	  capacity_  = _other.capacity_;
	  file_size_ = _other.file_size_;

	  // Do we need a temporary file?
	  if (is_temp_) {
	    // Allocate it
	    strcpy(temp_name_, "/tmp/vector.XXXXXX");
	    if ((fd_ = mkstemp(temp_name_)) == -1)
	      ttcl_sys_fire("Cannot create temporary file");
	  }
	  else {
	    // Copy file descriptor
	    fd_ = _other.fd_;
	  }

	  // Mmap it
	  void* raw = mmap(0, file_size_,
			   PROT_READ | PROT_WRITE, MAP_SHARED, fd_, 0);
	  if (raw == (void*)-1)
	    ttcl_sys_fire("Cannot map file");
	  data_ = static_cast<value_type*>(raw);

	  // Copy contents if it is a temporary
	  if (is_temp_) {
	    value_type* src = _other.data_;
	    value_type* dst = data_;
	    for (size_type i = 0; i < size_; ++i)
	      *dst++ = *src++;
	  }
	}

	// Return this
	return *this;
      }

      /// Empty?
      bool empty() const {
	return size_ == 0;
      }

      /// Size
      size_type size() const {
	return size_;
      }

      /// Maximum size
      size_type max_size() const {
	return size_type(-1) / sizeof(value_type);
      }

      /// Capacity
      size_type capacity() const {
	return capacity_;
      }

      /// Clear
      void clear() {
	// Set the size to 0
	size_ = 0;
      }

      /// Reserve
      void reserve(size_type _new_capacity)
	throw (exception) {
	// Check it is larger
	if (_new_capacity > capacity_) {
	  // Required size
	  size_type req_size     = file_size_;
	  size_type req_capacity = req_size / sizeof(value_type);
	  while (_new_capacity > req_capacity) {
	    req_size    *= 2;
	    req_capacity = req_size / sizeof(value_type);
	  }

	  // Resize
	  if (lseek(fd_, req_size - 1, SEEK_SET) == off_t(-1))
	    ttcl_sys_fire("Cannot resize file");
	  char null = '\0';
	  if (write(fd_, &null, 1) != 1)
	    ttcl_sys_fire("Cannot resize file");

	  // Remalloc
	  void* raw = mremap(data_, file_size_, req_size,
			     MREMAP_MAYMOVE);
	  if (raw == (void*)-1)
	    ttcl_sys_fire("Cannot remap file");

	  // Update attributes
	  file_size_ = req_size;
	  capacity_  = req_capacity;
	  data_      = static_cast<value_type*>(raw);
	}
      }

      /// Resize
      void resize(size_type _new_size,
		  const value_type _value = value_type())
	throw (exception) {
	// Relation?
	if (_new_size <= size_) {
	  // Shorter

	  // Just update size
	  size_ = _new_size;
	}
	else {
	  // Longer

	  // Need more capacity?
	  if (_new_size > capacity_)
	    reserve(_new_size);

	  // Fill and resize
	  iterator it = end();
	  for (size_type i = size_; i < _new_size; ++i)
	    *it++ = _value;
	  size_ = _new_size;
	}
      }

      /// Beginning of range
      /** Non-const version
       */
      iterator begin() {
	// Pointer to the data
	return data_;
      }

      /// Beginning of range
      /** Const version
       */
      const_iterator begin() const {
	// Pointer to the data
	return data_;
      }

      /// End of range
      /** Non-const version
       */
      iterator end() {
	// Pointer past the data
	return data_ + size_;
      }

      /// End of range
      /** Const version
       */
      const_iterator end() const {
	// Pointer past the data
	return data_ + size_;
      }

      /// Begin of reverse range
      /** Non-const version
       */
      reverse_iterator rbegin() {
	// Reverse of the end
	return reverse_iterator(end());
      }

      /// Begin of reverse range
      /** Const version
       */
      const_reverse_iterator rbegin() const {
	// Reverse of the end
	return reverse_iterator(end());
      }

      /// End of reverse range
      /** Non-const version
       */
      reverse_iterator rend() {
	// Reverse of the begin
	return reverse_iterator(begin());
      }

      /// End of reverse range
      /** Const version
       */
      const_reverse_iterator rend() const {
	// Reverse of the begin
	return reverse_iterator(begin());
      }

      /// Element access
      /** Non-const version
       */
      reference operator[](size_type _n) {
	// Access the data
	return data_[_n];
      }

      /// Element access
      /** const version
       */
      const_reference operator[](size_type _n) const {
	// Access the data
	return data_[_n];
      }
    
      /// Front
      /** Non-const version
       */
      reference front() {
	// Access the data
	return data_[0];
      }

      /// Front
      /** const version
       */
      const_reference front() const {
	// Access the data
	return data_[0];
      }

      /// Back
      /** Non-const version
       */
      reference back() {
	// Access the data
	return data_[size_ - 1];
      }

      /// Back
      /** const version
       */
      const_reference back() const {
	// Access the data
	return data_[size_ - 1];
      }

      /// Push back
      void push_back(const value_type _value) {
	resize(size_ + 1, _value);
      }

      /// Pop back
      void pop_back() {
	resize(size_ - 1);
      }
    };
  }
}

#endif
