#ifndef TTCL_CO_MULTI_BLOCK_HXX
#define TTCL_CO_MULTI_BLOCK_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Multi block
    @author Edgar Gonzlez i Pellicer
*/

#include <ttcl/types.hxx>
#include <ttcl/co/iterator_pair.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Multi block
    template <typename T>
    class multi_block {
    public:
      /// Value type
      typedef T value_type;

      /// Iterator
      typedef iterator_pair<T*> iterator;

      /// Iterator
      typedef iterator_pair<const T*> const_iterator;
      
    private:
      /// Data block
      struct block {
	/// Next
	block* next;
	
	/// Previous
	block* prev;
	
	/// End
	value_type* fill;
      };

      /// Phantom node
      block phantom_;

      /// Current block fill
      value_type* tail_fill_;
      
      /// Current block end
      value_type* tail_end_;
      
      /// Current block size;
      uint tail_size_;

      /// Current sector start
      value_type* sector_start_;

      /// Current sector end
      value_type* sector_end_;

    public:
      /// Constructor
      multi_block() {
	// Link the phantom
	phantom_.prev = phantom_.next = &phantom_;
	phantom_.fill = 0;
	
	// Create starting block
	tail_size_ = 2;
	new_block();
      }

      /// Destructor
      ~multi_block() {
	// Free the list
	free_list();
      }

      /// Clear
      void clear() {
	// Free the list
	free_list();

	// Relink the phantom
	phantom_.prev = phantom_.next = &phantom_;

	// Create the starting block
	tail_size_ = 2;
	new_block();
      }

      /// Start a new sector
      void new_sector() {
	// Reset
	sector_start_ = sector_end_ = tail_fill_;
      }
      
      /// Add a value to the current sector
      iterator add_to_sector(const value_type& _value) {
	// Can we add?
	if (tail_fill_ == tail_end_) {
	  // Remember old start as fill
	  phantom_.prev->fill = const_cast<value_type*>(sector_start_);

	  // Reallocate
	  new_block();

	  // New start
	  value_type* it = sector_start_;
	  sector_start_ = tail_fill_;

	  // Copy
	  for (; it != sector_end_; ++it) {
	    // Copy construct and delete
	    new (tail_fill_++) value_type(*it);
	    it->~value_type();
	  }
	}

	// Copy construct
	new (tail_fill_++) value_type(_value);
	sector_end_ = tail_fill_;
	phantom_.prev->fill = tail_fill_;

	// Return the pair
	return iterator(sector_start_, sector_end_);
      }
      
    private:
      /// Free the list
      void free_list() {
	// Follow the list
	block* b = phantom_.next;
	while (b != &phantom_) {
	  // Destroy every object
	  for (value_type* it = reinterpret_cast<value_type*>(b + 1);
	       it != b->fill; ++it)
	    // Destructor
	    it->~value_type();

	  // Save next and destroy it
	  block* d = b;
	  b = b->next;
	  ::operator delete(d);
	}
      }

      /// Get memory
      void new_block() {
	// Duply size
	tail_size_ *= 2;

	// Reserve the raw memory
	block* raw =
	  static_cast<block*>(::operator new(sizeof(block) +
					     tail_size_ * sizeof(value_type)));

	// Link
	raw->prev = phantom_.prev;
	raw->next = phantom_.prev->next;
	phantom_.prev->next = raw;
	phantom_.prev       = raw;

	// Set the tail info
	tail_fill_ = raw->fill = reinterpret_cast<value_type*>(raw + 1);
	tail_end_  = tail_fill_ + tail_size_;
      }
    };
  }
}

#endif
