#ifndef TTCL_CO_MULTI_MAP_HXX
#define TTCL_CO_MULTI_MAP_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Multi map
    @author Edgar Gonzlez i Pellicer
*/

#include <map>

#include <ttcl/exception.hxx>
#include <ttcl/types.hxx>
#include <ttcl/co/multi_block.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Multi map
    template <typename Key, typename Data>
    class multi_map {
    public:
      /// Key type
      typedef Key key_type;

      /// Data type
      typedef Data data_type;

      /// Block type
      typedef multi_block<data_type> block_type;

      /// Value type
      typedef typename block_type::iterator value_type;

      /// Const value type
      typedef typename block_type::const_iterator const_value_type;

      /// Map type
      typedef std::map<key_type, value_type> map_type;
      
      /// Const map type
      typedef std::map<key_type, const_value_type> const_map_type;

      /// Pointer type
      typedef data_type* pointer;

      /// Const pointer type
      typedef const data_type* const_pointer;

      /// Reference type
      typedef data_type& reference;

      /// Const reference
      typedef const data_type& const_reference;

      /// Iterator type
      typedef typename map_type::const_iterator iterator;

      /// Const iterator type
      typedef typename const_map_type::const_iterator const_iterator;

      /// Reverse iterator type
      typedef std::reverse_iterator<iterator> reverse_iterator;

      /// Const reverse iterator type
      typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

      /// Distance type
      typedef typename map_type::difference_type difference_type;

      /// Size type
      typedef typename map_type::size_type size_type;
      
    private:
      /// Multi block
      block_type blocks_;

      /// Map
      map_type map_;

      /// Last key
      key_type last_key_;

    public:
      /// Empty constructor
      multi_map() {
      }

      /// Size
      size_type size() const {
	return map_.size();
      }

      /// Empty?
      bool empty() const {
	return map_.empty();
      }

      /// Clear
      void clear() {
	// Clear the blocks and map part
	blocks_.clear();
	map_.clear();
      }

      /// Insert
      void insert(const key_type& _key, const data_type& _data) {
	// Is it already inserted?
	typename map_type::iterator it = map_.find(_key);
	if (it == map_.end()) {
	  // New sector and insert into map
	  blocks_.new_sector();
	  it = map_.insert(make_pair(_key, value_type())).first;
	  last_key_ = _key;
	}
	else if (_key != last_key_) {
	  // It's not the last
	  ttcl_fire("Cannot insert unsorted");
	}

	// Set
	it->second = blocks_.add_to_sector(_data);
      }
      
      /// Access
      value_type operator[](const key_type& _key) {
	// Just access the map
	typename map_type::const_iterator it = map_.find(_key);
	if (it == map_.end())
	  return value_type();
	else
	  return it->second;
      }

      /// Const access
      const_value_type operator[](const key_type& _key) const {
	// Just access the map
	typename map_type::const_iterator it = map_.find(_key);
	if (it == map_.end())
	  return const_value_type();
	else
	  return it->second;
      }

      /// Begin
      iterator begin() {
	// Just access the map
	return map_.begin();
      }

      /// Const begin
      const_iterator begin() const {
	// Just access the map
	return map_.begin();
      }

      /// End
      iterator end() {
	// Just access the map
	return map_.end();
      }

      /// Const end
      const_iterator end() const {
	// Just access the map
	return map_.end();
      }

      /// Reverse begin
      reverse_iterator rbegin() {
	// Just access the map
	return map_.rbegin();
      }

      /// Const reverse begin
      const_reverse_iterator rbegin() const {
	// Just access the map
	return map_.rbegin();
      }

      /// Reverse end
      reverse_iterator rend() {
	// Just access the map
	return map_.rend();
      }

      /// Const reverse end
      const_reverse_iterator rend() const {
	// Just access the map
	return map_.rend();
      }
    };
  }
}

#endif
