#ifndef TTCL_CO_MULTI_VECTOR_HXX
#define TTCL_CO_MULTI_VECTOR_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Multi vector
    @author Edgar Gonzlez i Pellicer
*/

#include <vector>

#include <ttcl/types.hxx>
#include <ttcl/co/multi_block.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Multi vector
    template <typename T>
    class multi_vector {
    public:
      /// Data type
      typedef T data_type;

      /// Block type
      typedef multi_block<data_type> block_type;

      /// Value type
      typedef typename block_type::iterator value_type;

      /// Const value type
      typedef typename block_type::const_iterator const_value_type;

      /// Vector type
      typedef std::vector<value_type> vector_type;
      
      /// Const vector type
      typedef std::vector<const_value_type> const_vector_type;

      /// Pointer type
      typedef T* pointer;

      /// Const pointer type
      typedef const T* const_pointer;

      /// Reference type
      typedef T& reference;

      /// Const reference
      typedef const T& const_reference;

      /// Iterator type
      typedef typename vector_type::const_iterator iterator;

      /// Const iterator type
      typedef typename const_vector_type::const_iterator const_iterator;

      /// Reverse iterator type
      typedef std::reverse_iterator<iterator> reverse_iterator;

      /// Const reverse iterator type
      typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

      /// Distance type
      typedef typename vector_type::difference_type difference_type;

      /// Size type
      typedef typename vector_type::size_type size_type;
      
    private:
      /// Multi block
      block_type blocks_;

      /// Vector
      vector_type vector_;

    public:
      /// Empty constructor
      multi_vector() {
      }

      /// Size
      size_type size() const {
	return vector_.size();
      }

      /// Empty?
      bool empty() const {
	return vector_.empty();
      }

      /// Clear
      void clear() {
	// Clear the blocks and vector part
	blocks_.clear();
	vector_.clear();
      }

      /// Insert
      void insert(const size_type _index, const value_type& _value) {
	// Index value
	if (_index < vector_.size() - 1)
	  // Error!
	  ttcl_fire("Cannot insert before the end");
	
	// Need to resize?
	if (_index >= vector_.size()) {
	  // Resize and start a new sector
	  vector_.resize(_index + 1);
	  blocks_.new_sector();
	}

	// Insert
	vector_.back() = blocks_.add_to_sector(_value);
      }
      
      /// Access
      value_type operator[](const size_type _index) {
	// Just access the vector
	return vector_[_index];
      }

      /// Const access
      const_value_type operator[](const size_type _index) const {
	// Just access the vector
	return vector_[_index];
      }

      /// Begin
      iterator begin() {
	// Just access the vector
	return vector_.begin();
      }

      /// Const begin
      const_iterator begin() const {
	// Just access the vector
	return vector_.begin();
      }

      /// End
      iterator end() {
	// Just access the vector
	return vector_.end();
      }

      /// Const end
      const_iterator end() const {
	// Just access the vector
	return vector_.end();
      }

      /// Reverse begin
      reverse_iterator rbegin() {
	// Just access the vector
	return vector_.rbegin();
      }

      /// Const reverse begin
      const_reverse_iterator rbegin() const {
	// Just access the vector
	return vector_.rbegin();
      }

      /// Reverse end
      reverse_iterator rend() {
	// Just access the vector
	return vector_.rend();
      }

      /// Const reverse end
      const_reverse_iterator rend() const {
	// Just access the vector
	return vector_.rend();
      }
    };
  }
}

#endif
