#ifndef TTCL_CO_NON_OWNER_ARRAY_HXX
#define TTCL_CO_NON_OWNER_ARRAY_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Fixed-size Non-Owner Array
    @author Edgar Gonzlez i Pellicer
*/

#include <algorithm>
#include <iterator>
#include <memory>

#include <boost/compressed_pair.hpp>


/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Fixed-size non-owner array
    /** (Almost) a model of STL Random Access Container
	@param T Value type
    */
    template <typename T>
    class non_owner_array {
    public:
      /// Value type
      typedef T value_type;

      /// Pointer type
      typedef T* pointer;

      /// Const pointer type
      typedef const T* const_pointer;

      /// Reference type
      typedef T& reference;

      /// Const reference
      typedef const T& const_reference;

      /// Iterator type
      typedef T* iterator;

      /// Const iterator type
      typedef const T* const_iterator;

      /// Reverse iterator type
      typedef std::reverse_iterator<iterator> reverse_iterator;

      /// Const reverse iterator type
      typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

      /// Distance type
      typedef ptrdiff_t difference_type;

      /// Size type
      typedef size_t size_type;
    
    protected:
      /// Size
      size_type size_;

      /// Data
      const value_type* data_;

    public:
      /// Default constructor
      non_owner_array() :
	size_(0), data_(0) {
      }

      /// Constructor
      /** @param _size Size
	  @param _data Data
       */
      non_owner_array(size_type _size, const value_type* _data) :
	size_(_size), data_(_data) {
      }

      /// Copy constructor
      /** @param _other Source array
       */
      non_owner_array(const non_owner_array& _other) :
	size_(_other.size_), data_(_other.data_) {
      }

      /// Constructor from other
      template <typename Container>
      non_owner_array(const Container& _other) :
	size_(_other.size()), data_(&_other.front()) {
      }

      /// Assignment operator
      non_owner_array& operator=(const non_owner_array& _other) {
	// Assign
	size_ = _other.size_;
	data_ = _other.data_;

	// Return this
	return *this;
      }

      /// Assignment from other  operator
      template <typename Container>
      non_owner_array& operator=(const Container& _other) {
	// Assign
	size_ = _other.size();
	data_ = &_other.front();

	// Return this
	return *this;
      }

      /// Beginning of range
      /** Const version
       */
      const_iterator begin() const {
	// Pointer to the data
	return data_;
      }

      /// End of range
      /** Const version
       */
      const_iterator end() const {
	// Pointer past the data
	return data_ + size_;
      }

      /// Begin of reverse range
      /** Const version
       */
      const_reverse_iterator rbegin() const {
	// Reverse of the end
	return reverse_iterator(end());
      }

      /// End of reverse range
      /** Const version
       */
      const_reverse_iterator rend() const {
	// Reverse of the begin
	return reverse_iterator(begin());
      }
      
      /// Size
      size_type size() const {
	// Size
	return size_;
      }

      /// Maximum size
      size_type max_size() const {
	// Size
	return size_;
      }

      /// Empty container
      bool empty() const {
	// Size is zero?
	return size_ == 0;
      }

      /// Clear
      void clear() {
	// Set
	size_ = 0;
	data_ = 0;
      }

      /// Element access
      /** const version
       */
      const_reference operator[](size_type _n) const {
	// Access the data
	return data_[_n];
      }
    
      /// Front
      /** const version
       */
      const_reference front() const {
	// Access the data
	return data_[0];
      }

      /// Back
      /** const version
       */
      const_reference back() const {
	// Access the data
	return data_[size_ - 1];
      }

      /// Swap
      /** @param _other Other array
       */
      void swap(non_owner_array& _other) {
	// Swap sizes and data
	std::swap(size_, _other.size_);
	std::swap(data_, _other.data_);
      }
    };

    /// Equality
    /** @param _a First array
	@param _b Second array
    */
    template <typename T>
    inline bool operator==(const non_owner_array<T>& _a,
			   const non_owner_array<T>& _b) {
      // Equal in size and content
      return (_a.size() == _b.size() and
	      std::equal(_a.begin(), _a.end(), _b.begin()));
    }

    /// Inequality
    /** Based on operator==
	@param _a First non_owner_array
	@param _b Second non_owner_array
    */
    template <typename T>
    inline bool operator!=(const non_owner_array<T>& _a,
			   const non_owner_array<T>& _b) {
      // Not equal
      return not (_a == _b);
    }

    /// Less
    /** @param _a First array
	@param _b Second array
    */
    template <typename T>
    inline bool operator<(const non_owner_array<T>& _a,
			  const non_owner_array<T>& _b) {
      // Lexicographical comparison
      return std::lexicographical_compare(_a.begin(), _a.end(),
					  _b.begin(), _b.end());
    }

    /// Greater
    /** Based on operator<
	@param _a First array
	@param _b Second array
    */
    template <typename T>
    inline bool operator>(const non_owner_array<T>& _a,
			  const non_owner_array<T>& _b) {
      // Inverted less
      return _b < _a;
    }

    /// Less or equal
    /** Based on operator<
	@param _a First array
	@param _b Second array
    */
    template <typename T>
    inline bool operator<=(const non_owner_array<T>& _a,
			   const non_owner_array<T>& _b) {
      // Not inverted less
      return not (_b < _a);
    }

    /// Greater or equal
    /** Based on operator<
	@param _a First array
	@param _b Second array
    */
    template <typename T>
    inline bool operator>=(const non_owner_array<T>& _a,
			   const non_owner_array<T>& _b) {
      // Not less
      return not (_a < _b);
    }
  }
};

/// Standard namespace
namespace std {
  /// Swap
  /** Template specialization
      @param _a First array
      @param _b Second array
  */
  template<typename T>
  void swap(ttcl::co::non_owner_array<T>& _a,
	    ttcl::co::non_owner_array<T>& _b) {
    // Swap
    _a.swap(_b);
  }
}

#endif
