#ifndef TTCL_RO_MMAP_VECTOR_HXX
#define TTCL_RO_MMAP_VECTOR_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Read-only vector through mmap
    @author Edgar Gonzlez i Pellicer
*/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <cstring>
#include <iterator>

#include <ttcl/sys_exception.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// Read-only vector through mmap
    /** Not a model of STL Random Access Container,
	due to constructors
	@param T Value type. Must be a simple type.
    */
    template <typename T>
    class ro_mmap_vector {
    public:
      /// Value type
      typedef T value_type;

      /// Const pointer type
      typedef const T* const_pointer;

      /// Const reference
      typedef const T& const_reference;

      /// Const iterator type
      typedef const T* const_iterator;

      /// Const reverse iterator type
      typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

      /// Distance type
      typedef ptrdiff_t difference_type;

      /// Size type
      typedef size_t size_type;

    private:
      /// File descriptor
      int fd_;

      /// Size
      size_type size_;

      /// File size
      size_type file_size_;

      /// Data pointer
      value_type* data_;

    public:
      /// Constructor
      ro_mmap_vector(int _fd)
	throw (exception) :
	fd_(_fd), size_(0), file_size_(0), data_(0) {
	// Check mode
	int flags = fcntl(fd_, F_GETFL);
	if (flags == -1)
	  ttcl_sys_fire("Cannot access file flags");
	if ((flags & O_ACCMODE) != O_RDONLY and
	    (flags & O_ACCMODE) != O_RDWR)
	  ttcl_fire("File is not open in read or read/write mode");

	// Get the file size
	off_t signed_size = lseek(fd_, 0, SEEK_END);
	if (signed_size == off_t(-1))
	  ttcl_sys_fire("Cannot find file size");
	file_size_ = signed_size;
	
	// Adjust size
	size_ = file_size_ / sizeof(value_type);

	// Mmap it
	void* raw = mmap(0, file_size_, PROT_READ, MAP_SHARED, fd_, 0);
	if (raw == (void*)-1)
	  ttcl_sys_fire("Cannot map file");
	data_ = static_cast<value_type*>(raw);
      }

      /// Copy constructor
      ro_mmap_vector(const ro_mmap_vector& _other) :
	fd_(_other.fd_), size_(_other.size_),
	file_size_(_other.file_size_), data_(0) {
	// Mmap it
	void* raw = mmap(0, file_size_, PROT_READ, MAP_SHARED, fd_, 0);
	if (raw == (void*)-1)
	  ttcl_sys_fire("Cannot map file");
	data_ = static_cast<value_type*>(raw);
      }

      /// Destructor
      ~ro_mmap_vector()
      throw (exception) {
	// Unmap
	if (munmap(data_, file_size_) == -1)
	  ttcl_sys_fire("Cannot unmap file");
      }

      /// Assignment operation
      ro_mmap_vector& operator=(const ro_mmap_vector& _other) {
	// Check for auto-assignment
	if (this != &_other) {
	  // Unmap
	  if (munmap(data_, file_size_) == -1)
	    ttcl_sys_fire("Cannot unmap file");
	  
	  // Field copy
	  fd_        = _other.fd_;
	  size_      = _other.size_;
	  file_size_ = _other.file_size_;

	  // Mmap it
	  void* raw = mmap(0, file_size_, PROT_READ, MAP_SHARED, fd_, 0);
	  if (raw == (void*)-1)
	    ttcl_sys_fire("Cannot map file");
	  data_ = static_cast<value_type*>(raw);
	}

	// Return this
	return *this;
      }

      /// Empty?
      bool empty() const {
	return size_ == 0;
      }

      /// Size
      size_type size() const {
	return size_;
      }

      /// Maximum size
      size_type max_size() const {
	return size_;
      }

      /// Capacity
      size_type capacity() const {
	return size_;
      }

      /// Beginning of range
      /** Const version
       */
      const_iterator begin() const {
	// Pointer to the data
	return data_;
      }

      /// End of range
      /** Const version
       */
      const_iterator end() const {
	// Pointer past the data
	return data_ + size_;
      }

      /// Begin of reverse range
      /** Const version
       */
      const_reverse_iterator rbegin() const {
	// Reverse of the end
	return reverse_iterator(end());
      }

      /// End of reverse range
      /** Const version
       */
      const_reverse_iterator rend() const {
	// Reverse of the begin
	return reverse_iterator(begin());
      }

      /// Element access
      /** const version
       */
      const_reference operator[](size_type _n) const {
	// Access the data
	return data_[_n];
      }
    
      /// Front
      /** const version
       */
      const_reference front() const {
	// Access the data
	return data_[0];
      }

      /// Back
      /** const version
       */
      const_reference back() const {
	// Access the data
	return data_[size_ - 1];
      }
    };
  }
}

#endif
