#ifndef TTCL_CO_SYMMETRIC_MATRIX_HXX
#define TTCL_CO_SYMMETRIC_MATRIX_HXX

// TTCL: The Template Clustering Library

/** @file
    Containers - Symmetric matrix
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/assert.hxx>
#include <ttcl/co/array.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Container Namespace
  namespace co {

    /// A symmetric matrix
    /** Storage order:
	(1,0)
	(2,0), (2,1)
	(3,0), (3,1), (3,2)
	...
	@param Container
     */
    template < typename Container = array<double> >
    class symmetric_matrix : public Container {
    protected:
      /// Base type
      typedef Container base_type;

    public:
      /// Value type
      typedef typename base_type::value_type value_type;

      /// Size type
      typedef typename base_type::size_type size_type;

      /// Reference
      typedef typename base_type::reference reference;

      /// Const reference
      typedef typename base_type::const_reference const_reference;

    protected:
      /// Size
      size_type size_;

    public:
      /// Empty constructor
      symmetric_matrix() :
	base_type(), size_(0) {
      }

      /// Constructor
      /** @param _size  Size
	  @param _value Value
       */
      symmetric_matrix(size_type _size,
		       const value_type& _value = value_type()) :
	base_type(_size * (_size - 1) / 2, _value), size_(_size) {
      }

      /// Copy constructor
      /** @param _other Source matrix
       */
      symmetric_matrix(const symmetric_matrix& _other) :
	base_type(_other), size_(_other.size_) {
      }

      /// Assignment operator
      symmetric_matrix& operator=(const symmetric_matrix& _other) {
	// Assign as a container
	this->base_type::operator=(_other);
	size_ = _other.size_;

	// Return this
	return *this;
      }

      /// Size
      size_type size() const {
	// Size
	return size_;
      }

      /// Data size
      size_type data_size() const {
	// The size of the container
	return base_type::size();
      }

      /// Resize
      void resize(size_type _size, const value_type& _value = value_type()) {
	// Base resize
	base_type::resize(_size * (_size - 1) / 2, _value);

	// Set
	size_ = _size;
      }

      /// Element access
      /** Non-const version
	  @param _row Row
	  @param _col Column
      */
      reference operator()(size_type _row, size_type _col) {
	ttcl_assert_ne(_row, _col);

	if (_row < _col)
	  return (*this)[_col * (_col - 1) / 2 + _row];
	else
	  return (*this)[_row * (_row - 1) / 2 + _col];
      }

      /// Element access
      /** const version
	  @param _row Row
	  @param _col Column
      */
      const_reference operator()(size_type _row, size_type _col) const {
	ttcl_assert_ne(_row, _col);

	if (_row < _col)
	  return (*this)[_col * (_col - 1) / 2 + _row];
	else
	  return (*this)[_row * (_row - 1) / 2 + _col];
      }
    };
  }
}

#endif
