#ifndef TTCL_CR_CALINSKI_CENTROID_SCORE_DATA_HXX
#define TTCL_CR_CALINSKI_CENTROID_SCORE_DATA_HXX

// TTCL: The Template Clustering Library

/** @file
    Criteria - Calinski-Harabasz Score - Centroids Based (For Data)
    @author Edgar Gonzlez i Pellicer
*/

#include <limits>
#include <list>
#include <vector>

#include <ttcl/types.hxx>
#include <ttcl/cr/choices.hxx>
#include <ttcl/cr/distance_data_criterion.hxx>
#include <ttcl/ut/arithmetical_traits.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/tu/tuning.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Criteria Namespace
  namespace cr {

    /// Calinski-Harabasz Score - Centroids Based (For Data)
    template <typename Data,
	      typename Distance,
	      typename Traits =
	        ut::arithmetical_traits<typename Data::value_type>,
	      typename Choices = tu::default_choices>
    class calinski_centroid_score_data :
      public distance_data_criterion<Data, Distance, Choices> {
    public:
      /// Base type
      typedef distance_data_criterion<Data, Distance, Choices> base_type;

      /// Data type
      typedef typename base_type::data_type data_type;

      /// Distance type
      typedef typename base_type::distance_type distance_type;

      /// Traits type
      typedef Traits traits_type;

      /// Value type
      typedef typename base_type::value_type value_type;

      /// Hard clustering type
      typedef typename base_type::hard_clustering_type hard_clustering_type;

      /// Dendrogram type
      typedef typename base_type::dendrogram_type dendrogram_type;

    protected:
      /// Use non-clustered
      bool use_non_clustered_;

    public:
      /// Constructor
      calinski_centroid_score_data(const distance_type& _distance =
				     distance_type()) :
	base_type(_distance), use_non_clustered_(true) {
      }

      /// Tune
      void
      tune(const tu::tuning& _tuning) {
	// Extract the option
	use_non_clustered_ = _tuning.get<use_non_clustered>();
      }

      /// Find the criterion value
      virtual value_type
      hard_apply(const hard_clustering_type& _hard_clustering,
		 const data_type& _data) {
	// Object type
	typedef typename data_type::value_type object_type;

	// Number of clusters
	uint k = _hard_clustering.clusters();

	// Equal to one or data?
	if (k == 1 or k == _data.size())
	  return std::numeric_limits<value_type>::quiet_NaN();

	// Centroids
	object_type metacentroid = traits_type::zero();
	std::vector<object_type> centroids(k, traits_type::zero());
	std::vector<uint>        sizes    (k, 0);

	// Find them
	uint i     = 0;
	uint total = 0;
	for (typename data_type::const_iterator it = _data.begin();
	     it != _data.end(); ++it, ++i) {
	  int cl = _hard_clustering[i];
	  if (cl != -1) {
	    // Add to metacentroid and centroid
	    traits_type::plus_eq(metacentroid,  *it);
	    ++total;
	    traits_type::plus_eq(centroids[cl], *it);
	    ++sizes[cl];
	  }
	  else if (use_non_clustered_) {
	    // Add to metacentroid
	    traits_type::plus_eq(metacentroid,  *it);
	    ++total;
	  }
	}

	// Normalize
	traits_type::div_eq(metacentroid, total);
	for (uint i = 0; i < k; ++i)
	  traits_type::div_eq(centroids[i], sizes[i]);

	// Between distances
	value_type between = value_type();
	for (uint i = 0; i < k; ++i) {
	  value_type dist = this->distance_(metacentroid, centroids[i]);
	  between        += value_type(sizes[i]) * dist * dist;
	}

	// Between distances
	value_type within = value_type();
	i     = 0;
	total = 0;
	for (typename data_type::const_iterator it = _data.begin();
	     it != _data.end(); ++it, ++i) {
	  int cl = _hard_clustering[i];
	  if (cl != -1) {
	    value_type dist = this->distance_(centroids[cl], *it);
	    within         += dist * dist;
	    ++total;
	  }
	}

	// Return the value
	return (between * value_type(total - k)) /
	       (within  * value_type(k - 1));
      }

      /// Find the criterion value for a tree
      virtual std::vector<value_type>
      hier_apply(const dendrogram_type& _dendrogram,
		 const data_type& _data) {
	// Object type
	typedef typename data_type::value_type object_type;

	// Size
	int total = _data.size();

	// Between/Within
	value_type between = value_type();
	value_type within  = value_type();

	// Centroids and elements
	object_type metacentroid = traits_type::zero();
	std::vector<object_type>       centroids;
	std::vector<object_type>       raw_centroids;
	std::vector<uint>              sizes(total, 1);
	std::vector< std::list<uint> > elements;
	std::vector<uint>              mapping;

	// Reserve
	centroids    .reserve(total);
	raw_centroids.reserve(total);
	elements     .reserve(total);
	mapping      .reserve(total);

	// Initialize
	int i = 0;
	for (typename data_type::const_iterator it = _data.begin();
	     it != _data.end(); ++it, ++i) {
	  // Add to metacentroid
	  traits_type::plus_eq(metacentroid, *it);

	  // Each object is its own cluster;
	  centroids    .push_back(*it);
	  raw_centroids.push_back(*it);
	  // sizes     .push_back(1);
	  elements     .push_back(std::list<uint>(1, i));
	  mapping      .push_back(i);
	}

	// Normalize
	traits_type::div_eq(metacentroid, total);

	// Cluster betweens and withins
	std::vector<value_type> cl_between;
	std::vector<value_type> cl_within(total, value_type());

	// Reserve
	cl_between.reserve(total);

	// Starting value
	for (typename data_type::const_iterator it = _data.begin();
	     it != _data.end(); ++it) {
	  value_type dist = this->distance_(metacentroid, *it);
	  dist    *= dist;
	  between += dist;
	  cl_between.push_back(dist);
	}

	// Return vector
	std::vector<value_type> ret
	  (total, std::numeric_limits<value_type>::quiet_NaN());

	// For each level of the dendrogram
	uint remain = total - 1;
	for (uint j = _dendrogram.size();
	     j < _dendrogram.branches() and remain > 1; ++j, --remain) {
	  // The two to merge
	  int map1 = mapping[_dendrogram[j].first];
	  int map2 = mapping[_dendrogram[j].second];
	  mapping.push_back(map1);

	  // Remove
	  between -= cl_between[map1] + cl_between[map2];
	  within  -= cl_within [map1] + cl_within [map2];

	  // Merge
	  traits_type::plus_eq(raw_centroids[map1], raw_centroids[map2]);
	  sizes[map1]    += sizes[map2];
	  centroids[map1] = raw_centroids[map1];
	  traits_type::div_eq(centroids[map1], sizes[map1]);
	  elements[map1].splice(elements[map1].end(), elements[map2]);

	  // Clear unused
	  traits_type::zero(raw_centroids[map2]);
	  traits_type::zero(centroids[map2]);

	  // New between
	  value_type dist  = this->distance_(metacentroid, centroids[map1]);
	  dist            *= value_type(sizes[map1]) * dist;
	  between         += dist;
	  cl_between[map1] = dist;

	  // New within
	  value_type new_within = value_type();
	  for (std::list<uint>::const_iterator it = elements[map1].begin();
	       it != elements[map1].end(); ++it) {
	    dist        = this->distance_(centroids[map1], _data[*it]);
	    new_within += dist * dist;
	  }
	  within         += new_within;
	  cl_within[map1] = new_within;

	  // Unmap
	  mapping[_dendrogram[j].first]  = -1;
	  mapping[_dendrogram[j].second] = -1;

	  // Now, calinski
	  value_type cal = (between * value_type(total - remain)) /
	                   (within  * value_type(remain - 1));
	  ret[remain - 1] = cal;
	}

	// Return
	return ret;
      }
    };
  }
}

#endif
