#ifndef TTCL_CR_CALINSKI_MEDOID_MATRIX_SCORE_HXX
#define TTCL_CR_CALINSKI_MEDOID_MATRIX_SCORE_HXX

// TTCL: The Template Clustering Library

/** @file
    Criteria - Calinski-Harabasz Score - Medoids Based (For Matrix)
    @author Edgar Gonzlez i Pellicer
*/

#include <limits>
#include <list>
#include <vector>

#include <ttcl/assert.hxx>
#include <ttcl/types.hxx>
#include <ttcl/cr/choices.hxx>
#include <ttcl/cr/matrix_criterion.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/tu/tuning.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Criteria Namespace
  namespace cr {

    /// Calinski-Harabasz Score - Medoids Based (For Matrix)
    template <typename Matrix = co::symmetric_matrix< co::array<double> >,
	      typename Choices = tu::default_choices>
    class calinski_medoid_score_matrix :
      public matrix_criterion<Matrix, Choices> {
    public:
      /// Base type
      typedef matrix_criterion<Matrix, Choices> base_type;

      /// Matrix type
      typedef typename base_type::matrix_type matrix_type;

      /// Value type
      typedef typename base_type::value_type value_type;

      /// Hard clustering type
      typedef typename base_type::hard_clustering_type hard_clustering_type;

      /// Dendrogram type
      typedef typename base_type::dendrogram_type dendrogram_type;

    protected:
      /// Use non-clustered
      bool use_non_clustered_;

    public:
      /// Constructor
      calinski_medoid_score_matrix() :
	base_type(), use_non_clustered_(true) {
      }

      /// Tune
      void
      tune(const tu::tuning& _tuning) {
	// Extract the option
	use_non_clustered_ = _tuning.get<use_non_clustered>();
      }

      /// Find the criterion value
      virtual value_type
      hard_apply(const hard_clustering_type& _hard_clustering,
		 const matrix_type& _matrix) {
	// Asserts
	ttcl_assert_eq(_hard_clustering.size(), _matrix.size());

	// Cluster and size type
	typedef typename hard_clustering_type::value_type cluster_type;
	typedef typename hard_clustering_type::size_type  size_type;

	// Number of elements and clusters
	size_type n = _hard_clustering.size();
	size_type k = _hard_clustering.clusters();

	// Equal to one or data?
	if (k == 1 or k == n)
	  return std::numeric_limits<value_type>::quiet_NaN();

	// Cluster sizes
	std::vector<size_type> sizes_cl(k, size_type());

	// Sums of distances
	std::vector<value_type> sum_dist_all  (n, value_type());
	std::vector<value_type> sum_dist_cl   (n, value_type());
	std::vector<value_type> sum_dist_cl_sq(n, value_type());

	// Global best -> Metacentroid
	value_type best_dist_all = std::numeric_limits<value_type>::max();
	size_type  best_idx_all  = size_type(-1);

	// Best in each cluster -> Centroids
	std::vector<value_type> best_dist_cl
	  (k, std::numeric_limits<value_type>::max());
	std::vector<size_type>  best_idx_cl(k, size_type(-1));

	// Total
	size_type total = size_type();

	// For each one
	for (size_type i = 0; i < n; ++i) {
	  // Clustered?
	  cluster_type cl_i = _hard_clustering[i];
	  if (cl_i != -1) {
	    // Clustered

	    // Count for the size
	    ++total;
	    ++sizes_cl[cl_i];

	    // For each other
	    for (size_type j = i + 1; j < n; ++j) {
	      // Distance between them
	      sum_dist_all[i] += _matrix(i, j);
	      sum_dist_all[j] += _matrix(i, j);

	      // In the same cluster?
	      if (_hard_clustering[j] == cl_i) {
		// Square
		double sq_matrix_ij = _matrix(i, j) * _matrix(i, j);

		// Distance between them
		sum_dist_cl   [i] += _matrix(i, j);
		sum_dist_cl   [j] += _matrix(i, j);
		sum_dist_cl_sq[i] += sq_matrix_ij;
		sum_dist_cl_sq[j] += sq_matrix_ij;
	      }
	    }

	    // Is it the best one?

	    // Overall
	    if (sum_dist_all[i] < best_dist_all) {
	      best_dist_all = sum_dist_all[i];
	      best_idx_all  = i;
	    }

	    // For its cluster
	    if (sum_dist_cl[i] < best_dist_cl[cl_i]) {
	      best_dist_cl[cl_i] = sum_dist_cl[i];
	      best_idx_cl [cl_i] = i;
	    }
	  }
	  else if (use_non_clustered_) {
	    // Non clustered, but we use them anyway

	    // Count for the size
	    ++total;

	    // For each other
	    for (size_type j = i + 1; j < n; ++j) {
	      // Distance between them
	      sum_dist_all[i] += _matrix(i, j);
	      sum_dist_all[j] += _matrix(i, j);
	    }

	    // Is it the best one?

	    // Overall
	    if (sum_dist_all[i] < best_dist_all) {
	      best_dist_all = sum_dist_all[i];
	      best_idx_all  = i;
	    }
	  }
	}

	// Between and within distances
	value_type between = value_type();
	value_type within  = value_type();

	// For each cluster
	for (cluster_type c = 0; c < k; ++c) {
	  // Add to within (sum of distances to centroid)
	  within += sum_dist_cl_sq[best_idx_cl[c]];

	  // Add to between (sum of distances to metacentroid)
	  if (best_idx_all != best_idx_cl[c])
	    between += value_type(sizes_cl[c]) *
	               _matrix(best_idx_all, best_idx_cl[c]) *
	               _matrix(best_idx_all, best_idx_cl[c]);
	}

	// Return the value
	return (between * value_type(total - k)) /
	       (within  * value_type(k - 1));
      }

      /// Find the criterion value for a tree
      virtual std::vector<value_type>
      hier_apply(const dendrogram_type& _dendrogram,
		 const matrix_type& _matrix) {
	// Asserts
	ttcl_assert_eq(_hard_clustering.size(), _matrix.size());

	// Cluster and size type
	typedef typename hard_clustering_type::value_type cluster_type;
	typedef typename hard_clustering_type::size_type  size_type;

	// Size
	size_type total = _matrix.size();

	// Sums of distances
	std::vector<value_type> sum_dist_all  (total, value_type());
	std::vector<value_type> sum_dist_cl   (total, value_type());
	std::vector<value_type> sum_dist_cl_sq(total, value_type());

	// Global best -> Metacentroid
	value_type best_dist_all = std::numeric_limits<value_type>::max();
	size_type  best_idx_all  = size_type(-1);

	// Cluster information
	std::vector<size_type>         sizes_cl(total, 1);
	std::vector< std::list<uint> > elements;
	std::vector<uint>              mapping;

	// Reserve
	elements.reserve(total);
	mapping .reserve(total);

	// Initialize, and find metacentroid
	for (size_type i = 0; i < total; ++i) {
	  // Each object is its own cluster;
	  elements.push_back(std::list<uint>(1, i));
	  mapping .push_back(i);

	  // For each other
	  for (size_type j = i + 1; j < total; ++j) {
	    // Add
	    sum_dist_all[i] += _matrix(i, j);
	    sum_dist_all[j] += _matrix(i, j);
	  }

	  // Is it the best one overall?
	  if (sum_dist_all[i] < best_dist_all) {
	    best_dist_all = sum_dist_all[i];
	    best_idx_all  = i;
	  }
	}

	// Between/Within
	value_type between = value_type();
	value_type within  = value_type();

	// Cluster betweens and withins
	std::vector<value_type> cl_between;
	std::vector<value_type> cl_within(total, value_type());

	// Reserve
	cl_between.reserve(total);

	// Initialize
	for (size_type i = 0; i < total; ++i) {
	  // Distance to meta-centroid
	  double new_between;
	  if (best_idx_all == i)
	    new_between = value_type();
	  else
	    new_between =_matrix(best_idx_all, i) * _matrix(best_idx_all, i);

	  // Add to between
	  between += new_between;
	  cl_between.push_back(new_between);
	}

	// Return vector
	std::vector<value_type> ret
	  (total, std::numeric_limits<value_type>::quiet_NaN());

	// For each level of the dendrogram
	uint remain = total - 1;
	for (uint j = _dendrogram.size();
	     j < _dendrogram.branches() and remain > 1; ++j, --remain) {
	  // The two to merge
	  int map1 = mapping[_dendrogram[j].first];
	  int map2 = mapping[_dendrogram[j].second];
	  mapping.push_back(map1);

	  // Remove
	  between -= cl_between[map1] + cl_between[map2];
	  within  -= cl_within [map1] + cl_within [map2];

	  // Update distances
	  for (std::list<uint>::const_iterator it1 = elements[map1].begin();
	       it1 != elements[map1].end(); ++it1) {
	    for (std::list<uint>::const_iterator it2 = elements[map2].begin();
		 it2 != elements[map2].end(); ++it2) {
	      // Square
	      double sq_matrix_12 = _matrix(*it1, *it2) * _matrix(*it1, *it2);

	      // Distance between them
	      sum_dist_cl   [*it1] += _matrix(*it1, *it2);
	      sum_dist_cl   [*it2] += _matrix(*it1, *it2);
	      sum_dist_cl_sq[*it1] += sq_matrix_12;
	      sum_dist_cl_sq[*it2] += sq_matrix_12;
	    }
	  }

	  // Merge
	  sizes_cl[map1] += sizes_cl[map2];
	  elements[map1]  . splice(elements[map1].end(), elements[map2]);

	  // Reset best
	  value_type best_dist_cl = std::numeric_limits<value_type>::max();
	  size_type  best_idx_cl  = size_type(-1);

	  // Find it
	  for (std::list<uint>::const_iterator it = elements[map1].begin();
	       it != elements[map1].end(); ++it) {
	    if (sum_dist_cl[*it] < best_dist_cl) {
	      best_dist_cl = sum_dist_cl[*it];
	      best_idx_cl  = *it;
	    }
	  }

	  // The new within
	  within         += sum_dist_cl_sq[best_idx_cl];
	  cl_within[map1] = sum_dist_cl_sq[best_idx_cl];

	  // New between
	  double new_between;
	  if (best_idx_all == best_idx_cl)
	    new_between = value_type();
	  else
	    new_between = value_type(sizes_cl[map1]) *
	                  _matrix(best_idx_all, best_idx_cl) *
	                  _matrix(best_idx_all, best_idx_cl);

	  // Add it
	  between          += new_between;
	  cl_between[map1]  = new_between;

	  // Unmap
	  mapping[_dendrogram[j].first]  = -1;
	  mapping[_dendrogram[j].second] = -1;

	  // Now, calinski
	  value_type cal = (between * value_type(total - remain)) /
	                   (within  * value_type(remain - 1));
	  ret[remain - 1] = cal;
	}

	// Return
	return ret;
      }
    };
  }
}

#endif
