#ifndef TTCL_CR_SILHOUETTE_WIDTH_MATRIX_HXX
#define TTCL_CR_SILHOUETTE_WIDTH_MATRIX_HXX

// TTCL: The Template Clustering Library

/** @file
    Criteria - Silhouette Width (For Matrices)
    @author Edgar Gonzlez i Pellicer
*/

#include <algorithm>
#include <limits>
#include <list>
#include <vector>

#include <ttcl/types.hxx>
#include <ttcl/cr/choices.hxx>
#include <ttcl/cr/matrix_criterion.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Criteria Namespace
  namespace cr {

    /// Silhouette Width (For Matrices)
    template <typename Matrix = co::symmetric_matrix< co::array<double> >,
	      typename Choices = tu::default_choices>
    class silhouette_width_matrix :
      public matrix_criterion<Matrix, Choices> {
    public:
      /// Base type
      typedef matrix_criterion<Matrix, Choices> base_type;

      /// Data type
      typedef typename base_type::matrix_type matrix_type;

      /// Value type
      typedef typename base_type::value_type value_type;

      /// Hard clustering type
      typedef typename base_type::hard_clustering_type hard_clustering_type;

      /// Dendrogram type
      typedef typename base_type::dendrogram_type dendrogram_type;

      /// Point to cluster matrix type
      typedef ttcl_tu_get_choice_tem(Choices, point_to_cluster_template,
				     value_type) point_to_cluster_type;

    public:
      /// Find the criterion value
      virtual value_type
      hard_apply(const hard_clustering_type& _hard_clustering,
		 const matrix_type& _matrix) {
	// Point to cluster matrix
	uint size    = _hard_clustering.size();
	int clusters = _hard_clustering.clusters();
	point_to_cluster_type point_to_cluster(size, clusters);
	for (uint i = 0; i < size; ++i)
	  for (int c = 0; c < clusters; ++c)
	    point_to_cluster[i][c] = value_type();

	// Cluster sizes
	std::vector<uint> cl_size(clusters, 0);

	// Loop
	for (uint i = 0; i < size - 1; ++i) {
	  int cli = _hard_clustering[i];
	  if (cli != -1) {
	    ++cl_size[cli];
	    for (uint j = i + 1; j < size; ++j) {
	      int clj = _hard_clustering[j];
	      if (clj != -1) {
		point_to_cluster[i][clj] += _matrix(i, j);
		point_to_cluster[j][cli] += _matrix(i, j);
	      }
	    }
	  }
	}

	// Now, silhouette for each one
	value_type silhouette = value_type();
	uint total = 0;
	for (uint i = 0; i < size; ++i) {
	  int cli = _hard_clustering[i];
	  if (cli != -1) {
	    // Find the average within its cluster
	    value_type a = point_to_cluster[i][cli];
	    if (cl_size[cli] > 1)
	      a /= value_type(cl_size[cli] - 1);

	    // Find the minimum average between other clusters
	    value_type b = std::numeric_limits<value_type>::max();
	    for (int cl = 0; cl < clusters; ++cl) {
	      if (cl != cli) {
		value_type dist =
		  point_to_cluster[i][cl] / value_type(cl_size[cl]);
		if (dist < b)
		  b = dist;
	      }
	    }

	    // Add the silhouette value
	    silhouette += (b - a) / std::max(a, b);
	    ++total;
	  }
	}

	// Average silhouette
	return silhouette / value_type(total);
      }
    };
  }
}

#endif
