#ifndef TTCL_CR_WR_DATA_CRITERION_WRAP_HXX
#define TTCL_CL_WR_DATA_CRITERION_WRAP_HXX

// TTCL: The Template Clustering Library

/** @file
    Criteria - Data criterion wrapper
    @author Edgar Gonzalez i Pellicer
*/

#include <vector>

#include <ttcl/types.hxx>
#include <ttcl/cl/choices.hxx>
#include <ttcl/cl/hard_clustering.hxx>
#include <ttcl/cl/wr/data_wrap.hxx>
#include <ttcl/hj/choices.hxx>
#include <ttcl/hj/dendrogram.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/wr/interface_wrap.hxx>
#include <ttcl/wr/shared_referee.hxx>
#include <ttcl/wr/typed_shared_referee.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Criteria Namespace
  namespace cr {

    /// Wrappers Namespace
    namespace wr {

      /// Data criterion interface
      template <typename Choices>
      class _data_criterion_interface :
	public ttcl::wr::shared_referee {
      public:
	/// Value type
	typedef double value_type;

	/// Hard clustering container
	typedef ttcl_tu_get_choice(Choices, cl::hard_clustering_container)
	  hard_clustering_container_type;

	/// Hard clustering type
	typedef cl::hard_clustering<hard_clustering_container_type>
	  hard_clustering_type;

	/// Dendrogram value type
	typedef ttcl_tu_get_choice(Choices, hj::dendrogram_value)
	  dendrogram_value_type;

	/// Dendrogram type
	typedef hj::dendrogram<value_type> dendrogram_type;

	/// Constructor
	_data_criterion_interface(bool _owned = true,
				      uint _references = 0) :
	  ttcl::wr::shared_referee(_owned, _references) {
	}

	/// Find the criterion value
	virtual value_type
	hard_apply(const hard_clustering_type& _hard_clustering,
		   const cl::wr::data_wrap& _data) = 0;

	/// Find the criterion value for a tree
	virtual std::vector<value_type>
	hier_apply(const dendrogram_type& _dendrogram,
		   const cl::wr::data_wrap& _data) = 0;
      };

      /// Data criterion referee
      template <typename T,
		typename Interface>
      class _data_criterion_referee :
	public ttcl::wr::typed_shared_referee<T, Interface> {
      private:
	// Value type
	typedef typename T::value_type value_type;

	// Data type
	typedef typename T::data_type data_type;

	/// Hard clustering type
	typedef typename Interface::hard_clustering_type hard_clustering_type;

	/// Dendrogram type
	typedef typename Interface::dendrogram_type dendrogram_type;

      public:
	/// Constructor
	_data_criterion_referee(T* _ptr = 0, bool _owned = true) :
	  ttcl::wr::typed_shared_referee<T, Interface>(_ptr, _owned) {
	}

	/// Find the criterion value
	virtual value_type
	hard_apply(const hard_clustering_type& _hard_clustering,
		   const cl::wr::data_wrap& _data) {
	  // Convert to data type
	  const data_type* data = _data.get<data_type>();

	  // Call
	  return this->ptr_->hard_apply(_hard_clustering, *data);
	}

	/// Find the criterion value for a tree
	virtual std::vector<value_type>
	hier_apply(const dendrogram_type& _dendrogram,
		   const cl::wr::data_wrap& _data) {
	  // Convert to data type
	  const data_type* data = _data.get<data_type>();

	  // Call
	  return this->ptr_->hier_apply(_dendrogram, *data);
	}
      };

      /// Base data criterion wrapper
      template <typename Interface>
      class _data_criterion_wrap :
	public ttcl::wr::interface_wrap<Interface> {
      public:
	/// Base type
	typedef ttcl::wr::interface_wrap<Interface> base_type;

	// Value type
	typedef typename Interface::value_type value_type;

	/// Hard clustering type
	typedef typename Interface::hard_clustering_type hard_clustering_type;

	/// Dendrogram type
	typedef typename Interface::dendrogram_type dendrogram_type;

      protected:
	/// Empty constructor
	_data_criterion_wrap() :
	  base_type() {
	}

	/// Constructor
	_data_criterion_wrap(Interface* _ptr) :
	  base_type(_ptr) {
	}

      public:
	/// Find the criterion value
	value_type
	hard_apply(const hard_clustering_type& _hard_clustering,
		   const cl::wr::data_wrap& _data) {
	  // Call
	  return this->ptr_->hard_apply(_hard_clustering, _data);
	}

	/// Find the criterion value for a tree
	std::vector<value_type>
	hier_apply(const dendrogram_type& _dendrogram,
		   const cl::wr::data_wrap& _data) {
	  // Call
	  return this->ptr_->hier_apply(_dendrogram, _data);
	}
      };

      /// Hard data clusterer wrapper
      template <typename Choices = tu::default_choices>
      class data_criterion_wrap :
	public _data_criterion_wrap
          < _data_criterion_interface<Choices> > {
      private:
	/// Interface
	typedef _data_criterion_interface<Choices> interface;

	/// Base type
	typedef _data_criterion_wrap<interface> base_type;

      public:
	/// Empty Constructor
	data_criterion_wrap() :
	  base_type() {
	}

	/// Constructor
	template <typename T>
	data_criterion_wrap(T* _value, bool _owned = true) :
	  base_type(new _data_criterion_referee<T, interface>(_value,
							      _owned)) {
	}

	/// Constructor
	template <typename T>
	data_criterion_wrap(T& _value, bool _owned = false) :
	  base_type(new _data_criterion_referee<T, interface>(&_value,
							      _owned)) {
	}
      };
    }
  }
};

#endif
