#ifndef TTCL_HJ_AG_AGGLOMERATIVE_CLUSTERER_HXX
#define TTCL_HJ_AG_AGGLOMERATIVE_CLUSTERER_HXX

// TTCL: The Template Clustering Library

/** @file
    Hierarchical agglomerative clusterer
    @author Edgar Gonzalez i Pellicer
*/

#include <vector>

#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>

#include <ttcl/hj/hier_matrix_clusterer.hxx>
#include <ttcl/hj/ag/choices.hxx>
#include <ttcl/hj/ag/single_link.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/tu/tuning.hxx>
#include <ttcl/types.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Hierarchical Clustering Namespace
  namespace hj {

    /// Agglomerative Clustering Namespace
    namespace ag {

      /// Hierarchical agglomerative clusterer
      /** Matricial version
	  @param Matrix Matrix type
	  @param Policy Merging policy
      */
      template <typename Matrix,
		typename Policy = single_link,
		typename Choices = tu::default_choices>
      class agglomerative_clusterer :
      public hier_matrix_clusterer<Matrix, Choices> {
      public:
	/// Base type
	typedef hier_matrix_clusterer<Matrix, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Dendrogram type
	typedef typename base_type::dendrogram_type dendrogram_type;

	/// Distance matrix type
	typedef typename base_type::matrix_type matrix_type;

	/// Policy type
	typedef Policy policy_type;

      private:
	/// Policy
	policy_type policy_;

	/// Merging threshold
	value_type merging_threshold_;

      public:
	/// Constructor
	agglomerative_clusterer(const policy_type& _policy = policy_type()) :
	  policy_(_policy), merging_threshold_(default_merging_threshold) {
	}

	/// Tune
	void
	tune(const tu::tuning& _tuning) {
	  // Get
	  merging_threshold_ = _tuning.get<merging_threshold>();
	}

	/// Perform the clustering
	virtual void
	hier_cluster_inplace_on(dendrogram_type& _dendrogram,
				matrix_type& _distances,
				const uint _n_clusters = 1) const {
	  // Data size
	  typename matrix_type::size_type n_data = _distances.size();

	  // Heap types
	  typedef boost::tuple<value_type, uint, uint> element_type;
	  typedef ttcl_tu_get_choice_tem(Choices,
					 queue_template,
					 element_type) heap_type;

	  // Fill the heap
	  heap_type heap;
	  typename matrix_type::const_iterator it_dist =
	    _distances.begin();
	  for (uint i = 0; i < n_data; ++i)
	    for (uint j = 0; j < i; ++j)
	      heap.push(boost::make_tuple(*it_dist++, i, j));

	  // Create the mapping and the sizes
	  std::vector<int> mapping;
	  std::vector<uint> sizes;
	  for (uint i = 0; i < n_data; ++i) {
	    mapping.push_back(i);
	    sizes.push_back(1);
	  }

	  // Initialize the dendrogram
	  _dendrogram.reset(n_data);

	  // Loop while something remains
	  uint remaining = n_data;
	  while (remaining > _n_clusters) {
	    // Get the top
	    value_type dist;
	    uint first, second;
	    do {
	      boost::tie(dist, first, second) = heap.top();
	      heap.pop();
	    } while (mapping[first] == -1 or mapping[second] == -1);

	    // Above merging threshold?
	    if (dist > merging_threshold_)
	      break;

	    // Record the merge
	    uint parent = _dendrogram.merge(first, second, dist);
	    mapping.push_back(mapping[first]);

	    // Update distances
	    for (uint i = 0; i < parent; ++i) {
	      if (i != first and i != second and mapping[i] != -1) {
		// New distance
		value_type new_dist = policy_(_distances(mapping[first],
							 mapping[i]),
					      value_type(sizes[mapping[first]]),
					      _distances(mapping[second],
							 mapping[i]),
					      value_type(sizes[mapping[second]]));
		heap.push(boost::make_tuple((_distances(mapping[i],
							mapping[parent]) =
					     new_dist), parent, i));
	      }
	    }

	    // Add sizes and unmap
	    sizes[mapping[parent]] =
	      sizes[mapping[first]] + sizes[mapping[second]];
	    mapping[first]  = -1;
	    mapping[second] = -1;

	    // One less
	    --remaining;
	  }
	}
      };
    }
  }
}

#endif
