#ifndef TTCL_HJ_AG_LIMIT_AGGLOMERATIVE_DISTANCE_CLUSTERER_HXX
#define TTCL_HJ_AG_LIMIT_AGGLOMERATIVE_DISTANCE_CLUSTERER_HXX

// TTCL: The Template Clustering Library

/** @file
    Hierarchical agglomerative clusterer distance version
    @author Edgar Gonzalez i Pellicer
*/

#include <vector>

#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>

#include <ttcl/co/followable_mf_set.hxx>
#include <ttcl/co/limit_priority_queue.hxx>
#include <ttcl/hj/hier_distance_clusterer.hxx>
#include <ttcl/hj/ag/choices.hxx>
#include <ttcl/hj/ag/single_link.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/tu/tuning.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Hierarchical Clustering Namespace
  namespace hj {
    
    /// Agglomerative Clustering Namespace
    namespace ag {

      /// Hierarchical agglomerative clusterer
      /** Distance version with a limit_priority_queue
       */
      template <typename Data,
		typename Dist,
		typename Policy = single_link,
		typename Choices = tu::default_choices>
      class limit_agglomerative_distance_clusterer :
	public hier_distance_clusterer<Data, Dist, Choices> {
      public:
	/// Base type
	typedef hier_distance_clusterer<Data, Dist, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Dendrogram type
	typedef typename base_type::dendrogram_type dendrogram_type;

	/// Distance type
	typedef typename base_type::distance_type distance_type;

	/// Policy type
	typedef Policy policy_type;
      
      protected:
	/// Queue element type
	typedef typename boost::tuple<value_type, uint, uint>
	  queue_element_type;
	
	/// Queue type
	typedef ttcl_tu_get_choice_tem(Choices,
				       limit_queue_template,
				       queue_element_type) queue_type;

	/// Policy
	policy_type policy_;
	
	/// Queue size
	typename queue_type::size_type queue_size_;
	
      public:
	/// Constructor
	limit_agglomerative_distance_clusterer(const distance_type& _distance =
					         distance_type(),
					       const policy_type& _policy =
					         policy_type(),
					       const tu::tuning& _tuning =
					         tu::tuning()) :
	  base_type(_distance), policy_(_policy) {
	  // Tune
	  tune(_tuning);
	}

	/// Tune
	void
	tune(const tu::tuning& _tuning) {
	  // Extract the option
	  queue_size_ = _tuning.get<limit_queue_size>();
	}

	/// Perform the clustering
	/** @param _dendrogram Target dendrogram
	    @param _data Data
	*/
	virtual void
	hier_cluster_on(dendrogram_type& _dendrogram,
			const data_type& _data,
			const uint _n_clusters = 1) const {
	  // Data size
	  typename data_type::size_type n_data = _data.size();

	  // Create the mapping and the sizes
	  std::vector<int> mapping;
	  for (uint i = 0; i < n_data; ++i)
	    mapping.push_back(i);

	  // Create the partition
	  co::followable_mf_set partition(n_data);

	  // Fill the heap
	  queue_type heap(queue_size_);
	  fill_queue(heap, mapping, partition, _data);

	  // Initialize the dendrogram
	  _dendrogram.reset(n_data);

	  // Loop while something remains
	  uint remaining = n_data;
	  while (remaining > _n_clusters) {
	    // Get the top
	    value_type dist;
	    uint first, second;
	    do {
	      // Refill
	      if (heap.empty())
		fill_queue(heap, mapping, partition, _data);

	      // Look for a top
	      do {
		boost::tie(dist, first, second) = heap.top();
		heap.pop();
	      } while (not heap.empty() and
		       (mapping[first] == -1 or mapping[second] == -1));
	    } while (mapping[first] == -1 or mapping[second] == -1);

	    // Record the merge
	    uint parent = _dendrogram.merge(first, second, dist);
	    partition.merge(mapping[first], mapping[second]);
	    mapping.push_back(mapping[first]);

	    // Update distances
	    for (uint i = 0; i < parent; ++i) {
	      if (i != first and i != second and mapping[i] != -1) {
		// New distance
		value_type new_dist = policy_(this->distance_,
					      _data,
					      partition.begin(mapping[parent]),
					      partition.end  (mapping[parent]),
					      partition.begin(mapping[i]),
					      partition.end  (mapping[i]));
		heap.push(boost::make_tuple(new_dist, parent, i));
	      }
	    }

	    // Unmap
	    mapping[first]  = -1;
	    mapping[second] = -1;

	    // One less
	    --remaining;
	  }
	}

      private:
	/// Fill the queue
	void
	fill_queue(queue_type& _heap,
		   const std::vector<int> _mapping,
		   const co::followable_mf_set& _partition,
		   const data_type& _data) const {
	  // Find distances for each pair
	  _heap.clear();
	  for (uint i = 0; i < _data.size(); ++i) {
	    if (_mapping[i] != -1) {
	      for (uint j = 0; j < i; ++j) {
		if (_mapping[j] != -1) {
		  // New distance
		  value_type new_dist = policy_(this->distance_,
						_data,
						_partition.begin(_mapping[i]),
						_partition.end  (_mapping[i]),
						_partition.begin(_mapping[j]),
						_partition.end  (_mapping[j]));
		  _heap.push(boost::make_tuple(new_dist, i, j));
		}
	      }
	    }
	  }
	}
      };
    }
  }
}


#endif
