#ifndef TTCL_HJ_DENDROGRAM_HXX
#define TTCL_HJ_DENDROGRAM_HXX

// TTCL: The Template Clustering Library

/** @file
    Hierarchical Clustering - Dendrogram
    @author Edgar Gonzalez i Pellicer
*/

#include <map>
#include <stack>
#include <vector>

#include <ttcl/types.hxx>
#include <ttcl/cl/hard_clustering.hxx>
#include <ttcl/ut/null_type.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Hierarchical Clustering Namespace
  namespace hj {
    
    /// Dendrogram
    /** @param T Value type
     */
    template <typename T = ut::null_type>
    class dendrogram {
    public:
      /// Value type
      typedef T value_type;

      /// Size type
      typedef size_t size_type;

      /// Node Information
      struct node_info {
	/// Parent
	int parent;

	/// First child
	int first;

	/// Second child
	int second;

	/// Size
	value_type size;

	/// Distance at merge
	value_type distance;
      };
	
    private:
      /// Data
      std::vector<node_info> data_;

      /// Number of points
      size_type size_;

      /// Number of top clusters
      size_type clusters_;

    public:
      /// Constructor
      /** @param _size Number of points
       */
      dendrogram(size_type _size = 0) {
	// Reset to size
	reset(_size);
      }

      /// Get the number of points
      size_type size() const {
	// Size
	return size_;
      }

      /// Get the number of clusters
      size_type clusters() const {
	// Clusters
	return clusters_;
      }

      /// Get the number of branches
      size_type branches() const {
	// Branches
	return data_.size();
      }

      /// Reset
      /** @param _size Number of points
       */
      void reset(size_type _size) {
	// Fill the data with singletons
	data_.clear();
	data_.resize(_size);
	for (size_type i = 0; i < _size; ++i) {
	  data_[i].parent   = -1;
	  data_[i].first    = -1;
	  data_[i].second   = -1;
	  data_[i].size     = value_type(1);
	  data_[i].distance = value_type();
	}

	// Remember the size
	size_     = _size;
	clusters_ = _size;
      }

      /// Merge
      /** @param _first First child
	  @param _second Second child
	  @param _distance Distance at merge
      */
#ifdef TTCL_MORE_CHECKS
      size_type merge(size_type _first, size_type _second,
		      value_type _distance = value_type())
	throw (exception) {
	// Not already merged
	if (data_[_first ].parent != -1 or
	    data_[_second].parent != -1)
	  ttcl_fire("Already merged");
#else
      size_type merge(size_type _first, size_type _second,
		      value_type _distance = value_type()) {
#endif
	// Add the new node
	data_.push_back(node_info());
	size_type new_node = data_.size() - 1;

	// Link to the parent
	data_[_first ].parent   = new_node;
	data_[_second].parent   = new_node;

	// Fill parent information
	data_[new_node].parent   = -1;
	data_[new_node].first    = _first;
	data_[new_node].second   = _second;	
	data_[new_node].size     =
	  data_[_first].size + data_[_second].size;
	data_[new_node].distance = _distance;

	// One less cluster
	--clusters_;

	// Return the index of the new node
	return new_node;
      }

      /// Accessor
      /** @param _index Index
       */
      const node_info& operator[](size_type _index) const {
	return data_[_index];
      }

      /// Cut the dendrogram
      /** @param _clustering Target hard_clustering
	  @param _n_clusters Number of clusters
      */
#ifdef TTCL_MORE_CHECKS
      template <typename Container>
      void cut_on(cl::hard_clustering<Container>& _clustering,
		  size_type _n_clusters) const
	throw (exception) {
	// Check size
	if (_clustering.size() != size_)
	  ttcl_fire("Size mismatch");
#else
      template <typename Container>
      void cut_on(cl::hard_clustering<Container>& _clustering,
		  size_type _n_clusters) const {
#endif
	// More clusters than data
	if (_n_clusters >= size_) {
	  // Fill all
	  for (size_type i = 0; i < size_; ++i)
	    _clustering[i] = i;
	  _clustering.update_clusters(size_);
	  return;
	}

	// Selected clusters
	std::map<size_type, size_type> selection;
	size_type clusters = 0;

	// Special case of a single node
	if (clusters_ == 1) {
	  // One node
	  selection.insert(std::make_pair(data_.size() - 1, clusters++));
	}
	else {
	  // Find the orphaned ones
	  for (size_type i = 0; i < data_.size(); ++i)
	    if (data_[i].parent == -1)
	      selection.insert(std::make_pair(i, clusters++));
	}

	// Is the tree joint enough?
	while (clusters < _n_clusters) {
	  // Find the last one
	  std::map<size_type, size_type>::iterator it = selection.end(); --it;
	  size_type idx = it->first;

	  // Insert the two children
	  selection.insert(std::make_pair(data_[idx].first,  it->second));
	  selection.insert(std::make_pair(data_[idx].second, clusters++));

	  // Remove the parent
	  selection.erase(it);
	}

	// Mark them
	for (std::map<size_type, size_type>::iterator it = selection.begin();
	     it != selection.end(); ++it) {
	  // Stack
	  std::stack<size_type> pending;
	  pending.push(it->first);

	  // Empty?
	  while (not pending.empty()) {
	    // Top
	    size_type top = pending.top();
	    pending.pop();

	    // Leaf or branch?
	    if (top < size_) {
	      // Mark
	      _clustering[top] = it->second;
	    } 
	    else {
	      // Add the children
	      pending.push(data_[top].first);
	      pending.push(data_[top].second);
	    }
	  }
	}

	// Update
	_clustering.update_clusters(clusters);
      }

      /// Cut the dendrogram
      /** Functional convenience version
	  @param _n_clusters
      */
      cl::hard_clustering<> cut(size_type _n_clusters) const {
	// Build and return
	cl::hard_clustering<> ret(size_);
	cut_on(ret, _n_clusters);
	return ret;
      }
    };
  }
}

#endif
