#ifndef TTCL_CL_WR_HIER_DATA_CLUSTERER_WRAP_HXX
#define TTCL_CL_WR_HIER_DATA_CLUSTERER_WRAP_HXX

// TTCL: The Template Clustering Library

/** @file
    Clustering - Hier data clusterer wrapper
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/cl/wr/data_wrap.hxx>
#include <ttcl/cl/wr/hard_data_clusterer_wrap.hxx>
#include <ttcl/hj/choices.hxx>
#include <ttcl/hj/hier_data_clusterer.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/wr/interface_wrap.hxx>
#include <ttcl/wr/shared_referee.hxx>
#include <ttcl/wr/typed_shared_referee.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Hierarchical Clustering Namespace
  namespace hj {

    /// Wrappers Namespace
    namespace wr {

      /// Hierarchical data clusterer interface
      template <typename Choices>
      class _hier_data_clusterer_interface :
	public cl::wr::_hard_data_clusterer_interface<Choices> {
      public:
	/// Value type
	typedef ttcl_tu_get_choice(Choices, dendrogram_value) value_type;

	/// Dendrogram type
	typedef dendrogram<value_type> dendrogram_type;

	/// Constructor
	_hier_data_clusterer_interface(bool _owned = true,
				       uint _references = 0) :
	  cl::wr::_hard_data_clusterer_interface<Choices>(_owned, _references) {
	}

	/// Perform the clustering
	/** @param _dendrogram Target dendrogram
	    @param _data Data
	    @param _n_clusters Number of clusters
	*/
	virtual void
	hier_cluster_on(dendrogram_type& _dendrogram,
			const cl::wr::data_wrap& _data,
			const uint _n_clusters = 1) const = 0;

	/// Perform the clustering
	/** Functional convenience version
	    @param _data Data
	    @param _n_clusters Number of clusters
	*/
	virtual dendrogram_type
	hier_cluster(const cl::wr::data_wrap& _data,
		     const uint _n_clusters = 1) const = 0;
      };

      /// Hierarchical data clusterer referee
      template <typename T,
		typename Interface>
      class _hier_data_clusterer_referee :
	public cl::wr::_hard_data_clusterer_referee<T, Interface> {
      private:
	// Data type
	typedef typename T::data_type data_type;

	/// Dendrogram type
	typedef typename Interface::dendrogram_type dendrogram_type;

      public:
	/// Constructor
	_hier_data_clusterer_referee(T* _ptr = 0, bool _owned = true) :
	  cl::wr::_hard_data_clusterer_referee<T, Interface>(_ptr, _owned) {
	}

	/// Perform the hier clustering
	virtual void
	hier_cluster_on(dendrogram_type& _dendrogram,
			const cl::wr::data_wrap& _data,
			const uint _n_clusters = 1) const {
	  // Convert to data type
	  const data_type* data = _data.get<data_type>();

	  // Call
	  this->ptr_->hier_cluster_on(_dendrogram, *data, _n_clusters);
	}

	/// Perform the clustering
	virtual dendrogram_type
	hier_cluster(const cl::wr::data_wrap& _data,
		     const uint _n_clusters = 1) const {
	  // Convert to data type
	  const data_type* data = _data.get<data_type>();

	  // Call
	  return this->ptr_->hier_cluster(*data, _n_clusters);
	}
      };

      /// Base hierarchical data clusterer wrapper
      template <typename Interface>
      class _hier_data_clusterer_wrap :
	public cl::wr::_hard_data_clusterer_wrap<Interface> {
      public:
	/// Dendrogram type
	typedef typename Interface::dendrogram_type dendrogram_type;

      protected:
	/// Empty constructor
	_hier_data_clusterer_wrap() :
	  cl::wr::_hard_data_clusterer_wrap<Interface>() {
	}

	/// Constructor
	_hier_data_clusterer_wrap(Interface* _ptr) :
	  cl::wr::_hard_data_clusterer_wrap<Interface>(_ptr) {
	}

      public:
	/// Perform the hier clustering
	void
	hier_cluster_on(dendrogram_type& _dendrogram,
			const cl::wr::data_wrap& _data,
			const uint _n_clusters = 1) const {
	  // Call
	  this->ptr_->hier_cluster_on(_dendrogram, _data, _n_clusters);
	}

	/// Perform the clustering
	dendrogram_type
	hier_cluster(const cl::wr::data_wrap& _data,
		     const uint _n_clusters = 1) const {
	  // Call
	  return this->ptr_->hier_cluster(_data, _n_clusters);
	}
      };

      /// Hierarchical data clusterer wrapper
      template <typename Choices = tu::default_choices>
      class hier_data_clusterer_wrap :
	public _hier_data_clusterer_wrap
          < _hier_data_clusterer_interface<Choices> > {
      private:
	/// Interface
	typedef _hier_data_clusterer_interface<Choices> interface;

	/// Base type
	typedef _hier_data_clusterer_wrap<interface> base_type;

      public:
	/// Empty Constructor
	hier_data_clusterer_wrap() :
	  base_type() {
	}

	/// Constructor
	template <typename T>
	hier_data_clusterer_wrap(T* _value, bool _owned = true) :
	  base_type(new _hier_data_clusterer_referee<T, interface>(_value,
								   _owned)) {
	}

	/// Constructor
	template <typename T>
	hier_data_clusterer_wrap(T& _value, bool _owned = false) :
	  base_type(new _hier_data_clusterer_referee<T, interface>(&_value,
								   _owned)) {
	}
      };
    }
  }
};

#endif
