#ifndef TTCL_IO__BINSTREAM_HXX
#define TTCL_IO__BINSTREAM_HXX

#ifndef TTCL_IO_IBINSTREAM_HXX
#ifndef TTCL_IO_OBINSTREAM_HXX
#error "Cannot include _binstream.hxx directly"
#endif
#endif

// TTCL: The Template Clustering Library

/** @file
    Input/Output - Base binary stream
    @author Edgar Gonzalez i Pellicer
*/

#include <iostream>

/// TTCL Namespace
namespace ttcl {

  /// Input/Output Namespace
  namespace io {
    
    /// Base binary stream
    /** Uses the Curiously Recurring Template Pattern
     */
    template <typename Stream>
    class _binstream {
    public:
      /// Stream type
      typedef Stream stream_type;

    protected:
      /// Inner stream
      stream_type* stream_;

      /// Is stream owned?
      bool is_owned_;
      
    protected:
      /// Empty constructor
      _binstream() :
	stream_(0), is_owned_(false) {
      }

      /// Constructor
      _binstream(stream_type* _stream, bool _is_owned = false) :
	stream_(_stream), is_owned_(_is_owned) {
      }
      
    private:
      /// Copy constructor
      /** Disabled
       */
      _binstream(const _binstream& _other);

    protected:
      /// Destructor
      ~_binstream() {
	if (is_owned_)
	  delete stream_;
      }

    private:
      /// Assignment operator
      /** Disabled
       */
      _binstream& operator=(const _binstream& _other);

    public:
      /// @name std::[io]stream -like operations
      /// @{

      /// Synchronize
      int sync() {
	return stream_->sync();
      }

      /// Good
      bool good() const {
	return stream_->good();
      }

      /// Eof
      bool eof() const {
	return stream_->eof();
      }

      /// Fail
      bool fail() const {
	return stream_->fail();
      }

      /// Bad
      bool bad() const {
	return stream_->bad();
      }

      /// ! operator
      bool operator!() const {
	return !(*stream_);
      }

      /// Conversion to pointer
      operator void*() const {
	return (void*)(*stream_);
      }

      /// Get error state flags
      std::ios_base::iostate rdstate() const {
	return stream_->rdstate();
      }

      /// Set error state flags
      void setstate(std::ios_base::iostate _state) {
	stream_->setstate(_state);
      }

      /// Clear error state flags
      void clear(std::ios_base::iostate _state = std::ios_base::goodbit) {
	stream_->clear(_state);
      }

      /// Get exceptions
      std::ios_base::iostate exceptions() const {
	return stream_->exceptions();
      }

      /// Set exceptions
      void exceptions(std::ios_base::iostate _except) {
	stream_->exceptions(_except);
      }

      /// @}
    };
  }
}

#endif
