#ifndef TTCL_IO_ANYSTREAM_HXX
#define TTCL_IO_ANYSTREAM_HXX

// TTCL: The Template Clustering Library

/** @file
    Input/Output - Open any file stream
    @author Edgar Gonzalez i Pellicer
*/

#include <cstring>
#include <fstream>
#include <iostream>

#include <boost/format.hpp>

#include <ttcl/exception.hxx>
#include <ttcl/io/bzstream.hxx>
#include <ttcl/io/gzstream.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Input/Output Namespace
  namespace io {

    /// Any input stream
    class ianystream {
    public:
      /// Open a file
      static std::istream* open(const char* _filename,
				std::ios::openmode _mode = std::ios::in) {
	// Ensure mode is not out
	if (_mode & std::ios::out)
	  ttcl_fire("Can't open a ianystream in std::ios::out mode");

	// Ensure mode is in
	_mode |= std::ios::in;

	// Length
	int length = std::strlen(_filename);

	// Filename ends in .gz or .tgz?
	if ((length > 3 and
	     not std::strcmp(_filename + length - 3, ".gz")) or
	    (length > 4 and
	     not std::strcmp(_filename + length - 4, ".tgz"))) {
	  // Open gzip-compressed
	  igzstream* igzs = new igzstream(_filename, _mode);
	  if (not igzs->is_open()) {
	    delete igzs;
	    ttcl_fire(boost::format("Cannot open %s") % _filename);
	  }
	  return igzs;
	}
	// Filename ends in .bz2 or .tbz or .tb2?
	else if (length > 4 and
		 (not std::strcmp(_filename + length - 4, ".bz2") or
		  not std::strcmp(_filename + length - 4, ".tbz") or
		  not std::strcmp(_filename + length - 4, ".tb2"))) {
	  // Open bzip2-compressed
	  ibzstream* ibzs = new ibzstream(_filename, _mode);
	  if (not ibzs->is_open()) {
	    delete ibzs;
	    ttcl_fire(boost::format("Cannot open %s") % _filename);
	  }
	  return ibzs;
	}
	// Otherwise
	else {
	  // Just open
	  std::ifstream* ifs = new std::ifstream(_filename, _mode);
	  if (not ifs->is_open()) {
	    delete ifs;
	    ttcl_fire(boost::format("Cannot open %s") % _filename);
	  }
	  return ifs;
	}
      }

      /// Open a file
      static std::ifstream* open_raw(const char* _filename,
				     std::ios::openmode _mode = std::ios::in) {
	// Just open
	std::ifstream* ifs = new std::ifstream(_filename, _mode);
	if (not ifs->is_open()) {
	  delete ifs;
	  ttcl_fire(boost::format("Cannot open %s") % _filename);
	}
	return ifs;
      }
    };

    /// Any output stream
    class oanystream {
    public:
      /// Open a file
      static std::ostream* open(const char* _filename,
				std::ios::openmode _mode = std::ios::out) {
	// Ensure mode is not in
	if (_mode & std::ios::in)
	  ttcl_fire("Can't open a oanystream in std::ios::in mode");

	// Ensure mode is out
	_mode |= std::ios::out;

	// Length
	int length = std::strlen(_filename);

	// Filename ends in .gz?
	if (length > 3 and
	    not std::strcmp(_filename + length - 3, ".gz")) {
	  // Open gzip-compressed
	  ogzstream* ogzs = new ogzstream(_filename, _mode);
	  if (not ogzs->is_open()) {
	    delete ogzs;
	    ttcl_fire(boost::format("Cannot open %s") % _filename);
	  }
	  return ogzs;
	}
	// Filename ends in .bz2?
	else if (length > 4 and
		 not std::strcmp(_filename + length - 4, ".bz2")) {
	  // Open bzip2-compressed
	  obzstream* obzs = new obzstream(_filename, _mode);
	  if (not obzs->is_open()) {
	    delete obzs;
	    ttcl_fire(boost::format("Cannot open %s") % _filename);
	  }
	  return obzs;
	}
	// Otherwise
	else {
	  // Just open
	  std::ofstream* ofs = new std::ofstream(_filename, _mode);
	  if (not ofs->is_open()) {
	    delete ofs;
	    ttcl_fire(boost::format("Cannot open %s") % _filename);
	  }
	  return ofs;
	}
      }

      /// Open a file
      static std::ofstream* open_raw(const char* _filename,
				     std::ios::openmode _mode = std::ios::out) {
	// Just open
	std::ofstream* ofs = new std::ofstream(_filename, _mode);
	if (not ofs->is_open()) {
	  delete ofs;
	  ttcl_fire(boost::format("Cannot open %s") % _filename);
	}
	return ofs;
      }
    };
  }
}

#endif
