#ifndef TTCL_IO_IBINSTREAM_HXX
#define TTCL_IO_IBINSTREAM_HXX

// TTCL: The Template Clustering Library

/** @file
    Input/Output - Binary input stream
    @author Edgar Gonzalez i Pellicer
*/

#include <fstream>
#include <iostream>
#include <limits>

#include <ttcl/exception.hxx>
#include <ttcl/io/_binstream.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Input/Output Namespace
  namespace io {
    
    /// Binary input stream
    /// @bug Generated files are architecture dependent
    class ibinstream :
      public _binstream<std::istream> {
    public:
      /// Parent type
      typedef _binstream<std::istream> parent_type;

      /// Empty constructor
      ibinstream() :
	parent_type() {
      }

      /// Constructor
      ibinstream(std::istream& _stream, bool _is_owned = false) :
	parent_type(&_stream, _is_owned) {
      }

      /// Constructor from a filename
      /** @overload
       */
      ibinstream(const char* _filename,
		 std::ios_base::openmode _mode = std::ios_base::in) :
	parent_type(new std::ifstream(_filename,
				      _mode | std::ios_base::binary), true) {
      }

      /// @name std::ifstream -like operations
      /// @{

      /// Open
      void open(std::istream& _stream, bool _is_owned = false) {
	if (stream_)
	  ttcl_fire("Stream already open");

	stream_   = &_stream;
	is_owned_ = _is_owned;
      }

      /// Open
      /** @overload
       */
      void open(const char* _filename,
		std::ios_base::openmode _mode = std::ios_base::in) {
	if (stream_)
	  ttcl_fire("Stream already open");

	stream_   = new std::ifstream(_filename, _mode | std::ios_base::binary);
	is_owned_ = true;
      }

      /// Is it open?
      bool is_open() const {
	return stream_;
      }

      /// Close
      void close() {
	if (is_owned_)
	  delete stream_;

	stream_ = 0;
      }

      /// @}

      /// @name Deserialization
      /// @{
      
      /// Deserialize a boolean
      ibinstream& operator>>(bool& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(bool));
	return *this;
      }

      /// Deserialize a short
      ibinstream& operator>>(short& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(short));
	return *this;
      }

      /// Deserialize an unsigned short
      ibinstream& operator>>(unsigned short& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(unsigned short));
	return *this;
      }

      /// Deserialize an integer
      ibinstream& operator>>(int& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(int));
	return *this;
      }

      /// Deserialize an unsigned integer
      ibinstream& operator>>(unsigned int& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(unsigned int));
	return *this;
      }

      /// Deserialize a long
      ibinstream& operator>>(long& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(long));
	return *this;
      }

      /// Deserialize an unsigned long
      ibinstream& operator>>(unsigned long& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(unsigned long));
	return *this;
      }
      
#ifdef _GLIBCXX_USE_LONG_LONG

      /// Deserialize a long long
      ibinstream& operator>>(long long& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(long long));
	return *this;
      }

      /// Deserialize an unsigned long long
      ibinstream& operator>>(unsigned long long& _val) {
	stream_->read(reinterpret_cast<char*>(&_val),
		      sizeof(unsigned long long));
	return *this;
      }

#endif
      
      /// Deserialize a float
      ibinstream& operator>>(float& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(float));
	return *this;
      }

      /// Deserialize a double
      ibinstream& operator>>(double& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(double));
	return *this;
      }

      /// Deserialize a long double
      ibinstream& operator>>(long double& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(long double));
	return *this;
      }

      /// Deserialize a void pointer
      ibinstream& operator>>(void*& _val) {
	stream_->read(reinterpret_cast<char*>(&_val), sizeof(void*));
	return *this;
      }

      /// @}

      // Friend declarations
      friend ibinstream& operator>>(ibinstream& _ibs, char& _ch);
      friend ibinstream& operator>>(ibinstream& _ibs, signed char& _ch);
      friend ibinstream& operator>>(ibinstream& _ibs, unsigned char& _ch);
      friend ibinstream& operator>>(ibinstream& _ibs, char* _str);
      friend ibinstream& operator>>(ibinstream& _ibs, signed char* _str);
      friend ibinstream& operator>>(ibinstream& _ibs, unsigned char* _str);
    };

    /// @name Deserialization global functions

    /// Deserialize a char
    inline ibinstream& operator>>(ibinstream& _ibs, char& _ch) {
      _ibs.stream_->get(_ch);
      return _ibs;
    }

    /// Deserialize a signed char
    inline ibinstream& operator>>(ibinstream& _ibs, signed char& _ch) {
      _ibs.stream_->get(*reinterpret_cast<char*>(&_ch));
      return _ibs;
    }

    /// Deserialize an unsigned char
    inline ibinstream& operator>>(ibinstream& _ibs, unsigned char& _ch) {
      _ibs.stream_->get(*reinterpret_cast<char*>(&_ch));
      return _ibs;
    }
 
    /// Deserialize a char string
    /** @bug No size checking is done
     */
    inline ibinstream& operator>>(ibinstream& _ibs, char* _str) {
      _ibs.stream_->
	getline(_str, std::numeric_limits<std::streamsize>::max(), '\0');
      return _ibs;
    }

    /// Deserialize a signed char string
    /** @bug No size checking is done
     */
    inline ibinstream& operator>>(ibinstream& _ibs, signed char* _str) {
      _ibs.stream_->
	getline(reinterpret_cast<char*>(_str),
		std::numeric_limits<std::streamsize>::max(), '\0');
      return _ibs;
    }

    /// Deserialize an unsigned char string
    /** @bug No size checking is done
     */
    inline ibinstream& operator>>(ibinstream& _ibs, unsigned char* _str) {
      _ibs.stream_->
	getline(reinterpret_cast<char*>(_str),
		std::numeric_limits<std::streamsize>::max(), '\0');
      return _ibs;
    }

    /// @}
  }
}

#endif
