#ifndef TTCL_IO_OBINSTREAM_HXX
#define TTCL_IO_OBINSTREAM_HXX

// TTCL: The Template Clustering Library

/** @file
    Input/Output - Binary output stream
    @author Edgar Gonzalez i Pellicer
*/

#include <fstream>
#include <iostream>

#include <ttcl/exception.hxx>
#include <ttcl/io/_binstream.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Input/Output Namespace
  namespace io {
    
    /// Binary input stream
    /// @bug Generated files are architecture dependent
    class obinstream :
      public _binstream<std::ostream> {
    public:
      /// Parent type
      typedef _binstream<std::ostream> parent_type;

      /// Empty constructor
      obinstream() :
	parent_type() {
      }

      /// Constructor
      obinstream(std::ostream& _stream, bool _is_owned = false) :
	parent_type(&_stream, _is_owned) {
      }

      /// Constructor from a filename
      /** @overload
       */
      obinstream(const char* _filename,
		 std::ios::openmode _mode = std::ios::in) :
	parent_type(new std::ofstream(_filename,
				      _mode | std::ios::binary), true) {
      }

      /// @name std::ofstream -like operations
      /// @{

      /// Open
      void open(std::ostream& _stream, bool _is_owned = false) {
	if (stream_)
	  ttcl_fire("Stream already open");

	stream_   = &_stream;
	is_owned_ = _is_owned;
      }

      /// Open
      /** @overload
       */
      void open(const char* _filename,
		std::ios::openmode _mode = std::ios::out) {
	if (stream_)
	  ttcl_fire("Stream already open");

	stream_   = new std::ofstream(_filename, _mode | std::ios::binary);
	is_owned_ = true;
      }

      /// Is it open?
      bool is_open() const {
	return stream_;
      }

      /// Close
      void close() {
	if (is_owned_)
	  delete stream_;

	stream_ = 0;
      }

      /// @}

      /// @name Serialization
      /// @{
      
      /// Serialize a boolean
      obinstream& operator<<(const bool& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val), sizeof(bool));
	return *this;
      }

      /// Serialize a short
      obinstream& operator<<(const short& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val), sizeof(short));
	return *this;
      }

      /// Serialize an unsigned short
      obinstream& operator<<(const unsigned short& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val),
		       sizeof(unsigned short));
	return *this;
      }

      /// Serialize an integer
      obinstream& operator<<(const int& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val), sizeof(int));
	return *this;
      }

      /// Serialize an unsigned integer
      obinstream& operator<<(const unsigned int& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val),
		       sizeof(unsigned int));
	return *this;
      }

      /// Serialize a long
      obinstream& operator<<(const long& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val), sizeof(long));
	return *this;
      }

      /// Serialize an unsigned long
      obinstream& operator<<(const unsigned long& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val),
		       sizeof(unsigned long));
	return *this;
      }
      
#ifdef _GLIBCXX_USE_LONG_LONG

      /// Serialize a long long
      obinstream& operator<<(const long long& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val), sizeof(long long));
	return *this;
      }

      /// Serialize an unsigned long long
      obinstream& operator<<(const unsigned long long& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val),
		       sizeof(unsigned long long));
	return *this;
      }

#endif
      
      /// Serialize a float
      obinstream& operator<<(const float& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val), sizeof(float));
	return *this;
      }

      /// Serialize a double
      obinstream& operator<<(const double& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val), sizeof(double));
	return *this;
      }

      /// Serialize a long double
      obinstream& operator<<(const long double& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val), 
		       sizeof(long double));
	return *this;
      }

      /// Serialize a void pointer
      obinstream& operator<<(const void*& _val) {
	stream_->write(reinterpret_cast<const char*>(&_val), sizeof(void*));
	return *this;
      }

      /// @}

      // Friend declarations
      friend obinstream& operator<<(obinstream& _obs, const char& _ch);
      friend obinstream& operator<<(obinstream& _obs, const signed char& _ch);
      friend obinstream& operator<<(obinstream& _obs, const unsigned char& _ch);
      friend obinstream& operator<<(obinstream& _obs, const char* _str);
      friend obinstream& operator<<(obinstream& _obs, const signed char* _str);
      friend obinstream& operator<<(obinstream& _obs,
				    const unsigned char* _str);
    };

    /// @name Serialization global functions

    /// Serialize a char
    inline obinstream& operator<<(obinstream& _obs, const char& _ch) {
      _obs.stream_->put(_ch);
      return _obs;
    }

    /// Serialize a signed char
    inline obinstream& operator<<(obinstream& _obs, const signed char& _ch) {
      _obs.stream_->put(_ch);
      return _obs;
    }

    /// Serialize an unsigned char
    inline obinstream& operator<<(obinstream& _obs, const unsigned char& _ch) {
      _obs.stream_->put(_ch);
      return _obs;
    }
 
    /// Serialize a char string
    /** @bug No size checking is done
     */
    inline obinstream& operator<<(obinstream& _obs, const char* _str) {
      *_obs.stream_ << _str;
      _obs.stream_->put('\0');
      return _obs;
    }

    /// Serialize a signed char string
    /** @bug No size checking is done
     */
    inline obinstream& operator<<(obinstream& _obs, const signed char* _str) {
      *_obs.stream_ << _str;
      _obs.stream_->put('\0');
      return _obs;
    }

    /// Serialize an unsigned char string
    /** @bug No size checking is done
     */
    inline obinstream& operator<<(obinstream& _obs, const unsigned char* _str) {
      *_obs.stream_ << _str;
      _obs.stream_->put('\0');
      return _obs;
    }

    /// @}
  }
}

#endif
