#ifndef TTCL_IT_BLOCK_DIAGONAL_MODEL_HXX
#define TTCL_IT_BLOCK_DIAGONAL_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    Iterative Algorithms - Block Diagonal Model
    @author Edgar Gonzalez i Pellicer
*/

#include <boost/multi_array.hpp>
#include <boost/utility/enable_if.hpp>

#include <ttcl/types.hxx>
#include <ttcl/co/choices.hxx>
#include <ttcl/co/traits.hxx>
#include <ttcl/it/hard_iterative_model.hxx>
#include <ttcl/tu/choices.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {
    
    /// Base Block Diagonal Model
    template <typename Data,
	      typename Choices = tu::default_choices>
    class _block_diagonal_model :
      public hard_iterative_model<Data, Choices> {
    public:
      /// Base type
      typedef hard_iterative_model<Data, Choices> base_type;

      /// Data type
      typedef typename base_type::data_type data_type;

      /// Hard clustering type
      typedef typename base_type::hard_clustering_type hard_clustering_type;

      /// Value type
      typedef typename base_type::value_type value_type;
      
      /// Infinite distance
      static const uint infinite_distance = (1 << (sizeof(uint) - 2));

    protected:
      /// Features
      uint features_;

      /// bool allocator type
      typedef ttcl_tu_get_choice_tem(Choices, co::parameter_allocator,
				     bool) bool_allocator_type;

      /// Activeness
      boost::multi_array<bool, 2, bool_allocator_type> activeness_;

      /// uint allocator type
      typedef ttcl_tu_get_choice_tem(Choices, co::parameter_allocator,
				     uint) uint_allocator_type;

      /// Total active
      co::array<uint, uint_allocator_type> total_active_;

    public:
      /// Empty constructor
      _block_diagonal_model() :
	base_type(), features_(0),
	activeness_(), total_active_() {
      }

      /// Constructor
      _block_diagonal_model(const uint _clusters,
			    const uint _features) :
	base_type(_clusters), features_(_features),
	activeness_(boost::extents[_clusters][_features]),
	total_active_(_clusters) {
      }

      /// @name Setters/Getters
      /// @{

      /// Number of features
      uint features() const {
	  return features_;
      }
      
      /// Set the number of features
      void set_features(const uint _features) {
	features_ = _features;
	resized();
      }

      /// Get an activeness
      bool activeness(const uint _cluster, const uint _feature) const {
	return activeness_[_cluster][_feature];
      }

      /// Set an activeness
      void set_activeness(const uint _cluster, const uint _feature,
			  const bool _value) {
	activeness_[_cluster][_feature] = _value;
      }

      /// Total active
      uint total_active(const uint _cluster) const {
	return total_active_[_cluster];
      }

      /// Set the total active
      void set_total_active(const uint _cluster, const uint _value) {
	total_active_[_cluster] = _value;
      }

      /// @}

    protected:
      /// Resized
      virtual void resized() {
	// Base
	base_type::resized();
	
	// Resize weights and totals
	activeness_  .resize(boost::extents[this->clusters_][features_]);
	total_active_.resize(this->clusters_);
      }
    };

    /// Block Diagonal Model
    template <typename Data,
	      typename Choices = tu::default_choices,
	      typename Enable = void>
    class block_diagonal_model;

    /// Block Diagonal Model
    /** Specialization for sparse containers
     */
    template <typename Data,
	      typename Choices>
    class block_diagonal_model
      <Data, Choices,
       typename boost::enable_if< co::is_sparse_container<Data> >::type> :
      public _block_diagonal_model<Data, Choices> {
    private:
      /// Real base type
      typedef _block_diagonal_model<Data, Choices> _base_type;

    public:
      /// Base type
      typedef typename _base_type::base_type base_type;

      /// Data type
      typedef typename base_type::data_type data_type;

      /// Sample type
      typedef typename data_type::value_type sample_type;

      /// Hard clustering type
      typedef typename base_type::hard_clustering_type hard_clustering_type;

      /// Value type
      typedef typename base_type::value_type value_type;

    protected:
      /// uint allocator type
      typedef typename _base_type::uint_allocator_type uint_allocator_type;

    public:
      /// Constructor
      block_diagonal_model() :
	_base_type() {
      }

      /// Constructor
      block_diagonal_model(const uint _clusters,
			   const uint _features) :
	_base_type(_clusters, _features) {
      }

      /// Refine
      virtual value_type hard_refine(const data_type& _data,
				     hard_clustering_type& _clustering) {
	// Active Counts
	boost::multi_array<uint, 2, uint_allocator_type>
	  active_counts(boost::extents[this->clusters_][this->features_]);
	std::fill(active_counts.data(),
		  active_counts.data() + this->clusters_ * this->features_,
		  0u);

	// Sizes
	co::array<uint, uint_allocator_type> sizes(this->clusters_);
	
	// Divergence
	value_type divergence = value_type();

	// For each sample and feature
	uint s = 0;
	for (typename data_type::const_iterator dit = _data.begin();
	     dit != _data.end(); ++dit, ++s) {
	  // Is the clustering set
	  uint cl = _clustering[s];
	  if (cl != -1) {
	    // Count
	    ++sizes[cl];

	    // For each feature
	    for (typename sample_type::const_iterator sit = dit->begin();
		 sit != dit->end(); ++sit)
	      // Count
	      ++active_counts[cl][sit->first];
	  }
	}

	// Reset total active
	std::fill(this->total_active_.begin(), this->total_active_.end(), 0);

	// For each cluster
	for (uint cl = 0; cl < this->clusters_; ++cl) {
	  // Empty?
	  if (sizes[cl] == 0) {
	    // Inactivate it
	    this->total_active_[cl] = _base_type::infinite_distance;
	  }
	  else {
	    // Threshold
	    uint th = sizes[cl] / 2;
	    if (sizes[cl] % 2) ++th;

	    // Divergence factor
	    value_type div_f = value_type();

	    // For each feature
	    for (uint f = 0; f < this->features_; ++f) {
	      // Decide
	      if (active_counts[cl][f] >= th) {
		// Active!
		this->activeness_[cl][f] = true;
		++this->total_active_[cl];

		// Update divergence
		value_type fk = 
		  value_type(1.0) -
		  value_type(active_counts[cl][f]) / value_type(sizes[cl]);
		div_f += fk * fk;
	      }
	      else {
		// Inactive
		this->activeness_[cl][f] = false;
	      
		// Update divergence
		value_type fk = 
		  value_type(active_counts[cl][f]) / value_type(sizes[cl]);
		div_f += fk * fk;
	      }
	    }

	    // Add to the total
	    divergence += value_type(sizes[cl]) * div_f;
	  }
	}

	// Cluster and get the distance
	value_type dist = hard_cluster(_data, _clustering);

	// Return the sum of both terms
	return divergence + dist;
      }

      /// Cluster
      virtual value_type hard_cluster(const data_type& _data,
				      hard_clustering_type& _clustering) const {
	// Total distance
	uint total_distance = 0;

	// Distances
	co::array<uint, uint_allocator_type> distances(this->clusters_);

	// For each sample
	uint s = 0;
	for (typename data_type::const_iterator dit = _data.begin();
	     dit != _data.end(); ++dit, ++s) {
	  // Starting distances
	  std::copy(this->total_active_.begin(), this->total_active_.end(),
		    distances          .begin());

	  // For each feature
	  for (typename sample_type::const_iterator sit = dit->begin();
	       sit != dit->end(); ++sit) {
	    // For each cluster
	    for (uint cl = 0; cl < this->clusters_; ++cl) {
	      // Update distance
	      if (this->activeness_[cl][sit->first])
		--distances[cl];
	      else
		++distances[cl];
	    }
	  }

	  // Take the closest one
	  uint min_cl   = 0;
	  uint min_dist = distances[0];
	  for (uint cl = 1; cl < this->clusters_; ++cl) {
	    if (distances[cl] < min_dist) {
	      min_cl   = cl;
	      min_dist = distances[cl];
	    }
	  }

	  // Set it
	  _clustering[s]  = min_cl;
	  total_distance += min_dist;
	}

	// Return the distance
	return total_distance;
      }
    };
  }
}

#endif
