#ifndef TTCL_IT_EM_ALPHA_MODEL_HXX
#define TTCL_IT_EM_ALPHA_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Alpha Model
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/types.hxx>
#include <ttcl/co/choices.hxx>
#include <ttcl/it/em/choices.hxx>
#include <ttcl/it/em/em_model.hxx>
#include <ttcl/tu/choices.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {
    
    /// Expectation-Maximization Namespace
    namespace em {

      /// Alpha Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class alpha_model :
	public em_model<Data, Choices> {
      public:
	/// Base type
	typedef em_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Parameter allocator type
	typedef ttcl_tu_get_choice_tem(Choices, co::parameter_allocator,
				       value_type) parameter_allocator_type;

      protected:
	/// Alpha laplace term
	value_type alpha_laplace_term_;

	/// Alpha distribution
	/** Mixture a priori log probabilities
	 */
	co::array<value_type, parameter_allocator_type> alpha_;

      public:
	/// Empty constructor
	alpha_model() :
	  base_type(), alpha_laplace_term_(default_laplace_term), alpha_() {
	}

	/// Constructor
	alpha_model(const uint _clusters,
		    const value_type _alpha_laplace_term =
		      default_laplace_term) :
	  base_type(_clusters), alpha_laplace_term_(_alpha_laplace_term),
	  alpha_(_clusters) {
	}

	/// Copy constructor
	alpha_model(const alpha_model& _other) :
	  base_type(_other), alpha_laplace_term_(_other.alpha_laplace_term_),
	  alpha_(_other.alpha_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Alpha laplace term
	value_type alpha_laplace_term() const {
	  return alpha_laplace_term_;
	}

	/// Set alpha laplace term
	void set_alpha_laplace_term(value_type _value) {
	  alpha_laplace_term_ = _value;
	}

	/// Log alpha value
	value_type log_alpha(uint _cluster) const {
	  return alpha_[_cluster];
	}

	/// Set log alpha value
	virtual void set_log_alpha(uint _cluster, value_type _value) {
	  alpha_[_cluster] = _value;
	}

	/// Alpha value
	/** Convenience function
	 */
	value_type alpha(uint _cluster) const {
	  return std::exp(alpha_[_cluster]);
	}

	/// @}

      protected:
	/// Resized
	virtual void resized() {
	  // Base
	  base_type::resized();

	  // Resize alpha and theta
	  alpha_.resize(this->clusters_);
	}
      };
    }
  }
}

#endif
