#ifndef TTCL_IT_EM_ALPHA_THETA_MODEL_HXX
#define TTCL_IT_EM_ALPHA_THETA_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Alpha-Theta Model
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/it/em/choices.hxx>
#include <ttcl/it/em/alpha_model.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/types.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {
    
    /// Expectation-Maximization Namespace
    namespace em {

      /// Alpha-Theta Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class alpha_theta_model :
	public alpha_model<Data, Choices> {
      public:
	/// Base type
	typedef alpha_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Parameter allocator type
	typedef typename base_type::parameter_allocator_type
	  parameter_allocator_type;

      protected:
	/// Number of features
	uint features_;

	/// Theta laplace term
	value_type theta_laplace_term_;

	/// Theta distribution
	/** Feature per cluster probability ratio
	 */
	boost::multi_array<value_type, 2, parameter_allocator_type> theta_;

      public:
	/// Empty constructor
	alpha_theta_model() :
	  base_type(), features_(0),
	  theta_laplace_term_(default_laplace_term), theta_() {
	}

	/// Constructor
	alpha_theta_model(const uint _clusters, const uint _features,
			  const value_type _alpha_laplace_term =
			    default_laplace_term,
			  const value_type _theta_laplace_term =
			    default_laplace_term) :
	  base_type(_clusters, _alpha_laplace_term), features_(_features),
	  theta_laplace_term_(_theta_laplace_term),
	  theta_(boost::extents[_clusters][_features]) {
	}

	/// Copy constructor
	alpha_theta_model(const alpha_theta_model& _other) :
	  base_type(_other), features_(_other.features_),
	  theta_laplace_term_(_other.theta_laplace_term_),
	  theta_(_other.theta_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Number of features
	uint features() const {
	  return features_;
	}

	/// Set the number of features
	void set_features(uint _features) {
	  features_ = _features;
	  resized();
	}

	/// Theta laplace term
	value_type theta_laplace_term() const {
	  return theta_laplace_term_;
	}

	/// Set theta laplace term
	void set_theta_laplace_term(value_type _value) {
	  theta_laplace_term_ = _value;
	}

	/// Theta value
	/** Convenience function
	 */
	virtual value_type theta(uint _cluster, uint _feature) const = 0;

	/// @}

      protected:
	/// Resized
	virtual void resized() {
	  // Base
	  base_type::resized();

	  // Resize theta
	  theta_.resize(boost::extents[this->clusters_][features_]);
	}
      };
    }
  }
}

#endif
