#ifndef TTCL_IT_EM_BREGMAN_DIVERGENCE_SPACE_HXX
#define TTCL_IT_EM_BREGMAN_DIVERGENCE_SPACE_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Bregman Divergence Space Base Class
    @author Edgar Gonzalez i Pellicer
*/

#include <boost/mpl/not.hpp>
#include <boost/utility/enable_if.hpp>

#include <ttcl/co/traits.hxx>
#include <ttcl/ut/null_type.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {
    
    /// Expectation-Maximization Namespace
    namespace em {

      /// Bregman Divergence Space Base Types
      struct _bregman_divergence_types {

	// Centroid Stat Type
	template <typename CentroidType>
	struct centroid_stat_type {
	  typedef ut::null_type type;
	};

	// Sample Stat Type
	template <typename SampleType>
	struct sample_stat_type {
	  typedef ut::null_type type;
	};
      };

      /// Bregman Divergence Space Base Class
      /** Includes default implementations for the operations
       */
      template <typename Child,
		typename ChildTypes = _bregman_divergence_types>
      struct bregman_divergence_space : public ChildTypes {
	// Centroid stat
	template <typename Centroid>
	static typename ChildTypes::template centroid_stat_type<Centroid>::type
	centroid_stat(const Centroid& /* _centroid */) {
	  typedef typename ChildTypes::template
	    centroid_stat_type<Centroid>::type stat_type;
	  return stat_type();
	}

	// Sample stat
	template <typename Sample>
	static typename ChildTypes::template sample_stat_type<Sample>::type
	sample_stat(const Sample& /* _sample */) {
	  typedef typename ChildTypes::template
	    sample_stat_type<Sample>::type stat_type;
	  return stat_type();
	}

	// Divergence
	template <typename Centroid, typename Sample>
	static typename Centroid::value_type
	divergence(const Centroid& /* _centroid */,
		   const typename ChildTypes::template
		     centroid_stat_type<Centroid>::type /*_centroid_stat */,
		   const Sample& /* _sample */,
		   const typename ChildTypes::template
		     sample_stat_type<Sample>::type /* _sample_stat */) {
	  // Should be overriden
	  ttcl_pseudovirtual;
	}

	// Update centroid
	/* Specialization for non-sparse samples
	 */
	template <typename Centroid, typename Sample>
	static typename boost::enable_if
	  < boost::mpl::not_<co::is_sparse_sample<Sample> > >::type
	update_centroid(Centroid& _centroid,
			const typename Centroid::value_type _expectation,
			const Sample& _sample,
			const typename ChildTypes::template
			  sample_stat_type<Sample>::type /* _sample_stat */) {
	  // Value type
	  typedef typename Centroid::value_type value_type;

	  // Just add to each feature
	  for (uint f = 0; f < _centroid.size(); ++f)
	    _centroid[f] += _expectation * value_type(_sample[f]);
	}

	// Update centroid
	/* Specialization for sparse samples
	 */
	template <typename Centroid, typename Sample>
	static typename boost::enable_if
	  < co::is_sparse_sample<Sample> >::type
	update_centroid(Centroid& _centroid,
			const typename Centroid::value_type _expectation,
			const Sample& _sample,
			const typename ChildTypes::template
			  sample_stat_type<Sample>::type /* _sample_stat */) {
	  // Value type
	  typedef typename Centroid::value_type value_type;

	  // Add to present features
	  for (typename Sample::const_iterator it = _sample.begin();
	       it != _sample.end(); ++it)
	    _centroid[it->first] += _expectation * value_type(it->second);
	}
	
        // Normalize centroid
	/* @return The stat */
	template <typename Centroid>
	static typename ChildTypes::template centroid_stat_type<Centroid>::type
	normalize_centroid(Centroid& _centroid,
			   const typename Centroid::value_type _counts) {
	  // Value type
	  typedef typename Centroid::value_type value_type;

	  // Divide
	  if (_counts > value_type()) {
	    for (typename Centroid::iterator it = _centroid.begin();
		 it != _centroid.end(); ++it)
	      *it /= _counts;
	  }

	  // Return the usual centroid stat
	  return Child::centroid_stat(_centroid);
	}
      };
    }
  }
}

#endif
