#ifndef TTCL_IT_EM_BREGMAN_MODEL_HXX
#define TTCL_IT_EM_BREGMAN_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Bregman-Divergencies Prioriless Based Clustering Model
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>
#include <cmath>
#include <limits>

#include <ttcl/co/array.hxx>
#include <ttcl/it/em/centroids_model.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Bregman-Divergencies Based Clustering Model
      template <typename Data,
		typename Divergence,
		typename Choices = tu::default_choices>
      class bregman_model :
	public centroids_model<Data, Choices> {
      public:
	/// Base type
	typedef centroids_model<Data, Choices> base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Parameter allocator type
	typedef typename base_type::parameter_allocator_type
	  parameter_allocator_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Divergence type
	typedef Divergence divergence_type;

      protected:
	/// Divergence
	divergence_type divergence_;

      public:
	/// Empty Constructor
	bregman_model() :
	  base_type(), divergence_() {
	}

	/// Constructor
	bregman_model(const uint _clusters, const uint _features,
		      const divergence_type& _divergence = divergence_type()) :
	  base_type(_clusters, _features), divergence_(_divergence) {
	}

	/// Copy constructor
	bregman_model(const bregman_model& _other) :
	  base_type(_other), divergence_(_other.divergence_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Divergence
	const divergence_type& divergence() const {
	  return divergence_;
	}

	/// Set divergence
	void set_divergence(const divergence_type& _divergence) {
	  divergence_ = _divergence;
	}

	/// @}

	/// Soft maximization step
	virtual void
	soft_maximization(const soft_clustering_type& _soft_clustering,
			  const data_type& _data) {
	  // Counts
	  co::array<value_type, parameter_allocator_type>
	    counts(this->clusters_, value_type());

	  // Reset centroids
	  this->centroids_.reset();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::const_reference epos =
	      _soft_clustering[i];

	    // For each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // Update count
	      counts[c] += epos[c];

	      // Update centroid
	      this->centroids_.update(c, *dit, epos[c]);
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    this->centroids_.normalize(c, counts[c]);
	  }
	}

	/// Hard maximization step
	virtual void
	hard_maximization(const hard_clustering_type& _hard_clustering,
			  const data_type& _data) {
	  // Counts
	  co::array<value_type, parameter_allocator_type>
	    counts(this->clusters_, value_type());

	  // Reset centroids
	  this->centroids_.reset();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Cluster
	    int c = _hard_clustering[i];

	    // Assigned?
	    if (c != -1) {
	      // Update count
	      counts[c] += value_type(1);

	      // Update centroid
	      this->centroids_.update(c, *dit);
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c)
	    this->centroids_.normalize(c, counts[c]);
	}

	/// Soft expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	soft_expectation(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  /* Constant factor of uniform choice of the mixture for each
	     sample: n * log 1/k */
	  value_type log_like =
	    -value_type(_data.size()) * std::log(value_type(this->clusters_));

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Starting maximum
	    value_type max_xp = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // Expectation
	      /* Does it update maximum? */
	      if ((epos[c] =
		   -this->divergence_(this->centroids_[c], *dit)) > max_xp)
		max_xp = epos[c];
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max_xp);
	    log_like += (sum = max_xp + std::log(sum));

	    // Normalization
	    for (uint c = 0; c < this->clusters_; ++c)
	      epos[c] = std::exp(epos[c] - sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Hard expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	hard_expectation(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  /* Constant factor of uniform choice of the mixture for each
	     sample: n * log 1/k */
	  value_type log_like =
	    -value_type(_data.size()) * std::log(value_type(this->clusters_));

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Maximum
	    value_type max_xp = -std::numeric_limits<value_type>::infinity();
	    int        max_cl = -1;

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // Expectation
	      double xp =
		-this->divergence_(this->centroids_[c], *dit);

	      // Maximum expectation?
	      if (xp > max_xp) {
		max_xp = xp;
		max_cl = c;
	      }
	    }

	    // Set
	    log_like           += max_xp;
	    _hard_clustering[i] = max_cl;
	  }

	  // Log like
	  return log_like;
	}

	/// Log-likelihood of data
	virtual value_type
	log_likelihood(const data_type& _data) const {
	  // Log-likelihood
	  /* Constant factor of uniform choice of the mixture for each
	     sample: n * log 1/k */
	  value_type log_like =
	    -value_type(_data.size()) * std::log(value_type(this->clusters_));

	  // Temporary buffer
	  value_type epos[this->clusters_];

	  // For each sample
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit) {
	    // Starting maximum
	    value_type max_xp = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // Expectation
	      /* Does it update maximum? */
	      if ((epos[c] =
		   -this->divergence_(this->centroids_[c], *dit)) > max_xp)
		max_xp = epos[c];
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max_xp);
	    log_like += max_xp + std::log(sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Raw soft log-probability
	virtual void
	soft_log_probability(soft_clustering_type& _soft_clustering,
			     const data_type& _data) const {
	  // Log probability of choice
	  value_type log_choice = -std::log(value_type(this->clusters_));

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // Expectation
	      epos[c] =
		log_choice -
		this->divergence_(this->centroids_[c], *dit);
	    }
	  }
	}
      };
    }
  }
}

#endif
