#ifndef TTCL_IT_EM_DIAGONAL_GAUSSIAN_MODEL_HXX
#define TTCL_IT_EM_DIAGONAL_GAUSSIAN_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Diagonal gaussian model
    @author Edgar Gonzalez i Pellicer
*/

#include <boost/multi_array.hpp>

#include <ttcl/types.hxx>
#include <ttcl/co/array.hxx>
#include <ttcl/it/em/alpha_model.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/types.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Base diagonal gaussian model
      template <typename Data,
		typename Choices = tu::default_choices>
      class _diagonal_gaussian_model :
	public alpha_model<Data, Choices> {
      public:
	/// Base type
	typedef alpha_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Parameter allocator type
	typedef typename base_type::parameter_allocator_type
	  parameter_allocator_type;

      protected:
	/// Features
	uint features_;

	/// Probabilistic offset
	/** \f$ -\frac{f}{2} \cdot \log 2\pi \f$
	 */
	value_type offset_;

	/// Means
	boost::multi_array<value_type, 2, parameter_allocator_type> mean_;

	/// Variances
	boost::multi_array<value_type, 2, parameter_allocator_type> var_;

	/// Denominators
	co::array<value_type, parameter_allocator_type> den_;

	/// Alpha plus denom
	/** Cached for efficiency
	 */
	co::array<value_type, parameter_allocator_type> alpha_den_;

	/// Mean laplace term
	value_type mean_laplace_term_;

	/// Variance laplace term
	value_type variance_laplace_term_;

      public:
	/// Empty constructor
	_diagonal_gaussian_model() :
	  base_type(), features_(0u), offset_(0),
	  mean_(), var_(), den_(), alpha_den_() {
	}

	/// Constructor
	_diagonal_gaussian_model(const uint _clusters,
				 const uint _features,
				 const value_type _alpha_laplace_term =
				   default_laplace_term,
				 const value_type _mean_laplace_term =
				   default_laplace_term,
				 const value_type _variance_laplace_term =
				   default_laplace_term) :
	  base_type(_clusters, _alpha_laplace_term),
	  features_(_features),
	  mean_(boost::extents[_clusters][_features]),
	  var_(boost::extents[_clusters][_features]),
	  den_(_clusters), alpha_den_(_clusters),
	  mean_laplace_term_(_mean_laplace_term),
	  variance_laplace_term_(_variance_laplace_term) {
	  // Find offset
	  offset_ = -0.5 * value_type(_features) * std::log(2.0 * M_PI);
	}

	/// Copy constructor
	_diagonal_gaussian_model(const _diagonal_gaussian_model& _other) :
	  base_type(_other), features_(_other.features_),
	  offset_(_other.offset_), mean_(_other.mean_),
	  var_(_other.var_), den_(_other.den_), alpha_den_(_other.alpha_den_),
	  mean_laplace_term_(_other.mean_laplace_term_),
	  variance_laplace_term_(_other.variance_laplace_term_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Features
	uint features() const {
	  return features_;
	}

	/// Set features
	void set_features(const uint _features) {
	  features_ = _features;
	  offset_   = -0.5 * value_type(_features) * std::log(2.0 * M_PI);
	  resized();
	}

	/// Set log alpha value
	virtual void set_log_alpha(uint _cluster, value_type _value) {
	  base_type::set_log_alpha(_cluster, _value);
	  alpha_den_[_cluster] = this->alpha_[_cluster] + den_[_cluster];
	}

	/// Mean
	value_type mean(uint _cluster, uint _feature) const {
	  return mean_[_cluster][_feature];
	}

	/// Set mean
	void set_mean(uint _cluster, uint _feature, value_type _value) {
	  mean_[_cluster][_feature] = _value;
	}

	/// Variance
	value_type variance(uint _cluster, uint _feature) const {
	  return var_[_cluster][_feature];
	}

	/// Set variance
	void set_variance(uint _cluster, uint _feature, value_type _value) {
	  var_[_cluster][_feature] = _value;
	}

	/// Log denominator
	value_type log_denominator(uint _cluster) const {
	  return den_[_cluster];
	}

	/// Set log complementary theta value
	void set_log_denominator(uint _cluster, value_type _value) {
	  den_[_cluster] = _value;
	  alpha_den_[_cluster] = this->alpha_[_cluster] + den_[_cluster];
	}

	/// @}

	/// Number of free parameters
	virtual uint
	free_parameters() const {
	  return this->clusters_ * (1 + 2 * this->features_) - 1;
	}

      protected:
	/// Resized
	virtual void resized() {
	  // Base
	  base_type::resized();

	  // Resize mean, variance, den and alpha den
	  mean_     .resize(boost::extents[this->clusters_][this->features_]);
	  var_      .resize(boost::extents[this->clusters_][this->features_]);
	  den_      .resize(this->clusters_);
	  alpha_den_.resize(this->clusters_);
	}
      };

      /// Diagonal gaussian model
      /** Default version
       */
      template <typename Data,
		typename Choices = tu::default_choices,
		typename Enable = void>
      class diagonal_gaussian_model :
	public _diagonal_gaussian_model<Data, Choices> {
      private:
	/// Real base type
	typedef _diagonal_gaussian_model<Data, Choices> _base_type;

      public:
	/// Base type
	typedef typename _base_type::base_type base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Sample type
	typedef typename data_type::value_type sample_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Empty constructor
	diagonal_gaussian_model() :
	  _base_type() {
	}

	/// Constructor
	diagonal_gaussian_model(const uint _clusters,
				const uint _features,
				const value_type _alpha_laplace_term =
				  default_laplace_term,
				const value_type _mean_laplace_term =
				  default_laplace_term,
				const value_type _variance_laplace_term =
				  default_laplace_term) :
	  _base_type(_clusters, _features, _alpha_laplace_term,
		     _mean_laplace_term, _variance_laplace_term) {
	}

	/// Copy constructor
	diagonal_gaussian_model(const diagonal_gaussian_model& _other) :
	  _base_type(_other) {
	}

	/// Soft maximization step
	virtual void
	soft_maximization(const soft_clustering_type& _soft_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  value_type alpha_total = value_type();
	  std::fill(this->alpha_.begin(), this->alpha_.end(),
		    value_type());

	  // Reset means and variances
	  std::fill(this->mean_.data(),
		    this->mean_.data() + this->clusters_ * this->features_,
		    value_type());
	  std::fill(this->var_.data(),
		    this->var_.data() + this->clusters_ * this->features_,
		    value_type());

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator it = _data.begin();
	       it != _data.end(); ++it, ++i) {
	    // Expectation
	    typename soft_clustering_type::const_reference epos =
	      _soft_clustering[i];

	    // For each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // Sum to alpha
	      this->alpha_[c] += epos[c];
	      alpha_total     += epos[c];

	      // For each feature
	      uint f = 0;
	      for (typename sample_type::const_iterator sit = it->begin();
		   sit != it->end(); ++sit, ++f) {
		// Sufficient statistics
		this->mean_[c][f] += epos[c] * *sit;
		this->var_ [c][f] += epos[c] * *sit * *sit;
	      }
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    // Zero denominator
	    this->den_[c] = value_type();

	    // For each feature
	    for (uint f = 0; f < this->features_; ++f) {
	      // Mean and variance
	      this->mean_[c][f] /=
		(this->alpha_[c] + this->mean_laplace_term_);
	      this->var_ [c][f]  =
		(this->var_[c][f] -
		 this->alpha_[c] * this->mean_[c][f] * this->mean_[c][f] +
		 this->variance_laplace_term_) /
		(this->alpha_[c] + this->variance_laplace_term_);

	      // Add the log to den
	      this->den_[c] -= 0.5 * std::log(this->var_[c][f]);
	    }

	    // Alpha
	    this->alpha_[c] =
	      std::log((this->alpha_[c] + this->alpha_laplace_term_) /
		       (alpha_total +
			value_type(this->clusters_) *
			  this->alpha_laplace_term_));
	    this->alpha_den_[c] = this->alpha_[c] + this->den_[c];
	  }
	}

	/// Hard maximization step
	virtual void
	hard_maximization(const hard_clustering_type& _hard_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  // value_type alpha_total = value_type();
	  std::fill(this->alpha_.begin(), this->alpha_.end(),
		    value_type());

	  // Reset means and variances
	  std::fill(this->mean_.data(),
		    this->mean_.data() + this->clusters_ * this->features_,
		    value_type());
	  std::fill(this->var_.data(),
		    this->var_.data() + this->clusters_ * this->features_,
		    value_type());

	  // TODO
	  ttcl_todo;
	}

	/// Soft expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	soft_expectation(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      double xp = this->alpha_den_[c];

	      // Feature a posteriori probabilities
	      uint f = 0;
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit, ++f) {
		value_type d = *fit - this->mean_[c][f];
		xp -= 0.5 * d * d / this->var_[c][f];
	      }

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += (sum = max + std::log(sum));

	    // Normalization
	    for (uint c = 0; c < this->clusters_; ++c)
	      epos[c] = std::exp(epos[c] - sum);
	  }

	  // Log like plus offset
	  return log_like + value_type(_data.size()) * this->offset_;
	}

	/// Hard expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	hard_expectation(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Set maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();
	    int     max_cl = -1;

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      double xp = this->alpha_den_[c];

	      // Feature a posteriori probabilities
	      uint f = 0;
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit, ++f) {
		value_type d = *fit - this->mean_[c][f];
		xp -= 0.5 * d * d / this->var_[c][f];
	      }

	      // Maximum expectation?
	      if (xp > max) {
		max    = xp;
		max_cl = c;
	      }
	    }

	    // Set
	    log_like           += max;
	    _hard_clustering[i] = max_cl;
	  }

	  // Log like plus offset
	  return log_like + value_type(_data.size()) * this->offset_;
	}

	/// Log-likelihood of data
	virtual value_type
	log_likelihood(const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // Temporary buffer
	  value_type epos[this->clusters_];

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      double xp = this->alpha_den_[c];

	      // Feature a posteriori probabilities
	      uint f = 0;
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit, ++f) {
		value_type d = *fit - this->mean_[c][f];
		xp -= 0.5 * d * d / this->var_[c][f];
	      }

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += max + std::log(sum);
	  }

	  // Log like plus offset
	  return log_like + value_type(_data.size()) * this->offset_;
	}

	/// Raw soft log-probability
	virtual void
	soft_log_probability(soft_clustering_type& _soft_clustering,
			     const data_type& _data) const {
	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      epos[c] = this->alpha_den_[c];

	      // Feature a posteriori probabilities
	      uint f = 0;
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit, ++f) {
		value_type d = *fit - this->mean_[c][f];
		epos[c] -= 0.5 * d * d / this->var_[c][f];
	      }
	    }
	  }
	}
      };
    }
  }
}

#endif
