#ifndef TTCL_IT_EM_DIAGONAL_GAUSSIAN_NOISY_MODEL_HXX
#define TTCL_IT_EM_DIAGONAL_GAUSSIAN_NOISY_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Diagonal gaussian model
    @author Edgar Gonzalez i Pellicer
*/

#include <boost/multi_array.hpp>

#include <ttcl/types.hxx>
#include <ttcl/co/array.hxx>
#include <ttcl/it/em/diagonal_gaussian_model.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/types.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Base diagonal gaussian noisy model
      template <typename Data,
		typename Choices = tu::default_choices>
      class _diagonal_gaussian_noisy_model :
	public diagonal_gaussian_model<Data, Choices> {
      public:
	/// Base type
	typedef diagonal_gaussian_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Parameter allocator type
	typedef typename base_type::parameter_allocator_type
	  parameter_allocator_type;

      protected:
	// alpha_den_ now also contains the offset

	/// Noisy alpha
	value_type alpha_noisy_;

	/// Noisy term
	value_type noisy_term_;

	/// Noisy alpha plus term
	value_type alpha_noisy_term_;

      public:
	/// Empty constructor
	_diagonal_gaussian_noisy_model() :
	  base_type(), alpha_noisy_(0.0), noisy_term_(0.0),
	  alpha_noisy_term_(0.0) {
	}

	/// Constructor
	_diagonal_gaussian_noisy_model(const uint _clusters,
				       const uint _features,
				       const value_type _alpha_laplace_term =
				         default_laplace_term,
				       const value_type _mean_laplace_term =
				         default_laplace_term,
				       const value_type _variance_laplace_term =
				         default_laplace_term,
				       const value_type _noisy_term =
				         default_noisy_term) :
	  base_type(_clusters, _features, _alpha_laplace_term,
		    _mean_laplace_term, _variance_laplace_term),
	  alpha_noisy_(0.0), noisy_term_(std::log(_noisy_term)),
	  alpha_noisy_term_(0.0) {
	}

	/// Copy constructor
	_diagonal_gaussian_noisy_model(const _diagonal_gaussian_noisy_model&
				         _other) :
	  base_type(_other), alpha_noisy_(_other.alpha_noisy_),
	  noisy_term_(_other.noisy_term_),
	  alpha_noisy_term_(_other.alpha_noisy_term_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Log noisy alpha value
	value_type log_alpha_noisy() const {
	  return alpha_noisy_;
	}

	/// Set log noisy alpha value
	void set_log_alpha_noisy(value_type _value) {
	  alpha_noisy_      = _value;
	  alpha_noisy_term_ = alpha_noisy_ + noisy_term_;
	}

	/// Noisy alpha value
	/** Convenience function
	 */
	value_type alpha_noisy() const {
	  return std::exp(alpha_noisy_);
	}

	/// Log noisy term value
	value_type log_noisy_term() const {
	  return noisy_term_;
	}

	/// Set noisy term value
	void set_noisy_term(value_type _value) {
	  noisy_term_       = std::log(_value);
	  alpha_noisy_term_ = alpha_noisy_ + noisy_term_;
	}

	/// Noisy term value
	/** Convenience function
	 */
	value_type noisy_term() const {
	  return std::exp(noisy_term_);
	}

	/// @}

	/// Number of free parameters
	virtual uint
	free_parameters() const {
	  return this->clusters_ * (1 + 2 * this->features_);
	}
      };

      /// Diagonal gaussian model
      /** Default version
       */
      template <typename Data,
		typename Choices = tu::default_choices,
		typename Enable = void>
      class diagonal_gaussian_noisy_model :
	public _diagonal_gaussian_noisy_model<Data, Choices> {
      private:
	/// Real base type
	typedef _diagonal_gaussian_noisy_model<Data, Choices> _base_type;

      public:
	/// Base type
	typedef typename _base_type::base_type base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Sample type
	typedef typename data_type::value_type sample_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Empty constructor
	diagonal_gaussian_noisy_model() :
	  _base_type() {
	}

	/// Constructor
	diagonal_gaussian_noisy_model(const uint _clusters,
				      const uint _features,
				      const value_type _alpha_laplace_term =
				        default_laplace_term,
				      const value_type _mean_laplace_term =
				        default_laplace_term,
				      const value_type _variance_laplace_term =
				        default_laplace_term,
				      const value_type _noisy_term =
				        default_noisy_term) :
	  _base_type(_clusters, _features, _alpha_laplace_term,
		     _mean_laplace_term, _variance_laplace_term,
		     _noisy_term) {
	}

	/// Copy constructor
	diagonal_gaussian_noisy_model(const diagonal_gaussian_noisy_model&
				        _other) :
	  _base_type(_other) {
	}

	/// Soft maximization step
	virtual void
	soft_maximization(const soft_clustering_type& _soft_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  value_type alpha_total = value_type();
	  std::fill(this->alpha_.begin(), this->alpha_.end(),
		    value_type());

	  // Reset means and variances
	  std::fill(this->mean_.data(),
		    this->mean_.data() + this->clusters_ * this->features_,
		    value_type());
	  std::fill(this->var_.data(),
		    this->var_.data() + this->clusters_ * this->features_,
		    value_type());

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator it = _data.begin();
	       it != _data.end(); ++it, ++i) {
	    // Expectation
	    typename soft_clustering_type::const_reference epos =
	      _soft_clustering[i];

	    // For each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // Sum to alpha
	      this->alpha_[c] += epos[c];
	      alpha_total     += epos[c];

	      // For each feature
	      uint f = 0;
	      for (typename sample_type::const_iterator sit = it->begin();
		   sit != it->end(); ++sit, ++f) {
		// Sufficient statistics
		this->mean_[c][f] += epos[c] * *sit;
		this->var_ [c][f] += epos[c] * *sit * *sit;
	      }
	    }
	  }

	  // Noisy component alpha
	  this->alpha_noisy_ = value_type(_data.size()) - alpha_total;

	  // Corrected total
	  value_type alpha_corr =
	    value_type(_data.size()) +
	    value_type(this->clusters_ + 1) * this->alpha_laplace_term_;

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    // Zero denominator
	    this->den_[c] = value_type();

	    // For each feature
	    for (uint f = 0; f < this->features_; ++f) {
	      // Mean and variance
	      this->mean_[c][f] /=
		(this->alpha_[c] + this->mean_laplace_term_);
	      this->var_ [c][f]  =
		(this->var_[c][f] -
		 this->alpha_[c] * this->mean_[c][f] * this->mean_[c][f] +
		 this->variance_laplace_term_) /
		(this->alpha_[c] + this->variance_laplace_term_);

	      // Add the log to den
	      this->den_[c] -= 0.5 * std::log(this->var_[c][f]);
	    }

	    // Alpha
	    this->alpha_[c] =
	      std::log((this->alpha_[c] + this->alpha_laplace_term_) /
		       alpha_corr);
	    this->alpha_den_[c] =
	      this->alpha_[c] + this->den_[c] + this->offset_;
	  }

	  // Noisy alpha
	  this->alpha_noisy_ =
	    std::log((this->alpha_noisy_ + this->alpha_laplace_term_) /
		     alpha_corr);
	  this->alpha_noisy_term_ = this->alpha_noisy_ + this->noisy_term_;
	}

	/// Hard maximization step
	virtual void
	hard_maximization(const hard_clustering_type& _hard_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  // value_type alpha_total = value_type();
	  std::fill(this->alpha_.begin(), this->alpha_.end(),
		    value_type());

	  // Reset means and variances
	  std::fill(this->mean_.data(),
		    this->mean_.data() + this->clusters_ * this->features_,
		    value_type());
	  std::fill(this->var_.data(),
		    this->var_.data() + this->clusters_ * this->features_,
		    value_type());

	  // TODO
	  ttcl_todo;
	}

	/// Soft expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	soft_expectation(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // A priori expectation for the noisy cluster
	    value_type max = this->alpha_noisy_term_;

	    // Expectation for the others
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_den_[c];

	      // Feature a posteriori probabilities
	      uint f = 0;
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit, ++f) {
		value_type d = *fit - this->mean_[c][f];
		xp -= 0.5 * d * d / this->var_[c][f];
	      }

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = std::exp(this->alpha_noisy_term_ - max);
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += (sum = max + std::log(sum));

	    // Normalization
	    for (uint c = 0; c < this->clusters_; ++c)
	      epos[c] = std::exp(epos[c] - sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Hard expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	hard_expectation(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Set maximum
	    value_type max = this->alpha_noisy_term_;
	    int     max_cl = -1;

	    // Expectation for the others
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_den_[c];

	      // Feature a posteriori probabilities
	      uint f = 0;
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit, ++f) {
		value_type d = *fit - this->mean_[c][f];
		xp -= 0.5 * d * d / this->var_[c][f];
	      }

	      // Maximum expectation?
	      if (xp > max) {
		max    = xp;
		max_cl = c;
	      }
	    }

	    // Set
	    log_like           += max;
	    _hard_clustering[i] = max_cl;
	  }

	  // Log like
	  return log_like;
	}

	/// Log-likelihood of data
	virtual value_type
	log_likelihood(const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // Temporary buffer
	  value_type epos[this->clusters_];

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // A priori expectation for the noisy cluster
	    value_type max = this->alpha_noisy_term_;

	    // Expectation for the others
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_den_[c];

	      // Feature a posteriori probabilities
	      uint f = 0;
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit, ++f) {
		value_type d = *fit - this->mean_[c][f];
		xp -= 0.5 * d * d / this->var_[c][f];
	      }

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = std::exp(this->alpha_noisy_term_ - max);
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += max + std::log(sum);
	  }

	  // Log like
	  return log_like;
	}
      };
    }
  }
}

#endif
