#ifndef TTCL_IT_EM_DITC_CLUSTERER_HXX
#define TTCL_IT_EM_DITC_CLUSTERER_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Divisive Information-Theoretical Clustering Clusterer
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/it/em/choices.hxx>
#include <ttcl/it/em/em_clusterer.hxx>
#include <ttcl/it/em/ditc_model.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/tu/tuning.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Divisive Information-Theoretical Clustering Clusterer
      template <typename Data,
		typename Choices = tu::default_choices>
      class ditc_clusterer :
	public em_clusterer<Data, Choices> {
      public:
	/// Base type
	typedef em_clusterer<Data, Choices> base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Random generator type
	typedef typename base_type::random_generator_type random_generator_type;

	/// Model type
	typedef ditc_model<data_type, Choices> model_type;

	/// Base model type
	typedef typename base_type::model_type base_model_type;

      protected:
	/// Alpha laplace term
	value_type alpha_laplace_term_;

	/// Starting prior
	value_type starting_prior_;

	/// Prior scaling
	value_type prior_scaling_;

	/// Prior threshold
	value_type prior_threshold_;

      public:
	/// Constructor
	ditc_clusterer(const random_generator_type& _random_generator =
		         random_generator_type()) :
	  base_type(_random_generator),
	  alpha_laplace_term_(default_laplace_term),
	  starting_prior_(default_starting_prior),
	  prior_scaling_(default_prior_scaling),
	  prior_threshold_(default_prior_threshold) {
	}

	/// Tune
	virtual void
	tune(const tu::tuning& _tuning) {
	  // Call parent version
	  base_type::tune(_tuning);

	  // Extract the options
	  alpha_laplace_term_ = _tuning.get<alpha_laplace_term>();
	  starting_prior_     = _tuning.get<starting_prior>    ();
	  prior_scaling_      = _tuning.get<prior_scaling>     ();
	  prior_threshold_    = _tuning.get<prior_threshold>   ();
	}

	/// Create a model
	virtual model_type*
	create_model(const data_type& _data,
		     const uint _n_clusters) const {
	  uint n_features = _data.features();
	  return new model_type(_n_clusters, n_features,
				starting_prior_,
				alpha_laplace_term_);
	}

      protected:
	/// Soft Modelling
   	/** @param _soft_clustering Target soft clustering
	    @param _data Data
	*/
 	virtual model_type*
	_soft_model_from(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const {
	  // Create a model
	  model_type* model =
	    create_model(_data, _soft_clustering.clusters());

	  // Maximization step
	  model->soft_maximization(_soft_clustering, _data);

	  // Scale the prior
	  model->scale_prior(prior_scaling_);

	  // Log-likelihood
	  value_type log_likelihood;
	  uint i;

	  // Less than one iteration?
	  if (this->max_iterations_ < 1) {
	    // Just a log-likelihood step
	    log_likelihood = model->log_likelihood(_data);
	    i = 0;
	  }
	  else {
	    // Expectation step
	    log_likelihood = model->soft_expectation(_soft_clustering, _data);

	    // Loop
	    for (i = 1; i < this->max_iterations_; ++i) {
	      // Old
	      value_type old_log_likelihood = log_likelihood;

	      // Maximization step
	      model->soft_maximization(_soft_clustering, _data);

	      // Scale the prior
	      model->scale_prior(prior_scaling_);

	      // Expectation step
	      log_likelihood =
		model->soft_expectation(_soft_clustering, _data);

	      // Change ratio below threshold, and prior small?
	      if ((old_log_likelihood - log_likelihood) / old_log_likelihood <
		  this->stop_threshold_ and model->prior() < prior_threshold_)
		break;
	    }
	  }

	  // Set training information
	  model->set_train_info(i, log_likelihood);

	  // Return the model,
	  return model;
	}

	/// Hard Modelling
   	/** @param _hard_clustering Target hard clustering
	    @param _data Data
	*/
 	virtual model_type*
	_hard_model_from(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const {
	  // Create a model
	  model_type* model =
	    create_model(_data, _hard_clustering.clusters());

	  // Maximization step
	  model->hard_maximization(_hard_clustering, _data);

	  // Scale the prior
	  model->scale_prior(prior_scaling_);

	  // Log-likelihood
	  value_type log_likelihood;
	  uint i;

	  // Less than one iteration?
	  if (this->max_iterations_ < 1) {
	    // Just a log-likelihood step
	    log_likelihood = model->log_likelihood(_data);
	    i = 0;
	  }
	  else {
	    // Expectation step
	    log_likelihood = model->hard_expectation(_hard_clustering, _data);

	    // Loop
	    for (i = 1; i < this->max_iterations_; ++i) {
	      // Old
	      value_type old_log_likelihood = log_likelihood;

	      // Maximization step
	      model->hard_maximization(_hard_clustering, _data);

	      // Scale the prior
	      model->scale_prior(prior_scaling_);

	      // Expectation step
	      log_likelihood =
		model->hard_expectation(_hard_clustering, _data);

	      // Change ratio below threshold, and prior small?
	      if ((old_log_likelihood - log_likelihood) / old_log_likelihood <
		  this->stop_threshold_ and model->prior() < prior_threshold_)
		break;
	    }
	  }

	  // Set training information
	  model->set_train_info(i, log_likelihood);

	  // Return the model,
	  return model;
	}
      };
    }
  }
}

#endif
