#ifndef TTCL_IT_EM_DITC_MODEL_HXX
#define TTCL_IT_EM_DITC_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Divisive Information-Theoretical Clustering Model

    I. Dhillon, Y. Guan,
    "Information-Theoretical Clustering of Sparse
     Co-Ocurrence Data",
    UTCS Technical Report #TR-03-39, 2003

    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>
#include <cmath>
#include <limits>

#include <boost/utility/enable_if.hpp>

#include <ttcl/exception.hxx>
#include <ttcl/co/traits.hxx>
#include <ttcl/it/em/alpha_centroids_model.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Base Divisive Information-Theoretical Clustering Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class _ditc_model :
	public alpha_centroids_model<Data, Choices> {
      public:
	/// Base type
	typedef alpha_centroids_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

      protected:
	/// Prior
	value_type prior_;

      public:
	/// Empty Constructor
	_ditc_model() :
	  base_type(), prior_() {
	}

	/// Constructor
	_ditc_model(const uint _clusters, const uint _features,
		    const value_type _prior =
		      default_starting_prior,
		    const value_type _alpha_laplace_term =
		      default_laplace_term) :
	  base_type(_clusters, _features, _alpha_laplace_term),
	  prior_(_prior) {
	}

	/// Copy constructor
	_ditc_model(const _ditc_model& _other) :
	  base_type(_other), prior_(_other.prior_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Prior
	value_type prior() const {
	  return prior_;
	}

	/// Set prior
	void set_prior(const value_type _prior) {
	  prior_ = _prior;
	}

	/// Scale prior
	void scale_prior(const value_type _factor) {
	  prior_ *= _factor;
	}

	/// @}
      };

      /// Divisive Information-Theoretical Clustering Model
      template <typename Data,
		typename Choices = tu::default_choices,
		typename Enable = void>
      class ditc_model;

      /// Divisive Information-Theoretical Clustering Model
      /** Specialization for sparse containers
       */
      template <typename Data,
		typename Choices>
      class ditc_model
        <Data, Choices,
	 typename boost::enable_if<co::is_sparse_container<Data> >::type> :
	public _ditc_model<Data, Choices> {
      private:
	/// Real base type
	typedef _ditc_model<Data, Choices> _base_type;

      public:
	/// Base type
	typedef typename _base_type::base_type base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Sample type
	typedef typename data_type::value_type sample_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Empty Constructor
	ditc_model() :
	  _base_type() {
	}

	/// Constructor
	ditc_model(const uint _clusters, const uint _features,
		    const value_type _prior =
		      default_starting_prior,
		   const value_type _alpha_laplace_term =
		     default_laplace_term) :
	  _base_type(_clusters, _features, _prior, _alpha_laplace_term) {
	}

	/// Copy constructor
	ditc_model(const ditc_model& _other) :
	  _base_type(_other) {
	}

	/// Soft maximization step
	virtual void
	soft_maximization(const soft_clustering_type& _soft_clustering,
			  const data_type& _data) {
	  // Precalculated values
	  value_type denom = this->prior_ + value_type(1);
	  value_type corr  = this->prior_ / value_type(this->features_);

	  // Reset alphas
	  std::fill(this->alpha_.begin(), this->alpha_.end(), value_type());
	  value_type alpha_total  = value_type();

	  // Reset centroids
	  std::fill(this->centroids_.data(),
		    this->centroids_.data() + this->clusters_ * this->features_,
		    value_type());

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::const_reference epos =
	      _soft_clustering[i];

	    // For each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      this->alpha_[c] += epos[c];
	      alpha_total     += epos[c];

	      // For each feature
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		this->centroids_[c][fit->first] += epos[c] * fit->second;
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    // Find the centroid
	    /* Suitably perturbated */
	    for (uint f = 0; f < this->features_; ++f) {
	      if (this->alpha_[c] > value_type())
		this->centroids_[c][f] /= this->alpha_[c];
	      this->centroids_[c][f] =
		(this->centroids_[c][f] + corr) / denom;
	    }

	    // Normalize alpha
	    this->alpha_[c] =
	      std::log((this->alpha_[c] + this->alpha_laplace_term_) /
		       (alpha_total +
			value_type(this->clusters_) *
			  this->alpha_laplace_term_));
	  }
	}

	/// Hard maximization step
	virtual void
	hard_maximization(const hard_clustering_type& _hard_clustering,
			  const data_type& _data) {
	  // Precalculated values
	  value_type denom = this->prior_ + value_type(1);
	  value_type corr  = this->prior_ / value_type(this->features_);

	  // Reset alphas
	  value_type alpha_total  = value_type();
	  std::fill(this->alpha_.begin(), this->alpha_.end(), value_type());

	  // Reset centroids
	  std::fill(this->centroids_.data(),
		    this->centroids_.data() + this->clusters_ * this->features_,
		    value_type());

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Cluster
	    int c = _hard_clustering[i];

	    // Assigned?
	    if (c != -1) {
	      this->alpha_[c] += 1.0;
	      alpha_total     += 1.0;

	      // For each feature
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		this->centroids_[c][fit->first] += 1.0 * fit->second;
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    // Find the centroid
	    /* Suitably perturbated */
	    for (uint f = 0; f < this->features_; ++f) {
	      if (this->alpha_[c] > value_type())
		this->centroids_[c][f] /= this->alpha_[c];
	      this->centroids_[c][f] =
		(this->centroids_[c][f] + corr) / denom;
	    }

	    // Normalize alpha
	    this->alpha_[c] =
	      std::log((this->alpha_[c] + this->alpha_laplace_term_) /
		       (alpha_total +
			value_type(this->clusters_) *
			  this->alpha_laplace_term_));
	  }
	}

	/// Soft expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	soft_expectation(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // Expectation
	      /* Does it update maximum? */
	      if ((epos[c] = this->alpha_[c] - divergence(c, *dit)) > max)
		max = epos[c];
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += (sum = max + std::log(sum));

	    // Normalization
	    for (uint c = 0; c < this->clusters_; ++c)
	      epos[c] = std::exp(epos[c] - sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Hard expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	hard_expectation(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();
	    int     max_cl = -1;

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // Expectation
	      double xp = this->alpha_[c] - divergence(c, *dit);

	      // Maximum expectation?
	      if (xp > max) {
		max    = xp;
		max_cl = c;
	      }
	    }

	    // Set
	    log_like           += max;
	    _hard_clustering[i] = max_cl;
	  }

	  // Log like
	  return log_like;
	}

	/// Log-likelihood of data
	virtual value_type
	log_likelihood(const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // Temporary buffer
	  value_type epos[this->clusters_];

	  // For each sample
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit) {
	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // Expectation
	      /* Does it update maximum? */
	      if ((epos[c] = this->alpha_[c] - divergence(c, *dit)) > max)
		max = epos[c];
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += max + std::log(sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Raw soft log-probability
	virtual void
	soft_log_probability(soft_clustering_type& _soft_clustering,
			     const data_type& _data) const {
	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c)
	      epos[c] = this->alpha_[c] - divergence(c, *dit);
	  }
	}

      protected:
	/// Divergence from centroid
	value_type divergence(uint c, const sample_type& _sample) const {
	  // Sum
	  value_type sum = value_type();

	  // Advance
	  for (typename sample_type::const_iterator it = _sample.begin();
	       it != _sample.end(); ++it)
	    sum += value_type(it->second) *
	      std::log(value_type(it->second) / this->centroids_[c][it->first]);

	  // Return the sum
	  return sum;
	}
      };
    }
  }
}

#endif
