#ifndef TTCL_IT_EM_EM_CLUSTERER_HXX
#define TTCL_IT_EM_EM_CLUSTERER_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Expectation-Maximization Clusterer
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>
#include <vector>

#include <ttcl/types.hxx>
#include <ttcl/cl/soft_data_model_clusterer.hxx>
#include <ttcl/it/choices.hxx>
#include <ttcl/it/randomized_clusterer.hxx>
#include <ttcl/it/em/choices.hxx>
#include <ttcl/it/em/em_model.hxx>
#include <ttcl/ut/choices.hxx>
#include <ttcl/ut/range.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/tu/tuning.hxx>

#ifdef TTCL_VERBOSE
#include <ttcl/lg/logger.hxx>
#endif

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Generic Expectation-Maximization Clusterer
      template <typename Data,
		typename Choices = tu::default_choices>
      class em_clusterer :
	public cl::soft_data_model_clusterer<Data, Choices>,
	public randomized_clusterer<Choices> {
      public:
	/// Base type
	typedef cl::soft_data_model_clusterer<Data, Choices> base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Soft clustering container type
	typedef typename base_type::soft_clustering_container_type
	  soft_clustering_container_type;

	/// Value type
	typedef typename soft_clustering_container_type::element value_type;

	/// Randomized type
	typedef randomized_clusterer<Choices> randomized_type;

	/// Random generator type
	typedef typename randomized_type::random_generator_type
	  random_generator_type;

	/// Model type
	typedef em_model<Data, Choices> model_type;

      protected:
	/// Max iterations
	uint max_iterations_;

	/// Stop threshold
	value_type stop_threshold_;

	/// Inner softness
	softness_enum inner_softness_;

      public:
	/// Constructor
	em_clusterer(const random_generator_type& _random_generator =
		       random_generator_type()) :
	  randomized_type(_random_generator),
	  max_iterations_(default_max_iterations),
	  stop_threshold_(default_stop_threshold),
	  inner_softness_(DEFAULT_SOFTNESS) {
	}

	/// Tune
	virtual void
	tune(const tu::tuning& _tuning) {
	  // Parent
	  randomized_type::tune(_tuning);

	  // Extract the options
	  max_iterations_ = _tuning.get<max_iterations>();
	  stop_threshold_ = _tuning.get<stop_threshold>();
	  inner_softness_ = _tuning.get<inner_softness>();
	}

	/// Model
   	/** @param _soft_clustering Target soft clustering
	    @param _data Data
	*/
	model_type*
	soft_model_from(soft_clustering_type& _soft_clustering,
			const data_type& _data,
			const uint /* _n_clusters */) const {
	  // Return
	  model_type* ret;

	  // According to softness
	  if (inner_softness_ == SOFTNESS_SOFT) {
	    // Direct forward
	    ret = _soft_model_from(_soft_clustering, _data);
	  }
	  else { // inner_softness_ == SOFTNESS_HARD
	    // Convert to hard, model and convert back
	    hard_clustering_type hard = _soft_clustering;
	    ret = _hard_model_from(hard, _data);
	    _soft_clustering = hard;
	  }

	  // Return the result
	  return ret;
	}

	/// Model
   	/** @param _hard_clustering Target soft clustering
	    @param _data Data
	*/
	model_type*
	hard_model_from(hard_clustering_type& _hard_clustering,
			const data_type& _data,
			const uint /* _n_clusters */) const {
	  // Return
	  model_type* ret;

	  // According to softness
	  if (inner_softness_ == SOFTNESS_SOFT) {
	    // Convert to soft, model and convert back
	    soft_clustering_type soft = _hard_clustering;
	    ret = _soft_model_from(soft, _data);
	    _hard_clustering = soft;
	  }
	  else { // inner_softness_ == SOFTNESS_HARD
	    // Direct forward
	    ret = _hard_model_from(_hard_clustering, _data);
	  }

	  // Return the result
	  return ret;
	}

	/// Model
	/** @param _soft_clustering Target soft clustering
	    @param _data Data
	*/
	model_type*
	soft_model_on(soft_clustering_type& _soft_clustering,
		      const data_type& _data,
		      const uint _n_clusters) const {
	  // Return
	  model_type* ret;

	  // According to softness
	  if (inner_softness_ == SOFTNESS_SOFT) {
	    // Randomize
	    soft_randomize(_soft_clustering);

	    // Model from it
	    ret = _soft_model_from(_soft_clustering, _data);
	  }
	  else { // inner_softness_ == SOFTNESS_HARD
	    // Create and randomize a hard clustering
	    hard_clustering_type hard(_soft_clustering.size(), _n_clusters);
	    hard_randomize(hard, _n_clusters);

	    // Model from it, and convert back
	    ret = _hard_model_from(hard, _data);
	    _soft_clustering = hard;
	  }

	  // Return the result
	  return ret;
	}

	/// Model
	/** @param _hard_clustering Target hard clustering
	    @param _data Data
	*/
	model_type*
	hard_model_on(hard_clustering_type& _hard_clustering,
		      const data_type& _data,
		      const uint _n_clusters) const {
	  // Return
	  model_type* ret;

	  // According to softness
	  if (inner_softness_ == SOFTNESS_SOFT) {
	    // Create and randomize a soft clustering
	    soft_clustering_type soft(_hard_clustering.size(), _n_clusters);
	    soft_randomize(soft);

	    // Model from it, and convert back
	    ret = _soft_model_from(soft, _data);
	    _hard_clustering = soft;
	  }
	  else { // inner_softness_ == SOFTNESS_HARD
	    // Randomize
	    hard_randomize(_hard_clustering, _n_clusters);

	    // Model from it
	    ret = _hard_model_from(_hard_clustering, _data);
	  }

	  // Return the result
	  return ret;
	}

	/// Model
	/** @param _data Data
	    @param _n_clusters Number of clusters
	*/
	model_type*
	model(const data_type& _data,
	      const uint _n_clusters) const {
	  // Return
	  model_type* ret;

	  // According to softness
	  if (inner_softness_ == SOFTNESS_SOFT) {
	    // Create and randomize a soft clustering
	    soft_clustering_type soft(_data.size(), _n_clusters);
	    soft_randomize(soft);

	    // Model from it
	    ret = _soft_model_from(soft, _data);
	  }
 	  else { // inner_softness_ == SOFTNESS_HARD
	    // Create and randomize a hard clustering
	    hard_clustering_type hard(_data.size(), _n_clusters);
	    hard_randomize(hard, _n_clusters);

	    // Model from it
	    ret = _hard_model_from(hard, _data);
	  }

	  // Return the result
	  return ret;
	}

	/// Refine a soft clustering
	/** @param _soft_clustering Target soft clustering
	    @param _data Data
	*/
	void
	soft_cluster_from(soft_clustering_type& _soft_clustering,
			  const data_type& _data,
			  const uint /* _n_clusters */) const {
	  // According to softness
	  if (inner_softness_ == SOFTNESS_SOFT) {
	    // Cluster using a model
	    model_type* model = _soft_model_from(_soft_clustering, _data);
	    delete model;
	  }
	  else { // inner_softness_ == SOFTNESS_HARD
	    // Create a hard clustering
	    hard_clustering_type hard = _soft_clustering;

	    // Cluster using a model
	    model_type* model = _hard_model_from(hard, _data);
	    delete model;

	    // Convert back
	    _soft_clustering = hard;
	  }
	}

	/// Perform the soft clustering
	/** @param _soft_clustering Target soft clustering
	    @param _data Data
	    @param _n_clusters Number of clusters
	*/
	virtual void
	soft_cluster_on(soft_clustering_type& _soft_clustering,
			const data_type& _data,
			const uint _n_clusters) const {
	  // Cluster using a model
	  model_type* model =
	    soft_model_on(_soft_clustering, _data, _n_clusters);
	  delete model;
	}

	/// Refine a hard clustering
	/** @param _hard_clustering Target hard clustering
	    @param _data Data
	*/
	void
	hard_cluster_from(hard_clustering_type& _hard_clustering,
			  const data_type& _data,
			  const uint /* _n_clusters */) const {
	  // According to softness
	  if (inner_softness_ == SOFTNESS_SOFT) {
	    // Create a soft clustering
	    soft_clustering_type soft = _hard_clustering;

	    // Cluster using a model
	    model_type* model = _soft_model_from(soft, _data);
	    delete model;

	    // Convert back
	    _hard_clustering = soft;
	  }
	  else { // inner_softness_ == SOFTNESS_HARD
	    // Cluster using a model
	    model_type* model = _hard_model_from(_hard_clustering, _data);
	    delete model;
	  }
	}

	/// Perform the hard clustering
	/** @param _hard_clustering Target hard clustering
	    @param _data Data
	    @param _n_clusters Number of clusters
	*/
	virtual void
	hard_cluster_on(hard_clustering_type& _hard_clustering,
			const data_type& _data,
			const uint _n_clusters) const {
	  // Cluster using a model
	  model_type* model =
	    hard_model_on(_hard_clustering, _data, _n_clusters);
	  delete model;
	}

	/// Create a model
	virtual model_type*
	create_model(const data_type& _data,
		     const uint _n_clusters) const = 0;

      protected:
	/// Soft Modelling
   	/** @param _soft_clustering Target soft clustering
	    @param _data Data
	*/
 	virtual model_type*
	_soft_model_from(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const {
	  // Create a model
	  model_type* model =
	    create_model(_data, _soft_clustering.clusters());

	  // Maximization step
	  model->soft_maximization(_soft_clustering, _data);

	  // Log-likelihood
	  value_type log_likelihood;
	  uint i;

	  // Less than one iteration?
	  if (max_iterations_ < 1) {
	    // Just a log-likelihood step
	    log_likelihood = model->log_likelihood(_data);
	    i = 0;
	  }
	  else {
	    // Expectation step
	    log_likelihood = model->soft_expectation(_soft_clustering, _data);

	    // Loop
	    for (i = 1; i < max_iterations_; ++i) {
#ifdef TTCL_VERBOSE
	      // Verbose
	      ttcl_logger().info("Iteration %d: Log-Likelihood = %g",
				 i - 1, log_likelihood);
#endif
	      // Old
	      value_type old_log_likelihood = log_likelihood;

	      // Maximization step
	      model->soft_maximization(_soft_clustering, _data);

	      // Expectation step
	      log_likelihood =
		model->soft_expectation(_soft_clustering, _data);

	      // Change ratio below threshold?
	      if ((old_log_likelihood - log_likelihood) / old_log_likelihood <
		  stop_threshold_) break;
	    }
	  }

#ifdef TTCL_VERBOSE
	  // Verbose
	  ttcl_logger().info("Final iteration %d: Log-Likelihood = %g",
			     i, log_likelihood);
#endif

	  // Set training information
	  model->set_train_info(i, log_likelihood);

	  // Return the model
	  return model;
	}

	/// Hard Modelling
   	/** @param _hard_clustering Target hard clustering
	    @param _data Data
	*/
 	virtual model_type*
	_hard_model_from(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const {
	  // Create a model
	  model_type* model =
	    create_model(_data, _hard_clustering.clusters());

	  // Maximization step
	  model->hard_maximization(_hard_clustering, _data);

	  // Log-likelihood
	  value_type log_likelihood;
	  uint i;

	  // Less than one iteration?
	  if (max_iterations_ < 1) {
	    // Just a log-likelihood step
	    log_likelihood = model->log_likelihood(_data);
	    i = 0;
	  }
	  else {
	    // Expectation step
	    log_likelihood = model->hard_expectation(_hard_clustering, _data);

	    // Loop
	    for (i = 1; i < max_iterations_; ++i) {
#ifdef TTCL_VERBOSE
	      // Verbose
	      ttcl_logger().info("Iteration %d: Log-Likelihood = %g",
				 i - 1, log_likelihood);
#endif
	      // Old
	      value_type old_log_likelihood = log_likelihood;

	      // Maximization step
	      model->hard_maximization(_hard_clustering, _data);

	      // Expectation step
	      log_likelihood =
		model->hard_expectation(_hard_clustering, _data);

	      // Change ratio below threshold?
	      if ((old_log_likelihood - log_likelihood) / old_log_likelihood <
		  stop_threshold_) break;
	    }
	  }

#ifdef TTCL_VERBOSE
	  // Verbose
	  ttcl_logger().info("Final iteration %d: Log-Likelihood = %g",
			     i, log_likelihood);
#endif

	  // Set training information
	  model->set_train_info(i, log_likelihood);

	  // Return the model
	  return model;
	}
      };
    }
  }
}

#endif
