#ifndef TTCL_IT_EM_EM_MODEL_HXX
#define TTCL_IT_EM_EM_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Expectation-Maximization Model
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/cl/choices.hxx>
#include <ttcl/cl/soft_data_model.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/types.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Expectation-Maximization Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class em_model :
	public cl::soft_data_model<Data, Choices> {
      public:
	/// Base type
	typedef cl::soft_data_model<Data, Choices> base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type
	hard_clustering_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type
          soft_clustering_type;

      protected:
	/// Soft clustering container type
	typedef typename base_type::soft_clustering_container_type
          soft_clustering_container_type;

      public:
	/// Value type
	typedef typename soft_clustering_container_type::element value_type;

      protected:
	/// Training iterations
	uint train_iterations_;

	/// Training log-likelihood
	value_type train_log_likelihood_;

      public:
	/// Constructor
	em_model(const uint _clusters = 0) :
	  base_type(_clusters) {
	}

	/// Copy constructor
	em_model(const em_model& _other) :
	  base_type(_other.clusters_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Training iterations
	uint train_iterations() const {
	  return train_iterations_;
	}

	/// Training log-likelihood
	value_type train_log_likelihood() const {
	  return train_log_likelihood_;
	}

	/// Set training information
	void set_train_info(uint _train_iterations,
			    value_type _train_log_likelihood) {
	  train_iterations_     = _train_iterations;
	  train_log_likelihood_ = _train_log_likelihood;
	}

	/// @}

	/// Soft maximization step
	virtual void
	soft_maximization(const soft_clustering_type& _soft_clustering,
			  const data_type& _data) = 0;

	/// Hard maxization step
	virtual void
	hard_maximization(const hard_clustering_type& _hard_clustering,
			  const data_type& _data) = 0;

	/// Soft expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	soft_expectation(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const = 0;

	/// Hard expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	hard_expectation(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const = 0;

	/// Log-likelihood of data
	virtual value_type
	log_likelihood(const data_type& _data) const = 0;

	/// Raw soft log-probability
	virtual void
	soft_log_probability(soft_clustering_type& _soft_clustering,
			     const data_type& _data) const = 0;

	/// Perform the hard clustering
	/** @param _hard_clustering Target hard clustering
	    @param _data Data
	*/
	virtual void
	hard_cluster_on(hard_clustering_type& _hard_clustering,
			const data_type& _data) const {
	  hard_expectation(_hard_clustering, _data);
	}

	/// Perform the soft clustering
	/** @param _soft_clustering Target soft clustering
	    @param _data Data
	    @param _n_clusters Number of clusters
	*/
	virtual void
	soft_cluster_on(soft_clustering_type& _soft_clustering,
			const data_type& _data) const {
	  soft_expectation(_soft_clustering, _data);
	}
      };
    }
  }
};

#endif
