#ifndef TTCL_IT_EM_HUGE_NB_BERNOUILLI_MODEL_HXX
#define TTCL_IT_EM_HUGE_NB_BERNOUILLI_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Huge Bernouilli Naive Bayes Model
    @author Edgar Gonzalez i Pellicer
*/

#include <boost/utility/enable_if.hpp>

#include <ttcl/co/traits.hxx>
#include <ttcl/it/em/nb_bernouilli_model.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Huge Bernouilli Naive Bayes Model
      template <typename Data,
		typename Choices = tu::default_choices,
		typename Enable = void>
      class huge_nb_bernouilli_model;

      /// Huge Bernouilli Naive Bayes Model
      /** Specialization for sparse containers
       */
      template <typename Data,
		typename Choices>
      class huge_nb_bernouilli_model
        <Data, Choices,
	 typename boost::enable_if< co::is_sparse_container<Data> >::type> :
	public nb_bernouilli_model<Data, Choices> {
      public:
	/// Base type
	typedef nb_bernouilli_model<Data, Choices> base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Sample type
	typedef typename base_type::sample_type sample_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Parameter allocator type
	typedef typename base_type::parameter_allocator_type
	  parameter_allocator_type;

	/// Empty constructor
	huge_nb_bernouilli_model() :
	  base_type() {
	}

	/// Constructor
	huge_nb_bernouilli_model(const uint _clusters, const uint _features,
				 const value_type _alpha_laplace_term =
				   default_laplace_term,
				 const value_type _theta_laplace_term =
				   default_laplace_term) :
	  base_type(_clusters, _features,
		    _alpha_laplace_term, _theta_laplace_term) {
	}

	/// Copy constructor
	huge_nb_bernouilli_model(const huge_nb_bernouilli_model& _other) :
	  base_type(_other) {
	}

	/// Hard maximization step
	virtual void
	hard_maximization(const hard_clustering_type& _hard_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  std::fill(this->alpha_.begin(), this->alpha_.end(), value_type());
	  value_type alpha_total  = value_type();

	  // Reset thetas
	  std::fill(this->theta_.data(),
		    this->theta_.data() + this->clusters_ * this->features_,
		    value_type());

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Cluster
	    int c = _hard_clustering[i];

	    // Assigned?
	    if (c != -1) {
	      this->alpha_[c] += 1.0;
	      alpha_total     += 1.0;

	      // For each feature
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		this->theta_[c][fit->first] += 1.0;
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    this->comp_theta_[c] = value_type();
	    for (uint f = 0; f < this->features_; ++f) {
	      // Ratio and logarithms
	      value_type theta_val =
		(this->theta_[c][f] + this->theta_laplace_term_) /
		(this->alpha_[c] + value_type(2) * this->theta_laplace_term_);
	      value_type pos_log = std::log(theta_val);
	      value_type neg_log = std::log(value_type(1) - theta_val);

	      // Normalize theta and complementary
	      this->theta_[c][f]    = pos_log - neg_log;
	      this->comp_theta_[c] += neg_log;
	    }

	    // Normalize alpha
	    this->alpha_[c] =
	      std::log((this->alpha_[c] + this->alpha_laplace_term_) /
		       (alpha_total +
			value_type(this->clusters_) *
			  this->alpha_laplace_term_));
	    this->alpha_comp_[c] = this->alpha_[c] + this->comp_theta_[c];
	  }
	}

	/// One-shot refinement step
	/** @return Log-likelihood of data
	 */
	value_type refine(const data_type& _data) {
	  // Extra data for counts
	  value_type alpha_total = 0.0;
	  co::array<value_type, parameter_allocator_type>
	    new_alpha(this->clusters_);
	  boost::multi_array<value_type, 2, parameter_allocator_type>
	    new_theta(boost::extents[this->clusters_][this->features_]);

	  // Reset
	  fill(new_alpha.begin(), new_alpha.end(), value_type());
	  fill(new_theta.data(),
	       new_theta.data() + this->clusters_ * this->features_,
	       value_type());

	  // Expectations
	  co::array<value_type, parameter_allocator_type> epos(this->clusters_);

	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {

	    /***************/
	    /* Expectation */
	    /***************/

	    // Set maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_comp_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		xp += this->theta_[c][fit->first];

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += (sum = max + std::log(sum));

	    // Normalization
	    for (uint c = 0; c < this->clusters_; ++c)
	      epos[c] = std::exp(epos[c] - sum);


	    /****************/
	    /* Maximization */
	    /****************/

	    // For each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      new_alpha[c] += epos[c];
	      alpha_total  += epos[c];

	      // For each feature
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		new_theta[c][fit->first] += epos[c];
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    this->comp_theta_[c] = value_type();
	    for (uint f = 0; f < this->features_; ++f) {
	      // Ratio and logarithms
	      value_type theta_val =
		(new_theta[c][f] + this->theta_laplace_term_) /
		(new_alpha[c] + value_type(2) * this->theta_laplace_term_);
	      value_type pos_log = std::log(theta_val);
	      value_type neg_log = std::log(value_type(1) - theta_val);

	      // Normalize theta and complementary
	      this->theta_[c][f]    = pos_log - neg_log;
	      this->comp_theta_[c] += neg_log;
	    }

	    // Normalize alpha
	    this->alpha_[c] =
	      std::log((new_alpha[c] + this->alpha_laplace_term_) /
		       (alpha_total +
			value_type(this->clusters_) *
			this->alpha_laplace_term_));
	    this->alpha_comp_[c] = this->alpha_[c] + this->comp_theta_[c];
	  }

	  // Log like
	  return log_like;
	}
      };
    }
  }
}

#endif
