#ifndef TTCL_IT_EM_INDEPENDENT_ALPHA_THETA_MODEL_HXX
#define TTCL_IT_EM_INDEPENDENT_ALPHA_THETA_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Independent Alpha-Theta Model
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/it/em/choices.hxx>
#include <ttcl/it/em/alpha_model.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/types.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {
    
    /// Expectation-Maximization Namespace
    namespace em {

      /// Independent Alpha-Theta Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class independent_alpha_theta_model :
	public alpha_model<Data, Choices> {
      public:
	/// Base type
	typedef alpha_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Parameter allocator type
	typedef typename base_type::parameter_allocator_type
	  parameter_allocator_type;

      protected:
	/// Cardinalities type
	typedef co::array<uint, typename parameter_allocator_type::
			  template rebind<uint>::other> cardinalities_type;

	/// Number of features
	uint features_;

	/// Feature cardinalities
	cardinalities_type feature_cardinalities_;

	/// Theta laplace term
	value_type theta_laplace_term_;

	/// Theta distribution
	/** Feature per cluster probabilities
	 */
	typedef co::array<value_type, parameter_allocator_type> inner_type;
	boost::multi_array<inner_type, 2, typename parameter_allocator_type::
			     template rebind<inner_type>::other> theta_;

      public:
	/// Empty constructor
	independent_alpha_theta_model() :
	  base_type(), features_(0), feature_cardinalities_(), 
	  theta_laplace_term_(default_laplace_term), theta_() {
	}

	/// Constructor
	independent_alpha_theta_model
	  (const uint _clusters, const uint _features,
	   const uint* _feature_cardinalities,
	   const value_type _alpha_laplace_term = default_laplace_term,
	   const value_type _theta_laplace_term = default_laplace_term) :
	    base_type(_clusters, _alpha_laplace_term),
	    features_(_features),
	    feature_cardinalities_(_features, _feature_cardinalities),
	    theta_laplace_term_(_theta_laplace_term),
	    theta_(boost::extents[_clusters][_features]) {
	  // Initialize each theta
	  for (uint c = 0; c < this->clusters_; ++c)
	    for (uint f = 0; f < this->features_; ++f)
	      theta_[c][f].resize(feature_cardinalities_[f]);
	}

	/// Copy constructor
	independent_alpha_theta_model
	  (const independent_alpha_theta_model& _other) :
	    base_type(_other), features_(_other.features_),
	    feature_cardinalities_(_other.feature_cardinalities_),
	    theta_laplace_term_(_other.theta_laplace_term_),
	    theta_(_other.theta_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Number of features
	uint features() const {
	  return features_;
	}

	/// Set features
	void set_features(uint _features) {
	  // Set
	  features_ = _features;
	  resized();
	}

	/// Feature cardinality
	uint feature_cardinality(uint _feature) const {
	  return feature_cardinalities_[_feature];
	}

	/// Set cardinalities
	void set_feature_cardinality(uint _feature, uint _cardinality) {
	  // Set
	  feature_cardinalities_[_feature] = _cardinality;

	  // Resize each theta
	  for (uint c = 0; c < this->clusters_; ++c)
	    theta_[c][_feature].resize(_cardinality);
	}

	/// Log theta
	value_type log_theta(uint _cluster, uint _feature, uint _index) const {
	  return theta_[_cluster][_feature][_index];
	}

	/// Set log theta
	void set_log_theta(uint _cluster, uint _feature, uint _index,
		       value_type _value) {
	  theta_[_cluster][_feature][_index] = _value;
	}

	/// Theta
	/** Convenience function
	 */
	value_type theta(uint _cluster, uint _feature, uint _index) const {
	  return std::exp(theta_[_cluster][_feature][_index]);
	}

	/// @}

      protected:
	/// Resized
	virtual void resized() {
	  // Base
	  base_type::resized();

	  // Resize
	  feature_cardinalities_.resize(features_);
	  theta_.resize(boost::extents[this->clusters_][features_]);
	}
      };
    }
  }
}

#endif
