#ifndef TTCL_IT_EM_NB_BERNOUILLI_INCREMENTAL_MODEL_HXX
#define TTCL_IT_EM_NB_BERNOUILLI_INCREMENTAL_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Bernouilli Incremental Naive Bayes Model
    @author Edgar Gonzalez i Pellicer
*/

#include <cmath>
#include <vector>

#include <boost/format.hpp>
#include <boost/multi_array.hpp>
#include <boost/utility/enable_if.hpp>

#include <ttcl/co/traits.hxx>
#include <ttcl/it/em/nb_bernouilli_model.hxx>
#include <ttcl/tu/choices.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Base Bernouilli Incremental Naive Bayes Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class _nb_bernouilli_incremental_model :
	public nb_bernouilli_model<Data, Choices> {
      public:
	/// Base type
	typedef nb_bernouilli_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Parameter allocator type
	typedef typename base_type::parameter_allocator_type
	  parameter_allocator_type;

      protected:
	/// New alpha
	/** For estimation
	 */
	co::array<value_type, parameter_allocator_type> new_alpha_;

	/// New theta
	/** For estimation
	 */
	boost::multi_array<value_type, 2, parameter_allocator_type> new_theta_;

	/// Learning rate
	value_type learning_rate_;

      public:
	/// Empty constructor
	_nb_bernouilli_incremental_model() :
	  base_type(), new_alpha_(), new_theta_(),
	  learning_rate_(default_learning_rate) {
	}

	/// Constructor
	_nb_bernouilli_incremental_model(const uint _clusters,
					 const uint _features,
					 const value_type _alpha_laplace_term =
					   default_laplace_term,
					 const value_type _theta_laplace_term =
					    default_laplace_term,
					 const value_type _learning_rate =
					    default_learning_rate) :
	  base_type(_clusters, _features,
		    _alpha_laplace_term, _theta_laplace_term),
	  new_alpha_(_clusters),
	  new_theta_(boost::extents[_clusters][_features]),
	  learning_rate_(_learning_rate) {
	}

	/// Copy constructor
	_nb_bernouilli_incremental_model
	  (const _nb_bernouilli_incremental_model& _other) :
	    base_type(_other), new_alpha_(_other.new_alpha_),
	    new_theta_(_other.new_theta_),
	    learning_rate_(_other.learning_rate_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Learning rate
	value_type learning_rate() const {
	  return learning_rate_;
	}

	/// Set learning rate
	void set_learning_rate(const value_type _value) {
	  learning_rate_ = _value;
	}

	/// @}

      protected:
	/// Resized
	virtual void resized() {
	  // Base
	  base_type::resized();

	  // Resize new_alpha_ and new_theta_
	  new_alpha_.resize(this->clusters_);
	  new_theta_.resize(boost::extents[this->clusters_][this->features_]);
	}
      };

      /// Bernouilli Incremental Naive Bayes Model
      template <typename Data,
		typename Choices = tu::default_choices,
		typename Enable = void>
      class nb_bernouilli_incremental_model;

      /// Bernouilli Incremental Naive Bayes Model
      /** Specialization for sparse containers
       */
      template <typename Data,
		typename Choices>
      class nb_bernouilli_incremental_model
        <Data, Choices,
	 typename boost::enable_if< co::is_sparse_container<Data> >::type> :
	public _nb_bernouilli_incremental_model<Data, Choices> {
      private:
	/// Real base type
	typedef _nb_bernouilli_incremental_model<Data, Choices> _base_type;

      public:
	/// Base type
	typedef typename _base_type::base_type base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Sample type
	typedef typename data_type::value_type sample_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Empty constructor
	nb_bernouilli_incremental_model() :
	  _base_type() {
	}

	/// Constructor
	nb_bernouilli_incremental_model(const uint _clusters,
					const uint _features,
					const value_type _alpha_laplace_term =
					  default_laplace_term,
					const value_type _theta_laplace_term =
					  default_laplace_term,
					const value_type _learning_rate =
					  default_learning_rate) :
	  _base_type(_clusters, _features,
		     _alpha_laplace_term, _theta_laplace_term) {
	}

	/// Copy constructor
	nb_bernouilli_incremental_model
	  (const nb_bernouilli_incremental_model& _other) :
	    _base_type(_other) {
	}

	/// Maximization step
	virtual void
	soft_maximization(const soft_clustering_type& _soft_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  value_type alpha_total_ = value_type();
	  std::fill(this->new_alpha_.begin(), this->new_alpha_.end(),
		    value_type());

	  // Reset thetas
	  std::fill(this->new_theta_.data(),
		    this->new_theta_.data() + this->clusters_ * this->features_,
		    value_type());

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::const_reference epos =
	      _soft_clustering[i];

	    // For each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      this->new_alpha_[c] += epos[c];
	      alpha_total_        += epos[c];

	      // For each feature
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		this->new_theta_[c][fit->first] += epos[c];
	    }
	  }

	  // Stay rate
	  value_type stay_rate = value_type(1) - this->learning_rate_;

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    this->comp_theta_[c] = value_type();
	    for (uint f = 0; f < this->features_; ++f) {
	      // Ratio
	      value_type new_theta_val =
		(this->new_theta_[c][f] + this->theta_laplace_term_) /
		(this->new_alpha_[c] +
		 value_type(2) * this->theta_laplace_term_);

	      // Average with old
	      value_type old = std::exp(this->theta_[c][f]);
	      new_theta_val =
		old / (1.0 + old) * stay_rate +
		new_theta_val * this->learning_rate_;

	      // Set
	      value_type pos_log = std::log(new_theta_val);
	      value_type neg_log = std::log(value_type(1) - new_theta_val);

	      // Normalize theta and complementary
	      this->theta_[c][f]    = pos_log - neg_log;
	      this->comp_theta_[c] += neg_log;
	    }

	    // Normalize alpha
	    value_type new_alpha =
	      (this->alpha_[c] + this->alpha_laplace_term_) /
	      (alpha_total_ +
	       value_type(this->clusters_) * this->alpha_laplace_term_);

	    // Average with old
	    this->alpha_[c] =
	      log(exp(this->alpha_[c]) * stay_rate +
		  new_alpha * this->learning_rate_);
	    this->alpha_comp_[c] = this->alpha_[c] + this->comp_theta_[c];
	  }
	}

	/// Raw soft maximization step
	void
	raw_soft_maximization(const soft_clustering_type& _soft_clustering,
			      const data_type& _data) {
	  // Inherited from base
	  base_type::soft_maximization(_data, _soft_clustering);
	}

	/// Hard maximization step
	virtual void
	hard_maximization(const hard_clustering_type& _hard_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  // value_type alpha_total_ = value_type();
	  std::fill(this->new_alpha_.begin(), this->new_alpha_.end(),
		    value_type());

	  // Reset thetas
	  std::fill(this->new_theta_.data(),
		    this->new_theta_.data() + this->clusters_ * this->features_,
		    value_type());

	  // TODO
	  ttcl_todo;
	}

	/// Raw hard maximization step
	void
	raw_hard_maximization(const hard_clustering_type& _hard_clustering,
			      const data_type& _data) {
	  // Inherited from base
	  base_type::hard_maximization(_data, _hard_clustering);
	}
      };
    }
  }
}

#endif
