#ifndef TTCL_IT_EM_NB_BERNOUILLI_MODEL_HXX
#define TTCL_IT_EM_NB_BERNOUILLI_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Bernouilli Naive Bayes Model
    @author Edgar Gonzalez i Pellicer
*/

#include <cmath>
#include <limits>
#include <vector>

#include <boost/format.hpp>
#include <boost/multi_array.hpp>
#include <boost/utility/enable_if.hpp>

#include <ttcl/co/traits.hxx>
#include <ttcl/it/em/alpha_theta_model.hxx>
#include <ttcl/tu/choices.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Base Bernouilli Naive Bayes Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class _nb_bernouilli_model :
	public alpha_theta_model<Data, Choices> {
      public:
	/// Base type
	typedef alpha_theta_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Parameter allocator type
	typedef typename base_type::parameter_allocator_type
	  parameter_allocator_type;

      protected:
	/// Complementary theta values
	co::array<value_type, parameter_allocator_type> comp_theta_;

	/// Alpha plus comp theta
	/** Cached for efficiency
	 */
	co::array<value_type, parameter_allocator_type> alpha_comp_;

      public:
	/// Empty constructor
	_nb_bernouilli_model() :
	  base_type(), comp_theta_(), alpha_comp_() {
	}

	/// Constructor
	_nb_bernouilli_model(const uint _clusters, const uint _features,
			     const value_type _alpha_laplace_term =
			       default_laplace_term,
			     const value_type _theta_laplace_term =
			       default_laplace_term) :
	  base_type(_clusters, _features,
		    _alpha_laplace_term, _theta_laplace_term),
	  comp_theta_(_clusters), alpha_comp_(_clusters) {
	}

	/// Copy constructor
	_nb_bernouilli_model(const _nb_bernouilli_model& _other) :
	  base_type(_other),
	  comp_theta_(_other.comp_theta_), alpha_comp_(_other.alpha_comp_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Set log alpha value
	virtual void set_log_alpha(uint _cluster, value_type _value) {
	  base_type::set_log_alpha(_cluster, _value);
	  alpha_comp_[_cluster] =
	    this->alpha_[_cluster] + comp_theta_[_cluster];
	}

	/// Log complementary theta value
	value_type log_comp_theta(uint _cluster) const {
	  return comp_theta_[_cluster];
	}

	/// Set log complementary theta value
	void set_log_comp_theta(uint _cluster, value_type _value) {
	  comp_theta_[_cluster] = _value;
	  alpha_comp_[_cluster] =
	    this->alpha_[_cluster] + comp_theta_[_cluster];
	}

	/// Clog theta value
	value_type clog_theta(uint _cluster, uint _feature) const {
	  return this->theta_[_cluster][_feature];
	}

	/// Set clog theta value
	void set_clog_theta(uint _cluster, uint _feature, value_type _value) {
	  this->theta_[_cluster][_feature] = _value;
	}

	/// Theta value
	/** Convenience function
	 */
	virtual value_type theta(uint _cluster, uint _feature) const {
	  double e = std::exp(this->theta_[_cluster][_feature]);
	  return e / (1.0 + e);
	}

	/// @}

	/// Number of free parameters
	virtual uint
	free_parameters() const {
	  return this->clusters_ * (this->features_ + 1) - 1;
	}

      protected:
	/// Resized
	virtual void resized() {
	  // Base
	  base_type::resized();

	  // Resize comp_theta and alpha_comp
	  comp_theta_.resize(this->clusters_);
	  alpha_comp_.resize(this->clusters_);
	}
      };

      /// Bernouilli Naive Bayes Model
      template <typename Data,
		typename Choices = tu::default_choices,
		typename Enable = void>
      class nb_bernouilli_model;

      /// Bernouilli Naive Bayes Model
      /** Specialization for sparse containers
       */
      template <typename Data,
		typename Choices>
      class nb_bernouilli_model
        <Data, Choices,
	 typename boost::enable_if< co::is_sparse_container<Data> >::type> :
	public _nb_bernouilli_model<Data, Choices> {
      private:
	/// Real base type
	typedef _nb_bernouilli_model<Data, Choices> _base_type;

      public:
	/// Base type
	typedef typename _base_type::base_type base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Sample type
	typedef typename data_type::value_type sample_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Empty constructor
	nb_bernouilli_model() :
	  _base_type() {
	}

	/// Constructor
	nb_bernouilli_model(const uint _clusters, const uint _features,
			    const value_type _alpha_laplace_term =
			      default_laplace_term,
			    const value_type _theta_laplace_term =
			      default_laplace_term) :
	  _base_type(_clusters, _features,
		     _alpha_laplace_term, _theta_laplace_term) {
	}

	/// Copy constructor
	nb_bernouilli_model(const nb_bernouilli_model& _other) :
	  _base_type(_other) {
	}

	/// Soft maximization step
	virtual void
	soft_maximization(const soft_clustering_type& _soft_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  value_type alpha_total = value_type();
	  std::fill(this->alpha_.begin(), this->alpha_.end(), value_type());

	  // Reset thetas
	  std::fill(this->theta_.data(),
		    this->theta_.data() + this->clusters_ * this->features_,
		    value_type());

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::const_reference epos =
	      _soft_clustering[i];

	    // For each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      this->alpha_[c] += epos[c];
	      alpha_total     += epos[c];

	      // For each feature
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		this->theta_[c][fit->first] += epos[c];
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    this->comp_theta_[c] = value_type();
	    for (uint f = 0; f < this->features_; ++f) {
	      // Ratio and logarithms
	      value_type theta_val =
		(this->theta_[c][f] + this->theta_laplace_term_) /
		(this->alpha_[c] + value_type(2) * this->theta_laplace_term_);
	      value_type pos_log = std::log(theta_val);
	      value_type neg_log = std::log(value_type(1) - theta_val);

	      // Normalize theta and complementary
	      this->theta_[c][f]    = pos_log - neg_log;
	      this->comp_theta_[c] += neg_log;
	    }

	    // Normalize alpha
	    this->alpha_[c] =
	      std::log((this->alpha_[c] + this->alpha_laplace_term_) /
		       (alpha_total +
			value_type(this->clusters_) *
			  this->alpha_laplace_term_));
	    this->alpha_comp_[c] = this->alpha_[c] + this->comp_theta_[c];
	  }
	}

	/// Hard maximization step
	virtual void
	hard_maximization(const hard_clustering_type& _hard_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  value_type alpha_total = value_type();
	  std::fill(this->alpha_.begin(), this->alpha_.end(), value_type());

	  // Reset thetas
	  std::fill(this->theta_.data(),
		    this->theta_.data() + this->clusters_ * this->features_,
		    value_type());

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Cluster
	    int c = _hard_clustering[i];

	    // Assigned?
	    if (c != -1) {
	      this->alpha_[c] += 1.0;
	      alpha_total     += 1.0;

	      // For each feature
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		this->theta_[c][fit->first] += 1.0;
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    this->comp_theta_[c] = value_type();
	    for (uint f = 0; f < this->features_; ++f) {
	      // Ratio and logarithms
	      value_type theta_val =
		(this->theta_[c][f] + this->theta_laplace_term_) /
		(this->alpha_[c] + value_type(2) * this->theta_laplace_term_);
	      value_type pos_log = std::log(theta_val);
	      value_type neg_log = std::log(value_type(1) - theta_val);

	      // Normalize theta and complementary
	      this->theta_[c][f]    = pos_log - neg_log;
	      this->comp_theta_[c] += neg_log;
	    }

	    // Normalize alpha
	    this->alpha_[c] =
	      std::log((this->alpha_[c] + this->alpha_laplace_term_) /
		       (alpha_total +
			value_type(this->clusters_) *
			  this->alpha_laplace_term_));
	    this->alpha_comp_[c] = this->alpha_[c] + this->comp_theta_[c];
	  }
	}

	/// Soft expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	soft_expectation(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      double xp = this->alpha_comp_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		xp += this->theta_[c][fit->first];

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += (sum = max + std::log(sum));

	    // Normalization
	    for (uint c = 0; c < this->clusters_; ++c)
	      epos[c] = std::exp(epos[c] - sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Hard expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	hard_expectation(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();
	    int     max_cl = -1;

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_comp_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		xp += this->theta_[c][fit->first];

	      // Maximum expectation?
	      if (xp > max) {
		max    = xp;
		max_cl = c;
	      }
	    }

	    // Set
	    log_like           += max;
	    _hard_clustering[i] = max_cl;
	  }

	  // Log like
	  return log_like;
	}

	/// Log-likelihood of data
	virtual value_type
	log_likelihood(const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // Temporary buffer
	  value_type epos[this->clusters_];

	  // For each sample
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit) {
	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_comp_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		xp += this->theta_[c][fit->first];

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += max + std::log(sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Raw soft log-probability
	virtual void
	soft_log_probability(soft_clustering_type& _soft_clustering,
			     const data_type& _data) const {
	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      epos[c] = this->alpha_comp_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		epos[c] += this->theta_[c][fit->first];
	    }
	  }
	}
      };
    }
  }
}

#endif
