#ifndef TTCL_IT_EM_NB_BERNOUILLI_NOISY_MODEL_HXX
#define TTCL_IT_EM_NB_BERNOUILLI_NOISY_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Bernouilli Naive Bayes Noisy Model
    @author Edgar Gonzalez i Pellicer
*/

#include <cmath>
#include <vector>

#include <boost/multi_array.hpp>
#include <boost/utility/enable_if.hpp>

#include <ttcl/co/traits.hxx>
#include <ttcl/it/em/nb_bernouilli_model.hxx>
#include <ttcl/tu/choices.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Base Bernouilli Naive Bayes Noisy Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class _nb_bernouilli_noisy_model :
	public nb_bernouilli_model<Data, Choices> {
      public:
	/// Base type
	typedef nb_bernouilli_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

      protected:
	/// Noisy alpha
	value_type alpha_noisy_;

	/// Noisy alpha comp
	value_type alpha_comp_noisy_;

	/// Noisy theta
	/** Uniform log-probability of noisy distribution
	 */
	value_type comp_theta_noisy_;

      public:
	/// Empty constructor
	_nb_bernouilli_noisy_model() :
	  base_type(),
	  alpha_noisy_(0.0), alpha_comp_noisy_(0.0), comp_theta_noisy_(0.0) {
	}

	/// Constructor
	_nb_bernouilli_noisy_model(const uint _clusters, const uint _features,
				   const value_type _alpha_laplace_term =
				     default_laplace_term,
				   const value_type _theta_laplace_term =
				     default_laplace_term) :
	  base_type(_clusters, _features,
		    _alpha_laplace_term, _theta_laplace_term),
	  alpha_noisy_(0.0), alpha_comp_noisy_(0.0),
	  comp_theta_noisy_(-value_type(_features) * std::log(value_type(2))) {
	}

	/// Copy constructor
	_nb_bernouilli_noisy_model(const _nb_bernouilli_noisy_model& _other) :
	  base_type(_other), alpha_noisy_(_other.alpha_noisy_),
	  alpha_comp_noisy_(_other.alpha_comp_noisy_),
	  comp_theta_noisy_(_other.comp_theta_noisy_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Log noisy alpha value
	value_type log_alpha_noisy() const {
	  return alpha_noisy_;
	}

	/// Set log noisy alpha value
	void set_log_alpha_noisy(value_type _value) {
	  alpha_noisy_      = _value;
	  alpha_comp_noisy_ = alpha_noisy_ + comp_theta_noisy_;
	}

	/// Noisy alpha value
	/** Convenience function
	 */
	value_type alpha_noisy() const {
	  return std::exp(alpha_noisy_);
	}

	/// @}

	/// Number of free parameters
	virtual uint free_parameters() const {
	  return this->clusters_ * (this->features_ + 1);
	}

      protected:
	/// Resized
	virtual void resized() {
	  // Base
	  base_type::resized();

	  // Set theta noisy
	  comp_theta_noisy_ =
	    -value_type(this->features_) * value_type(M_LN2);
	}
      };

      /// Bernouilli Naive Bayes Noisy Model
      template <typename Data,
		typename Choices = tu::default_choices,
		typename Enable = void>
      class nb_bernouilli_noisy_model;

      /// Bernouilli Naive Bayes Noisy Model
      /** Specialization for sparse containers
       */
      template <typename Data,
		typename Choices>
      class nb_bernouilli_noisy_model
        <Data, Choices,
	 typename boost::enable_if< co::is_sparse_container<Data> >::type> :
	public _nb_bernouilli_noisy_model<Data, Choices> {
      private:
	/// Real base type
	typedef _nb_bernouilli_noisy_model<Data, Choices> _base_type;

      public:
	/// Base type
	typedef typename _base_type::base_type base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Sample type
	typedef typename data_type::value_type sample_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Value type
	typedef typename base_type::value_type value_type;

      public:
	/// Empty constructor
	nb_bernouilli_noisy_model() :
	  _base_type() {
	}

	/// Constructor
	nb_bernouilli_noisy_model(const uint _clusters, const uint _features,
				  const value_type _alpha_laplace_term =
				    default_laplace_term,
				  const value_type _theta_laplace_term =
				    default_laplace_term) :
	  _base_type(_clusters, _features,
		     _alpha_laplace_term, _theta_laplace_term) {
	}

	/// Copy constructor
	nb_bernouilli_noisy_model(const nb_bernouilli_noisy_model& _other) :
	  _base_type(_other) {
	}

	/// Soft maximization step
	virtual void
	soft_maximization(const soft_clustering_type& _soft_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  std::fill(this->alpha_.begin(), this->alpha_.end(), value_type());
	  value_type alpha_total_ = value_type();

	  // Reset thetas
	  std::fill(this->theta_.data(),
		    this->theta_.data() + this->clusters_ * this->features_,
		    value_type());

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::const_reference epos =
	      _soft_clustering[i];

	    // For each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      this->alpha_[c] += epos[c];
	      alpha_total_    += epos[c];

	      // For each feature
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		this->theta_[c][fit->first] += epos[c];
	    }
	  }

	  // Noisy component alpha
	  this->alpha_noisy_   = value_type(_data.size()) - alpha_total_;

	  // Corrected total
	  value_type alpha_corr =
	    value_type(_data.size()) +
	    value_type(this->clusters_ + 1) * this->alpha_laplace_term_;

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    this->comp_theta_[c] = value_type();
	    for (uint f = 0; f < this->features_; ++f) {
	      // Ratio and logarithms
	      value_type theta_val =
		(this->theta_[c][f] + this->theta_laplace_term_) /
		(this->alpha_[c] + value_type(2) * this->theta_laplace_term_);
	      value_type pos_log = std::log(theta_val);
	      value_type neg_log = std::log(value_type(1) - theta_val);

	      // Normalize theta and complementary
	      this->theta_[c][f]    = pos_log - neg_log;
	      this->comp_theta_[c] += neg_log;
	    }

	    // Normalize alpha
	    this->alpha_[c] =
	      std::log((this->alpha_[c] + this->alpha_laplace_term_) /
		       alpha_corr);
	    this->alpha_comp_[c] = this->alpha_[c] + this->comp_theta_[c];
	  }

	  // Normalize noisy alpha
	  this->alpha_noisy_ =
	    std::log((this->alpha_noisy_ + this->alpha_laplace_term_) /
		     alpha_corr);
	  this->alpha_comp_noisy_ =
	    this->alpha_noisy_ + this->comp_theta_noisy_;
	}

	/// Soft maximization step
	virtual void
	hard_maximization(const hard_clustering_type& _soft_clustering,
			  const data_type& _data) {
	  // TODO
	  ttcl_todo;
	}

	/// Soft expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	soft_expectation(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // A priori expectation for the noisy cluster
	    value_type max = this->alpha_comp_noisy_;

	    // Expectation for the others
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_comp_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		xp += this->theta_[c][fit->first];

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = std::exp(this->alpha_comp_noisy_ - max);
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += (sum = max + std::log(sum));

	    // Normalization
	    for (uint c = 0; c < this->clusters_; ++c)
	      epos[c] = std::exp(epos[c] - sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Hard expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	hard_expectation(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Set maximum
	    value_type max = this->alpha_comp_noisy_;
	    int     max_cl = -1;

	    // Expectation for the others
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_comp_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		xp += this->theta_[c][fit->first];

	      // Maximum expectation?
	      if (xp > max) {
		max    = xp;
		max_cl = c;
	      }
	    }

	    // Set
	    log_like           += max;
	    _hard_clustering[i] = max_cl;
	  }

	  // Log like
	  return log_like;
	}

	/// Log-likelihood of data
	virtual value_type
	log_likelihood(const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // Temporary buffer
	  value_type epos[this->clusters_];

	  // For each sample
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit) {
	    // A priori expectation for the noisy cluster
	    value_type max = this->alpha_comp_noisy_;

	    // Expectation for the others
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_comp_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		xp += this->theta_[c][fit->first];

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = std::exp(this->alpha_comp_noisy_ - max);
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += max + std::log(sum);
	  }

	  // Log like
	  return log_like;
	}
      };
    }
  }
}

#endif
