#ifndef TTCL_IT_EM_NB_INDEPENDENT_MULTINOMIAL_MODEL_HXX
#define TTCL_IT_EM_NB_INDEPENDENT_MULTINOMIAL_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Independent Multinomial Naive Bayes Model
    @author Edgar Gonzalez i Pellicer
*/

#include <cmath>
#include <vector>

#include <boost/multi_array.hpp>
#include <boost/utility/enable_if.hpp>

#include <ttcl/co/traits.hxx>
#include <ttcl/it/em/choices.hxx>
#include <ttcl/it/em/independent_alpha_theta_model.hxx>
#include <ttcl/tu/choices.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Base Independent Multinomial Naive Bayes Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class _nb_independent_multinomial_model :
	public independent_alpha_theta_model<Data, Choices> {
      public:
	/// Base type
	typedef independent_alpha_theta_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Empty constructor
	_nb_independent_multinomial_model() :
	  base_type() {
	}

	/// Constructor
	_nb_independent_multinomial_model
	  (const uint _clusters, const uint _features,
	   const uint* _feature_cardinalities,
	   const value_type _alpha_laplace_term = default_laplace_term,
	   const value_type _theta_laplace_term = default_laplace_term) :
	    base_type(_clusters, _features, _feature_cardinalities,
		      _alpha_laplace_term, _theta_laplace_term) {
	}

	/// Copy constructor
	_nb_independent_multinomial_model
	  (const _nb_independent_multinomial_model& _other) :
	    base_type(_other) {
	}

	/// Number of free parameters
	virtual uint
	free_parameters() const {
	  uint sum = this->clusters_ - 1;
	  for (uint f = 0; f < this->features_; ++f)
	    sum += this->clusters_ * (this->feature_cardinalities_[f] - 1);
	  return sum;
	}
      };

      /// Independent Multinomial Naive Bayes Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class nb_independent_multinomial_model :
	public _nb_independent_multinomial_model<Data, Choices> {
      private:
	/// Real base type
	typedef _nb_independent_multinomial_model<Data, Choices> _base_type;

      public:
	/// Base type
	typedef typename _base_type::base_type base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Sample type
	typedef typename data_type::value_type sample_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Empty constructor
	nb_independent_multinomial_model() :
	  _base_type() {
	}

	/// Constructor
	nb_independent_multinomial_model
	  (const uint _clusters, const uint _features,
	   const uint* _feature_cardinalities,
	   const value_type _alpha_laplace_term = default_laplace_term,
	   const value_type _theta_laplace_term = default_laplace_term) :
	    _base_type(_clusters, _features, _feature_cardinalities,
		       _alpha_laplace_term, _theta_laplace_term) {
	}

	/// Copy constructor
	nb_independent_multinomial_model
	  (const nb_independent_multinomial_model& _other) :
	  _base_type(_other) {
	}

	/// Soft maximization step
	virtual void
	soft_maximization(const soft_clustering_type& _soft_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  std::fill(this->alpha_.begin(), this->alpha_.end(),
		    value_type());
	  value_type alpha_total_ =
	    value_type(this->clusters_) * this->alpha_laplace_term_;

	  // Reset thetas
	  for (uint c = 0; c < this->clusters_; ++c)
	    for (uint f = 0; f < this->features_; ++f)
	      std::fill(this->theta_[c][f].begin(), this->theta_[c][f].end(),
			value_type());

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::const_reference epos =
	      _soft_clustering[i];

	    // For each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      this->alpha_[c]    += epos[c];
	      alpha_total_       += epos[c];

	      // For each feature
	      for (uint f = 0; f < this->features_; ++f)
		this->theta_[c][f][(*dit)[f]] += epos[c];
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    for (uint f = 0; f < this->features_; ++f) {
	      value_type corr_alpha =
		this->alpha_[c] +
		  value_type(this->feature_cardinalities_[f]) *
		  this->theta_laplace_term_;
	      for (uint v = 0; v < this->feature_cardinalities_[f]; ++v)
		this->theta_[c][f][v] =
		  std::log((this->theta_[c][f][v] + this->theta_laplace_term_) /
			   corr_alpha);
	    }
	    this->alpha_[c] =
	      std::log((this->alpha_[c] + this->alpha_laplace_term_) /
		       alpha_total_);
	  }
	}

	/// Hard maximization step
	virtual void
	hard_maximization(const hard_clustering_type& _hard_clustering,
			  const data_type& _data) {
	  // TODO
	  ttcl_todo;
	}

	/// Soft expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	soft_expectation(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_[c];

	      // Feature a posteriori probabilities
	      for (uint f = 0; f < this->features_; ++f)
		xp += this->theta_[c][f][(*dit)[f]];

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += (sum = max + std::log(sum));

	    // Normalization
	    for (uint c = 0; c < this->clusters_; ++c)
	      epos[c] = std::exp(epos[c] - sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Hard expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	hard_expectation(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();
	    int     max_cl = -1;

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_[c];

	      // Feature a posteriori probabilities
	      for (uint f = 0; f < this->features_; ++f)
		xp += this->theta_[c][f][(*dit)[f]];

	      // Maximum expectation?
	      if (xp > max) {
		max    = xp;
		max_cl = c;
	      }
	    }

	    // Set
	    log_like           += max;
	    _hard_clustering[i] = max_cl;
	  }

	  // Log like
	  return log_like;
	}

	/// Log-likelihood of data
	virtual value_type
	log_likelihood(const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // Temporary buffer
	  value_type epos[this->clusters_];

	  // For each sample
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit) {
	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_[c];

	      // Feature a posteriori probabilities
	      for (uint f = 0; f < this->features_; ++f)
		xp += this->theta_[c][f][(*dit)[f]];

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += max + std::log(sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Raw soft log-probability
	virtual void
	soft_log_probability(soft_clustering_type& _soft_clustering,
			     const data_type& _data) const {
	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      epos[c] = this->alpha_[c];

	      // Feature a posteriori probabilities
	      for (uint f = 0; f < this->features_; ++f)
		epos[c] += this->theta_[c][f][(*dit)[f]];
	    }
	  }
	}
      };
    }
  }
}

#endif
