#ifndef TTCL_IT_EM_NB_MULTINOMIAL_MODEL_HXX
#define TTCL_IT_EM_NB_MULTINOMIAL_MODEL_HXX

// TTCL: The Template Clustering Library

/** @file
    EM - Multinomial Naive Bayes Model
    @author Edgar Gonzalez i Pellicer
*/

#include <cmath>
#include <limits>
#include <vector>

#include <boost/multi_array.hpp>
#include <boost/utility/enable_if.hpp>

#include <ttcl/co/traits.hxx>
#include <ttcl/it/em/alpha_theta_model.hxx>
#include <ttcl/tu/choices.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Expectation-Maximization Namespace
    namespace em {

      /// Base Multinomial Naive Bayes Model
      template <typename Data,
		typename Choices = tu::default_choices>
      class _nb_multinomial_model :
	public alpha_theta_model<Data, Choices> {
      public:
	/// Base type
	typedef alpha_theta_model<Data, Choices> base_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Parameter allocator type
	typedef typename base_type::parameter_allocator_type
	  parameter_allocator_type;

      protected:
	/// Totals
	/** For theta estimation
	 */
	co::array<value_type, parameter_allocator_type> theta_total_;

      public:
	/// Empty constructor
	_nb_multinomial_model() :
	  base_type(), theta_total_() {
	}

	/// Constructor
	_nb_multinomial_model(const uint _clusters, const uint _features,
			      const value_type _alpha_laplace_term =
			        default_laplace_term,
			      const value_type _theta_laplace_term =
			        default_laplace_term) :
	  base_type(_clusters, _features,
		    _alpha_laplace_term, _theta_laplace_term),
	  theta_total_(_clusters) {
	}

	/// Copy constructor
	_nb_multinomial_model(const _nb_multinomial_model& _other) :
	  base_type(_other), theta_total_(_other.theta_total_) {
	}

	/// @name Setters/Getters
	/// @{

	/// Log theta value
	value_type log_theta(uint _cluster, uint _feature) const {
	  return this->theta_[_cluster][_feature];
	}

	/// Set log theta value
	void set_log_theta(uint _cluster, uint _feature,
			   value_type _value) {
	  this->theta_[_cluster][_feature] = _value;
	}

	/// Theta value
	/** Convenience function
	 */
	virtual value_type theta(uint _cluster, uint _feature) const {
	  return std::exp(this->theta_[_cluster][_feature]);
	}

	/// @}

	/// Number of free parameters
	virtual uint free_parameters() const {
	  return this->clusters_ * this->features_ - 1;
	}

      protected:
	/// Resized
	virtual void resized() {
	  // Base
	  base_type::resized();

	  // Resize theta_total
	  theta_total_.resize(this->clusters_);
	}
      };

      /// Multinomial Naive Bayes Model
      template <typename Data,
		typename Choices = tu::default_choices,
		typename Enable = void>
      class nb_multinomial_model;

      /// Multinomial Naive Bayes Model
      /** Specialization for sparse containers
       */
      template <typename Data,
		typename Choices>
      class nb_multinomial_model
        <Data, Choices,
	 typename boost::enable_if< co::is_sparse_container<Data> >::type> :
	public _nb_multinomial_model<Data, Choices> {
      private:
	/// Real base type
	typedef _nb_multinomial_model<Data, Choices> _base_type;

      public:
	/// Base type
	typedef typename _base_type::base_type base_type;

	/// Data type
	typedef typename base_type::data_type data_type;

	/// Sample type
	typedef typename data_type::value_type sample_type;

	/// Hard clustering type
	typedef typename base_type::hard_clustering_type hard_clustering_type;

	/// Soft clustering type
	typedef typename base_type::soft_clustering_type soft_clustering_type;

	/// Value type
	typedef typename base_type::value_type value_type;

	/// Parameter allocator type
	typedef typename base_type::parameter_allocator_type
	  parameter_allocator_type;

	/// Empty constructor
	nb_multinomial_model() :
	  _base_type() {
	}

	/// Constructor
	nb_multinomial_model(const uint _clusters, const uint _features,
			     const value_type _alpha_laplace_term =
			     default_laplace_term,
			     const value_type _theta_laplace_term =
			     default_laplace_term) :
	  _base_type(_clusters, _features,
		     _alpha_laplace_term, _theta_laplace_term) {
	}

	/// Copy constructor
	nb_multinomial_model(const nb_multinomial_model& _other) :
	  _base_type(_other) {
	}

	/// Soft maximization step
	virtual void
	soft_maximization(const soft_clustering_type& _soft_clustering,
			  const data_type& _data) {
	  // Reset alphas
	  value_type alpha_total_ =
	    value_type(this->clusters_) * this->alpha_laplace_term_;
	  std::fill(this->alpha_.begin(), this->alpha_.end(),
		    this->alpha_laplace_term_);

	  // Reset thetas
	  std::fill(this->theta_.data(),
		    this->theta_.data() + this->clusters_ * this->features_,
		    this->theta_laplace_term_);
	  std::fill(this->theta_total_.begin(),
		    this->theta_total_.end(),
		    value_type(this->features_) * this->theta_laplace_term_);

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::const_reference epos =
	      _soft_clustering[i];

	    // For each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      this->alpha_[c]    += epos[c];
	      alpha_total_       += epos[c];

	      // For each feature
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit) {
		this->theta_[c][fit->first] += epos[c] * fit->second;
		this->theta_total_[c]       += epos[c] * fit->second;
	      }
	    }
	  }

	  // Normalize
	  for (uint c = 0; c < this->clusters_; ++c) {
	    this->alpha_[c] = std::log(this->alpha_[c] / alpha_total_);
	    for (uint f = 0; f < this->features_; ++f)
	      this->theta_[c][f] =
		std::log(this->theta_[c][f] / this->theta_total_[c]);
	  }
	}

	/// Hard maximization step
	virtual void
	hard_maximization(const hard_clustering_type& _hard_clustering,
			  const data_type& _data) {
	  // TODO
	  ttcl_todo;
	}

	/// Soft expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	soft_expectation(soft_clustering_type& _soft_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		xp += fit->second * this->theta_[c][fit->first];

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += (sum = max + std::log(sum));

	    // Normalization
	    for (uint c = 0; c < this->clusters_; ++c)
	      epos[c] = std::exp(epos[c] - sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Hard expectation step
	/** @return Log-likelihood of data
	 */
	virtual value_type
	hard_expectation(hard_clustering_type& _hard_clustering,
			 const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();
	    uint    max_cl = -1;

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		xp += fit->second * this->theta_[c][fit->first];

	      // Maximum expectation?
	      if (xp > max) {
		max    = xp;
		max_cl = c;
	      }
	    }

	    // Set
	    log_like           += max;
	    _hard_clustering[i] = max_cl;
	  }

	  // Log like
	  return log_like;
	}

	/// Log-likelihood of data
	virtual value_type
	log_likelihood(const data_type& _data) const {
	  // Log-likelihood
	  value_type log_like = value_type();

	  // Temporary buffer
	  value_type epos[this->clusters_];

	  // For each sample
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit) {
	    // Starting maximum
	    value_type max = -std::numeric_limits<value_type>::infinity();

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      value_type xp = this->alpha_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		xp += fit->second * this->theta_[c][fit->first];

	      // Maximum expectation?
	      if ((epos[c] = xp) > max)
		max = xp;
	    }

	    // Sum
	    value_type sum = value_type();
	    for (uint c = 0; c < this->clusters_; ++c)
	      sum += std::exp(epos[c] - max);
	    log_like += max + std::log(sum);
	  }

	  // Log like
	  return log_like;
	}

	/// Raw soft log-probability
	virtual void
	soft_log_probability(soft_clustering_type& _soft_clustering,
			     const data_type& _data) const {
	  // For each sample
	  uint i = 0;
	  for (typename data_type::const_iterator dit = _data.begin();
	       dit != _data.end(); ++dit, ++i) {
	    // Expectation
	    typename soft_clustering_type::reference epos =
	      _soft_clustering[i];

	    // Expectation for each cluster
	    for (uint c = 0; c < this->clusters_; ++c) {
	      // A priori expectation
	      epos[c] = this->alpha_[c];

	      // Feature a posteriori probabilities
	      for (typename sample_type::const_iterator fit = dit->begin();
		   fit != dit->end(); ++fit)
		epos[c] += fit->second * this->theta_[c][fit->first];
	    }
	  }
	}
      };
    }
  }
}

#endif
