#ifndef TTCL_IT_HARD_ITERATIVE_CLUSTERER_HXX
#define TTCL_IT_HARD_ITERATIVE_CLUSTERER_HXX

// TTCL: The Template Clustering Library

/** @file
    Iterative Algorithms - Hard Iterative Clusterer
    @author Edgar Gonzalez i Pellicer
*/

#include <boost/tuple/tuple.hpp>

#include <ttcl/types.hxx>
#include <ttcl/cl/hard_clustering.hxx>
#include <ttcl/cl/hard_data_clusterer.hxx>
#include <ttcl/it/choices.hxx>
#include <ttcl/it/randomized_clusterer.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/tu/tuning.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {
    
    /// Hard Iterative Clusterer
    template <typename Data,
	      typename Choices = tu::default_choices>
    class hard_iterative_clusterer :
      public cl::hard_data_clusterer<Data, Choices>,
      public randomized_clusterer<Choices> {
    public:
      /// Base type
      typedef cl::hard_data_clusterer<Data, Choices> base_type;

      /// Data type
      typedef typename base_type::data_type data_type;
	
      /// Hard clustering type
      typedef typename base_type::hard_clustering_type hard_clustering_type;

      /// Randomized type
      typedef randomized_clusterer<Choices> randomized_type;

      /// Random number generator
      typedef typename randomized_type::random_generator_type
        random_generator_type;

      /// Model type
      typedef hard_iterative_model<Data, Choices> model_type;

      /// Value type
      typedef typename model_type::value_type value_type;
				     
      /// Model return type
      typedef boost::tuple<model_type*, value_type, uint> model_return_type;

    protected:
      /// Max iterations
      uint max_iterations_;
	
      /// Stop threshold
      value_type stop_threshold_;

    public:
      /// Constructor
      hard_iterative_clusterer(const random_generator_type& _random_generator =
			         random_generator_type()) :
	randomized_type(_random_generator),
	max_iterations_(default_max_iterations),
	stop_threshold_(default_stop_threshold) {
      }

      /// Tune
      virtual void
      tune(const tu::tuning& _tuning) {
	// Parent
	randomized_type::tune(_tuning);

	// Extract the options
	max_iterations_ = _tuning.get<max_iterations>();
	stop_threshold_ = _tuning.get<stop_threshold>();
      }

      /// Model
      /** @param _hard_clustering Target hard clustering
	  @param _data Data
      */
      model_return_type
      hard_model_from(hard_clustering_type& _hard_clustering,
		      const data_type& _data) const {
	// Create a model
	model_type* model = create_model(_data,
					 _hard_clustering.clusters());

	// Starting refinement
	value_type fitness = model->hard_refine(_data, _hard_clustering);

	// Loop
	uint i;
	for (i = 1; i < max_iterations_; ++i) {
	  // Old
	  value_type old_fitness = fitness;

	  // Next refinement
	  fitness = model->hard_refine(_data, _hard_clustering);

	  // Change ratio below threshold?
	  if ((old_fitness - fitness) / old_fitness < stop_threshold_)
	    break;
	}

	// Return the model, 
	return boost::make_tuple(model, fitness, i);
      }

      /// Model
      /** @param _hard_clustering Target hard clustering
	  @param _data Data
	  @param _n_clusters Number of clusters
      */
      model_return_type
      hard_model_on(hard_clustering_type& _hard_clustering,
		    const data_type& _data,
		    const uint _n_clusters) const {
	// Randomize
	hard_randomize(_hard_clustering, _n_clusters);

	// Model from it
	return hard_model_from(_hard_clustering, _data);
      }
	  
      /// Model
      /** @param _data Data
	  @param _n_clusters Number of clusters
      */
      model_return_type
      hard_model(const data_type& _data,
		 const uint _n_clusters) const {
	// Create a hard clustering
	hard_clustering_type hard_clustering(_data.size(), _n_clusters);
	  
	// Return a model
	return hard_model_on(hard_clustering, _data, _n_clusters);
      }

      /// Refine a hard clustering
      /** @param _hard_clustering Target hard clustering
	  @param _data Data
      */
      void
      hard_cluster_from(hard_clustering_type& _hard_clustering,
			const data_type& _data,
			const uint /* _n_clusters */) const {
	// Cluster using a model
	model_type* model =
	  boost::get<0>(hard_model_from(_hard_clustering, _data));
	
	// Delete the model
	delete model;
      }

      /// Perform the hard clustering
      /** @param _hard_clustering Target hard clustering
	  @param _data Data
	  @param _n_clusters Number of clusters
      */
      virtual void
      hard_cluster_on(hard_clustering_type& _hard_clustering,
		      const data_type& _data,
		      const uint _n_clusters) const {
	// Cluster using a model
	model_type* model =
	  boost::get<0>(hard_model_on(_hard_clustering, _data, _n_clusters));
	
	// Delete the model
	delete model;
      }

      /// Create a model
      virtual model_type*
      create_model(const data_type& _data,
		   const uint _n_clusters) const = 0;
    };
  }
}

#endif
