#ifndef TTCL_IT_RANDOMIZED_CLUSTERER_HXX
#define TTCL_IT_RANDOMIZED_CLUSTERER_HXX

// TTCL: The Template Clustering Library

/** @file
    Iterative Algorithms - Randomized Clusterer
    @author Edgar Gonzalez i Pellicer
*/

#include <algorithm>
#include <vector>

#include <ttcl/types.hxx>
#include <ttcl/cl/choices.hxx>
#include <ttcl/cl/hard_clustering.hxx>
#include <ttcl/cl/soft_clustering.hxx>
#include <ttcl/it/choices.hxx>
#include <ttcl/tu/choices.hxx>
#include <ttcl/tu/tuning.hxx>
#include <ttcl/ut/choices.hxx>
#include <ttcl/ut/range.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Iterative Clustering Namespace
  namespace it {

    /// Randomized clusterer
    template <typename Choices = ttcl::tu::default_choices>
    class randomized_clusterer {
    public:
      /// Hard clustering container
      typedef ttcl_tu_get_choice(Choices, cl::hard_clustering_container)
	hard_clustering_container_type;

      /// Hard clustering type
      typedef cl::hard_clustering<hard_clustering_container_type>
        hard_clustering_type;

      /// Soft clustering container
      typedef ttcl_tu_get_choice(Choices, cl::soft_clustering_container)
	soft_clustering_container_type;

      /// Soft clustering type
      typedef cl::soft_clustering<soft_clustering_container_type>
        soft_clustering_type;

      /// Value type
      typedef ttcl_tu_get_choice(Choices, model_value_type) value_type;

      /// Random number generator
      typedef ttcl_tu_get_choice_tem(Choices,
				     ut::random_generator_template,
				     value_type) random_generator_type;

    private:
      /// Start policy
      start_policy_enum start_policy_;

      /// Peak fraction
      value_type peak_fraction_;
 
    protected:
      /// Random generator
      mutable random_generator_type random_generator_;

    public:
      /// Constructor
      randomized_clusterer(const random_generator_type& _random_generator =
			     random_generator_type()) :
	start_policy_(DEFAULT_START),
	peak_fraction_(default_peak_fraction),
	random_generator_(_random_generator) {
      }

      /// Destructor
      virtual ~randomized_clusterer() {
      }

      /// Tune
      virtual void
      tune(const tu::tuning& _tuning) {
	// Extract the options
	start_policy_  = _tuning.get<start_policy>  ();
	peak_fraction_ = _tuning.get<peak_fraction> ();
      }

      /// Randomize a hard clustering
      void hard_randomize(hard_clustering_type& _hard_clustering,
			  const uint _n_clusters) const {
	// Start?
	switch (start_policy_) {
	case START_CENTROIDS: // Select centroids
	case START_PEAKS:     // (Equivalent for hard)
	  hard_select_centroids(_hard_clustering, _n_clusters);
	  break;

	case START_RANDOM: // Select at random
	  hard_select_random(_hard_clustering, _n_clusters);
	  break;
	}
      }

      /// Randomize a soft clustering
      void soft_randomize(soft_clustering_type& _soft_clustering) const {
	// Start?
	switch (start_policy_) {
	case START_CENTROIDS: // Select centroids
	  soft_select_centroids(_soft_clustering);
	  break;
	  
	case START_PEAKS: // Select peaks
	  soft_select_peaks(_soft_clustering);
	  break;

	case START_RANDOM: // Select at random
	  soft_select_random(_soft_clustering);
	  break;
	}
      }

    protected:
      /// Select centroids for a hard clustering
      virtual void
      hard_select_centroids(hard_clustering_type& _hard_clustering,
			    const uint _n_clusters) const {
	// Enough?
	if (_n_clusters >= _hard_clustering.size()) {
	  // Assign each to one
	  std::copy(ut::range(0u), ut::range(_n_clusters),
		    _hard_clustering.begin());
	}
	else {
	  // Take a range
	  std::vector<size_t> shuffle(ut::range(size_t(0)),
				      ut::range(_hard_clustering.size()));
	  std::random_shuffle(shuffle.begin(), shuffle.end());
	  shuffle.resize(_n_clusters);
	  std::sort(shuffle.begin(), shuffle.end());

	  // Now, reset and assign
	  std::fill(_hard_clustering.begin(), _hard_clustering.end(), -1);
	  for (uint i = 0; i < _n_clusters; ++i)
	    _hard_clustering[shuffle[i]] = i;
	}

	// Update the number
	_hard_clustering.update_clusters(_n_clusters);
      }

      /// Select at random for hard clustering
      virtual void
      hard_select_random(hard_clustering_type& _hard_clustering,
			 const uint _n_clusters) const {
	// For each sample
	for (typename hard_clustering_type::iterator it =
	       _hard_clustering.begin(); it != _hard_clustering.end(); ++it)
	  // Generate
	  *it = random_generator_(_n_clusters);

	// Update the number
	_hard_clustering.update_clusters(_n_clusters);
      }

      /// Select centroids for a soft clustering
      virtual void
      soft_select_centroids(soft_clustering_type& _soft_clustering) const {
	// Clear the clustering
	_soft_clustering.clear();
		    
	// Enough?
	if (_soft_clustering.clusters() >= _soft_clustering.size()) {
	  // Assign each to one
	  for (uint c = 0; c < _soft_clustering.clusters(); ++c)
	    _soft_clustering[c][c] = value_type(1.0);
	}
	else {
	  // Take a range
	  std::vector<size_t> shuffle(ut::range(size_t(0)),
				      ut::range(_soft_clustering.size()));
	  std::random_shuffle(shuffle.begin(), shuffle.end());
	  shuffle.resize(_soft_clustering.clusters());
	  std::sort(shuffle.begin(), shuffle.end());

	  // Now, assign
	  for (uint c = 0; c < _soft_clustering.clusters(); ++c)
	    _soft_clustering[shuffle[c]][c] = value_type(1.0);
	}
      }

      /// Select peaks for a soft clustering
      virtual void
      soft_select_peaks(soft_clustering_type& _soft_clustering) const {
	// Enough?
	if (_soft_clustering.clusters() >= _soft_clustering.size()) {
	  // Clear the clustering
	  _soft_clustering.clear();
		    
	  // Assign each to one
	  for (uint c = 0; c < _soft_clustering.clusters(); ++c)
	    _soft_clustering[c][c] = value_type(1.0);
	}
	else {
	  // Clear the clustering
	  _soft_clustering.clear(peak_fraction_ /
				 value_type(_soft_clustering.clusters()));

	  // Take a range
	  std::vector<size_t> shuffle(ut::range(size_t(0)),
				      ut::range(_soft_clustering.size()));
	  std::random_shuffle(shuffle.begin(), shuffle.end());
	  shuffle.resize(_soft_clustering.clusters());
	  std::sort(shuffle.begin(), shuffle.end());

	  // Now, assign
	  for (uint c = 0; c < _soft_clustering.clusters(); ++c) {
	    std::fill(_soft_clustering[shuffle[c]].begin(),
		      _soft_clustering[shuffle[c]].end(),
		      value_type());
	    _soft_clustering[shuffle[c]][c] = value_type(1.0);
	  }
	}
      }

      /// Select at random for a soft clustering
      virtual void
      soft_select_random(soft_clustering_type& _soft_clustering) const {
	// For each sample
	for (typename soft_clustering_type::size_type i = 0;
	     i < _soft_clustering.size(); ++i) {
	  // For each cluster
	  value_type sum = value_type();
	  for (typename soft_clustering_type::size_type cl = 0; 
	       cl < _soft_clustering.clusters(); ++cl) 
	    sum += (_soft_clustering[i][cl] = random_generator_());

	  // Normalize the sum
	  for (typename soft_clustering_type::size_type cl = 0; 
	       cl < _soft_clustering.clusters(); ++cl) 
	    _soft_clustering[i][cl] /= sum;
	}
      }
    };
  }
}

#endif
