#ifndef TTCL_SYS_EXCEPTION_HXX
#define TTCL_SYS_EXCEPTION_HXX

// TTCL: The Template Clustering Library

/** @file
    System Exceptions
    @author Edgar Gonzalez i Pellicer
*/

#include <errno.h>

#include <cstring>

#include <ttcl/exception.hxx>

/// TTCL Namespace
namespace ttcl {

  /// System Exception
  class sys_exception : public exception {
  protected:
    // The error number
    int error_;

  public:
    /// Constructor
    /** @param _file    Exception source file
	@param _line_no Exception source line number
	@param _error   The error number
	@param _message Description message
    */
    sys_exception(const std::string& _file, uint _line_no, int _error,
		  const std::string& _message) :
      exception(_file, _line_no, _message), error_(_error) {
    }

    /// Constructor from a boost::format
    /** @param _message Description message
	@param _file    Exception source file
	@param _line_no Exception source line number
	@param _error   The error number
    */
    sys_exception(const std::string& _file, uint _line_no, int _error,
		  const boost::format& _message) :
      exception(_file, _line_no, _message), error_(_error) {
    }
    
    /// Constructor from a char* and a variable argument list
    /** printf-style interpolation
     */
    sys_exception(const std::string& _file, uint _line_no, int _error,
		  const char* _format, ...) ttcl_printf_check(5, 6) :
      exception(_file, _line_no), error_(_error) {
      // Create the list
      std::va_list args;
      va_start(args, _format);

      // Call
      fill_printf(_format, args);

      // End
      va_end(args);
    }

    /// Copy Constructor
    /** @param _other Source exception
     */
    sys_exception(const sys_exception& _other) :
      exception(_other), error_(_other.error_) {
    }

    /// Get the error number
    int error() const {
      return error_;
    }

    /// Write to an ostream
    /** @param _os Target ostream
     */
    virtual void display(std::ostream& _os) const {
      _os << message_ << " caused by " << std::strerror(error_)
	  << " in " << file_ << ":" << line_no_ << std::endl;
#ifdef _EXECINFO_H
      for (int i = 0; i < n_filled_; ++i)
	_os << " from " << trace_[i] << std::endl;
#endif
    }

    /// Description
    virtual const char* what() const
      throw() {
      static char buffer[TTCL_WHAT_SIZE];
      int printed = snprintf(buffer, TTCL_WHAT_SIZE - 1,
			     "%s caused by %s in %s:%d",
			     message_.c_str(), std::strerror(error_),
			     file_.c_str(), line_no_);
#ifdef _EXECINFO_H
      int i = 0;
      while (printed < TTCL_WHAT_SIZE - 1 and
	     i < n_filled_) {
	printed += snprintf(buffer + printed, TTCL_WHAT_SIZE - printed,
			    " from %s", trace_[i]);
	++i;
      }
#endif
      return buffer;
    }
  };
}

/// Throw a system exception
#define ttcl_sys_fire(...)						\
  throw ttcl::sys_exception(__FILE__, __LINE__, errno, __VA_ARGS__)

/// Throw a custom system exception
#define ttcl_custom_sys_fire(...)				\
  throw ttcl::sys_exception(__FILE__, __LINE__, __VA_ARGS__)

#endif
