#ifndef TTCL_UT_TUNING_HXX
#define TTCL_UT_TUNING_HXX

// TTCL: The Template Clustering Library

/** @file
    Utilities - Options and Tuning
    @author Edgar Gonzlez i Pellicer
*/

#include <vector>

#include <ttcl/exception.hxx>
#include <ttcl/types.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Tuning Namespace
  namespace tu {

    /// Option
    /** Heavily based on std::locale::facet

	Children must have an option_id id field
     */
    class option {
    private:
      /// Number of references
      uint refs_;

    protected:
      /// Constructor
      explicit option(int _refs = 0) :
	refs_(_refs) {
      }

      /// Destructor
      virtual ~option() {
      }

      // Friends
      friend class tuning;
      friend class tuning_imp;
    };

    /// Option id
    /** Heavily based on std::locale::facet::id
     */
    class option_id {
    private:
      /// The index
      uint index_;

      /// Number of ids
      static uint total_;

    public:
      /// Constructor
      option_id() {}

      // Friend declarations
      friend class option;
      friend class tuning;
    };

    /// Tuning implementation
    /** Heavily based on std::locale::imp
     */
    struct tuning_imp {
      /// Number of references
      uint refs_;

      /// Options
      std::vector<option*> options_;

      /// Default constructor
      tuning_imp() :
	refs_(1) {
      }

      /// Copy constructor
      tuning_imp(const tuning_imp& _other) :
	refs_(1) {
	// Copy every option
	for (std::vector<option*>::const_iterator it =
	       _other.options_.begin();
	     it != _other.options_.end(); ++it) {
	  options_.push_back(*it);
	  if (*it)
	    ++(*it)->refs_;
	}
      }

      /// Destructor
      ~tuning_imp() {
	// Free every option
	for (std::vector<option*>::iterator it = options_.begin();
	     it != options_.end(); ++it)
	  if (*it and --(*it)->refs_ == 0)
	    delete *it;
      }
    };

    /// Tuning
    /** Heavily based on std::locale
     */
    class tuning {
    private:
      /// Implementation
      tuning_imp* imp_;

      /// Default implementation
      static tuning_imp default_imp_;

    public:
      /// Default constructor
      tuning() :
	imp_(&default_imp_) {
	// One more reference
	++default_imp_.refs_;
      }

      /// Copy constructor
      /** @param _other Source tuning
       */
      tuning(const tuning& _other) :
	imp_(_other.imp_) {
	// One more reference
	++imp_->refs_;
      }

      /// Destructor
      ~tuning() {
	// Free the implementation
	if (--imp_->refs_ == 0)
	  delete imp_;
      }

      /// Assignment operator
      /** @param _other Source tuning
       */
      tuning& operator=(const tuning& _other) {
	// Check it's not the same
	if (imp_ != _other.imp_) {
	  // Free the current one
	  if (--imp_->refs_ == 0)
	    delete imp_;

	  // Assign and increase
	  imp_ = _other.imp_;
	  ++imp_->refs_;
	}

	// Return this
	return *this;
      }

      /// Add an option
      /** @param _opt Option
       */
      template <class Option>
      tuning& set(Option* _opt) {
	// Sharing the implementation?
	if (imp_->refs_ > 1) {
	  // Detach
	  tuning_imp* new_imp = new tuning_imp(*imp_);
	  --imp_->refs_;
	  imp_ = new_imp;
	}
	  
	// Get the option index
	uint index = Option::id_.index_;
	if (not index)
	  // Set a new one
	  index = Option::id_.index_ = ++option_id::total_;

	// Do we have place?
	if (index >= imp_->options_.size())
	  imp_->options_.resize(index, 0);

	// Something to replace?
	if (imp_->options_[index - 1] and
	    --imp_->options_[index - 1]->refs_ == 0)
	  delete imp_->options_[index - 1];

	// Save it
	imp_->options_[index - 1] = _opt;
	++_opt->option::refs_;

	// Return this
	return *this;
      }

      /// Remove an option
      template <class Option>
      void remove() const {
	// Look for it
	uint index = Option::id_.index_;
	if (index - 1 < imp_->options_.size() and
	    imp_->options_[index - 1]) {
	  if (--imp_->options_[index - 1]->refs_ == 0)
	    delete imp_->options_[index - 1];
	  imp_->options_[index - 1] = 0;
	}
      }

      /// Retrieve an option
      template <class Option>
      const Option& get() const {
	// Get the option index
	uint index = Option::id_.index_;
	if (not index)
	  // Set a new one
	  index = Option::id_.index_ = ++option_id::total_;

	// Do we have place?
	if (index >= imp_->options_.size())
	  imp_->options_.resize(index, 0);

	// Is it null?
	if (not imp_->options_[index - 1]) {
	  // Does the default have place?
	  if (index >= default_imp_.options_.size())
	    default_imp_.options_.resize(index, 0);

	  // Is it null?
	  if (not default_imp_.options_[index - 1]) {
	    default_imp_.options_[index - 1] =
	      new Option(Option::default_value());
	    ++default_imp_.options_[index - 1]->refs_;
	  }

	  // Copy and set
	  imp_->options_[index - 1] =
	    default_imp_.options_[index - 1];
	  ++imp_->options_[index - 1]->refs_;
	}
	    
	// Cast it
	return static_cast<const Option&>(*(imp_->options_[index - 1]));
      }
    };
  }
}

/// Define an option
#define ttcl_tu_option(type, name, default)	\
  class name : public ttcl::tu::option {	\
  public:					\
    static ttcl::tu::option_id id_;		\
    						\
    static type default_value() {		\
      return default;				\
    }						\
    						\
  private:					\
    type value_;				\
						\
  public:					\
    name(type _value = type(0)) :		\
      value_(_value) {				\
    }						\
						\
    operator type() const {			\
      return value_;				\
    }						\
  };

/// Generic static
#define ttcl_tu_static					\
  ttcl::uint ttcl::tu::option_id::total_;		\
  ttcl::tu::tuning_imp ttcl::tu::tuning::default_imp_;

/// Option static vars
#define ttcl_tu_option_static(name)		\
  ttcl::tu::option_id name::id_;

#endif
