#ifndef TTCL_UT_ARRAY_AUTO_PTR_HXX
#define TTCL_UT_ARRAY_AUTO_PTR_HXX

// TTCL: The Template Clustering Library

/** @file
    Utilities - C++ Array Auto-Ptr
    @author Edgar Gonzalez i Pellicer
*/

#include <cstdlib>

/// TTCL Namespace
namespace ttcl {

  /// Utility Namespace
  namespace ut { 

    /// Custom Function Auto-Ptr
    /** With the same interface as std::auto_ptr
     */
    template <typename T>
    class array_auto_ptr {
    public:
      /// Element type
      typedef T element_type;

      /// Size type
      typedef std::size_t size_type;

    private:
      /// Pointer
      element_type* pointer_;

      /// Disabled default copy constructor
      array_auto_ptr(const array_auto_ptr&);

      /// Disabled default assignment operator
      array_auto_ptr& operator=(const array_auto_ptr&);

    public:
      /// Construction
      explicit array_auto_ptr(element_type* _pointer = 0) :
	pointer_(_pointer) {
      }

      /// Copy construction
      array_auto_ptr(array_auto_ptr& _other) :
	pointer_(_other.pointer_) {
	// Release the other
	_other.pointer_ = 0;
      }

      /// Assignment
      array_auto_ptr& operator=(array_auto_ptr& _other) {
	// Check for self-assignment
	if (this != &_other) {
	  // Delete
	  delete[] pointer_;

	  // Copy
	  pointer_ = _other.pointer_;

	  // Release the other
	  _other.pointer_ = 0;
	}

	// Return this
	return *this;
      }

      /// Destructor
      ~array_auto_ptr() throw () {
	// Delete
	delete[] pointer_;
      }

      /// Dereference
      element_type& operator*() const throw() {
	// Access
	return *pointer_;
      }

      /// Dereference
      element_type* operator->() const throw() {
	// Access
	return pointer_;
      }

      /// Dereference
      element_type& operator[](size_type _idx) const throw() {
	// Access
	return pointer_[_idx];
      }

      /// Get
      element_type* get() const throw() {
	// Access
	return pointer_;
      }

      /// Release
      element_type* release() throw() {
	// Save, release and return
	element_type p = pointer_;
	pointer_ = 0;
	return p;
      }

      /// Reset
      void reset(element_type* _pointer = 0) throw() {
	// Delete
	delete[] pointer_;

	// Set
	pointer_ = _pointer;
      }
    };
  }
}

#endif
