#ifndef TTCL_UT_BINARY_FILE_HXX
#define TTCL_UT_BINARY_FILE_HXX

// TTCL: The Template Clustering Library

/** @file
    Utility - Binary file (A front-end for libbfd)
    Adapted from addr2line.c, by Ulrich.Lauther@mchp.siemens.de
    from the GNU Binutils package v2.18
    @author Edgar Gonzalez i Pellicer
*/

#include <bfd.h>

#include <cstdlib>
#include <vector>

#include <boost/format.hpp>

#include <ttcl/exception.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Utility Namespace
  namespace ut { 

    /// Binary file
    /** A front-end for libbfd
     */
    class binary_file {
    public:
      /// Address type
      typedef bfd_vma address;

    private:
      /// BFD instance
      bfd* abfd_;

      /// Symbol table
      asymbol** symbol_table_;

      /// Symbol table size
      uint symbol_table_size_;

      /// Initialization
      static bool initialized_;

    public:
      /// Address information
      struct address_info {
      public:
	/// File name
	std::string file;

	/// Line number
	int line;

	/// Function name
	std::string function;

	/// Empty constructor
	address_info() :
	  file("??"), line(0), function("??") {
	}

	/// Constructor
	address_info(const std::string& _file, const int _line,
		     const std::string& _function) :
	  file(_file), line(_line), function(_function) {
	}
      };

      /// Empty constructor
      binary_file() :
	abfd_(0), symbol_table_(0) {
      }

      /// Constructor
      binary_file(const char* _file) :
	abfd_(0), symbol_table_(0) {
	// Open
	open(_file);
      }

      /// Destructor
      ~binary_file() {
	// Close
	close();
      }

      /// Open
      void open(const char* _file) {
	// Verify initialization
	verify_initialized();

	// Close, just in case
	close();

	// Open the file
	if (not (abfd_ = bfd_openr(_file, 0)))
	  ttcl_fire(boost::format("Cannot open file: %s") %
		    bfd_errmsg(bfd_get_error()));
	
	// Check format
	if (not bfd_check_format(abfd_, bfd_object))
	  ttcl_fire("Wrong file format");

	// Slurp the symbol table
	slurp_symbol_table();
      }

      /// Close
      void close() {
	// Open?
	if (is_open()) {
	  // Free the symbols
	  std::free(symbol_table_);
	  
	  // Close the bfd
	  bfd_close(abfd_);

	  // Set both to null
	  abfd_         = 0;
	  symbol_table_ = 0;
	}
      }

      /// Is it open?
      bool is_open() const {
	// Is the structure null?
	return abfd_;
      }

      /// Find address information
      std::vector<address_info> operator[](address _pc) {
	// Return value
	std::vector<address_info> result;

	// Results
	const char* file;
	const char* func;
	uint        line;

	// Iterate over the sections
	bfd_section* s = abfd_->sections;
	while (s and not find_address_in_section(s, _pc, file, func, line))
	  s = s->next;

	// Found?
	if (not s) {
	  // Set an empty value
	  result.push_back(address_info());
	}
	else {
	  // Expand inlining
	  bool more = true;
	  while (more) {
	    // Filename?
	    if (not file)
	      file = "??";

	    // Demangle the name
	    char* demangled =
	      bfd_demangle(abfd_, func, 0x03 /* DMGL_ANSI | DMGL_PARAMS */);

	    // Could demangle?
	    if (demangled) {
	      // Add and free
	      result.push_back(address_info(file, line, demangled));
	      std::free(demangled);
	    }
	    else {
	      // Add
	      result.push_back(address_info(file, line, func));
	    }

	    // More?
	    more =
	      bfd_find_inliner_info(abfd_, &file, &func, &line);
	  }
	}

	// Return the result
	return result;
      }
      
    private:
      /// Verify initialization
      static void verify_initialized() {
	if (not initialized_) {
	  bfd_init();
	  initialized_ = true;
	}
      }

      /// Slurp the symbol table
      void slurp_symbol_table() {
	// Does it have any symbol
	if ((bfd_get_file_flags(abfd_) & HAS_SYMS) == 0)
	  return;

	// Read
	long symcount =
	  bfd_read_minisymbols(abfd_, false,
			       reinterpret_cast<void**>(&symbol_table_),
			       &symbol_table_size_);

	// Retry with dynamic?
	if (symcount == 0)
	  symcount =
	    bfd_read_minisymbols(abfd_, true,
				 reinterpret_cast<void**>(&symbol_table_),
				 &symbol_table_size_);

	// Error?
	if (symcount < 0)
	  ttcl_fire(boost::format("Cannot slurp symbols: %s") %
		    bfd_errmsg(bfd_get_error()));
      }

      /// Find an address in a section
      bool find_address_in_section(bfd_section* _section, address _pc,
				   const char*& _file, const char*& _func,
				   uint& _line) {
	// Check flags
	if ((bfd_get_section_flags(abfd_, _section) & SEC_ALLOC) == 0)
	  return false;
	
	// Check within start and end
	bfd_vma base = bfd_get_section_vma(abfd_, _section);
	if (_pc < base)
	  return false;
	bfd_size_type size = bfd_get_section_size(_section);
	if (_pc >= base + size)
	  return false;

	// Find the nearest line
	return
	  bfd_find_nearest_line (abfd_, _section, symbol_table_, _pc - base,
				 &_file, &_func, &_line);
      }

    public:
      /// Helper
      static address parse(const std::string& _address) {
	// Parse
	return bfd_scan_vma(_address.c_str(), 0, 16);
      } 
    };
  }
}

/// Static variable
#define ttcl_ut_binary_file_static		\
  bool ttcl::ut::binary_file::initialized_;

#endif
