#ifndef TTCL_UT_FUN_AUTO_PTR_HXX
#define TTCL_UT_FUN_AUTO_PTR_HXX

// TTCL: The Template Clustering Library

/** @file
    Utilities - Custom Function Auto-Ptr
    @author Edgar Gonzalez i Pellicer
*/

#include <cstdlib>

/// TTCL Namespace
namespace ttcl {

  /// Utility Namespace
  namespace ut {

    /// Custom Function Auto-Ptr
    /** With the same interface as std::auto_ptr
     */
    template <typename T>
    class fun_auto_ptr {
    public:
      /// Element type
      typedef T element_type;

      /// Function type
      typedef void function_type(element_type*);

    private:
      /// Pointer
      element_type* pointer_;

      /// Deleting function
      function_type* function_;

      /// Disabled default copy constructor
      fun_auto_ptr(const fun_auto_ptr&);

      /// Disabled default assignment operator
      fun_auto_ptr& operator=(const fun_auto_ptr&);

    public:
      /// Construction
      explicit fun_auto_ptr(element_type* _pointer = 0,
			    function_type* _function =
			      (function_type*)&std::free) :
	pointer_(_pointer), function_(_function) {
      }

      /// Copy construction
      fun_auto_ptr(fun_auto_ptr& _other) :
	pointer_(_other.pointer_), function_(_other.function_) {
	// Release the other
	_other.pointer_ = 0;
      }

      /// Assignment
      fun_auto_ptr& operator=(fun_auto_ptr& _other) {
	// Check for self-assignment
	if (this != &_other) {
	  // Call the function
	  (*function_)(pointer_);

	  // Copy
	  pointer_  = _other.pointer_;
	  function_ = _other.function_;

	  // Release the other
	  _other.pointer_ = 0;
	}

	// Return this
	return *this;
      }

      /// Destructor
      ~fun_auto_ptr() throw () {
	// Call the function
	(*function_)(pointer_);
      }

      /// Dereference
      element_type& operator*() const throw() {
	// Access
	return *pointer_;
      }

      /// Dereference
      element_type* operator->() const throw() {
	// Access
	return pointer_;
      }

      /// Get
      element_type* get() const throw() {
	// Access
	return pointer_;
      }

      /// Release
      element_type* release() throw() {
	// Save, release and return
	element_type p = pointer_;
	pointer_ = 0;
	return p;
      }

      /// Reset
      void reset(element_type* _pointer = 0) throw() {
	// Call the function
	(*function_)(pointer_);

	// Set
	pointer_ = _pointer;
      }

      /// Reset
      void reset(element_type* _pointer, function_type* _function) throw() {
	// Call the function
	(*function_)(pointer_);

	// Set
	pointer_  = _pointer;
	function_ = _function;
      }
    };
  }
}

#endif
