#ifndef TTCL_UT_GET_OPTIONS_HXX
#define TTCL_UT_GET_OPTIONS_HXX

// TTCL: The Template Clustering Library

/** @file
    Utilities - Command-line option management
    @author Edgar Gonzalez i Pellicer
*/

#include <cctype>
#include <cstdarg>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <list>

#include <boost/format.hpp>

#include <ttcl/exception.hxx>
#include <ttcl/ut/fillable.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Utility Namespace
  namespace ut { 
  
    /// Command-line option management
    class get_options :
      public fillable<get_options> {
    private:
      /// Parameter types
      enum param_type {
	typeBool, typeBoolNeg, typeBoolNot, typeBoolMinus,
	typeChar, typeInt,  typeDouble,  typeString,
	typeBoolCall, typeBoolNegCall,
	typeBoolNotCall, typeBoolMinusCall,
	typeCharCall, typeIntCall, typeDoubleCall, typeStringCall,
	typeRadio = 32
	/* Values larger than 32 reserved for typeRadio */ };

      /// Information about a parameter
      struct param_info {
	char*      name;
	void*      target;
	param_type type;
      
	param_info(char* _name, void* _target, param_type _type) :
	  name(_name), target(_target), type(_type) {
	}
      };
        
      /// Information about the parameters
      std::list<param_info> params;
    
    public:
      /// Empty constructor
      get_options() {
      }

      /// Constructor
      get_options(const char* first, ...)
	throw (exception) {
	// Check for non-emptiness
	if (not first)
	  // No options defined
	  return;
      
	// Initialize the parameter list
	va_list param_list;
	va_start(param_list, first);
      
	// Follow it
	const char* name = first;
	do {
	  // Find the target
	  void* target = va_arg(param_list, void*);
	
	  // Add the option
	  fill(name, target);

	  // Next name
	  name = va_arg(param_list, const char*);
	} while (name);
      }
    
      /// Destructor
      ~get_options() {
	for (std::list<param_info>::iterator it = params.begin();
	     it != params.end(); ++it) {
	  std::free(it->name);
	}
      }

      /// Add an Option
      void fill(const char* _option, void* _target)
	throw (exception) {
	// Duplicate the option
	char* name = ::strdup(_option);

	// Empty name
	int length = std::strlen(name);
	if (length == 0) {
	  // Error
	  std::free(name);
	  ttcl_fire("Empty Name");
	}
	
	// Get the target
	if (not _target) {
	  // Error
	  std::free(name);
	  ttcl_fire("Missing Target for Option %s", _option);
	}
	
	// Parse the options
	if (name[length-1] == '!') {
	  // A negable boolean
	  name[length-1] = '\0';
	  params.push_back(param_info(name, _target, typeBoolNeg));
	}
	else if (name[length-1] == '~') {
	  // A negated boolean
	  name[length-1] = '\0';
	  params.push_back(param_info(name, _target, typeBoolNot));
	}
	else if (name[length-1] == '-') {
	  // A pseudo-negated boolean
	  name[length-1] = '\0';
	  params.push_back(param_info(name, _target, typeBoolMinus));
	}
	else if (name[length-1] == '&') {
	  // A function call
	  if (length > 2 and name[length-2] == '!') {
	    // From a negable boolean
	    name[length-2] = '\0';
	    params.push_back(param_info(name, _target, typeBoolNegCall));
	  }
	  else if (length > 2 and name[length-2] == '~') {
	    // A negated boolean
	    name[length-2] = '\0';
	    params.push_back(param_info(name, _target, typeBoolNotCall));
	  }
	  else if (length > 2 and name[length-2] == '-') {
	    // A pseudo-negated boolean
	    name[length-2] = '\0';
	    params.push_back(param_info(name, _target, typeBoolMinusCall));
	  }
	  else if (length > 3 and name[length-3] == '=') {
	    // From non-booleans
	    name[length-3] = '\0';
	    param_type type;
	    switch (name[length-2]) {
	    case 'c': type = typeCharCall;   break; // A char
	    case 'i': type = typeIntCall;    break; // An integer
	    case 'd': type = typeDoubleCall; break; // A double
	    case 's': type = typeStringCall; break; // A string
	    default:
	      std::free(name);
	      ttcl_fire("Wrong Type for Option %s", _option);
	    }
	    params.push_back(param_info(name, _target, type));
	  }
	  else if (strchr(name, '=')) {
	    // Another kind of option
	    ttcl_fire("Wrong option specifier: %s", name);
	  }
	  else {
	    // Raw
	    name[length-1] = '\0';
	    params.push_back(param_info(name, _target, typeBoolCall));
	  }
	}
	else if (length > 2 and name[length-2] == '=') {
	  // Non-booleans
	  name[length-2] = '\0';
	  param_type type;
	  switch (name[length-1]) {
	  case 'c': type = typeChar;   break; // A char
	  case 'i': type = typeInt;    break; // An integer
	  case 'd': type = typeDouble; break; // A double
	  case 's': type = typeString; break; // A string
	  default:
	    std::free(name);
	    ttcl_fire("Wrong Type for Option %s", _option);
	  }
	  params.push_back(param_info(name, _target, type));
	}
	else if (length > 3 and
		 name[length-3] == '=' and name[length-2] == 'r' and
		 std::isdigit(name[length-1])) {
	  // Radio button
	  name[length-3] = '\0';
	  params.push_back(param_info(name, _target,
				      param_type(typeRadio |
						 (name[length-1] - '0'))));
	}
	else if (length > 4 and
		 name[length-4] == '=' and name[length-3] == 'r' and
		 std::isdigit(name[length-2]) and
		 std::isdigit(name[length-1])) {
	  // Radio button
	  name[length-4] = '\0';
	  params.push_back(param_info(name, _target,
				      param_type(typeRadio |
						 ((name[length-2] - '0') * 10 +
						  (name[length-1] - '0')))));
	}
	else if (strchr(name, '=')) {
	  // Another kind of option
	  ttcl_fire("Wrong option specifier: %s", name);
	}
	else {
	  // A boolean
	  params.push_back(param_info(name, _target, typeBool));
	}
      }

      /// Parse
      /* Returns a list with the non-processed arguments
       */
      const char** parse(int argc, const char* argv[])
	throw (exception) {
	// Result
	const char** result = new const char*[argc];
	int resultPos       = 0;

	try {
	  // Skip program name
	  int pos = 1;

	  // Process every argument
	  while (pos < argc) {
	    // Check for:

	    // End of options marker
	    if (std::strcmp(argv[pos], "--") == 0) {
	      // Copy them all to result
	      ++pos;
	      while (pos < argc) {
		result[resultPos++] = argv[pos++];
	      }
	    }
	    else if (argv[pos][0] != '-'  or
		     argv[pos][0] == '\0' or
		     argv[pos][1] != '-') {
	      // Not an option
	      result[resultPos++] = argv[pos++];
	    }
	    else {
	      // Scan the possible parameters
	      // Unambiguous shortenings not yet supported
	      std::list<param_info>::iterator it = params.begin();
	      while (it != params.end()) {
		if (::strcasecmp(argv[pos] + 2, it->name) == 0) {
		  ++pos;

		  if (it->type == typeBool or it->type == typeBoolNeg) {
		    *((int*)it->target) = 1;
		  }
		  else if (it->type == typeBoolCall or
			   it->type == typeBoolNegCall) {
		    ((void(*)(int))it->target)(1);
		  }
		  else if (it->type == typeBoolNot) {
		    *((int*)it->target) = 0;
		  }
		  else if (it->type == typeBoolNotCall) {
		    ((void(*)(int))it->target)(0);
		  }
		  else if (it->type == typeBoolMinus) {
		    *((int*)it->target) = -1;
		  }
		  else if (it->type == typeBoolMinusCall) {
		    ((void(*)(int))it->target)(-1);
		  }
		  else if (it->type & typeRadio) {
		    *((int*)it->target) = it->type & ~typeRadio;
		  }
		  else if (pos == argc) {
		    ttcl_fire("Missing argument for option %s", argv[pos - 1]);
		  }
		  else {
		    switch(it->type) {
		    case typeChar: 
		    case typeCharCall: {
		      if (argv[pos][0] == '\0' or argv[pos][1] != '\0')
			ttcl_fire("Argument for option %s"
				  " must be a single char", argv[pos - 1]);

		      if (it->type == typeChar)
			*((char*)it->target) = argv[pos][0];
		      else // it->type == typeCharCall
			((void(*)(char))it->target)(argv[pos][0]);
		      break;
		    }

		    case typeInt: 
		    case typeIntCall: {
		      char* endptr;
		      int value = std::strtol(argv[pos], &endptr, 10);

		      while (std::isspace(*endptr)) ++endptr;
		      if (*endptr != '\0')
			ttcl_fire("Argument for option %s"
				  " must be an integer", argv[pos - 1]);

		      if (it->type == typeInt)
			*((int*)it->target) = value;
		      else // it->type == typeIntCall
			((void(*)(int))it->target)(value);
		      break;
		    }

		    case typeDouble:
		    case typeDoubleCall: {
		      char* endptr;
		      double value = std::strtod(argv[pos], &endptr);
		      
		      while (std::isspace(*endptr)) ++endptr;
		      if (*endptr != '\0')
			ttcl_fire("Argument for option %s"
				  " must be a real", argv[pos - 1]);

		      if (it->type == typeDouble)
			*((double*)it->target) = value;
		      else // it->type == typeDoubleCall
			((void(*)(double))it->target)(value);
		      break;
		    }

		    case typeString:
		      *((const char**)it->target) = argv[pos];
		      break;
		    case typeStringCall:
		      ((void(*)(const char*))it->target)(argv[pos]);
		      break;
		    default:
		      ;
		    }

		    ++pos;
		  }
		    
		  // Leave the loop
		  break;
		}
		else if ((it->type == typeBoolNeg or 
			  it->type == typeBoolNegCall) and
			 ::strncasecmp(argv[pos] + 2, "no", 2) == 0 and
			 (::strcasecmp(argv[pos] + 4, it->name) == 0 or
			  (argv[pos][4] == '-' and
			   ::strcasecmp(argv[pos] + 5, it->name) == 0))) {
		  ++pos;
		  if (it->type == typeBoolNeg)
		    *((int*)it->target) = 0;
		  else // it->type == typeBoolNegCall
		    ((void(*)(int))it->target)(0);
		
		  // Leave the loop
		  break;
		}

		++it;
	      }

	      if (it == params.end())
		ttcl_fire("Wrong option %s", argv[pos]);
	    }
	  }

	  // End the result list
	  result[resultPos] = 0;

	  // Return
	  return result;
	}
	catch (...) {
	  // Delete result and rethrow
	  delete result;
	  throw;
	}
      }

      /// Print help
      void printHelp(std::ostream& _out = std::cerr) const {
	// Header
	_out << "Available options:" << std::endl;
	// Iterate over params
	for (std::list<param_info>::const_iterator it = params.begin();
	     it != params.end(); ++it) {
	  switch (it->type) {
	  case typeInt:
	  case typeIntCall:
	    _out << " --"    << it->name << " <N>"   << std::endl; break;
	  case typeDouble:
	  case typeDoubleCall:
	    _out << " --"    << it->name << " <N.M>" << std::endl; break;
	  case typeString:
	  case typeStringCall:
	    _out << " --"    << it->name << " <Str>" << std::endl; break;
	  case typeBoolNeg:
	  case typeBoolNegCall:
	    _out << " --"    << it->name << std::endl
		 << " --no"  << it->name << std::endl
		 << " --no-" << it->name << std::endl; break;
	  default:
	    _out << " --"    << it->name << std::endl;
	  }
	}
      }

    };
  }
}  

#endif

