#ifndef TTCL_UT_LOG_VALUE_HXX
#define TTCL_UT_LOG_VALUE_HXX

// TTCL: The Template Clustering Library

/** @file
    Utilities - Number stored in log form
    @author Edgar Gonzalez i Pellicer
*/

#include <cmath>
#include <limits>

/// TTCL Namespace
namespace ttcl {

  /// Utility Namespace
  namespace ut { 

    /// Log-value
    template <typename T>
    class log_value {
    public:
      /// Value type
      typedef T value_type;

      /// Type for raw input
      struct raw {
	// Value
	value_type value;

	// Constructor
	raw(const value_type _value) :
	  value(_value) {
	}
      };

    private:
      /// Log-Value
      T log_;

    public:
      /// Empty constructor
      log_value() :
	log_(-std::numeric_limits<value_type>::infinity()) {
      }

      /// Constructor
      log_value(const value_type _value) :
	log_(std::log(_value)) {
      }

      /// Constructor
      log_value(const raw _raw) :
	log_(_raw.value) {
      }

      /// Copy constructor
      log_value(const log_value& _other) :
	log_(_other.log_) {
      }

      /// Assignment operator
      log_value& operator=(const value_type _value) {
	log_ = std::log(_value);
	return *this;
      }

      /// Assignment operator
      log_value& operator=(const raw _raw) {
	log_ = _raw.value;
	return *this;
      }

      /// Assignment operator
      log_value& operator=(const log_value& _other) {
	log_ = _other.log_;
	return *this;
      }

      /// Increment
      log_value& operator+=(const value_type _value) {
	value_type log_v = std::log(_value);
	if (log_ < log_v)
	  log_ = log_v + std::log(std::exp(log_ - log_v) + T(1.0));
	else
	  log_ = log_ + std::log(T(1.0) + std::exp(log_v - log_));
	return *this;
      }

      /// Increment
      log_value& operator+=(const log_value _other) {
	if (log_ < _other.log_)
	  log_ = _other.log_ + std::log(std::exp(log_ - _other.log_) + T(1.0));
	else
	  log_ = log_ + std::log(T(1.0) + std::exp(_other.log_ - log_));
	return *this;
      }

      /// Decrement
      log_value& operator-=(const value_type _value) {
	value_type log_v = std::log(_value);
	if (log_ < log_v)
	  log_ = log_v + std::log(std::exp(log_ - log_v) - T(1.0));
	else
	  log_ = log_ + std::log(T(1.0) - std::exp(log_v - log_));
	return *this;
      }

      /// Decrement
      log_value& operator-=(const log_value _other) {
	if (log_ < _other.log_)
	  log_ = _other.log_ + std::log(std::exp(log_ - _other.log_) - T(1.0));
	else
	  log_ = log_ + std::log(T(1.0) - std::exp(_other.log_ - log_));
	return *this;
      }

      /// Product
      log_value& operator*=(const value_type _value) {
	log_ += std::log(_value);
	return *this;
      }

      /// Product
      log_value& operator*=(const log_value _other) {
	log_ += _other.log_;
	return *this;
      }

      /// Quotient
      log_value& operator/=(const value_type _value) {
	log_ -= std::log(_value);
	return *this;
      }

      /// Quotient
      log_value& operator/=(const log_value _other) {
	log_ -= _other.log_;
	return *this;
      }

      /// Value
      value_type value() const {
	return std::exp(log_);
      }
     
      /// Log
      value_type log() const {
	return log_;
      }

      /// Conversion
      operator value_type() const {
	return value();
      }
    };

    /// Raw value constructor
    template <typename T>
    typename log_value<T>::raw raw_value(const T _value) {
      return typename log_value<T>::raw(_value);
    }

    /// Equality
    template <typename T>
    log_value<T> operator==(const log_value<T> _a, const log_value<T> _b) {
      return _a.log() == _b.log();
    }

    /// Equality
    template <typename T>
    log_value<T> operator==(const log_value<T> _a, const T _b) {
      return _a.log() == std::log(_b);
    }

    /// Equality
    template <typename T>
    log_value<T> operator==(const T _a, const log_value<T> _b) {
      return std::log(_a) == _b.log();
    }


    /// Inequality
    template <typename T>
    log_value<T> operator!=(const log_value<T> _a, const log_value<T> _b) {
      return _a.log() != _b.log();
    }

    /// Inequality
    template <typename T>
    log_value<T> operator!=(const log_value<T> _a, const T _b) {
      return _a.log() != std::log(_b);
    }

    /// Inequality
    template <typename T>
    log_value<T> operator!=(const T _a, const log_value<T> _b) {
      return std::log(_a) != _b.log();
    }


    /// Less than
    template <typename T>
    log_value<T> operator<(const log_value<T> _a, const log_value<T> _b) {
      return _a.log() < _b.log();
    }

    /// Less than
    template <typename T>
    log_value<T> operator<(const log_value<T> _a, const T _b) {
      return _a.log() < std::log(_b);
    }

    /// Less than
    template <typename T>
    log_value<T> operator<(const T _a, const log_value<T> _b) {
      return std::log(_a) < _b.log();
    }


    /// Less than or equal
    template <typename T>
    log_value<T> operator<=(const log_value<T> _a, const log_value<T> _b) {
      return _a.log() <= _b.log();
    }

    /// Less than or equal
    template <typename T>
    log_value<T> operator<=(const log_value<T> _a, const T _b) {
      return _a.log() <= std::log(_b);
    }

    /// Less than or equal
    template <typename T>
    log_value<T> operator<=(const T _a, const log_value<T> _b) {
      return std::log(_a) <= _b.log();
    }


    /// Greater than
    template <typename T>
    log_value<T> operator>(const log_value<T> _a, const log_value<T> _b) {
      return _a.log() > _b.log();
    }

    /// Greater than
    template <typename T>
    log_value<T> operator>(const log_value<T> _a, const T _b) {
      return _a.log() > std::log(_b);
    }

    /// Greater than
    template <typename T>
    log_value<T> operator>(const T _a, const log_value<T> _b) {
      return std::log(_a) > _b.log();
    }


    /// Greater than or equal
    template <typename T>
    log_value<T> operator>=(const log_value<T> _a, const log_value<T> _b) {
      return _a.log() >= _b.log();
    }

    /// Greater than or equal
    template <typename T>
    log_value<T> operator>=(const log_value<T> _a, const T _b) {
      return _a.log() >= std::log(_b);
    }

    /// Greater than or equal
    template <typename T>
    log_value<T> operator>=(const T _a, const log_value<T> _b) {
      return std::log(_a) >= _b.log();
    }


    /// Addition
    template <typename T>
    log_value<T> operator+(const log_value<T> _a, const log_value<T> _b) {
      log_value<T> result;
      if (_a.log() < _b.log())
	result = raw_value(_b.log() +
			   std::log(std::exp(_a.log() - _b.log()) + T(1.0)));
      else
	result = raw_value(_a.log() +
			   std::log(T(1.0) + std::exp(_b.log() - _a.log())));
      return result;
    }

    /// Addition
    template <typename T>
    log_value<T> operator+(const log_value<T> _a, const T _b) {
      log_value<T> result;
      T log_b = std::log(_b);
      if (_a.log() < log_b)
	result = raw_value(log_b +
			   std::log(std::exp(_a.log() - log_b) + T(1.0)));
      else
	result = raw_value(_a.log() +
			   std::log(T(1.0) + std::exp(log_b - _a.log())));
      return result;
    }

    /// Addition
    template <typename T>
    log_value<T> operator+(const T _a, const log_value<T> _b) {
      log_value<T> result;
      T log_a = std::log(_a);
      if (log_a < _b.log())
	result = raw_value(_b.log() +
			   std::log(std::exp(log_a - _b.log()) + T(1.0)));
      else
	result = raw_value(log_a +
			   std::log(T(1.0) + std::exp(_b.log() - log_a)));
      return result;
    }


    /// Substraction
    template <typename T>
    log_value<T> operator-(const log_value<T> _a, const log_value<T> _b) {
      log_value<T> result;
      if (_a.log() < _b.log())
	result = raw_value(_b.log() +
			   std::log(std::exp(_a.log() - _b.log()) - T(1.0)));
      else
	result = raw_value(_a.log() +
			   std::log(T(1.0) - std::exp(_b.log() - _a.log())));
      return result;
    }

    /// Substraction
    template <typename T>
    log_value<T> operator-(const log_value<T> _a, const T _b) {
      log_value<T> result;
      T log_b = std::log(_b);
      if (_a.log() < log_b)
	result = raw_value(log_b +
			   std::log(std::exp(_a.log() - log_b) - T(1.0)));
      else
	result = raw_value(_a.log() +
			   std::log(T(1.0) - std::exp(log_b - _a.log())));
      return result;
    }

    /// Substraction
    template <typename T>
    log_value<T> operator-(const T _a, const log_value<T> _b) {
      log_value<T> result;
      T log_a = std::log(_a);
      if (log_a < _b.log())
	result = raw_value(_b.log() +
			   std::log(std::exp(log_a - _b.log()) - T(1.0)));
      else
	result = raw_value(log_a +
			   std::log(T(1.0) - std::exp(_b.log() - log_a)));
      return result;
    }


    /// Product
    template <typename T>
    log_value<T> operator*(const log_value<T> _a, const log_value<T> _b) {
      return log_value<T>(raw_value(_a.log() + _b.log()));
    }

    /// Product
    template <typename T>
    log_value<T> operator*(const log_value<T> _a, const T _b) {
      return log_value<T>(raw_value(_a.log() + std::log(_b)));
    }

    /// Product
    template <typename T>
    log_value<T> operator*(const T _a, const log_value<T> _b) {
      return log_value<T>(raw_value(std::log(_a) + _b.log()));
    }


    /// Quotient
    template <typename T>
    log_value<T> operator/(const log_value<T> _a, const log_value<T> _b) {
      return log_value<T>(raw_value(_a.log() - _b.log()));
    }

    /// Quotient
    template <typename T>
    log_value<T> operator/(const log_value<T> _a, const T _b) {
      return log_value<T>(raw_value(_a.log() - std::log(_b)));
    }

    /// Quotient
    template <typename T>
    log_value<T> operator/(const T _a, const log_value<T> _b) {
      return log_value<T>(raw_value(std::log(_a) - _b.log()));
    }
  }
}

#endif
