#ifndef TTCL_UT_SKIP_POINTER_HXX
#define TTCL_UT_SKIP_POINTER_HXX

// TTCL: The Template Clustering Library

/** @file
    Utilities - Skip pointer
    @author Edgar Gonzlez i Pellicer
*/

#include <iterator>

/// TTCL Namespace
namespace ttcl {

  /// Utility Namespace
  namespace ut {

    /// Skip pointer
    template <typename T>
    class skip_pointer {
    public:
      /// Iterator category
      typedef std::random_access_iterator_tag iterator_category;

      /// Value type
      typedef T value_type;

      /// Difference type
      typedef ptrdiff_t difference_type;

      /// Pointer type
      typedef T* pointer;

      /// Reference type
      typedef T& reference;

    private:
      /// Undelying raw pointer
      T* ptr_;

      /// Skip
      ptrdiff_t skip_;

    public:
      /// Constructor
      skip_pointer(T* _ptr = 0, ptrdiff_t _skip = 1) : ptr_(_ptr), skip_(_skip) {}

      /// Constructor from another skip pointer
      template <typename T2>
      skip_pointer(const skip_pointer<T2>& _other) : ptr_(_other.ptr_), skip_(_other.skip_) {}

      /// @name Accessors
      /// @{

      T* get() const {
	return ptr_;
      }

      ptrdiff_t skip() const {
	return skip_;
      }

      /// @}

      /// @name Increment/Decrement
      /// @{

      /// Pre-Increment
      skip_pointer& operator++() {
	ptr_ += skip_;
	return *this;
      }

      /// Post-Increment
      skip_pointer operator++(int) {
	skip_pointer other = *this;
	++*this;
	return other;
      }
  
      /// Increment
      skip_pointer& operator+=(int _n) {
	ptr_ += _n * skip_;
	return *this;
      }

      /// Addition
      skip_pointer operator+(int _n) const {
	return skip_pointer(ptr_ + _n * skip_, skip_);
      }

      /// Pre-Decrement
      skip_pointer& operator--() {
	ptr_ -= skip_;
	return *this;
      }

      /// Post-Decrement
      skip_pointer operator--(int) {
	skip_pointer other = *this;
	--*this;
	return other;
      }

      /// Decrement
      skip_pointer& operator-=(int _n) {
	ptr_ -= _n * skip_;
	return *this;
      }

      /// Substraction
      skip_pointer operator-(int _n) const {
	return skip_pointer(ptr_ - _n * skip_, skip_);
      }

      /// @}

      /// Difference
      int operator-(const skip_pointer& _other) {
	return (ptr_ - _other.ptr_) / skip_;
      }

      /// @name Dereference/Indirection
      /// @{

      /// Dereference
      reference operator*() const {
	return *ptr_;
      }

      /// Indirection
      pointer operator->() const {
	return ptr_;
      }

      /// Array access
      reference operator[](int _n) const {
	return *(ptr_ + _n * skip_);
      }

      /// @}
    };


    /// @name Relational operators
    /// @{

    /// Equality
    /** Between skip_pointer
     */
    template <typename T>
    bool operator==(const skip_pointer<T>& _a, const skip_pointer<T>& _b) {
      return _a.get() == _b.get();
    }

    /// Inequality
    /** Between skip_pointer
     */
    template <typename T>
    bool operator!=(const skip_pointer<T>& _a, const skip_pointer<T>& _b) {
      return _a.get() != _b.get();
    }

    /// Less than
    /** Between skip_pointer
     */
    template <typename T>
    bool operator<(const skip_pointer<T>& _a, const skip_pointer<T>& _b) {
      return _a.get() < _b.get();
    }

    /// Greater than
    /** Between skip_pointer
     */
    template <typename T>
    bool operator>(const skip_pointer<T>& _a, const skip_pointer<T>& _b) {
      return _a.get() > _b.get();
    }

    /// Less than or equal
    /** Between skip_pointer
     */
    template <typename T>
    bool operator<=(const skip_pointer<T>& _a, const skip_pointer<T>& _b) {
      return _a.get() <= _b.get();
    }

    /// Greater than or equal
    /** Between skip_pointer
     */
    template <typename T>
    bool operator>=(const skip_pointer<T>& _a, const skip_pointer<T>& _b) {
      return _a.get() >= _b.get();
    }


    /// Equality
    /** Between a skip_pointer and a raw pointer
     */
    template <typename T>
    bool operator==(const skip_pointer<T>& _a, T* const _b) {
      return _a.get() == _b;
    }

    /// Inequality
    /** Between a skip_pointer and a raw pointer
     */
    template <typename T>
    bool operator!=(const skip_pointer<T>& _a, T* const _b) {
      return _a.get() != _b;
    }

    /// Less than
    /** Between a skip_pointer and a raw pointer
     */
    template <typename T>
    bool operator<(const skip_pointer<T>& _a, T* const _b) {
      return _a.get() < _b;
    }

    /// Greater than
    /** Between a skip_pointer and a raw pointer
     */
    template <typename T>
    bool operator>(const skip_pointer<T>& _a, T* const _b) {
      return _a.get() > _b;
    }

    /// Less than or equal
    /** Between a skip_pointer and a raw pointer
     */
    template <typename T>
    bool operator<=(const skip_pointer<T>& _a, T* const _b) {
      return _a.get() <= _b;
    }

    /// Greater than or equal
    /** Between a skip_pointer and a raw pointer
     */
    template <typename T>
    bool operator>=(const skip_pointer<T>& _a, T* const _b) {
      return _a.get() >= _b;
    }


    /// Equality
    /** Between a raw pointer and a skip_pointer
     */
    template <typename T>
    bool operator==(T* const _a, const skip_pointer<T>& _b) {
      return _a == _b.get();
    }

    /// Inequality
    /** Between a raw pointer and a skip_pointer
     */
    template <typename T>
    bool operator!=(T* const _a, const skip_pointer<T>& _b) {
      return _a != _b.get();
    }

    /// Less than
    /** Between a raw pointer and a skip_pointer
     */
    template <typename T>
    bool operator<(T* const _a, const skip_pointer<T>& _b) {
      return _a < _b.get();
    }

    /// Greater than
    /** Between a raw pointer and a skip_pointer
     */
    template <typename T>
    bool operator>(T* const _a, const skip_pointer<T>& _b) {
      return _a > _b.get();
    }

    /// Less than or equal
    /** Between a raw pointer and a skip_pointer
     */
    template <typename T>
    bool operator<=(T* const _a, const skip_pointer<T>& _b) {
      return _a <= _b.get();
    }

    /// Greater than or equal
    /** Between a raw pointer and a skip_pointer
     */
    template <typename T>
    bool operator>=(T* const _a, const skip_pointer<T>& _b) {
      return _a >= _b.get();
    }

    /// @}
  }
}

#endif
