#ifndef TTCL_UT_TAG_ALLOCATOR_HXX
#define TTCL_UT_TAG_ALLOCATOR_HXX

// TTCL: The Template Clustering Library

/** @file
    Utilities - A tagged allocator
    @author Edgar Gonzalez i Pellicer
*/

#include <cstddef>
#include <vector>

/// TTCL Namespace
namespace ttcl {

  /// Utility Namespace
  namespace ut { 

    /// Tag Block
    template <typename Tag>
    class tag_block {
    protected:
      /// Reserved bytes
      static std::size_t bytes_;

      /// Reserved objects
      static std::size_t objects_;

    public:
      /// Reserved bytes
      static std::size_t bytes() {
	return bytes_;
      }

      /// Reserved objects
      static std::size_t objects() {
	return objects_;
      }
    };
  
    /// Tagged Allocator
    template <typename T, typename Tag>
    class tag_allocator : tag_block<Tag> {
    public:
      /// Size Type
      typedef std::size_t size_type;

      /// Difference Type
      typedef std::ptrdiff_t difference_type;

      /// Pointer Type
      typedef T* pointer;

      /// Const Pointer Type
      typedef const T* const_pointer;

      /// Reference Type
      typedef T& reference;

      /// Const Reference Type
      typedef const T& const_reference;

      /// Value Type
      typedef T value_type;

      /// Rebind Template
      template<typename T1>
      struct rebind {
	/// The Rebound Type
	typedef tag_allocator<T1, Tag> other;
      };

      /// Default Constructor
      tag_allocator() throw() {
      }

      /// Copy Constructor
      tag_allocator(const tag_allocator& /* _other */) throw() {
      }

      /// Rebinding Constructor
      template<typename T1, typename Tag1>
      tag_allocator(const tag_allocator<T1, Tag1>& /* _other */) throw() {
      }

      /// Destructor
      ~tag_allocator() throw() {
      }

      /// Address Function
      pointer address(reference _x) const {
	return &_x;
      }

      /// Const Address Function
      const_pointer address(const_reference _x) const {
	return &_x;
      }

      // Allocate Function
      pointer allocate(size_type _n, const void* /* _p */ = 0) {
	tag_block<Tag>::bytes_   += _n * sizeof(T);
	tag_block<Tag>::objects_ += _n;
	return static_cast<pointer>(::operator new(_n * sizeof(T)));
      }

      /// Deallocate Function
      void deallocate(pointer _p, size_type _n) {
	tag_block<Tag>::bytes_   -= _n * sizeof(T);
	tag_block<Tag>::objects_ -= _n;
	::operator delete(_p);
      }
      
      /// Maximum Allocatable Size
      size_type max_size() const
	throw() {
	return size_t(-1) / sizeof(T);
      }

      /// Construct Function
      void construct(pointer _p) {
	::new(_p) T();
      }

      /// Construct Function
      void construct(pointer _p, const T& _val) {
	::new(_p) T(_val);
      }

      /// Destroy Function
      void  destroy(pointer _p) {
	_p->~T();
      }
    };

    /// Equality Operator
    template<typename T, typename Tag>
    bool operator==(const tag_allocator<T, Tag>& /* _a */,
		    const tag_allocator<T, Tag>& /* _b */) {
      return true;
    }

    /// Inequality Operator
    template<typename T, typename Tag>
    bool operator!=(const tag_allocator<T, Tag>& /* _a */,
		    const tag_allocator<T, Tag>& /* _b */) {
      return false;
    }
  }
}

/// Static vars
#define ttcl_ut_tag_allocator_static					\
  template <typename tag> std::size_t ttcl::ut::tag_block<tag>::bytes_;	\
  template <typename tag> std::size_t ttcl::ut::tag_block<tag>::objects_

#endif
