#ifndef TTCL_WR_SHARED_PTR_HXX
#define TTCL_WR_SHARED_PTR_HXX

// TTCL: The Template Clustering Library

/** @file
    Wrappers - A shared pointer
    @author Edgar Gonzalez i Pellicer
*/

#include <ttcl/wr/shared_referee.hxx>

/// TTCL Namespace
namespace ttcl {

  /// Wrapper Namespace
  namespace wr { 

    /// A shared pointer
    template <typename T = shared_referee>
    class shared_ptr {
    public:
      /// Element type
      typedef T element_type;

    private:
      /// Pointer to referee
      element_type* ptr_;

    public:
      /// Constructor
      shared_ptr(element_type* _ptr = 0) :
	ptr_(_ptr) {
	// Add a reference
	if (ptr_)
	  ptr_->add_reference();
      }

      /// Copy constructor
      shared_ptr(const shared_ptr& _other) :
	ptr_(_other.ptr_) {
	// Add a reference
	if (ptr_)
	  ptr_->add_reference();
      }

      /// Destructor
      ~shared_ptr() {
	// Remove a reference
	if (ptr_)
	  ptr_->del_reference();
      }

      /// Assignment operator
      shared_ptr& operator=(const shared_ptr& _other) {
	// Add a reference
	if (_other.ptr_)
	  _other.ptr_->add_reference();

	// Remove a reference
	if (ptr_)
	  ptr_->del_reference();

	// Assign
	ptr_ = _other.ptr_;

	// Return this
	return *this;
      }

      /// Indirection
      element_type& operator*() const {
	return *ptr_;
      }

      /// Indirection
      element_type* operator->() const {
	return ptr_;
      }

      /// Pointer
      element_type* ptr() const {
	return ptr_;
      }
    };

    /// Equality comparison
    template <typename T>
    bool operator==(const shared_ptr<T>& _a,
		    const shared_ptr<T>& _b) {
      return _a.ptr() == _b.ptr();
    }

    /// Equality comparison
    /** @overload
     */
    template <typename T>
    bool operator==(const shared_ptr<T>& _a, T* _b) {
      return _a.ptr() == _b;
    }

    /// Equality comparison
    /** @overload
     */
    template <typename T>
    bool operator==(T* _a, const shared_ptr<T>& _b) {
      return _a == _b.ptr();
    }

    /// Inequality comparison
    template <typename T>
    bool operator!=(const shared_ptr<T>& _a,
		    const shared_ptr<T>& _b) {
      return _a.ptr() != _b.ptr();
    }

    /// Inequality comparison
    /** @overload
     */
    template <typename T>
    bool operator!=(const shared_ptr<T>& _a, T* _b) {
      return _a.ptr() != _b;
    }

    /// Inequality comparison
    /** @overload
     */
    template <typename T>
    bool operator!=(T* _a, const shared_ptr<T>& _b) {
      return _a != _b.ptr();
    }
  }
}

#endif
