#include <cmath>
#include <vector>

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>

#include <ttcl/cl/hard_clustering.hxx>
#include <ttcl/hj/ag/agglomerative_distance_clusterer.hxx>
#include <ttcl/hj/ag/single_link.hxx>
#include <ttcl/hj/ag/complete_link.hxx>
#include <ttcl/hj/ag/upgma.hxx>
#include <ttcl/tu/tuning.hxx>

// Options
ttcl_tu_static;
ttcl_tu_option_static(ttcl::hj::ag::merging_threshold);

// Data format
struct point {
  double x;
  double y;
  point(double _x, double _y) : x(_x), y(_y) {}
};

// Distance
struct euclidean {
  typedef double result_type;

  double operator()(const point& _a, const point& _b) const {
    double dx = _a.x - _b.x;
    double dy = _a.y - _b.y;
    return std::sqrt(dx * dx + dy * dy);
  }
};

// Number of samples
const unsigned int n_samples = 4;

// Data
const point points[] = { point(0, 1), point(1, 0), point(3, 0), point(2, 2) };

// Single link merges
const int single_link_merges[][3] = { {  4, -1, -1 },
				      {  4, -1, -1 },
				      {  5, -1, -1 },
				      {  6, -1, -1 },
				      {  5,  1,  0 },
				      {  6,  4,  2 },
				      { -1,  5,  3 } };

// Complete link merges
const int complete_link_merges[][3] = { {  4, -1, -1 },
					{  4, -1, -1 },
					{  5, -1, -1 },
					{  5, -1, -1 },
					{  6,  1,  0 },
					{  6,  3,  2 },
					{ -1,  5,  4 } };

// UPGMA merges
const int upgma_merges[][3] = { {  4, -1, -1 },
				{  4, -1, -1 },
				{  5, -1, -1 },
				{  5, -1, -1 },
				{  6,  1,  0 },
				{  6,  3,  2 },
				{ -1,  5,  4 } };

// Single link
BOOST_AUTO_TEST_CASE( single_link_cluster ) {
  // Create vector
  std::vector<point> v;
  for (unsigned int i = 0; i < n_samples; ++i) v.push_back(points[i]);
  
  // Single link clusterer
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::single_link> clusterer;
  clusterer cl;
  clusterer::dendrogram_type d1 = cl.hier_cluster(v);

  // Check
  BOOST_CHECK_EQUAL(d1.size(), n_samples);
  BOOST_CHECK_EQUAL(d1.clusters(), 1u);
  BOOST_REQUIRE_EQUAL(d1.branches(), 2 * n_samples - 1);
  for (uint i = 0; i < d1.branches(); ++i) {
    BOOST_CHECK_EQUAL(d1[i].parent, single_link_merges[i][0]);
    BOOST_CHECK_EQUAL(d1[i].first,  single_link_merges[i][1]);
    BOOST_CHECK_EQUAL(d1[i].second, single_link_merges[i][2]);
  }
}

// Complete link
BOOST_AUTO_TEST_CASE( complete_link_cluster ) {
  // Create vector
  std::vector<point> v;
  for (unsigned int i = 0; i < n_samples; ++i) v.push_back(points[i]);
  
  // Complete link clusterer
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::complete_link> clusterer;
  clusterer cl;
  clusterer::dendrogram_type d1 = cl.hier_cluster(v);

  // Check
  BOOST_CHECK_EQUAL(d1.size(), n_samples);
  BOOST_CHECK_EQUAL(d1.clusters(), 1u);
  BOOST_REQUIRE_EQUAL(d1.branches(), 2 * n_samples - 1);
  for (uint i = 0; i < d1.branches(); ++i) {
    BOOST_CHECK_EQUAL(d1[i].parent, complete_link_merges[i][0]);
    BOOST_CHECK_EQUAL(d1[i].first,  complete_link_merges[i][1]);
    BOOST_CHECK_EQUAL(d1[i].second, complete_link_merges[i][2]);
  }
}

// UPGMA
BOOST_AUTO_TEST_CASE( upgma_cluster ) {
  // Create vector
  std::vector<point> v;
  for (unsigned int i = 0; i < n_samples; ++i) v.push_back(points[i]);
  
  // UPGMA clusterer
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::upgma> clusterer;
  clusterer cl;
  clusterer::dendrogram_type d1 = cl.hier_cluster(v);

  // Check
  BOOST_CHECK_EQUAL(d1.size(), n_samples);
  BOOST_CHECK_EQUAL(d1.clusters(), 1u);
  BOOST_REQUIRE_EQUAL(d1.branches(), 2 * n_samples - 1);
  for (uint i = 0; i < d1.branches(); ++i) {
    BOOST_CHECK_EQUAL(d1[i].parent, upgma_merges[i][0]);
    BOOST_CHECK_EQUAL(d1[i].first,  upgma_merges[i][1]);
    BOOST_CHECK_EQUAL(d1[i].second, upgma_merges[i][2]);
  }
}

// Through pointer
BOOST_AUTO_TEST_CASE( pointer_cluster ) {
  // Create vector
  std::vector<point> v;
  for (unsigned int i = 0; i < n_samples; ++i) v.push_back(points[i]);
  
  // UPGMA clusterer
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::upgma> clusterer;
  typedef ttcl::hj::hier_data_clusterer< std::vector<point> > base_clusterer;
  base_clusterer* cl = new clusterer();
  base_clusterer::dendrogram_type d1 = cl->hier_cluster(v);

  // Check
  BOOST_CHECK_EQUAL(d1.size(), n_samples);
  BOOST_CHECK_EQUAL(d1.clusters(), 1u);
  BOOST_REQUIRE_EQUAL(d1.branches(), 2 * n_samples - 1);
  for (uint i = 0; i < d1.branches(); ++i) {
    BOOST_CHECK_EQUAL(d1[i].parent, upgma_merges[i][0]);
    BOOST_CHECK_EQUAL(d1[i].first,  upgma_merges[i][1]);
    BOOST_CHECK_EQUAL(d1[i].second, upgma_merges[i][2]);
  }
}
