#include <cmath>
#include <vector>

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include <ttcl/co/mmap_vector.hxx>
#include <ttcl/hj/dendrogram.hxx>
#include <ttcl/hj/ag/agglomerative_distance_clusterer.hxx>
#include <ttcl/hj/ag/limit_agglomerative_distance_clusterer.hxx>
#include <ttcl/hj/ag/single_link.hxx>
#include <ttcl/hj/ag/complete_link.hxx>
#include <ttcl/hj/ag/upgma.hxx>

// Static vars
ttcl_tu_static;
ttcl_tu_option_static(ttcl::hj::ag::limit_queue_size);
ttcl_tu_option_static(ttcl::hj::ag::merging_threshold);

// Data format
struct point {
  double x;
  double y;
  point() : x(0), y(0) {}
  point(double _x, double _y) : x(_x), y(_y) {}
};

// Distance
struct euclidean {
  typedef double result_type;

  double operator()(const point& _a, const point& _b) const {
    double dx = _a.x - _b.x;
    double dy = _a.y - _b.y;
    return std::sqrt(dx * dx + dy * dy);
  }
};

// Number of samples
const unsigned int n_samples = 1000;

// Data
std::vector<point> points;

// Mmap symmetric matrix
template <typename Value>
struct mmap_symmetric_matrix :
  public ttcl::co::symmetric_matrix< ttcl::co::mmap_vector<Value> > {
  // Base type
  typedef ttcl::co::symmetric_matrix< ttcl::co::mmap_vector<Value> > base_type;

  // Constructor
  mmap_symmetric_matrix(const typename base_type::size_type _size) :
    base_type(_size) {
  }
};

/// Mmap queue
template <typename Value>
class mmap_queue :
  public std::priority_queue<Value,
			     ttcl::co::mmap_vector<Value>,
			     std::greater<Value> > {
};

// Choices
typedef  
ttcl_tu_add_choice_tem(
  ttcl_tu_add_choice_tem(ttcl::tu::default_choices,
			 ttcl::cl::distance_matrix_template,
			 mmap_symmetric_matrix),
  ttcl::hj::ag::queue_template,
  mmap_queue) mmap_choices;


// Single link
BOOST_AUTO_TEST_CASE( single_link_cluster ) {
  // Fill the samples
  for (unsigned int i = 0; i < n_samples; ++i)
    points.push_back(point(double(rand()) / (double(RAND_MAX) + 1.0),
			   double(rand()) / (double(RAND_MAX) + 1.0)));

  // Standard
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<
    std::vector<point>, euclidean, ttcl::hj::ag::single_link>
    std_clusterer_type;
  std_clusterer_type std_clusterer;
  std_clusterer_type::dendrogram_type std_dendro =
    std_clusterer.hier_cluster(points);

  // Limited
  typedef ttcl::hj::ag::limit_agglomerative_distance_clusterer<
    std::vector<point>, euclidean, ttcl::hj::ag::single_link>
    limit_clusterer_type;
  limit_clusterer_type limit_clusterer;
  limit_clusterer_type::dendrogram_type limit_dendro =
    limit_clusterer.hier_cluster(points);
  
  // Check equal
  BOOST_REQUIRE_EQUAL(std_dendro.branches(), limit_dendro.branches());
  for (uint i = 0; i < std_dendro.branches(); ++i) {
    BOOST_CHECK_EQUAL(std_dendro[i].parent,   limit_dendro[i].parent);
    BOOST_CHECK_EQUAL(std_dendro[i].first,    limit_dendro[i].first);
    BOOST_CHECK_EQUAL(std_dendro[i].second,   limit_dendro[i].second);
    BOOST_CHECK_CLOSE(std_dendro[i].distance, limit_dendro[i].distance, 1e-4);
  }

  // Mmap
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<
    std::vector<point>, euclidean, ttcl::hj::ag::single_link, mmap_choices>
    mmap_clusterer_type;
  mmap_clusterer_type mmap_clusterer;
  mmap_clusterer_type::dendrogram_type mmap_dendro =
    mmap_clusterer.hier_cluster(points);

  // Check equal
  BOOST_REQUIRE_EQUAL(std_dendro.branches(), mmap_dendro.branches());
  for (uint i = 0; i < std_dendro.branches(); ++i) {
    BOOST_CHECK_EQUAL(std_dendro[i].parent,   mmap_dendro[i].parent);
    BOOST_CHECK_EQUAL(std_dendro[i].first,    mmap_dendro[i].first);
    BOOST_CHECK_EQUAL(std_dendro[i].second,   mmap_dendro[i].second);
    BOOST_CHECK_CLOSE(std_dendro[i].distance, mmap_dendro[i].distance, 1e-4);
  }
}
