#include <algorithm>
#include <cmath>
#include <functional>
#include <iostream>
#include <vector>

#define BOOST_TEST_MAIN
#include <boost/test/floating_point_comparison.hpp>
#include <boost/test/included/unit_test.hpp>

#include <ttcl/cl/wr/data_wrap.hxx>
#include <ttcl/cl/wr/matrix_wrap.hxx>
#include <ttcl/co/distance_matrix.hxx>
#include <ttcl/cr/calinski_centroid_score_data.hxx>
#include <ttcl/cr/calinski_medoid_score_data.hxx>
#include <ttcl/cr/calinski_medoid_score_matrix.hxx>
#include <ttcl/cr/wr/data_criterion_wrap.hxx>
#include <ttcl/cr/wr/matrix_criterion_wrap.hxx>
#include <ttcl/hj/dendrogram.hxx>

#define NaN NAN
#include "calinski_data.hxx"

// Precision
const double precision = 1e-6;

// Data type
struct data_vector : public std::vector<double> {
  // Empty constructor
  data_vector() :
    std::vector<double>() {
  }

  // Sized constructor
  data_vector(size_type _size, double _value = 0.0) :
    std::vector<double>(_size, _value) {
  }

  // Add
  data_vector& operator+=(const data_vector& _other) {
    // Resize?
    if (_other.size() > this->size())
      resize(_other.size(), 0.0);

    // Add
    std::transform(this ->begin(), this ->end(),
		   _other.begin(),
		   this ->begin(),
		   std::plus<double>());

    // Return this
    return *this;
  }

  // Normalize
  data_vector& operator/=(double s) {
    // Divide each
    std::transform(this->begin(), this->end(),
		   this->begin(),
		   std::bind2nd(std::divides<double>(), s));

    // Return this
    return *this;
  }
};

// Distance
struct euclidean {
  typedef double result_type;

  double operator()(const data_vector& _a, const data_vector& _b) const {
    double sum = 0.0;

    // Follow each
    data_vector::const_iterator it_a = _a.begin();
    data_vector::const_iterator it_b = _b.begin();
    while (it_a != _a.end() and it_b != _b.end()) {
      // Update sum
      double diff = *it_a - *it_b;
      sum += diff * diff;

      // Next
      ++it_a;
      ++it_b;
    }

    // Remains in a?
    while (it_a != _a.end()) {
      // Update sum and next
      sum += *it_a * *it_a;
      ++it_a;
    }

    // Remains in b?
    while (it_b != _b.end()) {
      // Update sum and next
      sum += *it_b * *it_b;
      ++it_b;
    }

    // Return the square root of that
    return std::sqrt(sum);
  }
};

// Calinski
BOOST_AUTO_TEST_CASE( calinski ) {
  // Data
  typedef std::vector<data_vector> data_type;
  data_type data(N_DATA, data_vector(N_DIMS));
  for (int i = 0; i < N_DATA; ++i)
    for (int j = 0; j < N_DIMS; ++j)
      data[i][j] = DATA[i][j];

  // Wrap
  ttcl::cl::wr::data_wrap data_w = data;


  // Distance matrix
  typedef ttcl::co::distance_matrix<> matrix_type;
  matrix_type matrix(data, euclidean());

  // Wrap
  ttcl::cl::wr::matrix_wrap matrix_w = matrix;


  // Check they are the same
  for (int i = 0; i < N_DATA; ++i)
    for (int j = i + 1; j < N_DATA; ++j)
      BOOST_CHECK_CLOSE(matrix(i, j), DISTANCES[i][j], precision);


  // Dendrogram
  typedef ttcl::hj::dendrogram<double> dendrogram_type;
  dendrogram_type dendrogram(N_DATA);
  for (int i = 0; i < N_MERGES; ++i)
    BOOST_CHECK_EQUAL(dendrogram.merge(DENDROGRAM[i][1], DENDROGRAM[i][2]),
		      DENDROGRAM[i][0]);


  // Calinski centroids
  typedef ttcl::cr::calinski_centroid_score_data
    <data_type, euclidean> cal_cd_type;
  cal_cd_type cal_cd;

  // Wrap
  ttcl::cr::wr::data_criterion_wrap<> cal_cd_w = cal_cd;


  // Calinski medoids (data)
  typedef ttcl::cr::calinski_medoid_score_data
    <data_type, euclidean> cal_md_type;
  cal_md_type cal_md;

  // Wrap
  ttcl::cr::wr::data_criterion_wrap<> cal_md_w = cal_md;


  // Calinski medoids (matrix)
  typedef ttcl::cr::calinski_medoid_score_matrix
    <matrix_type> cal_mm_type;
  cal_mm_type cal_mm;

  // Wrap
  ttcl::cr::wr::matrix_criterion_wrap<> cal_mm_w = cal_mm;


  // Call for the trees
  std::vector<double> tree_cd = cal_cd_w.hier_apply(dendrogram, data_w);
  std::vector<double> tree_md = cal_md_w.hier_apply(dendrogram, data_w);
  std::vector<double> tree_mm = cal_mm_w.hier_apply(dendrogram, matrix_w);

  // At each level
  for (int k = 1; k <= N_DATA; ++k) {
    // Cut
    ttcl::cl::hard_clustering<> hard = dendrogram.cut(k);

    // Find
    double hard_cd = cal_cd_w.hard_apply(hard, data_w);
    double hard_md = cal_md_w.hard_apply(hard, data_w);
    double hard_mm = cal_mm_w.hard_apply(hard, matrix_w);

    // Nan?
    if (std::isnan(CENT_CALINSKIS[k - 1])) {
      // Is Nan?
      BOOST_CHECK_PREDICATE(isnan, (tree_cd[k-1]));
      BOOST_CHECK_PREDICATE(isnan, (tree_md[k-1]));
      BOOST_CHECK_PREDICATE(isnan, (tree_mm[k-1]));

      BOOST_CHECK_PREDICATE(isnan, (hard_cd));
      BOOST_CHECK_PREDICATE(isnan, (hard_md));
      BOOST_CHECK_PREDICATE(isnan, (hard_mm));
    }
    else {
      // Close
      BOOST_CHECK_CLOSE(tree_cd[k - 1], CENT_CALINSKIS[k - 1], precision);
      BOOST_CHECK_CLOSE(tree_md[k - 1], MED_CALINSKIS [k - 1], precision);
      BOOST_CHECK_CLOSE(tree_mm[k - 1], MED_CALINSKIS [k - 1], precision);

      BOOST_CHECK_CLOSE(hard_cd, CENT_CALINSKIS[k - 1], precision);
      BOOST_CHECK_CLOSE(hard_md, MED_CALINSKIS [k - 1], precision);
      BOOST_CHECK_CLOSE(hard_mm, MED_CALINSKIS [k - 1], precision);
    }
  }
}
