#include <iostream>
#include <utility>
#include <vector>

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>
#include <boost/tuple/tuple.hpp>

#include <ttcl/cl/wr/data_wrap.hxx>
#include <ttcl/cl/wr/hard_data_clusterer_wrap.hxx>
#include <ttcl/cl/wr/hard_data_model_clusterer_wrap.hxx>
#include <ttcl/cl/wr/hard_data_model_wrap.hxx>
#include <ttcl/cl/wr/soft_data_clusterer_wrap.hxx>
#include <ttcl/cl/wr/soft_data_model_clusterer_wrap.hxx>
#include <ttcl/cl/wr/soft_data_model_wrap.hxx>
#include <ttcl/co/array.hxx>
#include <ttcl/co/centroid_set.hxx>
#include <ttcl/co/fake_pair.hxx>
#include <ttcl/co/feature_container.hxx>
#include <ttcl/hj/ag/agglomerative_distance_clusterer.hxx>
#include <ttcl/hj/ag/single_link.hxx>
#include <ttcl/hj/ag/complete_link.hxx>
#include <ttcl/hj/ag/upgma.hxx>
#include <ttcl/hj/ag/upgma.hxx>
#include <ttcl/hj/wr/hier_data_clusterer_wrap.hxx>
#include <ttcl/it/em/alpha_bregman_clusterer.hxx>
#include <ttcl/it/em/bregman_clusterer.hxx>
#include <ttcl/it/em/bregman_divergence_space.hxx>
#include <ttcl/it/em/ditc_clusterer.hxx>
#include <ttcl/it/em/nb_bernouilli_clusterer.hxx>
#include <ttcl/it/em/nb_bernouilli_incremental_clusterer.hxx>
#include <ttcl/it/em/nb_bernouilli_noisy_clusterer.hxx>
#include <ttcl/it/em/nb_multinomial_clusterer.hxx>

// Static vars
ttcl_tu_static;
ttcl_tu_option_static(ttcl::hj::ag::merging_threshold);
ttcl_tu_option_static(ttcl::it::max_iterations);
ttcl_tu_option_static(ttcl::it::stop_threshold);
ttcl_tu_option_static(ttcl::it::start_policy);
ttcl_tu_option_static(ttcl::it::peak_fraction);
ttcl_tu_option_static(ttcl::it::em::alpha_laplace_term);
ttcl_tu_option_static(ttcl::it::em::theta_laplace_term);
ttcl_tu_option_static(ttcl::it::em::learning_rate);
ttcl_tu_option_static(ttcl::it::em::inner_softness);
ttcl_tu_option_static(ttcl::it::em::prior_scaling);
ttcl_tu_option_static(ttcl::it::em::prior_threshold);
ttcl_tu_option_static(ttcl::it::em::starting_prior);

/* Data from nb_clustering.cxx */

// Number of points
const unsigned int n_points = 6;

// Points
const unsigned int  point1[] = { 3, /* -> */ 0, 1, 2 };
const unsigned int  point2[] = { 3, /* -> */ 0, 1, 3 };
const unsigned int  point3[] = { 2, /* -> */ 4, 5 };
const unsigned int  point4[] = { 5, /* -> */ 2, 3, 4, 5 };
const unsigned int  point5[] = { 3, /* -> */ 3, 4, 5 };
const unsigned int  point6[] = { 3, /* -> */ 0, 2, 3 };
const unsigned int* points[] = { point1, point2, point3,
				 point4, point5, point6 };

// Types
typedef ttcl::co::centroid_set<double>::const_centroid_type centroid_type;
typedef ttcl::cl::hard_clustering<> hard_clustering_type;
typedef ttcl::cl::soft_clustering<> soft_clustering_type;
typedef ttcl::hj::dendrogram<double> dendrogram_type;


/*************/
/* Distances */
/*************/

// Squared euclidean distance
/* Dense vs Sparse version */
template <typename Sample>
struct sq_euclidean_distance {
  // Result type
  typedef double result_type;

  // Apply
  result_type
  operator()(const centroid_type& _a, const Sample& _b) const {
    // Sum so far
    double sum = 0.0;

    // Merge sortish
    typename Sample::const_iterator it_b = _b.begin();
    for (ttcl::uint i = 0; i < _a.size(); ++i) {
      if (it_b == _b.end() || i < it_b->first) {
	sum += _a[i] * _a[i];
      }
      else { // i == it_b->first
	double diff = (_a[i] - it_b->second);
	sum += diff * diff;
	++it_b;
      }
    }

    // Return the sum
    return sum;
  }
};

// Squared euclidean distance
/* Sparse version */
template <typename Sample>
struct sq_euclidean_distance_sparse {
  // Result type
  typedef double result_type;

  // Apply
  result_type
  operator()(const Sample& _a, const Sample& _b) const {
    // Sum so far
    double sum = 0.0;

    // Merge sortish
    typename Sample::const_iterator it_a = _a.begin();
    typename Sample::const_iterator it_b = _b.begin();
    while (it_a != _a.end() and it_b != _b.end()) {
      if (it_a->first < it_b->first) {
	sum += it_a->second * it_a->second;
	++it_a;
      }
      else if (it_a->first > it_b->first) {
	sum += it_b->second * it_b->second;
	++it_b;
      }
      else { // it_a->first == it_b->first
	double diff = (it_a - it_b);
	sum += diff * diff;
	++it_a;
	++it_b;
      }
    }

    // Finish what remains
    while (it_a != _a.end()) {
      sum += it_a->second * it_a->second;
      ++it_a;
    }
    while (it_b != _b.end()) {
      sum += it_b->second * it_b->second;
      ++it_b;
    }

    // Return the sum
    return sum;
  }
};


/***********/
/* Types 1 */
/***********/

// Sample
typedef ttcl::co::array<
  ttcl::co::fake_pair_second<unsigned int,
			     unsigned int, 1> > sample_type_1;

// Data
typedef ttcl::co::feature_container< std::vector<sample_type_1> > data_type_1;

// Distance
typedef sq_euclidean_distance<sample_type_1> distance_type_1;

// Sparse distance
typedef sq_euclidean_distance_sparse<sample_type_1> sp_distance_type_1;

// Model clusterers
typedef ttcl::it::em::nb_multinomial_clusterer<data_type_1> multi_1;
typedef ttcl::it::em::nb_bernouilli_clusterer<data_type_1> berni_1;
typedef ttcl::it::em::nb_bernouilli_noisy_clusterer<data_type_1> nberni_1;
typedef ttcl::it::em::nb_bernouilli_incremental_clusterer<data_type_1> iberni_1;
typedef ttcl::it::em::bregman_clusterer<data_type_1, distance_type_1> bg_1;
typedef ttcl::it::em::alpha_bregman_clusterer<data_type_1,
					      distance_type_1> abg_1;
typedef ttcl::it::em::ditc_clusterer<data_type_1> ditc_1;

// Hierarchical clusterers
typedef ttcl::hj::ag::agglomerative_distance_clusterer
  <data_type_1, sp_distance_type_1, ttcl::hj::ag::single_link> ag_sl_1;
typedef ttcl::hj::ag::agglomerative_distance_clusterer
  <data_type_1, sp_distance_type_1, ttcl::hj::ag::complete_link> ag_cl_1;
typedef ttcl::hj::ag::agglomerative_distance_clusterer
  <data_type_1, sp_distance_type_1, ttcl::hj::ag::upgma> ag_upgma_1;


/***********/
/* Types 2 */
/***********/

// Sample
typedef std::vector< std::pair<unsigned int, unsigned int> > sample_type_2;

// Data
typedef ttcl::co::feature_container< std::vector<sample_type_2> > data_type_2;

// Distance
typedef sq_euclidean_distance<sample_type_2> distance_type_2;

// Distance
typedef sq_euclidean_distance<sample_type_2> distance_type_2;

// Sparse distance
typedef sq_euclidean_distance_sparse<sample_type_2> sp_distance_type_2;

// Model clusterers
typedef ttcl::it::em::nb_multinomial_clusterer<data_type_2> multi_2;
typedef ttcl::it::em::nb_bernouilli_clusterer<data_type_2> berni_2;
typedef ttcl::it::em::nb_bernouilli_noisy_clusterer<data_type_2> nberni_2;
typedef ttcl::it::em::nb_bernouilli_incremental_clusterer<data_type_2> iberni_2;
typedef ttcl::it::em::bregman_clusterer<data_type_2, distance_type_2> bg_2;
typedef ttcl::it::em::alpha_bregman_clusterer<data_type_2,
					      distance_type_2> abg_2;
typedef ttcl::it::em::ditc_clusterer<data_type_2> ditc_2;

// Hierarchical clusterers
typedef ttcl::hj::ag::agglomerative_distance_clusterer
  <data_type_2, sp_distance_type_2, ttcl::hj::ag::single_link> ag_sl_2;
typedef ttcl::hj::ag::agglomerative_distance_clusterer
  <data_type_2, sp_distance_type_2, ttcl::hj::ag::complete_link> ag_cl_2;
typedef ttcl::hj::ag::agglomerative_distance_clusterer
  <data_type_2, sp_distance_type_2, ttcl::hj::ag::upgma> ag_upgma_2;


/********/
/* Data */
/********/

// Data
data_type_1  data_1;
data_type_2* data_2;

// Wrappers
ttcl::cl::wr::data_wrap dw_1;
ttcl::cl::wr::data_wrap dw_2;

// Ensure data
BOOST_AUTO_TEST_CASE( t1_create_data ) {
  // Create data 1
  for (unsigned int i = 0; i < n_points; ++i) {
    data_1.push_back(sample_type_1(points[i][0]));
    for (unsigned int f = 0; f < points[i][0]; ++f)
      data_1.back()[f].first = points[i][f + 1];
  }

  // Create data 2
  data_2 = new data_type_2();
  for (unsigned int i = 0; i < n_points; ++i) {
    data_2->push_back(sample_type_2(points[i][0]));
    for (unsigned int f = 0; f < points[i][0]; ++f) {
      data_2->back()[f].first  = points[i][f + 1];
      data_2->back()[f].second = 1;
    }
  }

  // Wrap
  dw_1 = data_1;
  dw_2 = data_2;
}


/*********/
/* Wraps */
/*********/

// Test
BOOST_AUTO_TEST_CASE( t2_hard_data_clusterer ) {
  // Error
  bool error_ok;

  // Hard clusterer wrappers (type 1)
  std::vector< ttcl::cl::wr::hard_data_clusterer_wrap<> > wr1(10);
  wr1[0] = new multi_1();
  wr1[1] = new berni_1();
  wr1[2] = new nberni_1();
  wr1[3] = new iberni_1();
  wr1[4] = new bg_1();
  wr1[5] = new abg_1();
  wr1[6] = new ditc_1();
  wr1[7] = new ag_sl_1();
  wr1[8] = new ag_cl_1();
  wr1[9] = new ag_upgma_1();

  // Hard clusterer wrappers (type 2)
  std::vector< ttcl::cl::wr::hard_data_clusterer_wrap<> > wr2(10);
  wr2[0] = new multi_2();
  wr2[1] = new berni_2();
  wr2[2] = new nberni_2();
  wr2[3] = new iberni_2();
  wr2[4] = new bg_2();
  wr2[5] = new abg_2();
  wr2[6] = new ditc_2();
  wr2[7] = new ag_sl_2();
  wr2[8] = new ag_cl_2();
  wr2[9] = new ag_upgma_2();

  // Find hard clusterings
  for (uint i = 0; i < wr1.size(); ++i) {
    // OK
    hard_clustering_type hard1 = wr1[i].hard_cluster(dw_1, 2);
    hard_clustering_type hard2 = wr2[i].hard_cluster(dw_2, 2);

    // Errors
    error_ok = false;
    try   { wr1[i].hard_cluster(dw_2, 2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { wr2[i].hard_cluster(dw_1, 2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);
  }
}

// Test
BOOST_AUTO_TEST_CASE( t3_soft_data_clusterer ) {
  // Error
  bool error_ok;

  // Hard clusterer wrappers (type 1)
  std::vector< ttcl::cl::wr::soft_data_clusterer_wrap<> > wr1(7);
  wr1[0] = new multi_1();
  wr1[1] = new berni_1();
  wr1[2] = new nberni_1();
  wr1[3] = new iberni_1();
  wr1[4] = new bg_1();
  wr1[5] = new abg_1();
  wr1[6] = new ditc_1();

  // Hard clusterer wrappers (type 2)
  std::vector< ttcl::cl::wr::soft_data_clusterer_wrap<> > wr2(7);
  wr2[0] = new multi_2();
  wr2[1] = new berni_2();
  wr2[2] = new nberni_2();
  wr2[3] = new iberni_2();
  wr2[4] = new bg_2();
  wr2[5] = new abg_2();
  wr2[6] = new ditc_2();

  // Find hard clusterings
  for (uint i = 0; i < wr1.size(); ++i) {
    // OK
    hard_clustering_type hard1 = wr1[i].hard_cluster(dw_1, 2);
    hard_clustering_type hard2 = wr2[i].hard_cluster(dw_2, 2);
    soft_clustering_type soft1 = wr1[i].soft_cluster(dw_1, 2);
    soft_clustering_type soft2 = wr2[i].soft_cluster(dw_2, 2);

    // Errors
    error_ok = false;
    try   { wr1[i].hard_cluster(dw_2, 2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { wr2[i].hard_cluster(dw_1, 2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { wr1[i].soft_cluster(dw_2, 2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { wr2[i].soft_cluster(dw_1, 2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);
  }
}

// Test
BOOST_AUTO_TEST_CASE( t4_hard_data_model_clusterer ) {
  // Error
  bool error_ok;

  // Hard clusterer wrappers (type 1)
  std::vector< ttcl::cl::wr::hard_data_model_clusterer_wrap<> > wr1(7);
  wr1[0] = new multi_1();
  wr1[1] = new berni_1();
  wr1[2] = new nberni_1();
  wr1[3] = new iberni_1();
  wr1[4] = new bg_1();
  wr1[5] = new abg_1();
  wr1[6] = new ditc_1();

  // Hard clusterer wrappers (type 2)
  std::vector< ttcl::cl::wr::hard_data_model_clusterer_wrap<> > wr2(7);
  wr2[0] = new multi_2();
  wr2[1] = new berni_2();
  wr2[2] = new nberni_2();
  wr2[3] = new iberni_2();
  wr2[4] = new bg_2();
  wr2[5] = new abg_2();
  wr2[6] = new ditc_2();

  // Find hard clusterings
  for (uint i = 0; i < wr1.size(); ++i) {
    // Models
    ttcl::cl::wr::hard_data_model_wrap<> model1 = wr1[i].model(dw_1, 2);
    ttcl::cl::wr::hard_data_model_wrap<> model2 = wr2[i].model(dw_2, 2);

    // OK
    hard_clustering_type hard1 = model1.hard_cluster(dw_1);
    hard_clustering_type hard2 = model2.hard_cluster(dw_2);

    // Errors
    error_ok = false;
    try   { model1.hard_cluster(dw_2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { model2.hard_cluster(dw_1); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);
  }
}

// Test
BOOST_AUTO_TEST_CASE( t5_soft_data_model_clusterer ) {
  // Error
  bool error_ok;

  // Hard clusterer wrappers (type 1)
  std::vector< ttcl::cl::wr::soft_data_model_clusterer_wrap<> > wr1(7);
  wr1[0] = new multi_1();
  wr1[1] = new berni_1();
  wr1[2] = new nberni_1();
  wr1[3] = new iberni_1();
  wr1[4] = new bg_1();
  wr1[5] = new abg_1();
  wr1[6] = new ditc_1();

  // Hard clusterer wrappers (type 2)
  std::vector< ttcl::cl::wr::soft_data_model_clusterer_wrap<> > wr2(7);
  wr2[0] = new multi_2();
  wr2[1] = new berni_2();
  wr2[2] = new nberni_2();
  wr2[3] = new iberni_2();
  wr2[4] = new bg_2();
  wr2[5] = new abg_2();
  wr2[6] = new ditc_2();

  // Find hard clusterings
  for (uint i = 0; i < wr1.size(); ++i) {
    // Models
    ttcl::cl::wr::soft_data_model_wrap<> model1 = wr1[i].model(dw_1, 2);
    ttcl::cl::wr::soft_data_model_wrap<> model2 = wr2[i].model(dw_2, 2);

    // OK
    hard_clustering_type hard1 = model1.hard_cluster(dw_1);
    hard_clustering_type hard2 = model2.hard_cluster(dw_2);
    hard_clustering_type soft1 = model1.soft_cluster(dw_1);
    hard_clustering_type soft2 = model2.soft_cluster(dw_2);

    // Errors
    error_ok = false;
    try   { model1.hard_cluster(dw_2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { model2.hard_cluster(dw_1); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { model1.soft_cluster(dw_2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { model2.soft_cluster(dw_1); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);
  }
}

// Test
BOOST_AUTO_TEST_CASE( t6_hier_data_clusterer ) {
  // Error
  bool error_ok;

  // Hard clusterer wrappers (type 1)
  std::vector< ttcl::hj::wr::hier_data_clusterer_wrap<> > wr1(3);
  wr1[0] = new ag_sl_1();
  wr1[1] = new ag_cl_1();
  wr1[2] = new ag_upgma_1();

  // Hard clusterer wrappers (type 2)
  std::vector< ttcl::hj::wr::hier_data_clusterer_wrap<> > wr2(3);
  wr2[0] = new ag_sl_2();
  wr2[1] = new ag_cl_2();
  wr2[2] = new ag_upgma_2();

  // Find hard clusterings
  for (uint i = 0; i < wr1.size(); ++i) {
    // OK
    hard_clustering_type hard1 = wr1[i].hard_cluster(dw_1, 2);
    hard_clustering_type hard2 = wr2[i].hard_cluster(dw_2, 2);
    dendrogram_type    dendro1 = wr1[i].hier_cluster(dw_1, 2);
    dendrogram_type    dendro2 = wr2[i].hier_cluster(dw_2, 2);

    // Errors
    error_ok = false;
    try   { wr1[i].hard_cluster(dw_2, 2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { wr2[i].hard_cluster(dw_1, 2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { wr1[i].hier_cluster(dw_2, 2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);

    error_ok = false;
    try   { wr2[i].hier_cluster(dw_1, 2); }
    catch (ttcl::exception& _e) { error_ok = _e.message() == "Bad cast"; }
    catch (...) {}
    BOOST_CHECK(error_ok);
  }
}
