#include <algorithm>
#include <functional>
#include <utility>
#include <vector>

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>

#include <ttcl/co/extended_priority_queue.hxx>
#include <ttcl/ut/range.hxx>

// Number of values
const int n_values = 8;

// Starting element values
const double start_values[] = { 3.0, 1.0, 2.0, 8.0, 5.0, 7.0, 4.0, 6.0 };

// Starting sorted
const int start_sorted[] = { 3, 5, 7, 4, 6, 0, 2, 1 };

// Second element values
const double second_values[] = { 5.5, 2.5, 6.5, 3.5, 1.5, 8.5, 7.5, 4.5 };

// Second sorted
const int second_sorted[] = { 5, 6, 2, 0, 7, 3, 1, 4 };

// Types
typedef ttcl::co::extended_priority_queue<double> max_heap_type;
typedef ttcl::co::extended_priority_queue<double,
  std::vector<double>, std::greater<double> > min_heap_type;

// Simple max heap
BOOST_AUTO_TEST_CASE(simple_max_heap) {
  // Max heap
  max_heap_type max_heap;
  for (int i = 0; i < n_values; ++i)
    max_heap.push(start_values[i]);

  // Check
  BOOST_CHECK_EQUAL(max_heap.empty(), n_values == 0);
  BOOST_CHECK_EQUAL(max_heap.size(),  n_values);

  // Access
  for (int i = 0; i < n_values; ++i)
    BOOST_CHECK_EQUAL(max_heap[i], start_values[i]);

  // Top
  int i = 0;
  while (i < n_values and not max_heap.empty()) {
    // Check
    BOOST_CHECK_EQUAL(max_heap.top(), start_values[start_sorted[i]]);
    BOOST_CHECK_EQUAL(max_heap.top_element(), start_sorted[i]);
    
    // Pop
    max_heap.pop();
    ++i;
    BOOST_CHECK_EQUAL(max_heap.size(), n_values - i);
  }
}

// Simple min heap
BOOST_AUTO_TEST_CASE(simple_min_heap) {
  // Min heap
  min_heap_type min_heap;
  for (int i = 0; i < n_values; ++i)
    min_heap.push(start_values[i]);

  // Check
  BOOST_CHECK_EQUAL(min_heap.empty(), n_values == 0);
  BOOST_CHECK_EQUAL(min_heap.size(),  n_values);

  // Access
  for (int i = 0; i < n_values; ++i)
    BOOST_CHECK_EQUAL(min_heap[i], start_values[i]);

  // Top
  int i = n_values - 1;
  while (i >= 0 and not min_heap.empty()) {
    // Check
    BOOST_CHECK_EQUAL(min_heap.top(), start_values[start_sorted[i]]);
    BOOST_CHECK_EQUAL(min_heap.top_element(), start_sorted[i]);
    
    // Pop
    min_heap.pop();
    BOOST_CHECK_EQUAL(min_heap.size(), i);
    --i;
  }
}

// Updated max heap
BOOST_AUTO_TEST_CASE(updated_max_heap) {
  // Max heap
  max_heap_type max_heap;
  for (int i = 0; i < n_values; ++i)
    max_heap.push(start_values[i]);

  // Check
  BOOST_CHECK_EQUAL(max_heap.empty(), n_values == 0);
  BOOST_CHECK_EQUAL(max_heap.size(),  n_values);

  // Now, update
  for (int i = 0; i < n_values; ++i)
    max_heap[i] = second_values[i];

  // Access
  for (int i = 0; i < n_values; ++i)
    BOOST_CHECK_EQUAL(max_heap[i], second_values[i]);

  // Top
  int i = 0;
  while (i < n_values and not max_heap.empty()) {
    // Check
    BOOST_CHECK_EQUAL(max_heap.top(), second_values[second_sorted[i]]);
    BOOST_CHECK_EQUAL(max_heap.top_element(), second_sorted[i]);
    
    // Pop
    max_heap.pop();
    ++i;
    BOOST_CHECK_EQUAL(max_heap.size(), n_values - i);
  }
}

// Updated min heap
BOOST_AUTO_TEST_CASE(updated_min_heap) {
  // Min heap
  min_heap_type min_heap;
  for (int i = 0; i < n_values; ++i)
    min_heap.push(start_values[i]);

  // Check
  BOOST_CHECK_EQUAL(min_heap.empty(), n_values == 0);
  BOOST_CHECK_EQUAL(min_heap.size(),  n_values);

  // Now, update
  for (int i = 0; i < n_values; ++i)
    min_heap[i] = second_values[i];

  // Access
  for (int i = 0; i < n_values; ++i)
    BOOST_CHECK_EQUAL(min_heap[i], second_values[i]);

  // Top
  int i = n_values - 1;
  while (i >= 0 and not min_heap.empty()) {
    // Check
    BOOST_CHECK_EQUAL(min_heap.top(), second_values[second_sorted[i]]);
    BOOST_CHECK_EQUAL(min_heap.top_element(), second_sorted[i]);
    
    // Pop
    min_heap.pop();
    BOOST_CHECK_EQUAL(min_heap.size(), i);
    --i;
  }
}
