-- Data generator for Calinski and Harabasz scores
module Main (main) where

-- Imports
import Control.Monad (liftM, replicateM, sequence_)
import Data.Array
import Data.Function (on)
import Data.List (intersperse, minimumBy, sort, transpose)
import Data.Ord (comparing)
import System.Environment (getArgs)
import System.Exit (ExitCode(..), exitWith)
import System.Random (getStdRandom, mkStdGen, random, randomR, setStdGen)
import Text.Printf (printf)


-- Vector type
type Vector  = [ Double ]
type DataSet = Array Int Vector

-- Generate a random data set
randDataSet :: Int -> Int -> IO DataSet
randDataSet n d = liftM (listArray (1, n)) $ replicateM n randVector
    where
      randVector = replicateM d randElem
      randElem   = getStdRandom random


-- Cluster
type Cluster = [ Int ]

-- Cluster elements
clusterElements :: DataSet -> Cluster -> [ Vector ]
clusterElements ds = map (ds !)


-- Matrix
type Matrix = Array Int (Array Int Double)

-- Square a value
square x = x * x

-- Euclidean distance
distance :: Vector -> Vector -> Double
distance x y = sqrt $ sum . map square $ zipWith (-) x y

-- Distance matrix
distanceMatrix :: DataSet -> Matrix
distanceMatrix ds = listArray bnds $ map row idxs
    where
      bnds  = bounds  ds
      idxs  = indices ds
      row r = listArray bnds [ distance (ds ! r) (ds ! c) | c <- idxs ]


-- Average function
type AvgFun = DataSet -> Cluster -> Vector

-- Find centroid
centroid:: AvgFun
centroid ds = map mean . transpose . clusterElements ds
    where
      mean l = sum l / fromIntegral (length l)

-- Find medoid
medoid :: AvgFun
medoid ds cl = minimumBy (comparing distanceSum) elems
    where
      elems         = clusterElements ds cl
      distanceSum v = sum $ map (distance v) elems

-- Sum of distances
sumOfDistances :: AvgFun -> DataSet -> Cluster -> Double
sumOfDistances f ds cl = sum $ map (distance avg) elems
    where
      elems = clusterElements ds cl
      avg   = f ds cl

-- Sum of Squared Distances
sumOfSqDistances :: AvgFun -> DataSet -> Cluster -> Double
sumOfSqDistances f ds cl = sum $ map (square . distance avg) elems
    where
      elems = clusterElements ds cl
      avg   = f ds cl


-- Clustering
type Clustering = [ Cluster ]

-- Calinski
calinski :: AvgFun -> DataSet -> Clustering -> Double
calinski f ds cls = (sumBetween * fromIntegral (n - k)) /
                    (sumWithin  * fromIntegral (k - 1))
    where
      (1, n) = bounds ds
      k      = length cls

      sumBetween = sum $ zipWith (*) (map (fromIntegral . length) cls)
                                     (map (square . distance metaAvg) avgs)

      metaAvg    = f ds [ 1 .. n ]
      avgs       = map (f ds) cls

      sumWithin  = sum $ map (sumOfSqDistances f ds) cls


-- Branch
data Branch = Branch Int Int Int
              deriving (Show)

-- Dendrogram
type Dendrogram = [ Branch ]

-- Random pair
randPair :: Int -> IO (Int, Int)
randPair n = do x1 <- getStdRandom $ randomR (1,  n     )
                x2 <- getStdRandom $ randomR (1, (n - 1))
                return $ normalize x1 x2
    where
      normalize x1 x2
          | x1 <= x2  = (x1, x2 + 1)
          | otherwise = (x2, x1)

-- Random dendrogram
randDendrogram :: Int -> IO Dendrogram
randDendrogram n = randDendrogram' n (n + 1) [1 .. n]
    where
      randDendrogram' 1 _    _     = return []
      randDendrogram' n next elems =
          do (x1, x2) <- randPair n

             let (left, (e1 : mid'))  = splitAt (x1 - 1)      elems
             let (mid,  (e2 : right)) = splitAt (x2 - x1 - 1) mid'

             liftM (Branch next e1 e2 :) $
                   randDendrogram' (n - 1) (next + 1)
                                   (left ++ mid ++ right ++ [ next ])

-- Dendrogram clusterings
dendrogramClusterings:: Int -> Dendrogram -> [ Clustering ]
dendrogramClusterings n d =
    singletons : dendrogramClusterings' d singletons
    where
      singletons = map (\i -> [i]) [1 .. n]

      dendrogramClusterings'  []                  _        = []
      dendrogramClusterings' (Branch _ e1 e2 : d) clusters =
          neClusters' : dendrogramClusterings' d clusters'
          where
            (left, (c1 : mid'))  = splitAt (e1 - 1)      clusters
            (mid,  (c2 : right)) = splitAt (e2 - e1 - 1) mid'

            newCluster = merge c1 c2
            clusters'  = left ++ [[]] ++ mid ++ [[]] ++ right ++ [ newCluster ]

            neClusters' = filter (not . null) clusters'

-- Merge two lists
merge :: (Ord a) => [a] -> [a] -> [a]
merge xs [] = xs
merge [] ys = ys
merge xs@(x : xs') ys@(y : ys')
    | x < y     = x : merge xs' ys
    | otherwise = y : merge xs  ys'


-- Perl join(...) equivalent
perlJoin :: (Show a) => String -> [a] -> String
perlJoin j xs = concat $ intersperse j $ map show xs

-- Print the dimensions
printDimensions :: Int -> Int -> IO ()
printDimensions n d =
    do printf "const int N_DATA = %d;\n" n
       printf "const int N_DIMS = %d;\n" d

-- Print the dataset
printDataset :: DataSet -> IO ()
printDataset ds =
    do printf "const double DATA[N_DATA][N_DIMS] = {"
       printDataset' $ elems ds

    where
      printDataset' [x] =
          printf "\n\t{ %s }\n};\n" $ perlJoin ", " x
      printDataset' (x : xs) =
          do printf "\n\t{ %s }," $ perlJoin ", " x
             printDataset' xs

-- Print the distances
printDistances :: Matrix -> IO()
printDistances dm =
    do printf "const double DISTANCES[N_DATA][N_DATA] = {"
       printDistances' $ elems dm

    where
      printDistances' [r] =
          printf "\n\t{ %s }\n};\n" $ perlJoin ", " $ elems r
      printDistances' (r : rs) =
          do printf "\n\t{ %s }," $ perlJoin ", " $ elems r
             printDistances' rs

-- Print the dendrogram
printDendrogram :: Dendrogram -> IO ()
printDendrogram d =
    do printf "const int N_MERGES = N_DATA - 1;\n"
       printf "const int DENDROGRAM[N_MERGES][3] = { "
       printDendrogram' d

    where
      printDendrogram' [Branch id e1 e2] =
          printf "\n\t{ %d, %d, %d }\n};\n" (id - 1) (e1 - 1) (e2 - 1)
      printDendrogram' (Branch id e1 e2 : d) =
          do printf "\n\t{ %d, %d, %d }," (id - 1) (e1 - 1) (e2 - 1)
             printDendrogram' d

-- Print the calinskis
printCalinskis :: String -> [ Double ] -> IO ()
printCalinskis label calinskis =
    do printf "const double %s[N_DATA] = {\n\t%s\n};\n"
              label (perlJoin ",\n\t" $ reverse calinskis)


-- Main
main :: IO ()
main = main' =<< getArgs
    where
      main' (strN : (strD : (strSeed : _))) =
          do let n    = read strN
             let d    = read strD
             let seed = read strSeed

             setStdGen $ mkStdGen seed

             dataSet    <- randDataSet n d
             dendrogram <- randDendrogram n

             let dists         = distanceMatrix dataSet
             let levels        = dendrogramClusterings n dendrogram
             let centCalinskis = map (calinski centroid dataSet) levels
             let medCalinskis  = map (calinski medoid   dataSet) levels

             printDimensions n d
             printDataset    dataSet
             printDistances  dists
             printDendrogram dendrogram
             printCalinskis  "CENT_CALINSKIS" centCalinskis
             printCalinskis  "MED_CALINSKIS"  medCalinskis

      main' _ =
          exitWith $ ExitFailure (-1)
