#include <vector>

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>

#include <ttcl/cl/hard_clustering.hxx>
#include <ttcl/cl/soft_clustering.hxx>
#include <ttcl/co/array.hxx>

// Number of samples
const unsigned int n_samples  = 8;

// Number of clusters in input data
const int in_clusters = 9;

// Input data
const int in_data[]  = { 3, 5, 2, 1, 1, 0, 8, 2 };

// Number of clusters in output data
const int out_clusters = 6;

// Output data
const int out_data[] = { 0, 1, 2, 3, 3, 4, 5, 2 };

// Number of samples in soft data
const unsigned int in_soft_samples = 5;

// Number of clusters in soft data
const int in_soft_clusters = 3;

// Input soft
const double in_soft_data[] = { 0.2, 0.3,  0.5,
				0.9, 0.05, 0.05,
				0.1, 0.7,  0.2,
				0.0, 0.0,  0.0,
				0.3, 0.2,  0.5 };

// Output hard
const double out_hard_data[] = { 2, 0, 1, -1, 2 };

// Clustering on array
BOOST_AUTO_TEST_CASE( array_clustering ) {
  typedef ttcl::cl::hard_clustering< ttcl::co::array<int> > clustering;

  // Construction
  clustering cl(n_samples);
  BOOST_CHECK_EQUAL(cl.size(), n_samples);
  BOOST_CHECK_EQUAL(cl.clusters(), 0);

  // Fill
  for (unsigned int i = 0; i < n_samples; ++i)
    cl[i] = in_data[i];

  // Find number of clusters
  cl.update_clusters();
  BOOST_CHECK_EQUAL(cl.clusters(), in_clusters);

  // Reindex
  cl.reindex_clusters();
  BOOST_CHECK_EQUAL(cl.clusters(), out_clusters);

  // Check
  for (unsigned int i = 0; i < n_samples; ++i)
    BOOST_CHECK_EQUAL(cl[i], out_data[i]);

  // Copy
  clustering cl2(cl);
  BOOST_CHECK_EQUAL(cl2.size(), n_samples);
  BOOST_CHECK_EQUAL(cl2.clusters(), out_clusters);
  for (unsigned int i = 0; i < n_samples; ++i)
    BOOST_CHECK_EQUAL(cl2[i], out_data[i]);

  // Assignment
  clustering cl3(5);
  BOOST_CHECK_EQUAL(cl3.size(), 5u);
  BOOST_CHECK_EQUAL(cl3.clusters(), 0);
  cl3 = cl;
  BOOST_CHECK_EQUAL(cl3.size(), n_samples);
  BOOST_CHECK_EQUAL(cl3.clusters(), out_clusters);
  for (unsigned int i = 0; i < n_samples; ++i)
    BOOST_CHECK_EQUAL(cl3[i], out_data[i]);
}

// Clustering on vector
BOOST_AUTO_TEST_CASE( vector_clustering ) {
  typedef ttcl::cl::hard_clustering< std::vector<int> > clustering;

  // Construction
  clustering cl(n_samples);
  BOOST_CHECK_EQUAL(cl.size(), n_samples);
  BOOST_CHECK_EQUAL(cl.clusters(), 0);

  // Fill
  for (unsigned int i = 0; i < n_samples; ++i)
    cl[i] = in_data[i];

  // Find number of clusters
  cl.update_clusters();
  BOOST_CHECK_EQUAL(cl.clusters(), in_clusters);

  // Reindex
  cl.reindex_clusters();
  BOOST_CHECK_EQUAL(cl.clusters(), out_clusters);

  // Check
  for (unsigned int i = 0; i < n_samples; ++i)
    BOOST_CHECK_EQUAL(cl[i], out_data[i]);

  // Copy
  clustering cl2(cl);
  BOOST_CHECK_EQUAL(cl2.size(), n_samples);
  BOOST_CHECK_EQUAL(cl2.clusters(), out_clusters);
  for (unsigned int i = 0; i < n_samples; ++i)
    BOOST_CHECK_EQUAL(cl2[i], out_data[i]);

  // Assignment
  clustering cl3(5);
  BOOST_CHECK_EQUAL(cl3.size(), 5u);
  BOOST_CHECK_EQUAL(cl3.clusters(), 0);
  cl3 = cl;
  BOOST_CHECK_EQUAL(cl3.size(), n_samples);
  BOOST_CHECK_EQUAL(cl3.clusters(), out_clusters);
  for (unsigned int i = 0; i < n_samples; ++i)
    BOOST_CHECK_EQUAL(cl3[i], out_data[i]);
}

// Conversion to soft
BOOST_AUTO_TEST_CASE( hard_to_soft_clustering ) {
  typedef ttcl::cl::hard_clustering<> hclustering;
  typedef ttcl::cl::soft_clustering<> sclustering;

  // Create
  hclustering hcl(n_samples, out_clusters);
  std::copy(out_data, out_data + n_samples,
	    hcl.begin());

  // Convert to soft
  sclustering scl1 = hcl;
  BOOST_CHECK_EQUAL(scl1.size(), n_samples);
  BOOST_CHECK_EQUAL(int(scl1.clusters()), out_clusters);

  // Check each
  for (unsigned int i = 0; i < n_samples; ++i)
    for (int j = 0; j < out_clusters; ++j)
      if (out_data[i] == j)
	BOOST_CHECK_EQUAL(scl1[i][j], 1.0);
      else
	BOOST_CHECK_EQUAL(scl1[i][j], 0.0);

  // Assign
  sclustering scl2;
  scl2 = hcl;

  // Check each
  for (unsigned int i = 0; i < n_samples; ++i)
    for (int j = 0; j < out_clusters; ++j)
      if (out_data[i] == j)
	BOOST_CHECK_EQUAL(scl2[i][j], 1.0);
      else
	BOOST_CHECK_EQUAL(scl2[i][j], 0.0);
}


// Conversion from soft
BOOST_AUTO_TEST_CASE( soft_to_hard_clustering ) {
  typedef ttcl::cl::hard_clustering<> hclustering;
  typedef ttcl::cl::soft_clustering<> sclustering;

  // Create
  sclustering scl(in_soft_samples, in_soft_clusters);
  std::copy(in_soft_data,
	    in_soft_data + in_soft_samples * in_soft_clusters,
	    scl.data());

  // Check
  BOOST_CHECK_EQUAL(scl.size(), in_soft_samples);
  BOOST_CHECK_EQUAL(int(scl.clusters()), in_soft_clusters);
  for (unsigned int i = 0; i < in_soft_samples; ++i)
    for (int j = 0; j < in_soft_clusters; ++j)
      BOOST_CHECK_EQUAL(scl[i][j], in_soft_data[i * in_soft_clusters + j]);

  // Convert to hard
  hclustering hcl1 = scl;
  BOOST_CHECK_EQUAL(hcl1.size(), in_soft_samples);
  BOOST_CHECK_EQUAL(hcl1.clusters(), in_soft_clusters);
  for (unsigned int i = 0; i < in_soft_samples; ++i)
    BOOST_CHECK_EQUAL(hcl1[i], out_hard_data[i]);

  // Assign
  hclustering hcl2;
  hcl2 = scl;
  BOOST_CHECK_EQUAL(hcl2.size(), in_soft_samples);
  BOOST_CHECK_EQUAL(hcl2.clusters(), in_soft_clusters);
  for (unsigned int i = 0; i < in_soft_samples; ++i)
    BOOST_CHECK_EQUAL(hcl2[i], out_hard_data[i]);
}
  
  
  
