#include <cmath>
#include <cstdlib>
#include <vector>

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include <ttcl/cl/hard_clustering.hxx>
#include <ttcl/hj/ag/agglomerative_distance_clusterer.hxx>
#include <ttcl/hj/ag/limit_agglomerative_distance_clusterer.hxx>
#include <ttcl/hj/ag/single_link.hxx>
#include <ttcl/hj/ag/complete_link.hxx>
#include <ttcl/hj/ag/upgma.hxx>
#include <ttcl/tu/tuning.hxx>

// Tuning static vars
ttcl_tu_static;
ttcl_tu_option_static(ttcl::hj::ag::limit_queue_size);
ttcl_tu_option_static(ttcl::hj::ag::merging_threshold);

// Data format
struct point {
  double x;
  double y;
  point(double _x = 0, double _y = 0) : x(_x), y(_y) {}
};

// Distance
struct euclidean {
  typedef double result_type;

  double operator()(const point& _a, const point& _b) const {
    double dx = _a.x - _b.x;
    double dy = _a.y - _b.y;
    return std::sqrt(dx * dx + dy * dy);
  }
};

// Number of samples
const int n_samples = 100;

// Single link
BOOST_AUTO_TEST_CASE( single_link_cluster ) {
  // Fill at random points
  std::vector<point> v;
  for (int i = 0; i < n_samples; ++i)
    v.push_back(point(10.0 * (rand() / (RAND_MAX + 1.0)),
		      10.0 * (rand() / (RAND_MAX + 1.0))));

  // Normal clustering
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::single_link> clusterer1;
  clusterer1 cl1;
  clusterer1::dendrogram_type d1 = cl1.hier_cluster(v);
  
  // Limited clustering
  ttcl::tu::tuning tu;
  tu.set(new ttcl::hj::ag::limit_queue_size(32));
  typedef ttcl::hj::ag::limit_agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::single_link> clusterer2;
  clusterer2 cl2;
  clusterer2::dendrogram_type d2 = cl2.hier_cluster(v);

  // Compare cuts
  BOOST_REQUIRE_EQUAL(d1.branches(), d2.branches());
  for (uint i = 0; i < d1.branches(); ++i) {
    BOOST_CHECK_EQUAL(d1[i].parent,   d2[i].parent);
    BOOST_CHECK_EQUAL(d1[i].first,    d2[i].first);
    BOOST_CHECK_EQUAL(d1[i].second,   d2[i].second);
    BOOST_CHECK_CLOSE(d1[i].distance, d2[i].distance, 1e-4);
  }
}

// Complete link
BOOST_AUTO_TEST_CASE( complete_link_cluster ) {
  // Fill at random points
  std::vector<point> v;
  for (int i = 0; i < n_samples; ++i)
    v.push_back(point(10.0 * (rand() / (RAND_MAX + 1.0)),
		      10.0 * (rand() / (RAND_MAX + 1.0))));

  // Normal clustering
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::complete_link> clusterer1;
  clusterer1 cl1;
  clusterer1::dendrogram_type d1 = cl1.hier_cluster(v);
  
  // Limited clustering
  ttcl::tu::tuning tu;
  tu.set(new ttcl::hj::ag::limit_queue_size(32));
  typedef ttcl::hj::ag::limit_agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::complete_link> clusterer2;
  clusterer2 cl2;
  clusterer2::dendrogram_type d2 = cl2.hier_cluster(v);

  // Compare cuts
  BOOST_REQUIRE_EQUAL(d1.branches(), d2.branches());
  for (uint i = 0; i < d1.branches(); ++i) {
    BOOST_CHECK_EQUAL(d1[i].parent,   d2[i].parent);
    BOOST_CHECK_EQUAL(d1[i].first,    d2[i].first);
    BOOST_CHECK_EQUAL(d1[i].second,   d2[i].second);
    BOOST_CHECK_CLOSE(d1[i].distance, d2[i].distance, 1e-4);
  }
}

// UPGMA
BOOST_AUTO_TEST_CASE( upgma_cluster ) {
  // Fill at random points
  std::vector<point> v;
  for (int i = 0; i < n_samples; ++i)
    v.push_back(point(10.0 * (rand() / (RAND_MAX + 1.0)),
		      10.0 * (rand() / (RAND_MAX + 1.0))));

  // Normal clustering
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::upgma> clusterer1;
  clusterer1 cl1;
  clusterer1::dendrogram_type d1 = cl1.hier_cluster(v);
  
  // Limited clustering
  ttcl::tu::tuning tu;
  tu.set(new ttcl::hj::ag::limit_queue_size(32));
  typedef ttcl::hj::ag::limit_agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::upgma> clusterer2;
  clusterer2 cl2;
  clusterer2::dendrogram_type d2 = cl2.hier_cluster(v);

  // Compare cuts
  BOOST_REQUIRE_EQUAL(d1.branches(), d2.branches());
  for (uint i = 0; i < d1.branches(); ++i) {
    BOOST_CHECK_EQUAL(d1[i].parent,   d2[i].parent);
    BOOST_CHECK_EQUAL(d1[i].first,    d2[i].first);
    BOOST_CHECK_EQUAL(d1[i].second,   d2[i].second);
    BOOST_CHECK_CLOSE(d1[i].distance, d2[i].distance, 1e-4);
  }
}

// UPGMA
BOOST_AUTO_TEST_CASE( extreme_upgma_cluster ) {
  // Fill at random points
  std::vector<point> v;
  for (int i = 0; i < n_samples; ++i)
    v.push_back(point(10.0 * (rand() / (RAND_MAX + 1.0)),
		      10.0 * (rand() / (RAND_MAX + 1.0))));

  // Normal clustering
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::upgma> clusterer1;
  clusterer1 cl1;
  clusterer1::dendrogram_type d1 = cl1.hier_cluster(v);
  
  // Limited clustering
  ttcl::tu::tuning tu;
  tu.set(new ttcl::hj::ag::limit_queue_size(4));
  typedef ttcl::hj::ag::limit_agglomerative_distance_clusterer<std::vector<point>,
    euclidean, ttcl::hj::ag::upgma> clusterer2;
  clusterer2 cl2;
  clusterer2::dendrogram_type d2 = cl2.hier_cluster(v);

  // Compare cuts
  BOOST_REQUIRE_EQUAL(d1.branches(), d2.branches());
  for (uint i = 0; i < d1.branches(); ++i) {
    BOOST_CHECK_EQUAL(d1[i].parent,   d2[i].parent);
    BOOST_CHECK_EQUAL(d1[i].first,    d2[i].first);
    BOOST_CHECK_EQUAL(d1[i].second,   d2[i].second);
    BOOST_CHECK_CLOSE(d1[i].distance, d2[i].distance, 1e-4);
  }
}

