#include <cmath>
#include <queue>
#include <utility>
#include <vector>

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include <ttcl/cl/wr/data_wrap.hxx>
#include <ttcl/cl/wr/matrix_wrap.hxx>
#include <ttcl/co/distance_matrix.hxx>
#include <ttcl/co/matrix_adapter.hxx>
#include <ttcl/co/mmap_vector.hxx>
#include <ttcl/cr/calinski_centroid_score_data.hxx>
#include <ttcl/cr/calinski_medoid_score_data.hxx>
#include <ttcl/cr/calinski_medoid_score_matrix.hxx>
#include <ttcl/cr/silhouette_width_data.hxx>
#include <ttcl/cr/silhouette_width_matrix.hxx>
#include <ttcl/cr/wr/data_criterion_wrap.hxx>
#include <ttcl/cr/wr/matrix_criterion_wrap.hxx>
#include <ttcl/hj/dendrogram.hxx>
#include <ttcl/hj/ag/agglomerative_clusterer.hxx>
#include <ttcl/hj/ag/agglomerative_distance_clusterer.hxx>
#include <ttcl/hj/ag/choices.hxx>
#include <ttcl/hj/ag/upgma.hxx>
#include <ttcl/hj/wr/hier_data_clusterer_wrap.hxx>
#include <ttcl/hj/wr/hier_matrix_clusterer_wrap.hxx>

// Static vars
ttcl_tu_static;
ttcl_tu_option_static(ttcl::cr::use_non_clustered);
ttcl_tu_option_static(ttcl::hj::ag::merging_threshold);

// Data format
struct point {
  double x;
  double y;
  point() : x(0), y(0) {}
  point(double _x, double _y) : x(_x), y(_y) {}
  point& operator+=(const point& _other) {
    x += _other.x;
    y += _other.y;
    return *this;
  }
  point& operator/=(double s) {
    x /= s;
    y /= s;
    return *this;
  }
};

// Distance
struct euclidean {
  typedef double result_type;

  double operator()(const point& _a, const point& _b) const {
    double dx = _a.x - _b.x;
    double dy = _a.y - _b.y;
    return std::sqrt(dx * dx + dy * dy);
  }
};

// Precision
const double precision = 1e-6;

// Number of samples (small)
const unsigned int small_n_samples = 4;

// Data
const point small_points[] = { point(0, 1), point(1, 0),
			       point(3, 0), point(2, 2) };

// UPGMA merges
const int small_upgma_merges[][3] = { {  4, -1, -1 },
				      {  4, -1, -1 },
				      {  5, -1, -1 },
				      {  5, -1, -1 },
				      {  6,  1,  0 },
				      {  6,  3,  2 },
				      { -1,  5,  4 } };

// Calinskis
const double small_calinskis[] = { NAN,
				   2.42857142857142838110,
				   3.375,
				   NAN };
const int small_optimal = 3;

// Small UPGMA tests
BOOST_AUTO_TEST_CASE( small_upgma ) {
  // Fill the samples
  std::vector<point> data;
  for (unsigned int i = 0; i < small_n_samples; ++i)
    data.push_back(small_points[i]);

  // Cluster
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<
    std::vector<point>, euclidean, ttcl::hj::ag::upgma>
    clusterer_type;
  clusterer_type clusterer;
  clusterer_type::dendrogram_type dendro = clusterer.hier_cluster(data);

  // Check
  BOOST_CHECK_EQUAL(dendro.size(), small_n_samples);
  BOOST_CHECK_EQUAL(dendro.clusters(), 1u);
  BOOST_REQUIRE_EQUAL(dendro.branches(), 2 * small_n_samples - 1);
  for (uint i = 0; i < dendro.branches(); ++i) {
    BOOST_CHECK_EQUAL(dendro[i].parent, small_upgma_merges[i][0]);
    BOOST_CHECK_EQUAL(dendro[i].first,  small_upgma_merges[i][1]);
    BOOST_CHECK_EQUAL(dendro[i].second, small_upgma_merges[i][2]);
  }


  // Calinski

  // Find the tree values
  typedef ttcl::cr::calinski_centroid_score_data<
    std::vector<point>, euclidean> calinski_type;
  calinski_type calinski;
  std::vector<double> tree_cals = calinski.hier_apply(dendro, data);
  BOOST_REQUIRE_EQUAL(tree_cals.size(), small_n_samples);

  // Cut and calinski
  for (uint i = 2; i < small_n_samples - 1; ++i) {
    ttcl::cl::hard_clustering<> h = dendro.cut(i);
    calinski_type::value_type v = calinski.hard_apply(h, data);
    BOOST_CHECK_CLOSE(v, small_calinskis[i - 1], precision);
    BOOST_CHECK_CLOSE(tree_cals[i], small_calinskis[i], precision);
  }
}

/**************************************

Octave code:
function dist = dist(x, y)
  diff = x - y;
  dist = sum(diff .* diff);
end

A = [ 0, 1 ]
B = [ 1, 0 ]
C = [ 3, 0 ]
D = [ 2, 2 ]
M = [ A ; B ; C ; D ]
Sz = 4

Cl = [ 1, 1, 2, 3 ]
k  = 3

B1 = (Cl == 1)' * ones(1, 2)
S1 = sum(Cl == 1)
C1 = sum(M .* B1) / S1

B2 = (Cl == 2)' * ones(1, 2)
S2 = sum(Cl == 2)
C2 = sum(M .* B2) / S2

B3 = (Cl == 3)' * ones(1, 2)
S3 = sum(Cl == 3)
C3 = sum(M .* B3) / S3

MC = sum(M) / Sz

B  = S1 * dist(C1, MC) + S2 * dist(C2, MC) + S3 * dist(C3, MC)
W1 = (M - ones(4,1) * C1) .* B1
W2 = (M - ones(4,1) * C2) .* B2
W3 = (M - ones(4,1) * C3) .* B3
W  = sum(sum(W1 .* W1)) + sum(sum(W2 .* W2)) + sum(sum(W3 .* W3))

Cal = B * (Sz - k) / W / (k - 1)
printf("%.20f\n", Cal)

**************************************/

// Number of samples (big)
const unsigned int big_n_samples = 500;

// Big UPGMA tests
BOOST_AUTO_TEST_CASE( big_upgma ) {
  // Fill the samples
  std::vector<point> data;
  for (unsigned int i = 0; i < big_n_samples; ++i)
    data.push_back(point(double(rand()) / (double(RAND_MAX) + 1.0),
			 double(rand()) / (double(RAND_MAX) + 1.0)));

  // Wrap
  ttcl::cl::wr::data_wrap dw = data;


  // Cluster
  typedef ttcl::hj::ag::agglomerative_distance_clusterer<
    std::vector<point>, euclidean, ttcl::hj::ag::upgma>
    clusterer_type;
  clusterer_type clusterer;
  clusterer_type::dendrogram_type dendro = clusterer.hier_cluster(data);


  // Wrap
  ttcl::hj::wr::hier_data_clusterer_wrap<> aw = clusterer;

  // Call
  clusterer_type::dendrogram_type dendro_w = aw.hier_cluster(dw);

  // Check
  BOOST_CHECK_EQUAL(dendro.size(), dendro_w.size());
  BOOST_CHECK_EQUAL(dendro.clusters(), dendro_w.clusters());
  BOOST_REQUIRE_EQUAL(dendro.branches(), dendro_w.branches());
  for (uint i = 0; i < dendro.branches(); ++i) {
    BOOST_CHECK_EQUAL(dendro[i].parent, dendro_w[i].parent);
    BOOST_CHECK_EQUAL(dendro[i].first,  dendro_w[i].first);
    BOOST_CHECK_EQUAL(dendro[i].second, dendro_w[i].second);
  }


  // Calinski centroids

  // Find the tree values
  typedef ttcl::cr::calinski_centroid_score_data<
    std::vector<point>, euclidean> calinski_type;
  calinski_type calinski;
  std::vector<double> tree_cals = calinski.hier_apply(dendro, data);
  BOOST_REQUIRE_EQUAL(tree_cals.size(), big_n_samples);

  // Cut and check
  for (unsigned int i = 2; i < big_n_samples; ++i) {
    ttcl::cl::hard_clustering<> h = dendro.cut(i);
    calinski_type::value_type v = calinski.hard_apply(h, data);
    BOOST_CHECK_CLOSE(tree_cals[i - 1], v, precision);
  }

  // Wrap
  ttcl::cr::wr::data_criterion_wrap<> cw = calinski;
  std::vector<double> tree_cals_w = cw.hier_apply(dendro, dw);
  BOOST_REQUIRE_EQUAL(tree_cals_w.size(), big_n_samples);

  // Check
  for (unsigned int i = 2; i < big_n_samples; ++i) {
    BOOST_CHECK_EQUAL(tree_cals[i - 1], tree_cals_w[i - 1]);
  }


  // Calinski medoids

  // Find the tree values
  typedef ttcl::cr::calinski_medoid_score_data<
    std::vector<point>, euclidean> calinski_m_type;
  calinski_m_type calinski_m;
  std::vector<double> tree_cals_m = calinski_m.hier_apply(dendro, data);
  BOOST_REQUIRE_EQUAL(tree_cals_m.size(), big_n_samples);

  // Cut and check
  for (unsigned int i = 2; i < big_n_samples; ++i) {
    ttcl::cl::hard_clustering<> h = dendro.cut(i);
    calinski_type::value_type v = calinski_m.hard_apply(h, data);
    BOOST_CHECK_CLOSE(tree_cals_m[i - 1], v, precision);
  }

  // Wrap
  ttcl::cr::wr::data_criterion_wrap<> cmw = calinski_m;
  std::vector<double> tree_cals_m_w = cmw.hier_apply(dendro, dw);
  BOOST_REQUIRE_EQUAL(tree_cals_m_w.size(), big_n_samples);

  // Check
  for (unsigned int i = 2; i < big_n_samples; ++i) {
    BOOST_CHECK_EQUAL(tree_cals_m[i - 1], tree_cals_m_w[i - 1]);
  }


  // Silhouette

  // Find the tree values
  typedef ttcl::cr::silhouette_width_data<
    std::vector<point>, euclidean> silhouette_type;
  silhouette_type silhouette;
  std::vector<double> tree_sil = silhouette.hier_apply(dendro, data);
  BOOST_REQUIRE_EQUAL(tree_sil.size(), big_n_samples);

  // Cut and check
  for (unsigned int i = 2; i < big_n_samples; ++i) {
    ttcl::cl::hard_clustering<> h = dendro.cut(i);
    silhouette_type::value_type v = silhouette.hard_apply(h, data);
    BOOST_CHECK_CLOSE(tree_sil[i - 1], v, precision);
  }

  // Wrap
  ttcl::cr::wr::data_criterion_wrap<> sw = silhouette;
  std::vector<double> tree_sil_w = sw.hier_apply(dendro, dw);
  BOOST_REQUIRE_EQUAL(tree_sil_w.size(), big_n_samples);

  // Check
  for (unsigned int i = 2; i < big_n_samples; ++i) {
    BOOST_CHECK_EQUAL(tree_sil[i - 1], tree_sil_w[i - 1]);
  }
}


/*************************************/

// Big Choices!!

// Mmap queue
template <typename Value>
class mmap_queue :
  public std::priority_queue<Value,
			     ttcl::co::mmap_vector<Value>,
			     std::greater<Value> > {
};

// Mmap matrix
template <typename Value>
class mmap_matrix :
  public ttcl::co::matrix_adapter< ttcl::co::mmap_vector<Value> > {
public:
  // Constructor
  mmap_matrix(size_t _rows, size_t _cols) :
    ttcl::co::matrix_adapter< ttcl::co::mmap_vector<Value> >(_rows, _cols) {
  }
};

// Mmap choices
typedef
ttcl_tu_add_choice_tem(
  ttcl_tu_add_choice_tem(ttcl::tu::default_choices,
			 ttcl::cr::point_to_cluster_template,
			 mmap_matrix),
  ttcl::hj::ag::queue_template,
  mmap_queue) mmap_choices;

// Big UPGMA tests (with a matrix distance)
BOOST_AUTO_TEST_CASE( mmap_big_upgma_matrix ) {
  // Fill the samples
  std::vector<point> data;
  for (unsigned int i = 0; i < big_n_samples; ++i)
    data.push_back(point(double(rand()) / (double(RAND_MAX) + 1.0),
			 double(rand()) / (double(RAND_MAX) + 1.0)));


  // Create the matrix
  typedef ttcl::co::distance_matrix<
    ttcl::co::symmetric_matrix< ttcl::co::mmap_vector<double> > >
    matrix_type;
  matrix_type matrix(data, euclidean());

  // Wrap
  ttcl::cl::wr::matrix_wrap mw = matrix;


  // Cluster
  typedef ttcl::hj::ag::agglomerative_clusterer<
    matrix_type, ttcl::hj::ag::upgma, mmap_choices> clusterer_type;
  clusterer_type clusterer;
  clusterer_type::dendrogram_type dendro = clusterer.hier_cluster(matrix);

  // Wrap
  ttcl::hj::wr::hier_matrix_clusterer_wrap<> aw = clusterer;

  // Call
  clusterer_type::dendrogram_type dendro_w = aw.hier_cluster(mw);

  // Check
  BOOST_CHECK_EQUAL(dendro.size(), dendro_w.size());
  BOOST_CHECK_EQUAL(dendro.clusters(), dendro_w.clusters());
  BOOST_REQUIRE_EQUAL(dendro.branches(), dendro_w.branches());
  for (uint i = 0; i < dendro.branches(); ++i) {
    BOOST_CHECK_EQUAL(dendro[i].parent, dendro_w[i].parent);
    BOOST_CHECK_EQUAL(dendro[i].first,  dendro_w[i].first);
    BOOST_CHECK_EQUAL(dendro[i].second, dendro_w[i].second);
  }


  // Silhouette 1

  // Find the tree values
  typedef ttcl::cr::silhouette_width_matrix<matrix_type> silhouette1_type;
  silhouette1_type silhouette1;
  std::vector<double> tree_sil1 = silhouette1.hier_apply(dendro, matrix);
  BOOST_REQUIRE_EQUAL(tree_sil1.size(), big_n_samples);

  // Cut and check
  for (unsigned int i = 2; i < big_n_samples; ++i) {
    ttcl::cl::hard_clustering<> h = dendro.cut(i);
    silhouette1_type::value_type v = silhouette1.hard_apply(h, matrix);
    BOOST_CHECK_CLOSE(tree_sil1[i - 1], v, precision);
  }

  // Wrap
  ttcl::cr::wr::matrix_criterion_wrap<> sw_1 = silhouette1;
  std::vector<double> tree_sil1_w = sw_1.hier_apply(dendro, mw);
  BOOST_REQUIRE_EQUAL(tree_sil1_w.size(), big_n_samples);

  // Check
  for (unsigned int i = 2; i < big_n_samples; ++i) {
    BOOST_CHECK_EQUAL(tree_sil1[i - 1], tree_sil1_w[i - 1]);
  }


  // Silhouette 2

  // Find the tree values with mmap
  typedef ttcl::cr::silhouette_width_matrix<
    matrix_type, mmap_choices> silhouette2_type;
  silhouette2_type silhouette2;
  std::vector<double> tree_sil2 = silhouette2.hier_apply(dendro, matrix);
  BOOST_REQUIRE_EQUAL(tree_sil1.size(), tree_sil2.size());

  // Cut and check
  for (unsigned int i = 0; i < tree_sil1.size(); ++i) {
    BOOST_CHECK_EQUAL(tree_sil1[i], tree_sil2[i]);
  }

  // Wrap
  ttcl::cr::wr::matrix_criterion_wrap<> sw_2 = silhouette2;
  std::vector<double> tree_sil2_w = sw_2.hier_apply(dendro, mw);
  BOOST_REQUIRE_EQUAL(tree_sil2_w.size(), big_n_samples);

  // Check
  for (unsigned int i = 2; i < big_n_samples; ++i) {
    BOOST_CHECK_EQUAL(tree_sil2[i - 1], tree_sil2_w[i - 1]);
  }
}
