#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <iostream>

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>
#include <boost/tuple/tuple.hpp>

#include <ttcl/co/mmap_vector.hxx>
#include <ttcl/co/ro_mmap_vector.hxx>

// Number of points
const unsigned int n_points = 1000000;

// Mmap vector
BOOST_AUTO_TEST_CASE( mmap_vector ) {
  // Create vector
  ttcl::co::mmap_vector<unsigned int> vec1(n_points, 0);

  // Check initialization
  BOOST_REQUIRE_EQUAL(vec1.size(), n_points);
  for (unsigned int i = 0; i < n_points; ++i) 
    BOOST_CHECK_EQUAL(vec1[i], 0u);
  
  // Fill them
  for (unsigned int i = 0; i < n_points; ++i) {
    vec1[i] = i;
    BOOST_CHECK_EQUAL(vec1[i], i);
  }

  // Check it again
  for (unsigned int i = 0; i < n_points; ++i)
    BOOST_CHECK_EQUAL(vec1[i], i);
}

// Tuple vector
BOOST_AUTO_TEST_CASE( tuple_mmap_vector ) {
  // Data
  typedef boost::tuple<double, unsigned int, unsigned int> data;
  
  // Raw vector
  /* data* raw_data = new data[2];
     BOOST_CHECK_EQUAL((char*)&(raw_data[1]) - (char*)&(raw_data[0]),
                       sizeof(data));
     delete[] raw_data; */

  // Create vector
  ttcl::co::mmap_vector<data> vec1(n_points);

  // Check initialization
  BOOST_REQUIRE_EQUAL(vec1.size(), n_points);
  for (unsigned int i = 0; i < n_points; ++i) { 
    BOOST_CHECK_EQUAL(vec1[i].get<0>(), 0.0);
    BOOST_CHECK_EQUAL(vec1[i].get<1>(), 0u);
    BOOST_CHECK_EQUAL(vec1[i].get<2>(), 0u);
  }
  
  // Fill them
  for (unsigned int i = 0; i < n_points; ++i) {
    vec1[i].get<0>() = 2.0 * i;
    vec1[i].get<1>() = i - 1;
    vec1[i].get<2>() = i + 1;
    BOOST_CHECK_EQUAL(vec1[i].get<0>(), 2.0 * i);
    BOOST_CHECK_EQUAL(vec1[i].get<1>(), i - 1);
    BOOST_CHECK_EQUAL(vec1[i].get<2>(), i + 1);
  }

  // Check it again
  for (unsigned int i = 0; i < n_points; ++i) {
    BOOST_CHECK_EQUAL(vec1[i].get<0>(), 2.0 * i);
    BOOST_CHECK_EQUAL(vec1[i].get<1>(), i - 1);
    BOOST_CHECK_EQUAL(vec1[i].get<2>(), i + 1);
  }
}

// Pushy tuple vector
BOOST_AUTO_TEST_CASE( pushy_tuple_mmap_vector ) {
  // Data
  typedef boost::tuple<double, unsigned int, unsigned int> data;
  
  // Create vector
  ttcl::co::mmap_vector<data> vec1;

  // Check initialization
  BOOST_REQUIRE_EQUAL(vec1.size(), 0u);
  
  // Fill them
  for (unsigned int i = 0; i < n_points; ++i) {
    vec1.push_back(boost::make_tuple(2.0 * i, i - 1, i + 1));
    BOOST_REQUIRE_EQUAL(vec1.size(), i + 1);
    BOOST_CHECK_EQUAL(vec1[i].get<0>(), 2.0 * i);
    BOOST_CHECK_EQUAL(vec1[i].get<1>(), i - 1);
    BOOST_CHECK_EQUAL(vec1[i].get<2>(), i + 1);
  }

  // Check it again
  BOOST_REQUIRE_EQUAL(vec1.size(), n_points);
  for (unsigned int i = 0; i < n_points; ++i) {
    BOOST_CHECK_EQUAL(vec1[i].get<0>(), 2.0 * i);
    BOOST_CHECK_EQUAL(vec1[i].get<1>(), i - 1);
    BOOST_CHECK_EQUAL(vec1[i].get<2>(), i + 1);
  }
}

// Read-only mmap vector
BOOST_AUTO_TEST_CASE( ro_mmap_vector ) {
  // Open the test file
  char tmp_file[20];
  strcpy(tmp_file, "/tmp/test.XXXXXX");
  int fd = mkstemp(tmp_file);
  BOOST_REQUIRE(fd != -1);

  {
    // Create vector
    ttcl::co::mmap_vector<unsigned int> vec1(n_points, 0, fd);

    // Check initialization
    BOOST_REQUIRE_EQUAL(vec1.size(), n_points);
    for (unsigned int i = 0; i < n_points; ++i) 
      BOOST_CHECK_EQUAL(vec1[i], 0u);
  
    // Fill them
    for (unsigned int i = 0; i < n_points; ++i) {
      vec1[i] = i;
      BOOST_CHECK_EQUAL(vec1[i], i);
    }

    // Check it again
    for (unsigned int i = 0; i < n_points; ++i)
      BOOST_CHECK_EQUAL(vec1[i], i);

    // Open the read only vector
    ttcl::co::ro_mmap_vector<unsigned int> vec2(fd);

    // Check equal
    BOOST_REQUIRE(vec1.size() <= vec2.size());
    for (unsigned int i = 0; i < vec1.size(); ++i) 
      BOOST_CHECK_EQUAL(vec1[i], vec2[i]);
  }

  // Delete file
  close(fd);
  unlink(tmp_file);
}
