#include <vector>

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include <boost/tuple/tuple.hpp>

#include <ttcl/co/array.hxx>
#include <ttcl/co/fake_pair.hxx>
#include <ttcl/co/feature_container.hxx>
#include <ttcl/it/em/nb_bernouilli_clusterer.hxx>
#include <ttcl/it/em/nb_bernouilli_incremental_clusterer.hxx>
#include <ttcl/it/em/nb_bernouilli_noisy_clusterer.hxx>
#include <ttcl/it/em/nb_multinomial_clusterer.hxx>

// Static vars
ttcl_tu_static;
ttcl_tu_option_static(ttcl::it::max_iterations);
ttcl_tu_option_static(ttcl::it::stop_threshold);
ttcl_tu_option_static(ttcl::it::start_policy);
ttcl_tu_option_static(ttcl::it::peak_fraction);
ttcl_tu_option_static(ttcl::it::em::alpha_laplace_term);
ttcl_tu_option_static(ttcl::it::em::theta_laplace_term);
ttcl_tu_option_static(ttcl::it::em::learning_rate);
ttcl_tu_option_static(ttcl::it::em::inner_softness);

// Number of points
const unsigned int n_points = 6;

// Points
const unsigned int  point1[] = { 3, /* -> */ 0, 1, 2 };
const unsigned int  point2[] = { 3, /* -> */ 0, 1, 3 };
const unsigned int  point3[] = { 2, /* -> */ 4, 5 };
const unsigned int  point4[] = { 5, /* -> */ 2, 3, 4, 5 };
const unsigned int  point5[] = { 3, /* -> */ 3, 4, 5 };
const unsigned int  point6[] = { 3, /* -> */ 0, 2, 3 };
const unsigned int* points[] = { point1, point2, point3,
				 point4, point5, point6 };

// Sample type
typedef ttcl::co::array<
  ttcl::co::fake_pair_second<unsigned int,
			     unsigned int, 1> > sample_type;

// Container type
typedef std::vector<sample_type> container_type;

// Data type
typedef ttcl::co::feature_container<container_type> data_type;

// Multinomial
BOOST_AUTO_TEST_CASE( multinomial ) {
  // Fill the data
  data_type data;
  for (unsigned int i = 0; i < n_points; ++i) {
    data.push_back(sample_type(points[i][0]));
    for (unsigned int f = 0; f < points[i][0]; ++f)
      data.back()[f].first = points[i][f + 1];
  }

  // Multinomial Clusterer
  typedef ttcl::it::em::nb_multinomial_clusterer<data_type>
    multi_clusterer_type;
  multi_clusterer_type multi_clusterer;

  // Sizes
  for (unsigned int k = 1; k < 10; ++k) {
    // Clustering result
    multi_clusterer_type::soft_clustering_type
      multi_soft_clustering(data.size(), k);

    // Model
    multi_clusterer_type::base_model_type* model =
      multi_clusterer.soft_model_on(multi_soft_clustering, data, k);

    multi_clusterer_type::value_type log_like = model->train_log_likelihood();
    // ttcl::uint iterations = model->train_iterations();

    // Log likelihood
    ttcl::uint free_params = model->free_parameters();
    double AIC = 2.0 * double(free_params) - log_like;

    // Write
    std::cerr << "M  " << k << ' ' << log_like << ' '
	      << free_params << ' ' << AIC << std::endl;

    // Delete the model
    delete model;
  }
}

// Bernouilli
BOOST_AUTO_TEST_CASE( bernouilli ) {
  // Fill the data
  data_type data;
  for (unsigned int i = 0; i < n_points; ++i) {
    data.push_back(sample_type(points[i][0]));
    for (unsigned int f = 0; f < points[i][0]; ++f)
      data.back()[f].first = points[i][f + 1];
  }

  // Bernouilli Clusterer
  typedef ttcl::it::em::nb_bernouilli_clusterer<data_type>
    berni_clusterer_type;
  berni_clusterer_type berni_clusterer;

  // Sizes
  for (unsigned int k = 1; k < 10; ++k) {
    // Clustering result
    berni_clusterer_type::soft_clustering_type
      berni_soft_clustering(data.size(), k);

    // Model
    berni_clusterer_type::base_model_type* model =
      berni_clusterer.soft_model_on(berni_soft_clustering, data, k);

    berni_clusterer_type::value_type log_like = model->train_log_likelihood();
    // ttcl::uint iterations = model->train_iterations();

    // Log likelihood
    ttcl::uint free_params = model->free_parameters();
    double AIC = 2.0 * double(free_params) - log_like;

    // Write
    std::cerr << "B  " << k << ' ' << log_like << ' '
	      << free_params << ' ' << AIC << std::endl;

    // Delete the model
    delete model;
  }
}

// Bernouilli Noisy
BOOST_AUTO_TEST_CASE( bernouilli_noisy ) {
  // Fill the data
  data_type data;
  for (unsigned int i = 0; i < n_points; ++i) {
    data.push_back(sample_type(points[i][0]));
    for (unsigned int f = 0; f < points[i][0]; ++f)
      data.back()[f].first = points[i][f + 1];
  }

  // Bernouilli Clusterer
  typedef ttcl::it::em::nb_bernouilli_noisy_clusterer<data_type>
    berni_clusterer_type;
  berni_clusterer_type berni_clusterer;

  // Sizes
  for (unsigned int k = 1; k < 10; ++k) {
    // Clustering result
    berni_clusterer_type::soft_clustering_type
      berni_soft_clustering(data.size(), k);

    // Model
    berni_clusterer_type::base_model_type* model =
      berni_clusterer.soft_model_on(berni_soft_clustering, data, k);

    berni_clusterer_type::value_type log_like = model->train_log_likelihood();
    // ttcl::uint iterations = model->train_iterations();

    // Log likelihood
    ttcl::uint free_params = model->free_parameters();
    double AIC = 2.0 * double(free_params) - log_like;

    // Write
    std::cerr << "BN " << k << ' ' << log_like << ' '
	      << free_params << ' ' << AIC << std::endl;

    // Delete the model
    delete model;
  }
}

// Bernouilli Incremental
BOOST_AUTO_TEST_CASE( bernouilli_incremental ) {
  // Fill the data
  data_type data;
  for (unsigned int i = 0; i < n_points; ++i) {
    data.push_back(sample_type(points[i][0]));
    for (unsigned int f = 0; f < points[i][0]; ++f)
      data.back()[f].first = points[i][f + 1];
  }

  // Bernouilli Clusterer
  typedef ttcl::it::em::nb_bernouilli_incremental_clusterer<data_type>
    berni_clusterer_type;
  berni_clusterer_type berni_clusterer;

  // Sizes
  for (unsigned int k = 1; k < 10; ++k) {
    // Clustering result
    berni_clusterer_type::soft_clustering_type
      berni_soft_clustering(data.size(), k);

    // Model
    berni_clusterer_type::base_model_type* model =
      berni_clusterer.soft_model_on(berni_soft_clustering, data, k);

    berni_clusterer_type::value_type log_like = model->train_log_likelihood();
    // ttcl::uint iterations = model->train_iterations();

    // Log likelihood
    ttcl::uint free_params = model->free_parameters();
    double AIC = 2.0 * double(free_params) - log_like;

    // Write
    std::cerr << "BI " << k << ' ' << log_like << ' '
	      << free_params << ' ' << AIC << std::endl;

    // Delete the model
    delete model;
  }
}
