#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>

#include <ttcl/co/array.hxx>
#include <ttcl/ut/tag_allocator.hxx>

// Tags
class tag1 {};
class tag2 {};

// Static vars
ttcl_ut_tag_allocator_static;

// Sizes
const int size1 =  16;
const int size2 = 256;

// Pointered
template <typename T>
struct pointered {
  T* data_;

  pointered() : data_(new T()) {}
  pointered(const T& _value) : data_(new T(_value)) {}
  pointered(const pointered& _other) : data_(new T(*_other.data_)) {}
  ~pointered() { delete data_; }
  pointered& operator=(const pointered& _other) {
    delete data_; data_ = new T(*_other.data_); }
};

// Typedefs
typedef pointered<int> pi;
typedef pointered<double> pd;

// Allocator
BOOST_AUTO_TEST_CASE( allocator ) {
  // Create vector1
  ttcl::co::array<pi, ttcl::ut::tag_allocator<pi, tag1> > v1;
  for (int i = 0; i < size1; ++i) v1.resize(i + 1, i);
  
  {
    // Create vector2
    ttcl::co::array<pi, ttcl::ut::tag_allocator<pi, tag2> > v2;
    for (int i = 0; i < size2; ++i) v2.resize(i + 1, i);
    
    // Check bytes
    BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag1>::bytes(), size1 * sizeof(pi));
    BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag2>::bytes(), size2 * sizeof(pi));
 
    // Check objects
    BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag1>::objects(), size1);
    BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag2>::objects(), size2);
  }

  // Check
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag1>::bytes(), size1 * sizeof(pi));
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag2>::bytes(),     0);

  // Check objects
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag1>::objects(), size1);
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag2>::objects(),     0);

  {
    // Create vector3 -> at tag1
    ttcl::co::array<pd, ttcl::ut::tag_allocator<pd, tag1> > v3;
    for (int i = 0; i < size2; ++i) v3.resize(i + 1, double(i));
    
    // Check bytes
    BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag1>::bytes(),
		      size1 * sizeof(pi) + size2 * sizeof(pd));
    BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag2>::bytes(), 0);
 
    // Check objects
    BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag1>::objects(), size1 + size2);
    BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag2>::objects(), 0);
  }

  // Check
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag1>::bytes(), size1 * sizeof(pi));
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag2>::bytes(),     0);

  // Check objects
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag1>::objects(), size1);
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag2>::objects(),     0);

  // Clear
  while (not v1.empty())
    v1.resize(v1.size() - 1);
  v1.clear();

  // Check
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag1>::bytes(), 0);
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag2>::bytes(), 0);

  // Check objects
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag1>::objects(), 0);
  BOOST_CHECK_EQUAL(ttcl::ut::tag_block<tag2>::objects(), 0);
}
